@@ -11,10 +11,9 @@ static int mousebuttons = 0;
 static g3d_gui *windowhit = NULL;
 
 //text field state - only one ever active/focused
-static string fieldtext;  //copy of while focused
-static int fieldpos = -1; //-1=no focus, -2=wanting to commit
+static string fieldname, fieldtext;  //copy of while focused
+static int fieldpos = -1, fieldlen = 0; //-1=no focus, -2=wanting to commit
 static bool fieldactive; 
-static char *fieldref;    //so can itentify the source of the field
 
 bool menukey(int code, bool isdown, int cooked) 
 {
@@ -28,6 +27,8 @@ bool menukey(int code, bool isdown, int cooked)
         case SDLK_KP_ENTER:
 			fieldpos = -2; //signal field commit
 			return false;
+        case SDLK_HOME:
+        case SDLK_END:
 		case SDLK_DELETE:
         case SDLK_BACKSPACE:
         case SDLK_LEFT:
@@ -40,7 +41,14 @@ bool menukey(int code, bool isdown, int cooked)
 	if(!isdown) return true;	
 	int len = strlen(fieldtext);
     if(fieldpos>len) fieldpos = len;
-	switch(code) {
+	switch(code) 
+    {
+        case SDLK_HOME:
+            fieldpos = 0;
+            break;
+        case SDLK_END:
+            fieldpos = len;
+            break;
 		case SDLK_LEFT:
 			if(fieldpos > 0) fieldpos--;
 			break;
@@ -49,21 +57,57 @@ bool menukey(int code, bool isdown, int cooked)
 			if(fieldpos < len) fieldpos++;
 			break;
 		case SDLK_DELETE:
-            if(fieldpos < len) memmove(fieldtext+fieldpos, fieldtext+fieldpos+1, len-fieldpos);
+            if(fieldpos < len) 
+            {
+                memmove(fieldtext+fieldpos, fieldtext+fieldpos+1, len-fieldpos);
+                fieldtext[len-1] = '\0';
+            }
 			break;
 		case SDLK_BACKSPACE:
             if(fieldpos > 0) 
-            {   fieldpos--;
+            {   
+                fieldpos--;
                 memmove(fieldtext+fieldpos, fieldtext+fieldpos+1, len-fieldpos);
+                fieldtext[len-1] = '\0';
             }
 			break;
 		default:
-            memmove(fieldtext+fieldpos+1, fieldtext+fieldpos, len-fieldpos); //length then limited inside field draw code
-			fieldtext[fieldpos++] = cooked;
+            if(fieldpos<fieldlen)
+            {
+                memmove(fieldtext+fieldpos+1, fieldtext+fieldpos, len-fieldpos); //length then limited inside field draw code
+			    fieldtext[fieldpos++] = cooked;
+                fieldtext[len+1] = '\0';
+            }
+            break;
 	}
 	return true;
 }
 
+VARP(gui2d, 0, 0, 1);
+
+static bool hascursor;
+static float cursorx = 0.5f, cursory = 0.5f;
+
+void g3d_cursorpos(float &x, float &y)
+{
+    if(gui2d) { x = cursorx; y = cursory; }
+    else x = y = 0.5f;
+}
+
+void g3d_resetcursor()
+{
+    cursorx = cursory = 0.5f;
+}
+
+bool g3d_movecursor(int dx, int dy)
+{
+    if(!gui2d || !hascursor) return false;
+    const float CURSORSCALE = 500.0f;
+    cursorx = max(0, min(1, cursorx+dx/CURSORSCALE));
+    cursory = max(0, min(1, cursory+dy/CURSORSCALE));
+    return true;
+}
+
 #define SHADOW 4
 #define ICON_SIZE (FONTH-SHADOW)
 #define SKIN_W 256
@@ -71,7 +115,7 @@ bool menukey(int code, bool isdown, int cooked)
 #define SKIN_SCALE 4
 #define INSERT (3*SKIN_SCALE)
 
-VARP(guiautotab, FONTH*6, FONTH*16, FONTH*40);
+VARP(guiautotab, 6, 16, 40);
 
 struct gui : g3d_gui
 {
@@ -98,7 +142,7 @@ struct gui : g3d_gui
         if(tcurrent)
         {
             if(layoutpass && !tpos) tcurrent = NULL; //disable tabs because you didn't start with one
-            if(!curdepth && (layoutpass ? 0 : cury) + ysize > guiautotab) tab(NULL, tcolor); 
+            if(!curdepth && (layoutpass ? 0 : cury) + ysize > guiautotab*FONTH) tab(NULL, tcolor); 
         }
     }
 
@@ -110,8 +154,12 @@ struct gui : g3d_gui
         if(curdepth != 0) return;
         tcolor = color;
         tpos++; 
-        s_sprintfd(title)("%d", tpos);
-        if(!name) name = title;
+        if(!name) 
+        {
+            static string title;
+            s_sprintf(title)("%d", tpos);
+            name = title;
+        }
         int w = text_width(name) - 2*INSERT;
         if(layoutpass) 
         {  
@@ -193,9 +241,9 @@ struct gui : g3d_gui
     void progress(float percent) { autotab(); line_(FONTH*2/5, percent); }
 
     //use to set min size (useful when you have progress bars)
-    void strut(int size) { layout(isvertical() ? size*FONTH : 0, isvertical() ? 0 : size*FONTH); }
+    void strut(int size) { layout(isvertical() ? size*FONTW : 0, isvertical() ? 0 : size*FONTH); }
     //add space between list items
-    void space(int size) { layout(isvertical() ? 0 : size*FONTH, isvertical() ? size*FONTH : 0); }
+    void space(int size) { layout(isvertical() ? 0 : size*FONTW, isvertical() ? size*FONTH : 0); }
 
     int layout(int w, int h)
     {
@@ -229,10 +277,8 @@ struct gui : g3d_gui
         return windowhit==this && hitx>=x && hity>=y && hitx<x+w && hity<y+h;
     }
 
-    int image(const char *path, float scale, bool overlaid)
+    int image(Texture *t, float scale, bool overlaid)
     {
-        Texture *t = textureload(path, 0, true, false);
-        if(t==crosshair) return 0;
         autotab();
         if(scale==0) scale = 1;
         int size = (int)(scale*2*FONTH)-SHADOW;
@@ -245,39 +291,44 @@ struct gui : g3d_gui
         autotab();
         if(scale==0) scale = 1;
         int size = (int)(scale*2*FONTH)-SHADOW;
-        if(t!=crosshair && visible()) icon_(t, true, true, curx, cury, size, ishit(size+SHADOW, size+SHADOW));
+        if(t!=notexture && visible()) icon_(t, true, true, curx, cury, size, ishit(size+SHADOW, size+SHADOW));
         return layout(size+SHADOW, size+SHADOW);
     }
 
-    void slider(int &val, int vmin, int vmax, int color)
-    {	
+    void slider(int &val, int vmin, int vmax, int color, char *label)
+    {
         autotab();
         int x = curx;
         int y = cury;
         line_(10);
-        if(visible()) 
+        if(visible())
         {
-            s_sprintfd(label)("%d", val);
+            if(!label)
+            {
+                static string s;
+                s_sprintf(s)("%d", val);
+                label = s;
+            }
             int w = text_width(label);
-        
+
             bool hit;
             int px, py;
-            if(ishorizontal()) 
+            if(ishorizontal())
             {
                 hit = ishit(FONTH, ysize, x, y);
                 px = x + (FONTH-w)/2;
-                py = y + (ysize-FONTH) - ((ysize-FONTH)*(val-vmin))/(vmax-vmin); //zero at the bottom
+                py = y + (ysize-FONTH) - ((ysize-FONTH)*(val-vmin))/max(vmax-vmin, 1); //zero at the bottom
             }
             else
             {
                 hit = ishit(xsize, FONTH, x, y);
-                px = x + ((xsize-w)*(val-vmin))/(vmax-vmin);
+                px = x + ((xsize-w)*(val-vmin))/max(vmax-vmin, 1);
                 py = y;
             }
-        
+
             if(hit) color = 0xFF0000;
             text_(label, px, py, color, hit && actionon);
-            if(hit && actionon) 
+            if(hit && actionon)
             {
                 int vnew = 1+vmax-vmin;
                 if(ishorizontal()) vnew = int(vnew*(y+ysize-hity)/ysize);
@@ -288,64 +339,70 @@ struct gui : g3d_gui
         }
     }
 
-    char *field(char *text, int color, int length)
+    char *field(const char *name, int color, int length, const char *initval)
 	{	
-        int w = FONTH*length;
+        length = min(length, (int)sizeof(string)-1);
+        int w = char_width('%')*length + FONTW;
         char *result = NULL;
-        if(!layoutpass)
+        if(visible() && !layoutpass)
 		{
-            bool hit = ishit(w, FONTH);            
-            if(hit && (mousebuttons & G3D_DOWN) && fieldref != text) //mouse request focus
+            bool hit = ishit(w, FONTH), editing = !strcmp(fieldname, name);            
+            if(hit && (mousebuttons&G3D_DOWN) && !editing) //mouse request focus
             {
-                s_strcpy(fieldtext, text);
-                fieldref = text;
+                s_strcpy(fieldname, name);
+                s_strcpy(fieldtext, initval);
+                fieldtext[length] = '\0';
+                fieldpos = strlen(fieldtext);
+                fieldlen = length;
+                editing = true;
             }
-            if(fieldref==text) 
+            if(editing && windowhit==this)
             {
-                if((int)strlen(fieldtext) > length) fieldtext[length] = '\0';
-                if(fieldpos > length) fieldpos = length;
-                if(fieldpos==-2) {
+                if(fieldpos==-2) 
+                {
                     result = fieldtext;
                     fieldpos = -1;
-                    fieldref = NULL;
+                    fieldname[0] = '\0';
+                    editing = false;
                 } 
-                text = fieldtext;
-                fieldactive = hit;
-                if(!hit) fieldpos = -1; //mouse wandered out of focus
+                else if(hit) fieldactive = true;
+                else fieldpos = -1; // mouse wandered out of focus
             }
-            if(hit && (mousebuttons & G3D_PRESSED)) //mouse request position
+            if(editing && hit && (mousebuttons&G3D_PRESSED)) //mouse request position
             {
-                extern int char_width(int c, int x);
-                int x = curx;
+                int x = curx+FONTW/2;
                 fieldpos = 0;
-                while(text[fieldpos] && fieldpos < length) {
-                    int nx = char_width(text[fieldpos], x);
+                while(fieldtext[fieldpos] && fieldpos < length)
+                {
+                    int nx = char_width(fieldtext[fieldpos], x);
                     if(nx > hitx) break;
                     x = nx;
                     fieldpos++;
                 }
             }
                            
-            bool editing = fieldref && (fieldtext==text);
             notextureshader->set();
-			if(editing) glColor3f(1, 0.5f, 0.5f);
-            else glColor3fv(light.v);
+            glDisable(GL_TEXTURE_2D);
+			if(editing) glColor3f(1, 0, 0);
+            else glColor3ub(color>>16, (color>>8)&0xFF, color&0xFF);
 			glBegin(GL_LINE_LOOP);
             rect_(curx, cury, w, FONTH);
 			glEnd();
+            glEnable(GL_TEXTURE_2D);
             defaultshader->set();
             
-            draw_text(text, curx, cury, color>>16, (color>>8)&0xFF, color&0xFF);
+            draw_text(editing ? fieldtext : (result ? result : initval), curx+FONTW/2, cury, color>>16, (color>>8)&0xFF, color&0xFF);
             
-            if(hit && fieldpos>=0 && (totalmillis/250)&1) 
+            if(editing && hit && fieldpos>=0 && (totalmillis/250)&1) 
             {
-                int fx = curx + text_width(text, fieldpos);
-                color = 0xFF0000;
-                glColor3ub(color>>16, (color>>8)&0xFF, color&0xFF);
+                int fx = curx+FONTW/2 + text_width(fieldtext, fieldpos);
+                glColor3f(1, 0, 0);
                 notextureshader->set();
+                glDisable(GL_TEXTURE_2D);
                 glBegin(GL_QUADS);
                 rect_(fx-2, cury, 4, FONTH);
                 glEnd();				
+                glEnable(GL_TEXTURE_2D);
                 defaultshader->set();
             }
         }
@@ -355,7 +412,7 @@ struct gui : g3d_gui
 
     void rect_(float x, float y, float w, float h, int usetc = -1) 
     {
-        GLint tc[4][2] = {{0, 0}, {1, 0}, {1, 1}, {0, 1}};
+        static const GLint tc[4][2] = {{0, 0}, {1, 0}, {1, 1}, {0, 1}};
         if(usetc>=0) glTexCoord2iv(tc[usetc]); 
         glVertex2f(x, y);
         if(usetc>=0) glTexCoord2iv(tc[(usetc+1)%4]);
@@ -373,6 +430,40 @@ struct gui : g3d_gui
         draw_text(text, x, y, color>>16, (color>>8)&0xFF, color&0xFF);
     }
 
+    void background(int color, int inheritw, int inherith)
+    {
+        if(layoutpass) return;
+        glDisable(GL_TEXTURE_2D);
+        notextureshader->set();
+        glColor4ub(color>>16, (color>>8)&0xFF, color&0xFF, 0x80);
+        int w = xsize, h = ysize;
+        if(inheritw>0) 
+        {
+            int parentw = curlist;
+            while(inheritw>0 && lists[parentw].parent>=0)
+            {
+                parentw = lists[parentw].parent;
+                inheritw--;
+            }
+            w = lists[parentw].w;
+        }
+        if(inherith>0)
+        {
+            int parenth = curlist;
+            while(inherith>0 && lists[parenth].parent>=0)
+            {
+                parenth = lists[parenth].parent;
+                inherith--;
+            }
+            h = lists[parenth].h;
+        }
+        glBegin(GL_QUADS);
+        rect_(curx, cury, w, h);
+        glEnd();
+        glEnable(GL_TEXTURE_2D);
+        defaultshader->set();
+    }
+
     void icon_(Texture *t, bool overlaid, bool tiled, int x, int y, int size, bool hit) 
     {
         float xs, ys, xt, yt;
@@ -496,12 +587,11 @@ struct gui : g3d_gui
         int gapx1 = INT_MAX, gapy1 = INT_MAX, gapx2 = INT_MAX, gapy2 = INT_MAX;
         float wscale = 1.0f/(SKIN_W*SKIN_SCALE), hscale = 1.0f/(SKIN_H*SKIN_SCALE);
         
-        loopj(2)
+        loopj(gui2d ? 1 : 2)
         {	
             bool quads = false;
-            glDepthFunc(j?GL_LEQUAL:GL_GREATER);
-            glColor4f(j?light.x:1.0f, j?light.y:1.0f, j?light.z:1.0f, j?0.80f:0.35f); //ghost when its behind something in depth
-           
+            if(!gui2d) glDepthFunc(j ? GL_LEQUAL : GL_GREATER);
+            glColor4f(j ? light.x : 1.0f, j ? light.y : 1.0f, j ? light.z : 1.0f, gui2d || j ? 0.80f : 0.35f); //ghost when its behind something in depth
             loopi(n)
             {
                 const patch &p = patches[start+i];
@@ -567,22 +657,37 @@ struct gui : g3d_gui
             if(quads) glEnd();
             else break; //if it didn't happen on the first pass, it won't happen on the second..
         }
-        glDepthFunc(GL_ALWAYS);
+        if(!gui2d) glDepthFunc(GL_ALWAYS);
     } 
 
-    vec origin;
+    vec origin, scale;
     float dist;
     g3d_callback *cb;
 
-    static float scale;
+    static float basescale;
     static bool passthrough;
     static vec light;
 
-    void start(int starttime, float basescale, int *tab, bool allowinput)
+    void adjustscale()
+    {
+        float aspect = float(screen->h)/float(screen->w), fit = 1.0f;
+        int w = xsize + (skinx[2]-skinx[1])*SKIN_SCALE + (skinx[10]-skinx[9])*SKIN_SCALE, h = ysize + (skiny[8]-skiny[6])*SKIN_SCALE;
+        if(tcurrent) h += ((skiny[5]-skiny[1])-(skiny[3]-skiny[2]))*SKIN_SCALE + FONTH-2*INSERT;
+        else h += (skiny[5]-skiny[3])*SKIN_SCALE;
+
+        if(w*aspect*basescale>1) fit = 1.0f/(w*aspect*basescale);
+        if(h*basescale*fit>1) fit *= 1.0f/(h*basescale*fit);
+        origin = vec(0.5f-((w-xsize)/2 - (skinx[2]-skinx[1])*SKIN_SCALE)*aspect*scale.x*fit, 0.5f + (0.5f*h-(skiny[8]-skiny[6])*SKIN_SCALE)*scale.y*fit, 0);
+        scale = vec(aspect*scale.x*fit, scale.y*fit, 1);
+    }
+
+    void start(int starttime, float initscale, int *tab, bool allowinput)
     {	
-        scale = basescale*min((totalmillis-starttime)/300.0f, 1.0f);
-        
-        passthrough = scale<basescale || !allowinput;
+        if(gui2d) initscale *= 0.025f; 
+        basescale = initscale;
+        if(layoutpass) scale.x = scale.y = scale.z = basescale*min((totalmillis-starttime)/300.0f, 1.0f);
+        passthrough = scale.x<basescale || !allowinput;
+        if(allowinput) hascursor = true;
         curdepth = -1;
         curlist = -1;
         tpos = 0;
@@ -598,18 +703,27 @@ struct gui : g3d_gui
             cury = -ysize; 
             curx = -xsize/2;
             
-            float yaw = atan2f(origin.y-camera1->o.y, origin.x-camera1->o.x);
             glPushMatrix();
-            glTranslatef(origin.x, origin.y, origin.z);
-            glRotatef(yaw/RAD-90, 0, 0, 1); 
-            glRotatef(-90, 1, 0, 0);
-            glScalef(-scale, scale, scale);
+            if(gui2d)
+            {
+                glTranslatef(origin.x, origin.y, origin.z);
+                glScalef(scale.x, scale.y, scale.z);
+                light = vec(1, 1, 1);
+            }
+            else
+            {
+                float yaw = atan2f(origin.y-camera1->o.y, origin.x-camera1->o.x);
+                glTranslatef(origin.x, origin.y, origin.z);
+                glRotatef(yaw/RAD-90, 0, 0, 1); 
+                glRotatef(-90, 1, 0, 0);
+                glScalef(-scale.x, scale.y, scale.z);
             
-            vec dir;
-            lightreaching(origin, light, dir, 0, 0.5f); 
-            float intensity = vec(yaw, 0.0f).dot(dir);
-            light.mul(1.0f + max(intensity, 0));
-       
+                vec dir;
+                lightreaching(origin, light, dir, 0, 0.5f); 
+                float intensity = vec(yaw, 0.0f).dot(dir);
+                light.mul(1.0f + max(intensity, 0));
+            }
+
             skin_(curx-skinx[2]*SKIN_SCALE, cury-skiny[5]*SKIN_SCALE, xsize, ysize, 0, 9);
             if(!tcurrent) skin_(curx-skinx[5]*SKIN_SCALE, cury-skiny[5]*SKIN_SCALE, xsize, 0, 9, 1);
         }
@@ -623,13 +737,23 @@ struct gui : g3d_gui
             ysize = max(ty, ysize);
             ysize = max(ysize, (skiny[6]-skiny[5])*SKIN_SCALE);
             if(tcurrent) *tcurrent = max(1, min(*tcurrent, tpos));
+            if(gui2d) adjustscale();
             if(!windowhit && !passthrough)
             {
-                vec planenormal = vec(origin).sub(camera1->o).set(2, 0).normalize(), intersectionpoint;
-                int intersects = intersect_plane_line(camera1->o, worldpos, origin, planenormal, intersectionpoint);
-                vec intersectionvec = vec(intersectionpoint).sub(origin), xaxis(-planenormal.y, planenormal.x, 0);
-                hitx = xaxis.dot(intersectionvec)/scale;
-                hity = -intersectionvec.z/scale;
+                int intersects = INTERSECT_MIDDLE;
+                if(gui2d)
+                {
+                    hitx = (cursorx - origin.x)/scale.x;
+                    hity = (cursory - origin.y)/scale.y;
+                }
+                else
+                {
+                    vec planenormal = vec(origin).sub(camera1->o).set(2, 0).normalize(), intersectionpoint;
+                    intersects = intersect_plane_line(camera1->o, worldpos, origin, planenormal, intersectionpoint);
+                    vec intersectionvec = vec(intersectionpoint).sub(origin), xaxis(-planenormal.y, planenormal.x, 0);
+                    hitx = xaxis.dot(intersectionvec)/scale.x;
+                    hity = -intersectionvec.z/scale.y;
+                }
                 if(intersects>=INTERSECT_MIDDLE && hitx>=-xsize/2 && hitx<=xsize/2 && hity<=0)
                 {
                     if(hity>=-ysize || (tcurrent && hity>=-ysize-(FONTH-2*INSERT)-((skiny[5]-skiny[1])-(skiny[3]-skiny[2]))*SKIN_SCALE && hitx<=tx-xsize/2))
@@ -649,8 +773,8 @@ struct gui : g3d_gui
 Texture *gui::skintex = NULL, *gui::overlaytex = NULL, *gui::slidertex = NULL;
 
 //chop skin into a grid
-const int gui::skiny[] = {0, 7, 21, 34, 48, 56, 104, 111, 128},
-          gui::skinx[] = {0, 11, 23, 37, 105, 119, 137, 151, 215, 229, 256};
+const int gui::skiny[] = {0, 7, 21, 34, 48, 56, 104, 111, 116, 128},
+          gui::skinx[] = {0, 11, 23, 37, 105, 119, 137, 151, 215, 229, 245, 256};
 //Note: skinx[3]-skinx[2] = skinx[7]-skinx[6]
 //      skinx[5]-skinx[4] = skinx[9]-skinx[8]		 
 const gui::patch gui::patches[] = 
@@ -691,7 +815,7 @@ const gui::patch gui::patches[] =
 };
 
 vector<gui::list> gui::lists;
-float gui::scale, gui::hitx, gui::hity;
+float gui::basescale, gui::hitx, gui::hity;
 bool gui::passthrough;
 vec gui::light;
 int gui::curdepth, gui::curlist, gui::xsize, gui::ysize, gui::curx, gui::cury;
@@ -700,7 +824,7 @@ static vector<gui> guis;
 
 void g3d_addgui(g3d_callback *cb, vec &origin, bool follow)
 {
-    if(follow) origin.z = camera1->o.z-(player->eyeheight-1);
+    if(follow) origin.z = player->o.z-(player->eyeheight-1);
     gui &g = guis.add();
     g.cb = cb;
     g.origin = origin;
@@ -714,17 +838,16 @@ bool g3d_windowhit(bool on, bool act)
     extern int cleargui(int n);
     if(act) mousebuttons |= (actionon=on) ? G3D_DOWN : G3D_UP;
     else if(!on && windowhit) cleargui(1);
-    return windowhit!=NULL;
+    return gui2d ? hascursor : windowhit!=NULL;
+}
+
+char *g3d_fieldname()
+{
+    return fieldname;
 }
 
 void g3d_render()   
 {
-    glMatrixMode(GL_MODELVIEW);
-    glDepthFunc(GL_ALWAYS);
-    glDepthMask(GL_FALSE);
-    glEnable(GL_BLEND);
-    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
-    
     windowhit = NULL;
     if(actionon) mousebuttons |= G3D_PRESSED;
     gui::reset();
@@ -741,18 +864,58 @@ void g3d_render()
     
     bool fieldfocus = (fieldpos>=0);
     fieldactive = false;
-    
+    hascursor = false;
+
+    if(guis.length())
+    {
+        if(gui2d)
+        {
+            glMatrixMode(GL_PROJECTION);
+            glPushMatrix();
+            glLoadIdentity();
+            glOrtho(0, 1, 1, 0, -1, 1);
+
+            glMatrixMode(GL_MODELVIEW);
+            glPushMatrix();
+            glLoadIdentity();
+
+            glDisable(GL_DEPTH_TEST);
+        }
+        else
+        {
+            glDepthFunc(GL_ALWAYS);
+            glDepthMask(GL_FALSE);
+        }
+        glEnable(GL_BLEND);
+        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+    }
+
     layoutpass = true;
     loopv(guis) guis[i].cb->gui(guis[i], true);
     layoutpass = false;
     loopvrev(guis) guis[i].cb->gui(guis[i], false);
+
+    if(guis.length())
+    {
+        glDisable(GL_BLEND);
+        if(gui2d)
+        {
+            glEnable(GL_DEPTH_TEST);
     
+            glMatrixMode(GL_PROJECTION);
+            glPopMatrix();
+            glMatrixMode(GL_MODELVIEW);
+            glPopMatrix();
+        }
+        else
+        { 
+            glDepthFunc(GL_LESS);
+            glDepthMask(GL_TRUE);
+        }
+    }
+
     if(!fieldactive) fieldpos = -1; //no hit fields, so loose focus - mainly for menu closed
     if((fieldpos>=0) != fieldfocus) SDL_EnableUNICODE(fieldpos>=0);
     
     mousebuttons = 0;
-	
-    glDisable(GL_BLEND);
-    glDepthFunc(GL_LESS);
-    glDepthMask(GL_TRUE);
 }
