@@ -7,7 +7,7 @@
 void itoa(char *s, int i) { s_sprintf(s)("%d", i); }
 char *exchangestr(char *o, const char *n) { delete[] o; return newstring(n); }
 
-typedef hashtable<char *, ident> identtable;
+typedef hashtable<const char *, ident> identtable;
 
 identtable *idents = NULL;        // contains ALL vars/commands/aliases
 
@@ -53,6 +53,7 @@ void clearoverrides()
 
 void pushident(ident &id, char *val)
 {
+    if(id._type != ID_ALIAS) return;
     identstack *stack = new identstack;
     stack->action = id._isexecuting==id._action ? newstring(id._action) : id._action;
     stack->next = id._stack;
@@ -62,7 +63,7 @@ void pushident(ident &id, char *val)
 
 void popident(ident &id)
 {
-    if(!id._stack) return;
+    if(id._type != ID_ALIAS || !id._stack) return;
     if(id._action != id._isexecuting) delete[] id._action;
     identstack *stack = id._stack;
     id._action = stack->action;
@@ -96,15 +97,14 @@ void pop(char *name)
 COMMAND(push, "ss");
 COMMAND(pop, "s");
 
-void aliasa(char *name, char *action)
+void aliasa(const char *name, char *action)
 {
     ident *b = idents->access(name);
     if(!b) 
     {
-        name = newstring(name);
-        ident b(ID_ALIAS, name, action, persistidents);
+        ident b(ID_ALIAS, newstring(name), action, persistidents);
         if(overrideidents) b._override = OVERRIDDEN;
-        idents->access(name, &b);
+        idents->access(b._name, &b);
     }
     else if(b->_type != ID_ALIAS)
     {
@@ -124,13 +124,13 @@ void aliasa(char *name, char *action)
     }
 }
 
-void alias(char *name, char *action) { aliasa(name, newstring(action)); }
+void alias(const char *name, const char *action) { aliasa(name, newstring(action)); }
 
 COMMAND(alias, "ss");
 
 // variable's and commands are registered through globals, see cube.h
 
-int variable(char *name, int min, int cur, int max, int *storage, void (*fun)(), bool persist)
+int variable(const char *name, int min, int cur, int max, int *storage, void (*fun)(), bool persist)
 {
     if(!idents) idents = new identtable;
     ident v(ID_VAR, name, min, cur, max, storage, (void *)fun, persist);
@@ -141,37 +141,37 @@ int variable(char *name, int min, int cur, int max, int *storage, void (*fun)(),
 #define GETVAR(id, name, retval) \
     ident *id = idents->access(name); \
     if(!id || id->_type!=ID_VAR) return retval;
-void setvar(char *name, int i, bool dofunc) 
+void setvar(const char *name, int i, bool dofunc) 
 { 
     GETVAR(id, name, );
     *id->_storage = i; 
     if(dofunc) id->changed();
 } 
-int getvar(char *name) 
+int getvar(const char *name) 
 { 
     GETVAR(id, name, 0);
     return *id->_storage;
 }
-int getvarmin(char *name) 
+int getvarmin(const char *name) 
 { 
     GETVAR(id, name, 0);
     return id->_min;
 }
-int getvarmax(char *name) 
+int getvarmax(const char *name) 
 { 
     GETVAR(id, name, 0);
     return id->_max;
 }
-bool identexists(char *name) { return idents->access(name)!=NULL; }
-ident *getident(char *name) { return idents->access(name); }
+bool identexists(const char *name) { return idents->access(name)!=NULL; }
+ident *getident(const char *name) { return idents->access(name); }
 
-const char *getalias(char *name)
+const char *getalias(const char *name)
 {
     ident *i = idents->access(name);
     return i && i->_type==ID_ALIAS ? i->_action : "";
 }
 
-bool addcommand(char *name, void (*fun)(), char *narg)
+bool addcommand(const char *name, void (*fun)(), const char *narg)
 {
     if(!idents) idents = new identtable;
     ident c(ID_COMMAND, name, narg, (void *)fun);
@@ -179,7 +179,7 @@ bool addcommand(char *name, void (*fun)(), char *narg)
     return false;
 }
 
-void addident(char *name, ident *id)
+void addident(const char *name, ident *id)
 {
     if(!idents) idents = new identtable;
     idents->access(name, id);
@@ -188,9 +188,9 @@ void addident(char *name, ident *id)
 static vector<vector<char> *> wordbufs;
 static int bufnest = 0;
 
-char *parseexp(char *&p, int right);
+char *parseexp(const char *&p, int right);
 
-void parsemacro(char *&p, int level, vector<char> &wordbuf)
+void parsemacro(const char *&p, int level, vector<char> &wordbuf)
 {
     int escape = 1;
     while(*p=='@') p++, escape++;
@@ -209,16 +209,15 @@ void parsemacro(char *&p, int level, vector<char> &wordbuf)
         }
         return;
     }
-    char *ident = p;
-    while(isalnum(*p) || *p=='_') p++;
-    int c = *p;
-    *p = 0;
-    const char *alias = getalias(ident);
-    *p = c;
+    static vector<char> ident;
+    ident.setsizenodelete(0);
+    while(isalnum(*p) || *p=='_') ident.add(*p++);
+    ident.add(0);
+    const char *alias = getalias(ident.getbuf());
     while(*alias) wordbuf.add(*alias++);
 }
 
-char *parseexp(char *&p, int right)          // parse any nested set of () or []
+char *parseexp(const char *&p, int right)          // parse any nested set of () or []
 {
     if(bufnest++>=wordbufs.length()) wordbufs.add(new vector<char>);
     vector<char> &wordbuf = *wordbufs[bufnest-1];
@@ -235,7 +234,7 @@ char *parseexp(char *&p, int right)          // parse any nested set of () or []
         if(c=='\"')
         {
             wordbuf.add(c);
-            char *end = p+strcspn(p, "\"\r\n\0");
+            const char *end = p+strcspn(p, "\"\r\n\0");
             while(p < end) wordbuf.add(*p++);
             if(*p=='\"') wordbuf.add(*p++);
             continue;
@@ -288,7 +287,7 @@ char *lookup(char *n)                           // find value of ident reference
     return n;
 }
 
-char *parseword(char *&p)                       // parse single argument, including expressions
+char *parseword(const char *&p)                       // parse single argument, including expressions
 {
     for(;;)
     {
@@ -299,7 +298,7 @@ char *parseword(char *&p)                       // parse single argument, includ
     if(*p=='\"')
     {
         p++;
-        char *word = p;
+        const char *word = p;
         p += strcspn(p, "\"\r\n\0");
         char *s = newstring(word, p-word);
         if(*p=='\"') p++;
@@ -307,7 +306,7 @@ char *parseword(char *&p)                       // parse single argument, includ
     }
     if(*p=='(') return parseexp(p, ')');
     if(*p=='[') return parseexp(p, ']');
-    char *word = p;
+    const char *word = p;
     for(;;)
     {
         p += strcspn(p, "/; \t\r\n\0");
@@ -343,7 +342,7 @@ char *commandret = NULL;
 
 extern const char *addreleaseaction(const char *s);
 
-char *executeret(char *p)               // all evaluation happens here, recursively
+char *executeret(const char *p)               // all evaluation happens here, recursively
 {
     const int MAXWORDS = 25;                    // limit, remove
     char *w[MAXWORDS];
@@ -354,7 +353,7 @@ char *executeret(char *p)               // all evaluation happens here, recursiv
         int numargs = MAXWORDS;
         loopi(MAXWORDS)                         // collect all argument values
         {
-            w[i] = "";
+            w[i] = (char *)"";
             if(i>numargs) continue;
             char *s = parseword(p);             // parse and evaluate exps
             if(s) w[i] = s;
@@ -385,26 +384,10 @@ char *executeret(char *p)               // all evaluation happens here, recursiv
             }
             else switch(id->_type)
             {
-                case ID_ICOMMAND:
-                {
-                    switch(id->_narg[0])
-                    {
-                        default: id->run(w+1); break;
-                        case 'D': id->run((char **)addreleaseaction(id->_name)); break;
-                        case 'C': 
-                        { 
-                            char *r = conc(w+1, numargs-1, true); 
-                            id->run(&r); 
-                            delete[] r; 
-                            break;
-                        }
-                    }
-                    setretval(commandret);
-                    break;
-                }
-
+                case ID_CCOMMAND:
                 case ID_COMMAND:                     // game defined commands
-                {    
+                {   
+                    SAUERBRATEN_COMMAND_ENTRY(c, w[1], w[2], w[3]);
                     void *v[MAXWORDS];
                     union
                     {
@@ -412,14 +395,16 @@ char *executeret(char *p)               // all evaluation happens here, recursiv
                         float f;
                     } nstor[MAXWORDS];
                     int n = 0, wn = 0;
-                    for(char *a = id->_narg; *a; a++) switch(*a)
+                    char *cargs = NULL;
+                    if(id->_type==ID_CCOMMAND) v[n++] = id->self;
+                    for(const char *a = id->_narg; *a; a++) switch(*a)
                     {
                         case 's':                                 v[n] = w[++wn];     n++; break;
                         case 'i': nstor[n].i = parseint(w[++wn]); v[n] = &nstor[n].i; n++; break;
                         case 'f': nstor[n].f = atof(w[++wn]);     v[n] = &nstor[n].f; n++; break;
                         case 'D': nstor[n].i = addreleaseaction(id->_name) ? 1 : 0; v[n] = &nstor[n].i; n++; break;
                         case 'V': v[n++] = w+1; nstor[n].i = numargs-1; v[n] = &nstor[n].i; n++; break;
-                        case 'C': v[n++] = conc(w+1, numargs-1, true);  break;
+                        case 'C': if(!cargs) cargs = conc(w+1, numargs-1, true); v[n++] = cargs; break;
                         default: fatal("builtin declared with illegal type");
                     }
                     switch(n)
@@ -431,15 +416,18 @@ char *executeret(char *p)               // all evaluation happens here, recursiv
                         case 4: ((void (__cdecl *)(void *, void *, void *, void *)        )id->_fun)(v[0], v[1], v[2], v[3]);       break;
                         case 5: ((void (__cdecl *)(void *, void *, void *, void *, void *))id->_fun)(v[0], v[1], v[2], v[3], v[4]); break;
                         case 6: ((void (__cdecl *)(void *, void *, void *, void *, void *, void *))id->_fun)(v[0], v[1], v[2], v[3], v[4], v[5]); break;
-
+                        case 7: ((void (__cdecl *)(void *, void *, void *, void *, void *, void *, void *))id->_fun)(v[0], v[1], v[2], v[3], v[4], v[5], v[6]); break;
+                        case 8: ((void (__cdecl *)(void *, void *, void *, void *, void *, void *, void *, void *))id->_fun)(v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7]); break;
                         default: fatal("builtin declared with too many args (use V?)");
                     }
-                    if(id->_narg[0]=='C') delete[] (char *)v[0];
+                    if(cargs) delete[] cargs;
                     setretval(commandret);
+                    SAUERBRATEN_COMMAND_RETURN(c);
                     break;
                 }
 
                 case ID_VAR:                        // game defined variables 
+                    SAUERBRATEN_VAR_ENTRY(c, w[1]);
                     if(!w[1][0]) conoutf("%s = %d", c, *id->_storage);      // var with no value just prints its current value
                     else if(id->_min>id->_max) conoutf("variable %s is read-only", id->_name);
                     else
@@ -463,14 +451,27 @@ char *executeret(char *p)               // all evaluation happens here, recursiv
                         *id->_storage = i1;
                         id->changed();                                             // call trigger function if available
                     }
+                    SAUERBRATEN_VAR_RETURN(c, *id->_storage);
                     break;
                     
                 case ID_ALIAS:                              // alias, also used as functions and (global) variables
                 {
+                    SAUERBRATEN_ALIAS_ENTRY(c, w[1], w[2], w[3]);
+                    static vector<ident *> argids;
                     for(int i = 1; i<numargs; i++)
                     {
-                        s_sprintfd(t)("arg%d", i);          // set any arguments as (global) arg values so functions can access them
-                        pusha(t, w[i]);
+                        if(i > argids.length())
+                        {
+                            s_sprintfd(argname)("arg%d", i);
+                            ident *id = idents->access(argname);
+                            if(!id)
+                            {
+                                ident init(ID_ALIAS, newstring(argname), newstring(""), persistidents);
+                                id = idents->access(init._name, &init);
+                            }
+                            argids.add(id);
+                        }
+                        pushident(*argids[i-1], w[i]); // set any arguments as (global) arg values so functions can access them
                         w[i] = NULL;
                     }
                     _numargs = numargs-1;
@@ -482,11 +483,8 @@ char *executeret(char *p)               // all evaluation happens here, recursiv
                     if(id->_isexecuting != id->_action && id->_isexecuting != wasexecuting) delete[] id->_isexecuting;
                     id->_isexecuting = wasexecuting;
                     overrideidents = wasoverriding;
-                    for(int i = 1; i<numargs; i++)
-                    {
-                        s_sprintfd(t)("arg%d", i);          // set any arguments as (global) arg values so functions can access them
-                        pop(t);
-                    }
+                    for(int i = 1; i<numargs; i++) popident(*argids[i-1]);
+                    SAUERBRATEN_ALIAS_RETURN(c);
                     break;
                 }
             }
@@ -496,7 +494,7 @@ char *executeret(char *p)               // all evaluation happens here, recursiv
     return retval;
 }
 
-int execute(char *p)
+int execute(const char *p)
 {
     char *ret = executeret(p);
     int i = 0; 
@@ -504,7 +502,7 @@ int execute(char *p)
     return i;
 }
 
-bool execfile(char *cfgfile)
+bool execfile(const char *cfgfile)
 {
     string s;
     s_strcpy(s, cfgfile);
@@ -515,18 +513,19 @@ bool execfile(char *cfgfile)
     return true;
 }
 
-void exec(char *cfgfile)
+void exec(const char *cfgfile)
 {
     if(!execfile(cfgfile)) conoutf("could not read \"%s\"", cfgfile);
 }
 
 void writecfg()
 {
-    FILE *f = fopen("config.cfg", "w");
+    FILE *f = openfile(path(cl->savedconfig(), true), "w");
     if(!f) return;
-    fprintf(f, "// automatically written on exit, DO NOT MODIFY\n// delete this file to have defaults.cfg overwrite these settings\n// modify settings in game, or put settings in autoexec.cfg to override anything\n\n");
+    fprintf(f, "// automatically written on exit, DO NOT MODIFY\n// delete this file to have %s overwrite these settings\n// modify settings in game, or put settings in %s to override anything\n\n", cl->defaultconfig(), cl->autoexec());
     cc->writeclientinfo(f);
     fprintf(f, "\n");
+    writecrosshairs(f);
     enumerate(*idents, ident, id,
         if(id._type==ID_VAR && id._persist)
         {
@@ -555,10 +554,10 @@ COMMAND(writecfg, "");
 void intset(char *name, int v) { string b; itoa(b, v); alias(name, b); }
 void intret            (int v) { string b; itoa(b, v); commandret = newstring(b); }
 
-ICOMMAND(if, "sss", commandret = executeret(args[0][0]!='0' ? args[1] : args[2]));
+ICOMMAND(if, "sss", (char *cond, char *t, char *f), commandret = executeret(cond[0]!='0' ? t : f));
 
-ICOMMAND(loop, "sss", { int n = parseint(args[1]); loopi(n) { intset(args[0], i); execute(args[2]); } });
-ICOMMAND(while, "ss", while(execute(args[0])) execute(args[1]));    // can't get any simpler than this :)
+ICOMMAND(loop, "sis", (char *var, int *n, char *body), loopi(*n) { intset(var, i); execute(body); });
+ICOMMAND(while, "ss", (char *cond, char *body), while(execute(cond)) execute(body));    // can't get any simpler than this :)
 
 void concat(const char *s) { commandret = newstring(s); }
 void result(const char *s) { commandret = newstring(s); }
@@ -595,6 +594,18 @@ void format(char **args, int *numargs)
 #define whitespaceskip s += strspn(s, "\n\t ")
 #define elementskip *s=='"' ? (++s, s += strcspn(s, "\"\n\0"), s += *s=='"') : s += strcspn(s, "\n\t \0")
 
+void explodelist(char *s, vector<char *> &elems)
+{
+    whitespaceskip;
+    while(*s)
+    {
+        char *elem = s;
+        elementskip;
+        elems.add(*elem=='"' ? newstring(elem+1, s-elem-(s[-1]=='"' ? 2 : 1)) : newstring(elem, s-elem));
+        whitespaceskip;
+    }
+}
+
 void listlen(char *s)
 {
     int n = 0;
@@ -612,7 +623,7 @@ void at(char *s, int *pos)
     if(*e=='"') 
     {
         e++;
-        if(s[-1] == '"') --s;
+        if(s[-1]=='"') --s;
     }
     *s = '\0';
     result(e);
@@ -645,6 +656,8 @@ void lte   (int *a, int *b) { intret((int)(*a <= *b)); } COMMANDN(<=, lte, "ii")
 void gte   (int *a, int *b) { intret((int)(*a >= *b)); } COMMANDN(>=, gte, "ii");
 void xora (int *a, int *b) { intret(*a ^ *b); }          COMMANDN(^, xora, "ii");
 void nota (int *a)         { intret(*a == 0); }          COMMANDN(!, nota, "i");
+void mina (int *a, int *b) { intret(min(*a, *b)); }      COMMANDN(min, mina, "ii");
+void maxa (int *a, int *b) { intret(max(*a, *b)); }      COMMANDN(max, maxa, "ii");
 
 void anda (char *a, char *b) { intret(execute(a)!=0 && execute(b)!=0); }
 void ora  (char *a, char *b) { intret(execute(a)!=0 || execute(b)!=0); }
@@ -656,7 +669,7 @@ void rndn(int *a)          { intret(*a>0 ? rnd(*a) : 0); }  COMMANDN(rnd, rndn,
 
 void strcmpa(char *a, char *b) { intret(strcmp(a,b)==0); }  COMMANDN(strcmp, strcmpa, "ss");
 
-ICOMMAND(echo, "C", conoutf("\f1%s", args[0]));
+ICOMMAND(echo, "C", (char *s), conoutf("\f1%s", s));
 
 void strstra(char *a, char *b) { char *s = strstr(a, b); intret(s ? s-a : -1); } COMMANDN(strstr, strstra, "ss");
 
@@ -664,6 +677,7 @@ struct sleepcmd
 {
     int millis;
     char *command;
+    bool override;
 };
 vector<sleepcmd> sleepcmds;
 
@@ -672,6 +686,7 @@ void addsleep(int *msec, char *cmd)
     sleepcmd &s = sleepcmds.add();
     s.millis = *msec+lastmillis;
     s.command = newstring(cmd);
+    s.override = overrideidents;
 }
 
 COMMANDN(sleep, addsleep, "is");
@@ -691,11 +706,21 @@ void checksleep(int millis)
     }
 }
 
-void clearsleep()
+void clearsleep(bool clearoverrides)
 {
-    loopv(sleepcmds) delete[] sleepcmds[i].command;
-    sleepcmds.setsize(0);
+    int len = 0;
+    loopv(sleepcmds) 
+    {
+        if(clearoverrides && !sleepcmds[i].override) sleepcmds[len++] = sleepcmds[i];
+        else delete[] sleepcmds[i].command;
+    }
+    sleepcmds.setsize(len);
 }
 
-COMMAND(clearsleep, "");
+void clearsleep_(int *clearoverrides)
+{
+    clearsleep(*clearoverrides!=0 || overrideidents);
+}
+
+COMMANDN(clearsleep, clearsleep_, "i");
 
