@@ -2,15 +2,10 @@
 
 #include "pch.h"
 #include "engine.h"
-#ifndef WIN32
-#include <dirent.h>
-#endif
 
 struct cline { char *cref; int outtime; };
 vector<cline> conlines;
 
-const int ndraw = 5;
-const int WORDWRAP = 80;
 int conskip = 0;
 
 bool saycommandon = false;
@@ -25,42 +20,62 @@ void setconskip(int *n)
 
 COMMANDN(conskip, setconskip, "i");
 
-void conline(const char *sf, bool highlight)        // add a line to the console buffer
+void conline(const char *sf)        // add a line to the console buffer
 {
     cline cl;
     cl.cref = conlines.length()>100 ? conlines.pop().cref : newstringbuf("");   // constrain the buffer size
     cl.outtime = totalmillis;                       // for how long to keep line on screen
-    conlines.insert(0,cl);
-    if(highlight)                                   // show line in a different colour, for chat etc.
-    {
-        cl.cref[0] = '\f';
-        cl.cref[1] = '0';
-        cl.cref[2] = 0;
-        s_strcat(cl.cref, sf);
-    }
-    else
-    {
-        s_strcpy(cl.cref, sf);
-    }
+    conlines.insert(0, cl);
+    s_strcpy(cl.cref, sf);
 }
 
-const int CONSPAD = FONTH/3;
+#define CONSPAD (FONTH/3)
 
 void conoutf(const char *s, ...)
 {
-    int w = scr_w, h = scr_h;
+    extern int scr_w, scr_h;
+    int w = screen ? screen->w : scr_w, h = screen ? screen->h : scr_h;
     gettextres(w, h);
     s_sprintfdv(sf, s);
-    puts(sf);
+    string sp;
+    filtertext(sp, sf);
+    puts(sp);
     s = sf;
-    int n = 0, visible;
-    while((visible = text_visible(s, 3*w - 2*CONSPAD - 2*FONTH/3))) // cut strings to fit on screen
+    int visible;
+
+    string cols;
+    s_strcpy(cols, "\f7");
+    int cpos = 1;
+     
+    while((visible = curfont ? text_visible(s, 3*w - 2*CONSPAD - 2*FONTH/3) : strlen(s))) // cut strings to fit on screen
     {
         const char *newline = (const char *)memchr(s, '\n', visible);
         if(newline) visible = newline+1-s;
-        string t;
-        s_strncpy(t, s, visible+1);
-        conline(t, n++!=0);
+        
+        s_strncpy(cols+cpos+1, s, visible+1);
+        conline(cols);
+    
+        for(int i = 0; s[i] && (i<=visible); i++) //process color change info
+            if(s[i]=='\f') 
+                switch(s[++i])
+                {
+                    case 's':
+                        if(cpos<4*8) //8 = stackdepth in textdraw
+                        { 
+                            cols[cpos+1] = '\f';
+                            cols[cpos+2] = 's';
+                            cols[cpos+3] = '\f';
+                            cols[cpos+4] = cols[cpos];
+                            cpos += 4;
+                        }
+                        break;
+                    case 'r': 
+                        if(cpos>4) cpos -= 4; 
+                        break;
+                    default:
+                        cols[cpos] = s[i];
+                }
+
         s += visible;
     }
 }
@@ -109,6 +124,8 @@ void blendbox(int x1, int y1, int x2, int y2, bool border)
     defaultshader->set();
 }
 
+VARP(consize, 0, 5, 100);
+
 int renderconsole(int w, int h)                   // render buffer taking into account time & scrolling
 {
     if(fullconsole)
@@ -119,20 +136,20 @@ int renderconsole(int w, int h)                   // render buffer taking into a
         loopi(numl) draw_text(offset+i>=conlines.length() ? "" : conlines[offset+i].cref, CONSPAD+FONTH/3, CONSPAD+FONTH*(numl-i-1)+FONTH/3); 
         return 2*CONSPAD+numl*FONTH+2*FONTH/3;
     }
-    else     
+    else
     {
-        int nd = 0;
-        char *refs[ndraw];
-        loopv(conlines) if(conskip ? i>=conskip-1 || i>=conlines.length()-ndraw : totalmillis-conlines[i].outtime<20000)
+        static vector<char *> refs;
+        refs.setsizenodelete(0);
+        if(consize) loopv(conlines) if(conskip ? i>=conskip-1 || i>=conlines.length()-consize : totalmillis-conlines[i].outtime<20000)
         {
-            refs[nd++] = conlines[i].cref;
-            if(nd==ndraw) break;
+            refs.add(conlines[i].cref);
+            if(refs.length()>=consize) break;
         }
-        loopj(nd)
+        loopvj(refs)
         {
-            draw_text(refs[j], CONSPAD+FONTH/3, CONSPAD+FONTH*(nd-j-1)+FONTH/3);
+            draw_text(refs[j], CONSPAD+FONTH/3, CONSPAD+FONTH*(refs.length()-j-1)+FONTH/3);
         }
-        return CONSPAD+nd*FONTH+2*FONTH/3;
+        return CONSPAD+refs.length()*FONTH+2*FONTH/3;
     }
 }
 
@@ -150,31 +167,45 @@ vector<keym> keyms;
 
 void keymap(char *code, char *key)
 {
-    if(overrideidents) conoutf("cannot override keymap %s", code);
+    if(overrideidents) { conoutf("cannot override keymap %s", code); return; }
     keym &km = keyms.add();
     km.code = atoi(code);
     km.name = newstring(key);
     km.action = newstring("");
     km.editaction = newstring("");
 }
-
+    
 COMMAND(keymap, "ss");
 
 keym *keypressed = NULL;
 char *keyaction = NULL;
 
+keym *findbind(char *key)
+{
+    loopv(keyms) if(!strcasecmp(keyms[i].name, key)) return &keyms[i];
+    return NULL;
+}   
+    
+void getbind(char *key)
+{
+    keym *km = findbind(key);
+    result(km ? km->action : "");
+}   
+ 
+void geteditbind(char *key)
+{
+    keym *km = findbind(key);
+    result(km ? km->editaction : "");
+}  
+
 void bindkey(char *key, char *action, bool edit)
 {
-    if(overrideidents) conoutf("cannot override %s \"%s\"", edit ? "editbind" : "bind", key);
-    for(char *x = key; *x; x++) *x = toupper(*x);
-    loopv(keyms) if(strcmp(keyms[i].name, key)==0)
-    {
-        char *&binding = edit ? keyms[i].editaction : keyms[i].action;
-        if(!keypressed || keyaction!=binding) delete[] binding;
-        binding = newstring(action);
-        return;
-    }
-    conoutf("unknown key \"%s\"", key);   
+    if(overrideidents) { conoutf("cannot override %s \"%s\"", edit ? "editbind" : "bind", key); return; }
+    keym *km = findbind(key);
+    if(!km) { conoutf("unknown key \"%s\"", key); return; }
+    char *&binding = edit ? km->editaction : km->action;
+    if(!keypressed || keyaction!=binding) delete[] binding;
+    binding = newstring(action);
 }
 
 void bindnorm(char *key, char *action) { bindkey(key, action, false); }
@@ -182,13 +213,14 @@ void bindedit(char *key, char *action) { bindkey(key, action, true);  }
 
 COMMANDN(bind,     bindnorm, "ss");
 COMMANDN(editbind, bindedit, "ss");
+COMMAND(getbind, "s");
+COMMAND(geteditbind, "s");
 
 void saycommand(char *init)                         // turns input to the command line on or off
 {
     SDL_EnableUNICODE(saycommandon = (init!=NULL));
     if(!editmode) keyrepeat(saycommandon);
-    if(!init) init = "";
-    s_strcpy(commandbuf, init);
+    s_strcpy(commandbuf, init ? init : "");
     commandpos = -1;
     player->stopmoving(); // prevent situations where player presses direction key, open command line, then releases key
 }
@@ -301,6 +333,14 @@ void keypress(int code, bool isdown, int cooked)
                 case SDLK_KP_ENTER:
                     break;
 
+                case SDLK_HOME: 
+                    if(strlen(commandbuf)) commandpos = 0; 
+                    break;
+
+                case SDLK_END: 
+                    commandpos = -1; 
+                    break;
+
                 case SDLK_DELETE:
                 {
                     int len = (int)strlen(commandbuf);
@@ -437,26 +477,30 @@ void writebinds(FILE *f)
 
 // tab-completion of all idents and base maps
 
+enum { FILES_DIR = 0, FILES_LIST };
+
 struct fileskey
 {
+    int type;
     const char *dir, *ext;
 
     fileskey() {}
-    fileskey(const char *dir, const char *ext) : dir(dir), ext(ext) {}
+    fileskey(int type, const char *dir, const char *ext) : type(type), dir(dir), ext(ext) {}
 };
 
 struct filesval
 {
+    int type;
     char *dir, *ext;
     vector<char *> files;
-
-    filesval(const char *dir, const char *ext) : dir(newstring(dir)), ext(ext[0] ? newstring(ext) : NULL) {}
+    
+    filesval(int type, const char *dir, const char *ext) : type(type), dir(newstring(dir)), ext(ext && ext[0] ? newstring(ext) : NULL) {}
     ~filesval() { DELETEA(dir); DELETEA(ext); loopv(files) DELETEA(files[i]); files.setsize(0); }
 };
 
 static inline bool htcmp(const fileskey &x, const fileskey &y)
 {
-    return !strcmp(x.dir, y.dir) && (x.ext == y.ext || (x.ext && y.ext && !strcmp(x.ext, y.ext)));
+    return x.type==y.type && !strcmp(x.dir, y.dir) && (x.ext == y.ext || (x.ext && y.ext && !strcmp(x.ext, y.ext)));
 }
 
 static inline uint hthash(const fileskey &k)
@@ -472,7 +516,7 @@ string lastcomplete;
 
 void resetcomplete() { completesize = 0; }
 
-void addcomplete(char *command, char *dir, char *ext)
+void addcomplete(char *command, int type, char *dir, char *ext)
 {
     if(overrideidents)
     {
@@ -485,16 +529,24 @@ void addcomplete(char *command, char *dir, char *ext)
         if(hasfiles) *hasfiles = NULL;
         return;
     }
-    int dirlen = (int)strlen(dir);
-    while(dirlen > 0 && (dir[dirlen-1] == '/' || dir[dirlen-1] == '\\'))
-        dir[--dirlen] = '\0';
-    if(strchr(ext, '*')) ext[0] = '\0';
-    fileskey key(dir, ext[0] ? ext : NULL);
+    if(type==FILES_DIR)
+    {
+        int dirlen = (int)strlen(dir);
+        while(dirlen > 0 && (dir[dirlen-1] == '/' || dir[dirlen-1] == '\\'))
+            dir[--dirlen] = '\0';
+        if(ext)
+        {
+            if(strchr(ext, '*')) ext[0] = '\0';
+            if(!ext[0]) ext = NULL;
+        }
+    }
+    fileskey key(type, dir, ext);
     filesval **val = completefiles.access(key);
     if(!val)
     {
-        filesval *f = new filesval(dir, ext);
-        val = &completefiles[fileskey(f->dir, f->ext)];
+        filesval *f = new filesval(type, dir, ext);
+        if(type==FILES_LIST) explodelist(dir, f->files); 
+        val = &completefiles[fileskey(type, f->dir, f->ext)];
         *val = f;
     }
     filesval **hasfiles = completions.access(command);
@@ -502,46 +554,19 @@ void addcomplete(char *command, char *dir, char *ext)
     else completions[newstring(command)] = *val;
 }
 
-COMMANDN(complete, addcomplete, "sss");
-
-void buildfilenames(filesval *f)
+void addfilecomplete(char *command, char *dir, char *ext)
 {
-    int extsize = f->ext ? (int)strlen(f->ext)+1 : 0;
-    #if defined(WIN32)
-    s_sprintfd(pathname)("%s\\*.%s", f->dir, f->ext ? f->ext : "*");
-    WIN32_FIND_DATA FindFileData;
-    HANDLE Find = FindFirstFile(path(pathname), &FindFileData);
-    if(Find != INVALID_HANDLE_VALUE)
-    {
-        do {
-            f->files.add(newstring(FindFileData.cFileName, (int)strlen(FindFileData.cFileName) - extsize));
-        } while(FindNextFile(Find, &FindFileData));
-    }
-    #elif defined(__GNUC__)
-    string pathname;
-    s_strcpy(pathname, f->dir);
-    DIR *d = opendir(path(pathname));
-    if(d)
-    {
-        struct dirent *dir;
-        while((dir = readdir(d)) != NULL)
-        {
-            if(!f->ext) f->files.add(newstring(dir->d_name));
-            else
-            {
-                int namelength = strlen(dir->d_name) - extsize;
-                if(namelength > 0 && dir->d_name[namelength] == '.' && strncmp(dir->d_name+namelength+1, f->ext, extsize-1)==0)
-                    f->files.add(newstring(dir->d_name, namelength));
-            }
-        }
-        closedir(d);
-    }
-    #else
-    if(0)
-    #endif
-    else conoutf("unable to read base folder for map autocomplete");
+    addcomplete(command, FILES_DIR, dir, ext);
 }
 
+void addlistcomplete(char *command, char *list)
+{
+    addcomplete(command, FILES_LIST, list, NULL);
+}
+
+COMMANDN(complete, addfilecomplete, "sss");
+COMMANDN(listcomplete, addlistcomplete, "ss");
+
 void complete(char *s)
 {
     if(*s!='/')
@@ -567,14 +592,14 @@ void complete(char *s)
         }
     }
 
-    char *nextcomplete = NULL;
+    const char *nextcomplete = NULL;
     string prefix;
     s_strcpy(prefix, "/");
     if(f) // complete using filenames
     {
         int commandsize = strchr(s, ' ')+1-s;
         s_strncpy(prefix, s, min(size_t(commandsize+1), sizeof(prefix)));
-        if(f->files.empty()) buildfilenames(f);
+        if(f->type==FILES_DIR && f->files.empty()) listfiles(f->dir, f->ext, f->files);
         loopi(f->files.length())
         {
             if(strncmp(f->files[i], s+commandsize, completesize+1-commandsize)==0 &&
@@ -584,7 +609,7 @@ void complete(char *s)
     }
     else // complete using command names
     {
-        extern hashtable<char *, ident> *idents;
+        extern hashtable<const char *, ident> *idents;
         enumerate(*idents, ident, id,
             if(strncmp(id._name, s+1, completesize)==0 &&
                strcmp(id._name, lastcomplete) > 0 && (!nextcomplete || strcmp(id._name, nextcomplete) < 0))
@@ -603,7 +628,7 @@ void complete(char *s)
 void writecompletions(FILE *f)
 {
     enumeratekt(completions, char *, k, filesval *, v,
-        if(v) fprintf(f, "complete \"%s\" \"%s\" \"%s\"\n", k, v->dir, v->ext ? v->ext : "*");
+        if(v) fprintf(f, "%scomplete \"%s\" \"%s\" \"%s\"\n", v->type==FILES_LIST ? "list" : "", k, v->dir, v->type==FILES_LIST ? "" : (v->ext ? v->ext : "*"));
     );
 }
 
