@@ -10,17 +10,36 @@ VARF(lightlod, 0, 0, 10, hdr.mapllod = lightlod);
 VARF(worldlod, 0, 0, 1,  hdr.mapwlod = worldlod);
 VARF(ambient, 1, 25, 64, hdr.ambient = ambient);
 
+void skylight(char *r, char *g, char *b)
+{
+	if(!r[0])
+	{
+		s_sprintfd(s)("%d %d %d", hdr.skylight[0], hdr.skylight[1], hdr.skylight[2]);
+		result(s);
+	}
+	else
+	{
+		hdr.skylight[0] = strtol(r, NULL, 0);
+		hdr.skylight[1] = g[0] ? strtol(g, NULL, 0) : hdr.skylight[0];
+		hdr.skylight[2] = b[0] ? strtol(b, NULL, 0) : hdr.skylight[1];
+	}
+}
+
+COMMAND(skylight, "sss");
+
 // quality parameters, set by the calclight arg
 int shadows = 1;
 int mmshadows = 0;
 int aalights = 3;
 
 static int lmtype, lmorient;
-static uchar lm[3 * LM_MAXW * LM_MAXH];
-static vec lm_ray[LM_MAXW * LM_MAXH];
-static uint lm_w, lm_h;
-static vector<const entity *> lights1, lights2;
+static uchar lm[3*LM_MAXW*LM_MAXH];
+static vec lm_ray[LM_MAXW*LM_MAXH];
+static int lm_w, lm_h;
+static vector<const extentity *> lights1, lights2;
 static uint progress = 0;
+static GLuint progresstex = 0;
+static int progresstexticks = 0;
 
 bool calclight_canceled = false;
 volatile bool check_calclight_progress = false;
@@ -53,7 +72,20 @@ void show_calclight_progress()
     s_sprintfd(text1)("%d%%", int(bar1 * 100));
     s_sprintfd(text2)("%d textures %d%% utilized", lightmaps.length(), int(bar2 * 100));
 
-    show_out_of_renderloop_progress(bar1, text1, bar2, text2);
+    if(LM_PACKW <= hwtexsize && !progresstex) 
+    {
+        glGenTextures(1, &progresstex); 
+        createtexture(progresstex, LM_PACKW, LM_PACKH, NULL, 3, false, GL_RGB);
+    }
+    // only update once a sec (4 * 250 ms ticks) to not kill performance
+    if(progresstex && !calclight_canceled) loopvrev(lightmaps) if(lightmaps[i].type==LM_DIFFUSE || lightmaps[i].type==LM_BUMPMAP0)
+    {
+        if(progresstexticks++ % 4) break;
+        glBindTexture(GL_TEXTURE_2D, progresstex);
+        glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, LM_PACKW, LM_PACKH, GL_RGB, GL_UNSIGNED_BYTE, lightmaps[i].data);
+        break;
+    }
+    show_out_of_renderloop_progress(bar1, text1, bar2, text2, progresstexticks ? progresstex : 0);
 }
 
 #define CHECK_PROGRESS(exit) CHECK_CALCLIGHT_PROGRESS(exit, show_calclight_progress)
@@ -234,26 +266,35 @@ void pack_lightmap(int type, surfaceinfo &surface)
     else insert_lightmap(type, surface.x, surface.y, surface.lmid);
 }
 
-void generate_lumel(const float tolerance, const vector<const entity *> &lights, const vec &target, const vec &normal, vec &sample, int x, int y)
+void generate_lumel(const float tolerance, const vector<const extentity *> &lights, const vec &target, const vec &normal, vec &sample, int x, int y)
 {
     vec avgray(0, 0, 0);
     float r = 0, g = 0, b = 0;
     loopv(lights)
     {
-        const entity &light = *lights[i];
+        const extentity &light = *lights[i];
         vec ray = target;
         ray.sub(light.o);
         float mag = ray.magnitude(),
-              attenuation = 1.0;
+              attenuation = 1;
         if(light.attr1)
         {
             attenuation -= mag / float(light.attr1);
-            if(attenuation <= 0.0) continue;
+            if(attenuation <= 0) continue;
         }
-        ray.mul(1.0f / mag);
-        if(shadows)
+        if(mag) ray.mul(1.0f / mag);
+        else { ray = normal; ray.neg(); }
+        if(light.attached && light.attached->type==ET_SPOTLIGHT)
         {
-            float dist = raycube(light.o, ray, mag - tolerance, RAY_SHADOW | (mmshadows > 1 ? RAY_ALPHAPOLY : (mmshadows ? RAY_POLY : 0)));
+            vec spot(vec(light.attached->o).sub(light.o).normalize());
+            float maxatten = 1-cosf(max(1, min(90, light.attached->attr1))*RAD);
+            float spotatten = 1-(1-ray.dot(spot))/maxatten;
+            if(spotatten <= 0) continue;
+            attenuation *= spotatten;
+        }
+        if(shadows && mag)
+        {
+            float dist = shadowray(light.o, ray, mag - tolerance, RAY_SHADOW | (mmshadows > 1 ? RAY_ALPHAPOLY : (mmshadows ? RAY_POLY : 0)));
             if(dist < mag - tolerance) continue;
         }
         float intensity;
@@ -306,10 +347,87 @@ bool lumel_sample(const vec &sample, int aasample, int stride)
     return false;
 }
 
+VAR(mmskylight, 0, 1, 1);
+
+void calcskylight(const vec &o, const vec &normal, float tolerance, uchar *skylight, int mmshadows = 1)
+{
+    static const vec rays[17] =
+    {
+        vec(cosf(21*RAD)*cosf(50*RAD), sinf(21*RAD)*cosf(50*RAD), sinf(50*RAD)),
+        vec(cosf(111*RAD)*cosf(50*RAD), sinf(111*RAD)*cosf(50*RAD), sinf(50*RAD)),
+        vec(cosf(201*RAD)*cosf(50*RAD), sinf(201*RAD)*cosf(50*RAD), sinf(50*RAD)),
+        vec(cosf(291*RAD)*cosf(50*RAD), sinf(291*RAD)*cosf(50*RAD), sinf(50*RAD)),
+
+        vec(cosf(66*RAD)*cosf(70*RAD), sinf(66*RAD)*cosf(70*RAD), sinf(70*RAD)),
+        vec(cosf(156*RAD)*cosf(70*RAD), sinf(156*RAD)*cosf(70*RAD), sinf(70*RAD)),
+        vec(cosf(246*RAD)*cosf(70*RAD), sinf(246*RAD)*cosf(70*RAD), sinf(70*RAD)),
+        vec(cosf(336*RAD)*cosf(70*RAD), sinf(336*RAD)*cosf(70*RAD), sinf(70*RAD)),
+       
+        vec(0, 0, 1),
+
+        vec(cosf(43*RAD)*cosf(60*RAD), sinf(43*RAD)*cosf(60*RAD), sinf(60*RAD)),
+        vec(cosf(133*RAD)*cosf(60*RAD), sinf(133*RAD)*cosf(60*RAD), sinf(60*RAD)),
+        vec(cosf(223*RAD)*cosf(60*RAD), sinf(223*RAD)*cosf(60*RAD), sinf(60*RAD)),
+        vec(cosf(313*RAD)*cosf(60*RAD), sinf(313*RAD)*cosf(60*RAD), sinf(60*RAD)),
+
+        vec(cosf(88*RAD)*cosf(80*RAD), sinf(88*RAD)*cosf(80*RAD), sinf(80*RAD)),
+        vec(cosf(178*RAD)*cosf(80*RAD), sinf(178*RAD)*cosf(80*RAD), sinf(80*RAD)),
+        vec(cosf(268*RAD)*cosf(80*RAD), sinf(268*RAD)*cosf(80*RAD), sinf(80*RAD)),
+        vec(cosf(358*RAD)*cosf(80*RAD), sinf(358*RAD)*cosf(80*RAD), sinf(80*RAD)),
+
+    };
+    int hit = 0;
+    loopi(17) if(normal.dot(rays[i])>=0)
+    {
+        if(shadowray(vec(rays[i]).mul(tolerance).add(o), rays[i], 1e16f, RAY_SHADOW | (!mmskylight || !mmshadows ? 0 : (mmshadows > 1 ? RAY_ALPHAPOLY : RAY_POLY)))>1e15f) hit++;
+    }
+
+    loopk(3) skylight[k] = uchar(ambient + (max(hdr.skylight[k], ambient) - ambient)*hit/17.0f);
+}
+
+VAR(blurlms, 0, 0, 2);
+VAR(blurskylight, 0, 0, 2);
+
+void blurlightmap(int n)
+{
+    static uchar blur[3*LM_MAXW*LM_MAXH];
+    static const int matrix3x3[9] =
+    {
+        1, 2, 1,
+        2, 4, 2,
+        1, 2, 1
+    };
+    static const int matrix3x3sum = 16;
+    static const int matrix5x5[25] =
+    {
+        1, 1, 2, 1, 1,
+        1, 2, 4, 2, 1,
+        2, 4, 8, 4, 2,
+        1, 2, 4, 2, 1,
+        1, 1, 2, 1, 1
+    };
+    static const int matrix5x5sum = 52;
+    uchar *src = lm, *dst = blur;
+    int stride = 3*lm_w;
+
+    loop(y, lm_h) loop(x, lm_w) loopk(3)
+    {
+        int c = *src, val = 0; 
+        const int *m = n>1 ? matrix5x5 : matrix3x3;
+        for(int t = -n; t<=n; t++) for(int s = -n; s<=n; s++, m++)
+        {
+            val += *m * (x+s>=0 && x+s<lm_w && y+t>=0 && y+t<lm_h ? src[t*stride+3*s] : c);
+        }
+        *dst++ = val/(n>1 ? matrix5x5sum : matrix3x3sum);
+        src++;
+    }
+    memcpy(lm, blur, 3*lm_w*lm_h);
+}
+
 VAR(edgetolerance, 1, 4, 8);
 VAR(adaptivesample, 0, 1, 1);
 
-bool generate_lightmap(float lpu, uint y1, uint y2, const vec &origin, const lerpvert *lv, int numv, const vec &ustep, const vec &vstep)
+bool generate_lightmap(float lpu, int y1, int y2, const vec &origin, const lerpvert *lv, int numv, const vec &ustep, const vec &vstep)
 {
     static uchar mincolor[3], maxcolor[3];
     static float aacoords[8][2] =
@@ -325,7 +443,7 @@ bool generate_lightmap(float lpu, uint y1, uint y2, const vec &origin, const ler
         {-0.6f, -0.3f},
     };
     float tolerance = 0.5 / lpu;
-    vector<const entity *> &lights = (y1 == 0 ? lights1 : lights2);
+    vector<const extentity *> &lights = (y1 == 0 ? lights1 : lights2);
     vec v = origin;
     vec offsets[8];
     loopi(8) loopj(3) offsets[i][j] = aacoords[i][0]*ustep[j] + aacoords[i][1]*vstep[j];
@@ -342,18 +460,26 @@ bool generate_lightmap(float lpu, uint y1, uint y2, const vec &origin, const ler
     int aasample = min(1 << aalights, 4);
     int stride = aasample*(lm_w+1);
     vec *sample = &samples[stride*y1];
+    uchar *skylight = &lm[3*lm_w*y1];
     lerpbounds start, end;
     initlerpbounds(lv, numv, start, end);
-    for(uint y = y1; y < y2; ++y, v.add(vstep)) 
+    for(int y = y1; y < y2; ++y, v.add(vstep)) 
     {
         vec normal, nstep;
         lerpnormal(y, lv, numv, start, end, normal, nstep);
         
         vec u(v);
-        for(uint x = 0; x < lm_w; ++x, u.add(ustep), normal.add(nstep)) 
+        for(int x = 0; x < lm_w; ++x, u.add(ustep), normal.add(nstep), skylight += 3) 
         {
             CHECK_PROGRESS(return false);
             generate_lumel(tolerance, lights, u, vec(normal).normalize(), *sample, x, y);
+            if(hdr.skylight[0]>ambient || hdr.skylight[1]>ambient || hdr.skylight[2]>ambient)
+            {
+                if(lmtype==LM_BUMPMAP0 || !adaptivesample || sample->x<hdr.skylight[0] || sample->y<hdr.skylight[1] || sample->z<hdr.skylight[2])
+                    calcskylight(u, normal, tolerance, skylight, mmshadows);
+                else loopk(3) skylight[k] = max(hdr.skylight[k], ambient);
+            }
+            else loopk(3) skylight[k] = ambient;
             sample += aasample;
         }
         sample += aasample;
@@ -361,13 +487,13 @@ bool generate_lightmap(float lpu, uint y1, uint y2, const vec &origin, const ler
     v = origin;
     sample = &samples[stride*y1];
     initlerpbounds(lv, numv, start, end);
-    for(uint y = y1; y < y2; ++y, v.add(vstep)) 
+    for(int y = y1; y < y2; ++y, v.add(vstep)) 
     {
         vec normal, nstep;
         lerpnormal(y, lv, numv, start, end, normal, nstep);
 
         vec u(v);
-        for(uint x = 0; x < lm_w; ++x, u.add(ustep), normal.add(nstep)) 
+        for(int x = 0; x < lm_w; ++x, u.add(ustep), normal.add(nstep)) 
         {
             vec &center = *sample++;
             if(adaptivesample && x > 0 && x+1 < lm_w && y > y1 && y+1 < y2 && !lumel_sample(center, aasample, stride))
@@ -380,7 +506,6 @@ bool generate_lightmap(float lpu, uint y1, uint y2, const vec &origin, const ler
      || (!y && aacoords[i][1] < 0) \
      || (y+1==lm_h && aacoords[i][1] > 0) \
      ? edgetolerance : 1)
-
                 vec n(normal);
                 n.normalize();
                 loopi(aasample-1)
@@ -414,7 +539,7 @@ bool generate_lightmap(float lpu, uint y1, uint y2, const vec &origin, const ler
             vec normal, nstep;
             lerpnormal(lm_h, lv, numv, start, end, normal, nstep);
 
-            for(uint x = 0; x <= lm_w; ++x, v.add(ustep), normal.add(nstep))
+            for(int x = 0; x <= lm_w; ++x, v.add(ustep), normal.add(nstep))
             {
                 CHECK_PROGRESS(return false);
                 vec n(normal);
@@ -426,6 +551,10 @@ bool generate_lightmap(float lpu, uint y1, uint y2, const vec &origin, const ler
             } 
         }
 
+        if(hdr.skylight[0]>ambient || hdr.skylight[1]>ambient || hdr.skylight[2]>ambient)
+        {
+            if(blurskylight && (lm_w>1 || lm_h>1)) blurlightmap(blurskylight);
+        }
         sample = samples;
         float weight = 1.0f / (1.0f + 4.0f*aalights),
               cweight = weight * (aalights == 3 ? 5.0f : 1.0f);
@@ -452,26 +581,40 @@ bool generate_lightmap(float lpu, uint y1, uint y2, const vec &origin, const ler
                     l.add(next[aasample+1]);
                 }
 
-                lumel[0] = int(center.x*cweight + l.x*weight);
-                lumel[1] = int(center.y*cweight + l.y*weight);
-                lumel[2] = int(center.z*cweight + l.z*weight);
+                int r = int(center.x*cweight + l.x*weight),
+                    g = int(center.y*cweight + l.y*weight),
+                    b = int(center.z*cweight + l.z*weight),
+                    ar = lumel[0], ag = lumel[1], ab = lumel[2];
+                lumel[0] = max(ar, r);
+                lumel[1] = max(ag, g);
+                lumel[2] = max(ab, b);
                 loopk(3)
                 {
                     mincolor[k] = min(mincolor[k], lumel[k]);
                     maxcolor[k] = max(maxcolor[k], lumel[k]);
                 }
-                lumel += 3;
 
                 if(lmtype == LM_BUMPMAP0)
                 {
                     bvec &n = ((bvec *)lm_ray)[ray-lm_ray];
-                    n = ray->iszero() ? bvec(128, 128, 255) : bvec(ray->normalize());
+                    if(ray->iszero()) n = bvec(128, 128, 255);
+                    else
+                    {
+                        ray->normalize();
+                        // bias the normals towards the amount of ambient/skylight in the lumel
+                        // this is necessary to prevent the light values in shaders from dropping too far below the skylight (to the ambient) if N.L is small
+                        int l = max(r, max(g, b)), a = max(ar, max(ag, ab));
+                        ray->mul(max(l-a, 0));
+                        ray->z += a;
+                        n = bvec(ray->normalize());
+                    }
                     loopk(3)
                     {
                         minray[k] = min(minray[k], n[k]);
                         maxray[k] = max(maxray[k], n[k]);
                     }
                 }
+                lumel += 3;
                 ray++;
             }
             sample += aasample;
@@ -497,6 +640,7 @@ bool generate_lightmap(float lpu, uint y1, uint y2, const vec &origin, const ler
                 lm_h = 1;
             }
         }
+        if(blurlms && (lm_w>1 || lm_h>1)) blurlightmap(blurlms);
     }
     return true;
 }
@@ -538,7 +682,7 @@ void clearlightcache(int e)
     }
     else
     {
-        const entity &light = *et->getents()[e];
+        const extentity &light = *et->getents()[e];
         int radius = light.attr1;
         for(int x = int(max(light.o.x-radius, 0))>>lightcachesize, ex = int(min(light.o.x+radius, hdr.worldsize-1))>>lightcachesize; x <= ex; x++)
         for(int y = int(max(light.o.y-radius, 0))>>lightcachesize, ey = int(min(light.o.y+radius, hdr.worldsize-1))>>lightcachesize; y <= ey; y++)
@@ -581,7 +725,7 @@ const vector<int> &checklightcache(int x, int y)
     return lce.lights;
 }
 
-static inline void addlight(const entity &light, int cx, int cy, int cz, int size, plane planes[2], int numplanes)
+static inline void addlight(const extentity &light, int cx, int cy, int cz, int size, plane planes[2], int numplanes)
 {
     int radius = light.attr1;
     if(radius > 0)
@@ -613,17 +757,17 @@ bool find_lights(int cx, int cy, int cz, int size, plane planes[2], int numplane
         const vector<int> &lights = checklightcache(cx, cy);
         loopv(lights)
         {
-            const entity &light = *ents[lights[i]];
+            const extentity &light = *ents[lights[i]];
             addlight(light, cx, cy, cz, size, planes, numplanes);
         }
     }
     else loopv(ents)
     {
-        const entity &light = *ents[i];
+        const extentity &light = *ents[i];
         if(light.type != ET_LIGHT) continue;
         addlight(light, cx, cy, cz, size, planes, numplanes);
     }
-    return lights1.length() || lights2.length();
+    return lights1.length() || lights2.length() || hdr.skylight[0]>ambient || hdr.skylight[1]>ambient || hdr.skylight[2]>ambient;
 }
 
 bool setup_surface(plane planes[2], const vec *p, const vec *n, const vec *n2, uchar texcoords[8])
@@ -850,7 +994,7 @@ void setup_surfaces(cube &c, int cx, int cy, int cz, int size, bool lodcube)
                 cn[i].normals[3] = bvec(numplanes < 2 ? n[3] : n2[2]);
             }
         }
-        if(lights1.empty() && lights2.empty()) continue;
+        if(lights1.empty() && lights2.empty() && hdr.skylight[0]<=ambient && hdr.skylight[1]<=ambient && hdr.skylight[2]<=ambient) continue;
         uchar texcoords[8];
         if(!setup_surface(planes, v, n, numplanes >= 2 ? n2 : NULL, texcoords))
             continue;
@@ -894,10 +1038,10 @@ static Uint32 calclight_timer(Uint32 interval, void *param)
     check_calclight_progress = true;
     return interval;
 }
-    
-void calclight(int *quality)
+
+bool setlightmapquality(int quality)
 {
-    switch(*quality)
+    switch(quality)
     {
         case  3: shadows = 1; aalights = 3; mmshadows = 2; break;
         case  2: shadows = 1; aalights = 3; mmshadows = 1; break;
@@ -905,14 +1049,25 @@ void calclight(int *quality)
         case  0: shadows = 1; aalights = 2; mmshadows = 0; break;
         case -1: shadows = 1; aalights = 1; mmshadows = 0; break;
         case -2: shadows = 0; aalights = 0; mmshadows = 0; break;
-        default: conoutf("valid range for calclight quality is -2..3"); return;
+        default: return false;
+    }
+    return true;
+}
+
+void calclight(int *quality)
+{
+    if(!setlightmapquality(*quality))
+    {
+        conoutf("valid range for calclight quality is -2..3"); 
+        return;
     }
     computescreen("computing lightmaps... (esc to abort)");
-    remipworld();
+    mpremip(true);
     resetlightmaps();
     clear_lmids(worldroot);
     curlumels = 0;
     progress = 0;
+    progresstexticks = 0;
     calclight_canceled = false;
     check_calclight_progress = false;
     SDL_TimerID timer = SDL_AddTimer(250, calclight_timer, NULL);
@@ -949,11 +1104,17 @@ COMMAND(calclight, "i");
 
 VAR(patchnormals, 0, 0, 1);
 
-void patchlight()
+void patchlight(int *quality)
 {
     if(noedit(true)) return;
+    if(!setlightmapquality(*quality))
+    {
+        conoutf("valid range for patchlight quality is -2..3"); 
+        return;
+    }
     computescreen("patching lightmaps... (esc to abort)");
     progress = 0;
+    progresstexticks = 0;
     int total = 0, lumels = 0;
     loopv(lightmaps)
     {
@@ -963,7 +1124,7 @@ void patchlight()
     curlumels = lumels;
     calclight_canceled = false;
     check_calclight_progress = false;
-    SDL_TimerID timer = SDL_AddTimer(500, calclight_timer, NULL);
+    SDL_TimerID timer = SDL_AddTimer(250, calclight_timer, NULL);
     if(patchnormals) show_out_of_renderloop_progress(0, "computing normals...");
     Uint32 start = SDL_GetTicks();
     if(patchnormals) calcnormals();
@@ -990,21 +1151,9 @@ void patchlight()
             (end - start) / 1000.0f); 
 }
 
-COMMAND(patchlight, "");
+COMMAND(patchlight, "i");
 
-VARF(fullbright, 0, 0, 1,
-    if(fullbright)
-    {
-        if(noedit(true))
-        {
-            fullbright = 0;
-            initlights();
-            return;
-        }
-        clearlights();
-    }
-    else initlights();
-);
+VARF(fullbright, 0, 0, 1, initlights());
 
 vector<GLuint> lmtexids;
 
@@ -1015,6 +1164,16 @@ void alloctexids()
 
 void clearlights()
 {
+    clearlightcache();
+    const vector<extentity *> &ents = et->getents();
+    loopv(ents)
+    {
+        extentity &e = *ents[i];
+        e.color = vec(1, 1, 1);
+        e.dir = vec(0, 0, 1);
+    }
+    if(nolights) return;
+
     uchar bright[3] = { 128, 128, 128 };
     bvec front(128, 128, 255);
     alloctexids();
@@ -1031,26 +1190,20 @@ void clearlights()
                 break;
         }
     }            
-    clearlightcache();
-    const vector<extentity *> &ents = et->getents();
-    loopv(ents) 
-    {
-        extentity &e = *ents[i];
-        e.color = vec(1, 1, 1);
-        e.dir = vec(0, 0, 1);
-    }
 }
 
 void lightent(extentity &e, float height)
 {
     if(e.type==ET_LIGHT) return;
+    float ambient = hdr.ambient/255.0f;
     if(e.type==ET_MAPMODEL)
     {
         model *m = loadmodel(NULL, e.attr2);
         if(m) height = m->above()*0.75f;
     }
+    else if(e.type>=ET_GAMESPECIFIC) ambient = 0.4f;
     vec target(e.o.x, e.o.y, e.o.z + height);
-    lightreaching(target, e.color, e.dir, &e, renderpath==R_FIXEDFUNCTION && e.type>=ET_GAMESPECIFIC ? 0.7f : 0.4f);
+    lightreaching(target, e.color, e.dir, &e, ambient);
 }
 
 void updateentlighting()
@@ -1113,11 +1266,15 @@ static void find_unlit(int i)
 
 void initlights()
 {
-    if(fullbright)
+    if(nolights || (fullbright && editmode) || lightmaps.empty())
     {
         clearlights();
         return;
     }
+
+    clearlightcache();
+    updateentlighting();
+
     alloctexids();
     uchar unlit[3] = { ambient, ambient, ambient };
     createtexture(lmtexids[LMID_AMBIENT], 1, 1, unlit, 0, false);
@@ -1141,13 +1298,11 @@ void initlights()
         }
         createtexture(lmtexids[i+LMID_RESERVED], LM_PACKW, LM_PACKH, data, 0, false);
     }
-    clearlightcache();
-    updateentlighting();
 }
 
 void lightreaching(const vec &target, vec &color, vec &dir, extentity *t, float ambient)
 {
-    if(fullbright)
+    if(nolights || (fullbright && editmode) || lightmaps.empty())
     {
         color = vec(1, 1, 1);
         dir = vec(0, 0, 1);
@@ -1159,7 +1314,7 @@ void lightreaching(const vec &target, vec &color, vec &dir, extentity *t, float
     const vector<int> &lights = checklightcache(int(target.x), int(target.y));
     loopv(lights)
     {
-        entity &e = *ents[lights[i]];
+        extentity &e = *ents[lights[i]];
         if(e.type != ET_LIGHT)
             continue;
     
@@ -1170,11 +1325,19 @@ void lightreaching(const vec &target, vec &color, vec &dir, extentity *t, float
             continue;
     
         ray.div(mag);
-        if(raycube(e.o, ray, mag, RAY_SHADOW | RAY_POLY, 0, t) < mag)
+        if(shadowray(e.o, ray, mag, RAY_SHADOW | RAY_POLY, t) < mag)
             continue;
-        float intensity = 1.0;
+        float intensity = 1;
         if(e.attr1)
             intensity -= mag / float(e.attr1);
+        if(e.attached && e.attached->type==ET_SPOTLIGHT)
+        {
+            vec spot(vec(e.attached->o).sub(e.o).normalize());
+            float maxatten = 1-cosf(max(1, min(90, e.attached->attr1))*RAD);
+            float spotatten = 1-(1-ray.dot(spot))/maxatten;
+            if(spotatten<=0) continue;
+            intensity *= spotatten;
+        }
 
         //if(target==player->o)
         //{
@@ -1185,24 +1348,34 @@ void lightreaching(const vec &target, vec &color, vec &dir, extentity *t, float
 
         intensity *= e.attr2*e.attr3*e.attr4;
 
-        dir.add(vec(e.o).sub(target).mul(intensity/mag));
+        if(fabs(mag)<1e-3) dir.add(vec(0, 0, 1));
+        else dir.add(vec(e.o).sub(target).mul(intensity/mag));
     }
 
-    color.x = min(1.5f, max(ambient, color.x));
-    color.y = min(1.5f, max(ambient, color.y));
-    color.z = min(1.5f, max(ambient, color.z));
-    if(!dir.iszero()) dir.normalize();
+    if(t && (hdr.skylight[0]>ambient || hdr.skylight[1]>ambient || hdr.skylight[2]>ambient))
+    {
+        uchar skylight[3];
+        calcskylight(target, vec(0, 0, 0), 0.5f, skylight);
+        loopk(3) color[k] = min(1.5f, max(max(skylight[k]/255.0f, ambient), color[k]));
+    }
+    else loopk(3)
+    {
+        float skylight = 0.75f*max(hdr.skylight[k]/255.0f, ambient) + 0.25f*ambient;
+        color[k] = min(1.5f, max(skylight, color[k]));
+    }
+    if(dir.iszero()) dir = vec(0, 0, 1);
+    else dir.normalize();
 }
 
 entity *brightestlight(const vec &target, const vec &dir)
 {
     const vector<extentity *> &ents = et->getents();
     const vector<int> &lights = checklightcache(int(target.x), int(target.y));
-    entity *brightest = NULL;
+    extentity *brightest = NULL;
     float bintensity = 0;
     loopv(lights)
     {
-        entity &e = *ents[lights[i]];
+        extentity &e = *ents[lights[i]];
         if(e.type != ET_LIGHT || vec(e.o).sub(target).dot(dir)<0)
             continue;
 
@@ -1213,11 +1386,20 @@ entity *brightestlight(const vec &target, const vec &dir)
              continue;
 
         ray.div(mag);
-        if(raycube(e.o, ray, mag, RAY_SHADOW | RAY_POLY) < mag)
+        if(shadowray(e.o, ray, mag, RAY_SHADOW | RAY_POLY) < mag)
             continue;
-        float intensity = 1.0;
+        float intensity = 1;
         if(e.attr1)
             intensity -= mag / float(e.attr1);
+        if(e.attached && e.attached->type==ET_SPOTLIGHT)
+        {
+            vec spot(vec(e.attached->o).sub(e.o).normalize());
+            float maxatten = 1-cosf(max(1, min(90, e.attached->attr1))*RAD);
+            float spotatten = 1-(1-ray.dot(spot))/maxatten;
+            if(spotatten<=0) continue;
+            intensity *= spotatten;
+        }
+
         if(!brightest || intensity > bintensity)
         {
             brightest = &e;
