@@ -72,22 +72,34 @@ struct QuadNode
     }
 };
 
-void renderwaterfall(materialsurface &m, Texture *t, float offset)
+void renderwaterfall(materialsurface &m, Texture *tex, float scale, float offset, uchar mat)
 {
-    float xf = 8.0f/t->xs;
-    float yf = 8.0f/t->ys;
-    float d = 16.0f*lastmillis/1000.0f;
+    float xf = 8.0f/(tex->xs*scale);
+    float yf = 8.0f/(tex->ys*scale);
+    float d = 16.0f*lastmillis;
     int dim = dimension(m.orient),
         csize = C[dim]==2 ? m.rsize : m.csize,
         rsize = R[dim]==2 ? m.rsize : m.csize;
-
+    float t = lastmillis;
+    switch(mat)
+    {
+        case MAT_WATER: 
+            t /= renderpath!=R_FIXEDFUNCTION ? 600.0f : 300.0f; 
+            d /= 1000.0f;
+            break;
+        case MAT_LAVA: 
+            t /= 2000.0f; 
+            d /= 3000.0f;
+            break;
+    }
+    float wave = m.ends&2 ? (vertwater ? WATER_AMPLITUDE*sinf(t)-WATER_OFFSET : -WATER_OFFSET) : 0;
     loopi(4)
     {
         vec v(m.o.tovec());
         v[dim] += dimcoord(m.orient) ? -offset : offset;
         if(i==1 || i==2) v[dim^1] += csize;
         if(i<=1) v.z += rsize;
-        if(m.ends&(i<=1 ? 2 : 1)) v.z -= 1.1f;
+        if(m.ends&(i<=1 ? 2 : 1)) v.z += i<=1 ? wave : -WATER_OFFSET-WATER_AMPLITUDE;
         glTexCoord2f(xf*v[dim^1], yf*(v.z+d));
         glVertex3fv(v.v);
     }
@@ -95,7 +107,7 @@ void renderwaterfall(materialsurface &m, Texture *t, float offset)
     xtraverts += 4;
 }
 
-void drawmaterial(int orient, int x, int y, int z, int csize, int rsize, float offset, bool usetc)
+void drawmaterial(int orient, int x, int y, int z, int csize, int rsize, float offset)
 {
     int dim = dimension(orient), c = C[dim], r = R[dim];
     loopi(4)
@@ -105,7 +117,6 @@ void drawmaterial(int orient, int x, int y, int z, int csize, int rsize, float o
         v[c] += cubecoords[coord][c]/8*csize;
         v[r] += cubecoords[coord][r]/8*rsize;
         v[dim] += dimcoord(orient) ? -offset : offset;
-        if(usetc) glTexCoord2f(v[c]/8, v[r]/8);
         glVertex3fv(v.v);
     }
     xtraverts += 4;
@@ -122,6 +133,8 @@ struct material
     {"clip", MAT_CLIP},
     {"glass", MAT_GLASS},
     {"noclip", MAT_NOCLIP},
+    {"lava", MAT_LAVA},
+    {"aiclip", MAT_AICLIP}
 };
 
 int findmaterial(const char *name)
@@ -140,8 +153,9 @@ int visiblematerial(cube &c, int orient, int x, int y, int z, int size)
     case MAT_AIR:
          break;
 
+    case MAT_LAVA:
     case MAT_WATER:
-        if(visibleface(c, orient, x, y, z, size, MAT_WATER))
+        if(visibleface(c, orient, x, y, z, size, c.ext->material))
             return (orient != O_BOTTOM ? MATSURF_VISIBLE : MATSURF_EDIT_ONLY);
         break;
 
@@ -268,9 +282,9 @@ int optimizematsurfs(materialsurface *matbuf, int matsurfs)
                cur->orient == start->orient &&
                cur->o[dim] == start->o[dim])
             ++cur;
-         if(start->material != MAT_WATER || start->orient != O_TOP || renderpath!=R_FIXEDFUNCTION)
+         if(!isliquid(start->material) || start->orient != O_TOP || !vertwater)
          {
-            if(start!=matbuf) memcpy(matbuf, start, (cur-start)*sizeof(materialsurface));
+            if(start!=matbuf) memmove(matbuf, start, (cur-start)*sizeof(materialsurface));
             matbuf += mergemats(matbuf, cur-start);
          }
          else if(cur-start>=4)
@@ -281,7 +295,7 @@ int optimizematsurfs(materialsurface *matbuf, int matsurfs)
          }
          else
          {
-            if(start!=matbuf) memcpy(matbuf, start, (cur-start)*sizeof(materialsurface));
+            if(start!=matbuf) memmove(matbuf, start, (cur-start)*sizeof(materialsurface));
             matbuf += cur-start;
          }
     }
@@ -298,6 +312,7 @@ struct waterinfo
 
 void setupmaterials()
 {
+    int hasmat = 0;
     vector<waterinfo> water;
     hashtable<ivec, int> watersets;
     unionfind uf;
@@ -326,13 +341,13 @@ void setupmaterials()
                 }
                 waterinfo &wi = water.add();
                 wi.m = &m;
-                vec center(m.o.x+m.rsize/2, m.o.y+m.csize/2, m.o.z-1.1f);
+                vec center(m.o.x+m.rsize/2, m.o.y+m.csize/2, m.o.z-WATER_OFFSET);
                 m.light = brightestlight(center, vec(0, 0, 1));
                 float depth = raycube(center, vec(0, 0, -1), 10000);
                 wi.depth = double(depth)*m.rsize*m.csize;
                 wi.area = m.rsize*m.csize; 
             }
-            else if(m.material==MAT_WATER && m.orient!=O_BOTTOM)
+            else if(isliquid(m.material) && m.orient!=O_BOTTOM && m.orient!=O_TOP)
             {
                 m.ends = 0;
                 int dim = dimension(m.orient), coord = dimcoord(m.orient);
@@ -343,7 +358,7 @@ void setupmaterials()
                 while(o[dim^1] < maxc)
                 {
                     cube &c = lookupcube(o.x, o.y, o.z);
-                    if(c.ext && c.ext->material==MAT_WATER) { m.ends |= 1; break; }
+                    if(c.ext && isliquid(c.ext->material)) { m.ends |= 1; break; }
                     o[dim^1] += lusize;
                 }
                 o[dim^1] = minc;
@@ -368,6 +383,7 @@ void setupmaterials()
                     m.envmap = closestenvmap(center);
                 }
             }
+            hasmat |= 1<<m.material;
         }
     }
     loopv(water)
@@ -385,6 +401,13 @@ void setupmaterials()
         water[i].m->light = water[root].m->light;
         water[i].m->depth = (short)(water[root].depth/water[root].area);
     }
+    if(hasmat&(1<<MAT_WATER))
+    {
+        extern void loadcaustics();
+        loadcaustics();
+        lookuptexture(-MAT_WATER);
+    }
+    if(hasmat&(1<<MAT_LAVA)) lookuptexture(-MAT_LAVA);
 }
 
 VARP(showmat, 0, 1, 1);
@@ -456,7 +479,7 @@ void sortmaterials(vector<materialsurface *> &vismats, float zclip, bool refract
 
 void rendermatgrid(vector<materialsurface *> &vismats)
 {
-    notextureshader->set();
+    glEnable(GL_POLYGON_OFFSET_LINE);
     glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
     static uchar cols[MAT_EDIT][3] =
     {
@@ -465,6 +488,8 @@ void rendermatgrid(vector<materialsurface *> &vismats)
         { 85, 0, 0 },  // MAT_CLIP - red,
         { 0, 85, 85 }, // MAT_GLASS - cyan,
         { 0, 85, 0 },  // MAT_NOCLIP - green
+        { 85, 40, 0 }, // MAT_LAVA - orange
+        { 85, 85, 0 }, // MAT_AICLIP - yellow
     };
     int lastmat = -1;
     glBegin(GL_QUADS);
@@ -481,6 +506,7 @@ void rendermatgrid(vector<materialsurface *> &vismats)
     }
     glEnd();
     glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
+    glDisable(GL_POLYGON_OFFSET_LINE);
 }
 
 VARP(glassenv, 0, 1, 1);
@@ -512,25 +538,28 @@ void rendermaterials(float zclip, bool refract)
     sortmaterials(vismats, zclip, refract);
     if(vismats.empty()) return;
 
-    if(!editmode || !showmat) glDepthMask(GL_FALSE);
     glDisable(GL_CULL_FACE);
-    glEnable(GL_BLEND);
 
-    Slot &wslot = lookuptexture(-MAT_WATER);
+    Slot &wslot = lookuptexture(-MAT_WATER), &lslot = lookuptexture(-MAT_LAVA);
     uchar wcol[4] = { 128, 128, 128, 192 };
     if(hdr.watercolour[0] || hdr.watercolour[1] || hdr.watercolour[2]) memcpy(wcol, hdr.watercolour, 3);
     int lastorient = -1, lastmat = -1;
     GLenum textured = GL_TEXTURE_2D;
-    bool begin = false;
+    bool begin = false, depth = true, blended = false, overbright = false;
     ushort envmapped = EMID_NONE;
     vec normal(0, 0, 0);
 
+    static float zerofog[4] = { 0, 0, 0, 1 };
+    float oldfogc[4];
+    glGetFloatv(GL_FOG_COLOR, oldfogc);
+    int lastfogtype = 1;
     loopv(vismats)
     {
         materialsurface &m = *vismats[editmode && showmat ? vismats.length()-1-i : i];
         int curmat = !editmode || !showmat || m.material>=MAT_EDIT ? m.material : m.material+MAT_EDIT;
         if(lastmat!=curmat || lastorient!=m.orient || (curmat==MAT_GLASS && envmapped && m.envmap != envmapped))
         {
+            int fogtype = lastfogtype;
             switch(curmat)
             {
                 case MAT_WATER:
@@ -538,6 +567,9 @@ void rendermaterials(float zclip, bool refract)
                     if(begin) { glEnd(); begin = false; }
                     if(lastmat!=MAT_WATER || (lastorient==O_TOP)!=(m.orient==O_TOP))
                     {
+                        if(depth) { glDepthMask(GL_FALSE); depth = false; }
+                        if(!blended) { glEnable(GL_BLEND); blended = true; }
+                        if(overbright) { resettmu(0); overbright = false; }
                         if(m.orient==O_TOP)
                         {
                             glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
@@ -548,15 +580,42 @@ void rendermaterials(float zclip, bool refract)
                             glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
                             glColor3f(wcol[0]/255.0f, wcol[1]/255.0f, wcol[2]/255.0f);
                         }
+                        foggedshader->set();
+                        fogtype = 0;
                     }
                     if(textured!=GL_TEXTURE_2D) 
                     { 
-                        if(textured) glDisable(textured); 
-                        glEnable(GL_TEXTURE_2D); 
+                        if(textured) glDisable(textured);
+                        glEnable(GL_TEXTURE_2D);
                         textured = GL_TEXTURE_2D; 
                     }
                     glBindTexture(GL_TEXTURE_2D, wslot.sts[m.orient==O_TOP ? 0 : 1].t->gl);
-                    defaultshader->set();
+                    break;
+
+                case MAT_LAVA:
+                    if(lastmat==MAT_LAVA && lastorient!=O_TOP && m.orient!=O_TOP) break;
+                    if(begin) { glEnd(); begin = false; }
+                    if(lastmat!=MAT_LAVA)
+                    {
+                        if(!depth) { glDepthMask(GL_TRUE); depth = true; }
+                        if(blended) { glDisable(GL_BLEND); blended = false; }
+                        if(renderpath==R_FIXEDFUNCTION && !overbright) { setuptmu(0, "C * T x 2"); overbright = true; }
+                        float t = lastmillis/2000.0f;
+                        t -= int(t);
+                        t = 1.0f - fabs(t-0.5f);
+                        glColor3f(t, t, t);
+                        static Shader *lavashader = NULL;
+                        if(!lavashader) lavashader = lookupshaderbyname("lava");
+                        lavashader->set();
+                        fogtype = 1;
+                    }
+                    if(textured!=GL_TEXTURE_2D)
+                    {
+                        if(textured) glDisable(textured);
+                        glEnable(GL_TEXTURE_2D);
+                        textured = GL_TEXTURE_2D;
+                    }
+                    glBindTexture(GL_TEXTURE_2D, lslot.sts[m.orient==O_TOP ? 0 : 1].t->gl);
                     break;
 
                 case MAT_GLASS:
@@ -584,12 +643,16 @@ void rendermaterials(float zclip, bool refract)
                     }
                     if(lastmat!=MAT_GLASS)
                     {
+                        if(!blended) { glEnable(GL_BLEND); blended = true; }
+                        if(overbright) { resettmu(0); overbright = false; }
+                        if(depth) { glDepthMask(GL_FALSE); depth = false; }
                         if(m.envmap!=EMID_NONE && glassenv)
                         {
                             if(renderpath==R_FIXEDFUNCTION)
                             {
                                 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                                 glColor4f(0.8f, 0.9f, 1.0f, 0.25f);
+                                fogtype = 1;
                             }
                             else
                             {
@@ -609,7 +672,8 @@ void rendermaterials(float zclip, bool refract)
                             }
                             glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
                             glColor3f(0.3f, 0.15f, 0.0f);
-                            notextureshader->set();
+                            foggednotextureshader->set();
+                            fogtype = 0;
                         }
                     }
                     break;
@@ -620,17 +684,23 @@ void rendermaterials(float zclip, bool refract)
                     if(lastmat<MAT_EDIT)
                     {
                         if(begin) { glEnd(); begin = false; }
+                        if(!depth) { glDepthMask(GL_TRUE); depth = true; }
+                        if(!blended) { glEnable(GL_BLEND); blended = true; }
+                        if(overbright) { resettmu(0); overbright = false; }
                         glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
                         if(textured) { glDisable(GL_TEXTURE_2D); textured = 0; }
-                        notextureshader->set();
+                        foggednotextureshader->set();
+                        fogtype = 0;
                     }
                     static uchar blendcols[MAT_EDIT][3] =
                     {
-                        { 0, 0, 0 },     // MAT_AIR - no edit volume,
-                        { 255, 128, 0 }, // MAT_WATER - blue,
-                        { 0, 255, 255 }, // MAT_CLIP - red,
-                        { 255, 0, 0 },   // MAT_GLASS - cyan,
-                        { 255, 0, 255 }, // MAT_NOCLIP - green
+                        { 0,   0,   0   }, // MAT_AIR - no edit volume,
+                        { 255, 128, 0   }, // MAT_WATER - blue,
+                        { 0,   255, 255 }, // MAT_CLIP - red,
+                        { 255, 0,   0   }, // MAT_GLASS - cyan,
+                        { 255, 0,   255 }, // MAT_NOCLIP - green
+                        { 0, 128,   255 }, // MAT_LAVA - orange
+                        { 0,   0,   255 }, // MAT_AICLIP - yellow
                     };
                     glColor3ubv(blendcols[curmat >= MAT_EDIT ? curmat-MAT_EDIT : curmat]);
                     break;
@@ -638,6 +708,11 @@ void rendermaterials(float zclip, bool refract)
             }
             lastmat = curmat;
             lastorient = m.orient;
+            if(fogtype!=lastfogtype)
+            {
+                glFogfv(GL_FOG_COLOR, fogtype ? oldfogc : zerofog);
+                lastfogtype = fogtype;
+            }
         }
         switch(curmat)
         {
@@ -645,7 +720,20 @@ void rendermaterials(float zclip, bool refract)
                 if(m.orient!=O_TOP)
                 {
                     if(!begin) { glBegin(GL_QUADS); begin = true; }
-                    renderwaterfall(m, wslot.sts[1].t, 0.1f);
+                    renderwaterfall(m, wslot.sts[1].t, wslot.sts[1].scale ? wslot.sts[1].scale : 1, 0.1f, MAT_WATER);
+                }
+                break;
+
+            case MAT_LAVA:
+                if(m.orient==O_TOP) 
+                {
+                    if(!vertwater && !begin) { glBegin(GL_QUADS); begin = true; }
+                    renderlava(m, lslot.sts[0].t, lslot.sts[0].scale ? lslot.sts[0].scale : 1);
+                }
+                else
+                {
+                    if(!begin) { glBegin(GL_QUADS); begin = true; }
+                    renderwaterfall(m, lslot.sts[1].t, lslot.sts[1].scale ? lslot.sts[1].scale : 1, 0.1f, MAT_LAVA);
                 }
                 break;
 
@@ -667,9 +755,15 @@ void rendermaterials(float zclip, bool refract)
     }
 
     if(begin) glEnd();
-    glDisable(GL_BLEND);
-    if(!editmode || !showmat) glDepthMask(GL_TRUE);
-    else rendermatgrid(vismats);
+    if(!depth) glDepthMask(GL_TRUE);
+    if(blended) glDisable(GL_BLEND);
+    if(overbright) resettmu(0);
+    if(!lastfogtype) glFogfv(GL_FOG_COLOR, oldfogc);
+    if(editmode && showmat)
+    {
+        foggednotextureshader->set();
+        rendermatgrid(vismats);
+    }
 
     glEnable(GL_CULL_FACE);
     if(textured!=GL_TEXTURE_2D)
