@@ -135,7 +135,7 @@ void validatec(cube *c, int size)
     {
         if(c[i].children)
         {
-            if(size<=4)
+            if(size<=(8>>VVEC_FRAC))
             {
                 solidfaces(c[i]);
                 discardchildren(c[i]);
@@ -155,13 +155,11 @@ void validatec(cube *c, int size)
 
 ivec lu;
 int lusize;
-bool luperfect;
 cube &lookupcube(int tx, int ty, int tz, int tsize)
 {
     int size = hdr.worldsize;
     int x = 0, y = 0, z = 0;
     cube *c = worldroot;
-    luperfect = true;
     for(;;)
     {
         size >>= 1;
@@ -175,17 +173,7 @@ cube &lookupcube(int tx, int ty, int tz, int tsize)
         {
             //if(!tsize) break;
             if(tsize<=0) break;
-            if(isempty(*c))
-            {
-                int mat = c->ext ? c->ext->material : MAT_AIR;
-                c->children = newcubes(F_EMPTY);
-                loopi(8)
-                {
-                    loopl(6) c->children[i].texture[l] = c->texture[l];
-                    if(mat!=MAT_AIR) ext(c->children[i]).material = mat;
-                }
-            }
-            else if(!subdividecube(*c)) luperfect = false;
+            subdividecube(*c);
         }
         c = c->children;
     }
@@ -210,6 +198,13 @@ cube &neighbourcube(int x, int y, int z, int size, int rsize, int orient)
     return lookupcube(x, y, z, rsize);
 }
 
+int lookupmaterial(const vec &o)
+{
+    if(!insideworld(o)) return MAT_AIR;
+    cube &c = lookupcube(int(o.x), int(o.y), int(o.z));
+    return c.ext ? c.ext->material : MAT_AIR;
+}
+
 uchar octantrectangleoverlap(const ivec &c, int size, const ivec &o, const ivec &s)
 {
     uchar p = 0xFF; // bitmask of possible collisions with octants. 0 bit = 0 octant, etc
@@ -257,6 +252,7 @@ int getmippedtexture(cube &p, int orient)
 void forcemip(cube &c)
 {   
     cube *ch = c.children;
+    emptyfaces(c);
 
     loopi(8) loopj(8)
     {
@@ -305,6 +301,17 @@ int midedge(const ivec &a, const ivec &b, int xd, int yd, bool &perfect)
 bool subdividecube(cube &c, bool fullcheck, bool brighten)
 {   
     if(c.children) return true;
+	if(isempty(c) || isentirelysolid(c))
+    {
+        int mat = c.ext ? c.ext->material : MAT_AIR;
+		c.children = newcubes(isempty(c) ? F_EMPTY : F_SOLID);
+        loopi(8)
+        {
+            loopl(6) c.children[i].texture[l] = c.texture[l];
+            if(mat!=MAT_AIR) ext(c.children[i]).material = mat;
+        }
+        return true;
+    }
     cube *ch = c.children = newcubes(F_SOLID);
     bool perfect = true, p1, p2;
     int mat = c.ext ? c.ext->material : MAT_AIR; 
@@ -452,19 +459,22 @@ bool remip(cube &c, int x, int y, int z, int size)
     return true;
 }
 
-void remipworld()
+void mpremip(bool local)
 {
+    extern selinfo sel;
+    if(local) cl->edittrigger(sel, EDIT_REMIP);
     loopi(8)
     {
         ivec o(i, 0, 0, 0, hdr.worldsize>>1);
         remip(worldroot[i], o.x, o.y, o.z, hdr.worldsize>>2);
     }
     calcmerges();
+    if(!local) allchanged();
 }
 
 void remip_()
 {
-    remipworld();
+    mpremip(true);
     allchanged();
 }
 
@@ -806,12 +816,17 @@ bool collapsedface(uint cfe)
     return ((cfe >> 4) & 0x0F0F) == (cfe & 0x0F0F) ||
            ((cfe >> 20) & 0x0F0F) == ((cfe >> 16) & 0x0F0F);
 }  
-    
-bool occludesface(cube &c, int orient, const ivec &o, int size, const ivec &vo, int vsize, uchar vmat, const facevec *vf)
+   
+bool occludesface(cube &c, int orient, const ivec &o, int size, const ivec &vo, int vsize, uchar vmat, uchar nmat, const facevec *vf)
 {
     int dim = dimension(orient);
     if(!c.children)
     {
+         if(nmat != MAT_AIR && c.ext && c.ext->material == nmat)
+         {
+            facevec nf[8];
+            return clipfacevecs(vf, o[C[dim]], o[R[dim]], size, nf) < 3;
+         }
          if(isentirelysolid(c)) return true;
          if(vmat != MAT_AIR && c.ext && c.ext->material == vmat) return true;
          if(touchingface(c, orient) && faceedges(c, orient) == F_SOLID) return true;
@@ -828,13 +843,13 @@ bool occludesface(cube &c, int orient, const ivec &o, int size, const ivec &vo,
     int coord = dimcoord(orient);
     loopi(8) if(octacoord(dim, i) == coord)
     {
-        if(!occludesface(c.children[i], orient, ivec(i, o.x, o.y, o.z, size), size, vo, vsize, vmat, vf)) return false;
+        if(!occludesface(c.children[i], orient, ivec(i, o.x, o.y, o.z, size), size, vo, vsize, vmat, nmat, vf)) return false;
     }
 
     return true;
 }
 
-bool visibleface(cube &c, int orient, int x, int y, int z, int size, uchar mat, bool lodcube)
+bool visibleface(cube &c, int orient, int x, int y, int z, int size, uchar mat, uchar nmat, bool lodcube)
 {
     uint cfe = faceedges(c, orient);
     if(mat != MAT_AIR)
@@ -852,8 +867,9 @@ bool visibleface(cube &c, int orient, int x, int y, int z, int size, uchar mat,
 
     if(lusize > size || (lusize == size && (!o.children || lodcube)))
     {
+        if(nmat != MAT_AIR && o.ext && o.ext->material == nmat) return true;
         if(isentirelysolid(o)) return false;
-        if(mat != MAT_AIR && o.ext && (o.ext->material == mat || (mat == MAT_WATER && o.ext->material == MAT_GLASS))) return false;
+        if(mat != MAT_AIR && o.ext && (o.ext->material == mat || (o.ext->material == MAT_GLASS && isliquid(mat)))) return false;
         if(isempty(o) || !touchingface(o, opposite(orient))) return true;
         if(faceedges(o, opposite(orient)) == F_SOLID) return false;
         facevec cf[4], of[4];
@@ -864,7 +880,7 @@ bool visibleface(cube &c, int orient, int x, int y, int z, int size, uchar mat,
 
     facevec cf[4];
     genfacevecs(c, orient, ivec(x, y, z), size, mat != MAT_AIR, cf);
-    return !occludesface(o, opposite(orient), lu, lusize, ivec(x, y, z), size, mat, cf);
+    return !occludesface(o, opposite(orient), lu, lusize, ivec(x, y, z), size, mat, nmat, cf);
 }
 
 void calcvert(cube &c, int x, int y, int z, int size, vvec &v, int i, bool solid)
@@ -882,7 +898,7 @@ void calcvert(cube &c, int x, int y, int z, int size, vvec &v, int i, bool solid
 void calcverts(cube &c, int x, int y, int z, int size, vvec *verts, bool *usefaces, int *vertused, bool lodcube)
 {
     loopi(8) vertused[i] = 0;
-    loopi(6) if(usefaces[i] = visibleface(c, i, x, y, z, size, MAT_AIR, lodcube)) loopk(4) vertused[faceverts(c,i,k)]++;
+    loopi(6) if(usefaces[i] = visibleface(c, i, x, y, z, size, MAT_AIR, MAT_AIR, lodcube)) loopk(4) vertused[faceverts(c,i,k)]++;
     loopi(8) if(vertused[i]) calcvert(c, x, y, z, size, verts[i], i);
 }
 
@@ -895,12 +911,6 @@ int genclipplane(cube &c, int orient, vec *v, plane *clip)
     if(p[0]==p[2]) return 0;
     if(p[0]!=p[1] && p[1]!=p[2]) clip[planes++].toplane(p[0], p[1], p[2]);
     if(p[0]!=p[3] && p[2]!=p[3] && (!planes || faceconvexity(c, orient))) clip[planes++].toplane(p[0], p[2], p[3]);
-/* this will break lightmap/normal calcs... should never need to do this!
-    if(planes == 0 && dimension(orient)!=dimension(O_TOP)) 
-    {   // simple hack for now. should probably find more robust solution
-        clip[planes++].toplane(p[0], p[1].add(vec(0,0,orient==0 ? -1 : 1)), p[2]);
-    }
-*/
     return planes;
 }
 
