@@ -67,12 +67,11 @@ int gridsize = 8;
 ivec cor, lastcor;
 ivec cur, lastcur;
 
+extern int entediting;
 bool editmode = false;
 bool havesel = false;
-
-int *hmap = NULL; // heightmap
-ushort *htex = NULL; // textures for heightmap
-ushort htexture = 0; // single texture for heightmap
+bool hmapsel = false;
+int horient  = 0;
 
 extern int entmoving;
 
@@ -84,8 +83,6 @@ VARF(dragging, 0, 0, 1,
     sel.orient = orient;
 );
 
-VAR(selectcorners, 0, 0, 1);
-
 VARF(moving, 0, 0, 1,
     if(!moving) return;
     vec v(cur.v); v.add(1);
@@ -93,33 +90,16 @@ VARF(moving, 0, 0, 1,
     if(moving) havesel = false; // tell cursorupdate to create handle
 );
 
-void clearheighttexture()
-{
-    if(htex != NULL) delete[] htex;
-    htex = NULL;
-}
-
-void clearheightmap()
-{
-    clearheighttexture();
-    if(hmap != NULL) delete[] hmap;
-    hmap = NULL;
-}
-
 void forcenextundo() { lastsel.orient = -1; }
 
 void cubecancel()
 {
     havesel = false;
     moving = dragging = 0;
-    clearheightmap();
     forcenextundo();
 }
 
-void entcancel()
-{
-    entgroup.setsize(0);
-}
+extern void entcancel();
 
 void cancelsel()
 {
@@ -127,7 +107,7 @@ void cancelsel()
     entcancel();
 }
 
-VARF(gridpower, 2, 3, VVEC_INT-1,
+VARF(gridpower, 3-VVEC_FRAC, 3, VVEC_INT-1,
 {
     if(dragging) return;
     gridsize = 1<<gridpower;
@@ -135,9 +115,10 @@ VARF(gridpower, 2, 3, VVEC_INT-1,
     cancelsel();
 });
 
-VAR(passthroughcube, 0, 0, 1);
 VAR(passthroughsel, 0, 0, 1);
 VAR(editing, 1, 0, 0);
+VAR(selectcorners, 0, 0, 1);
+VARF(hmapedit, 0, 0, 1, horient = sel.orient);
 
 void toggleedit()
 {
@@ -156,7 +137,10 @@ void toggleedit()
     }
     cancelsel();
     keyrepeat(editmode);
-    editing = editmode;
+    editing = entediting = editmode;
+    extern int fullbright;
+    if(fullbright) initlights();
+    cc->edittoggled(editmode);
 }
 
 bool noedit(bool view)
@@ -184,7 +168,6 @@ void reorient()
     sel.cxs = sel.s[R[dimension(orient)]]*2;
     sel.cys = sel.s[C[dimension(orient)]]*2;
     sel.orient = orient;
-    if(hmap) createheightmap();
 }
 
 void selextend()
@@ -229,7 +212,7 @@ cube &blockcube(int x, int y, int z, const block3 &b, int rgrid) // looks up a w
 
 int selchildcount=0;
 
-ICOMMAND(havesel, "", intret(havesel ? selchildcount : 0));
+ICOMMAND(havesel, "", (), intret(havesel ? selchildcount : 0));
 
 void countselchild(cube *c, const ivec &cor, int size)
 {
@@ -289,9 +272,14 @@ void editmoveplane(const vec &o, const vec &ray, int d, float off, vec &handle,
     }
 }
 
+inline bool isheightmap(int orient, int d, bool empty, cube *c);
 extern void entdrag(const vec &ray);
-extern bool hoveringonent(const vec &o, const vec &ray);
+extern bool hoveringonent(int ent, int orient);
 extern void renderentselection(const vec &o, const vec &ray, bool entmoving);
+extern float rayent(const vec &o, vec &ray, vec &hitpos, float radius, int mode, int size, int &orient, int &ent);
+
+VAR(gridlookup, 0, 0, 1);
+VAR(passthroughcube, 0, 1, 1);
 
 void cursorupdate()
 {
@@ -302,12 +290,12 @@ void cursorupdate()
         target[i] = max(min(target[i], hdr.worldsize), 0);
     vec ray(target);
     ray.sub(player->o).normalize();
-    int g2  = gridsize/2,
-        d   = dimension(sel.orient),
+    int d   = dimension(sel.orient),
         od  = dimension(orient),
         odc = dimcoord(orient);
 
     bool hovering = false;
+    hmapsel = false;
            
     if(moving)
     {       
@@ -330,91 +318,103 @@ void cursorupdate()
     {
         entdrag(ray);       
     }
-    else if(hovering = hoveringonent(player->o, ray))
-    {
-       if(!havesel) {
-           selchildcount = 0;
-           sel.s = vec(0);
-       }
-    }
     else
     {  
         vec v;
         ivec w;
         float sdist = 0, wdist = 0, t;
+        int entorient = 0, ent = -1;
        
-        wdist = raycubepos(player->o, ray, v, 0,  (editmode && showmat ? RAY_EDITMAT : 0)   // select cubes first
-                                                | (passthroughcube ? RAY_PASS : 0)
-                                                | RAY_SKIPFIRST, gridsize);
-        if(dragging) // update selection 
-        {
-            w = v;
-            lookupcube(w.x, w.y, w.z);
-            normalizelookupcube(w.x, w.y, w.z);
-            cur = lu;
-            updateselection();
-        }
-
-        if(havesel && !passthroughsel)     // now try selecting the selection
+        wdist = rayent(player->o, ray, v, 0, (editmode && showmat ? RAY_EDITMAT : 0)   // select cubes first
+                                           | (!dragging && entediting ? RAY_ENTS : 0)
+                                           | RAY_SKIPFIRST 
+                                           | (passthroughcube==1 ? RAY_PASS : 0), gridsize, entorient, ent);
+     
+        if((havesel || dragging) && !passthroughsel)     // now try selecting the selection
             if(rayrectintersect(sel.o.tovec(), vec(sel.s.tovec()).mul(sel.grid), player->o, ray, sdist, orient))
-                wdist = min(sdist, wdist);  // and choose the nearest of the two
-
-        v = ray;
-        v.mul(wdist+0.1f);
-        v.add(player->o);
-        w = v;
-        lookupcube(w.x, w.y, w.z);
-        int mag = lusize / gridsize;
-        normalizelookupcube(w.x, w.y, w.z);
-        if(sdist == 0 || sdist > wdist) rayrectintersect(lu.tovec(), vec(gridsize), player->o, ray, t=0, orient); // just getting orient     
-        cur = lu;
-        cor = w;
-        cor.div(g2);
-        od = dimension(orient);
-        d = dimension(sel.orient);
-
-        if(dragging) 
-        {
-            updateselection();
-            sel.cx   = min(cor[R[d]], lastcor[R[d]]);
-            sel.cy   = min(cor[C[d]], lastcor[C[d]]);
-            sel.cxs  = max(cor[R[d]], lastcor[R[d]]);
-            sel.cys  = max(cor[C[d]], lastcor[C[d]]);
-
-            if(!selectcorners)
-            {
-                sel.cx &= ~1;
-                sel.cy &= ~1;
-                sel.cxs &= ~1;
-                sel.cys &= ~1;
-                sel.cxs -= sel.cx-2;
-                sel.cys -= sel.cy-2;
-            }
-            else
-            {
-                sel.cxs -= sel.cx-1;
-                sel.cys -= sel.cy-1;
+            {   // and choose the nearest of the two
+                if(sdist < wdist) 
+                {
+                    wdist = sdist;
+                    ent   = -1;
+                }
             }
 
-            sel.cx  &= 1;
-            sel.cy  &= 1;
-            havesel = true;
-        }
-        else if(!havesel && hmap == NULL)
+        if(hovering = hoveringonent(ent, entorient))
         {
-            sel.o = lu;
-            sel.s.x = sel.s.y = sel.s.z = 1;
-            sel.cx = sel.cy = 0;
-            sel.cxs = sel.cys = 2;
-            sel.grid = gridsize;
-            sel.orient = orient;
-            d = od;
+           if(!havesel) {
+               selchildcount = 0;
+               sel.s = vec(0);
+           }
         }
+        else 
+        {
+       
+            v = ray;
+            v.mul(wdist+0.1f);
+            v.add(player->o);
+            w = v;
+            cube *c = &lookupcube(w.x, w.y, w.z);            
+            if(gridlookup && !dragging && !moving && !havesel && hmapedit!=1) gridsize = lusize;
+            int mag = lusize / gridsize;
+            normalizelookupcube(w.x, w.y, w.z);
+            if(sdist == 0 || sdist > wdist) rayrectintersect(lu.tovec(), vec(gridsize), player->o, ray, t=0, orient); // just getting orient     
+            cur = lu;
+            cor = w;
+            cor.mul(2).div(gridsize);
+            od = dimension(orient);
+            d = dimension(sel.orient);
+            
+            if(mag > 0 && hmapedit==1 && dimcoord(horient) == ray[dimension(horient)]<0)
+            {
+                hmapsel = isheightmap(horient, dimension(horient), false, c);     
+                if(hmapsel)
+                    od = dimension(orient = horient);
+            }
 
-        sel.corner = (cor[R[d]]-lu[R[d]]/g2)+(cor[C[d]]-lu[C[d]]/g2)*2;
-        selchildcount = 0;
-        countselchild(worldroot, vec(0), hdr.worldsize/2);
-        if(mag>1 && selchildcount==1) selchildcount = -mag;
+            if(dragging) 
+            {
+                updateselection();
+                sel.cx   = min(cor[R[d]], lastcor[R[d]]);
+                sel.cy   = min(cor[C[d]], lastcor[C[d]]);
+                sel.cxs  = max(cor[R[d]], lastcor[R[d]]);
+                sel.cys  = max(cor[C[d]], lastcor[C[d]]);
+
+                if(!selectcorners)
+                {
+                    sel.cx &= ~1;
+                    sel.cy &= ~1;
+                    sel.cxs &= ~1;
+                    sel.cys &= ~1;
+                    sel.cxs -= sel.cx-2;
+                    sel.cys -= sel.cy-2;
+                }
+                else
+                {
+                    sel.cxs -= sel.cx-1;
+                    sel.cys -= sel.cy-1;
+                }
+
+                sel.cx  &= 1;
+                sel.cy  &= 1;
+                havesel = true;
+            }
+            else if(!havesel)
+            {
+                sel.o = lu;
+                sel.s.x = sel.s.y = sel.s.z = 1;
+                sel.cx = sel.cy = 0;
+                sel.cxs = sel.cys = 2;
+                sel.grid = gridsize;
+                sel.orient = orient;
+                d = od;
+            }
+
+            sel.corner = (cor[R[d]]-(lu[R[d]]*2)/gridsize)+(cor[C[d]]-(lu[C[d]]*2)/gridsize)*2;
+            selchildcount = 0;
+            countselchild(worldroot, vec(0), hdr.worldsize/2);
+            if(mag>1 && selchildcount==1) selchildcount = -mag;
+        }
     }
     
     glEnable(GL_BLEND);
@@ -424,40 +424,39 @@ void cursorupdate()
 
     renderentselection(player->o, ray, entmoving!=0);
 
+    glEnable(GL_POLYGON_OFFSET_LINE);
+
     if(!moving && !hovering)
     {
-        glColor3ub(120,120,120);
+        if(hmapedit==1)
+            glColor3ub(0, hmapsel ? 255 : 40, 0);
+        else
+            glColor3ub(120,120,120);
         boxs(orient, lu.tovec(), vec(lusize));
     }
 
     // selections
-    if(hmap != NULL)
-    {
-        d = dimension(sel.orient);
-        glColor3ub(0,200,0);
-        boxs3D(sel.o.tovec(), sel.s.tovec(), sel.grid);
-        glColor3ub(20,20,20);   // grid
-        boxsgrid(sel.orient, sel.o.tovec(), sel.s.tovec(), sel.grid);
-    }
-    else if(havesel)
+    if(havesel)
     {
         d = dimension(sel.orient);
         glColor3ub(50,50,50);   // grid
         boxsgrid(sel.orient, sel.o.tovec(), sel.s.tovec(), sel.grid);
         glColor3ub(200,0,0);    // 0 reference
-        boxs3D(sel.o.tovec().sub(1), vec(2), 1);
+        boxs3D(sel.o.tovec().sub(0.5f*min(gridsize*0.25f, 2)), vec(min(gridsize*0.25f, 2)), 1);
         glColor3ub(200,200,200);// 2D selection box
         vec co(sel.o.v), cs(sel.s.v);
-        co[R[d]] += sel.cx*g2;
-        co[C[d]] += sel.cy*g2;
-        cs[R[d]]  = sel.cxs*g2;
-        cs[C[d]]  = sel.cys*g2;       
+        co[R[d]] += 0.5f*(sel.cx*gridsize);
+        co[C[d]] += 0.5f*(sel.cy*gridsize);
+        cs[R[d]]  = 0.5f*(sel.cxs*gridsize);
+        cs[C[d]]  = 0.5f*(sel.cys*gridsize);       
         cs[D[d]] *= gridsize;
         boxs(sel.orient, co, cs);
         glColor3ub(0,0,120);     // 3D selection box
         boxs3D(sel.o.tovec(), sel.s.tovec(), sel.grid);
     }
-    
+   
+    glDisable(GL_POLYGON_OFFSET_LINE);
+
     glDisable(GL_BLEND);
 }
 
@@ -481,7 +480,7 @@ void readychanges(block3 &b, cube *c, const ivec &cor, int size)
         }
         if(c[i].children)
         {
-            if(size<=4)
+            if(size<=(8>>VVEC_FRAC))
             {
                 solidfaces(c[i]);
                 discardchildren(c[i]);
@@ -512,6 +511,7 @@ void changed(const block3 &sel)
     octarender();
     inbetweenframes = true;
     setupmaterials();
+    invalidatereflections();
     entitiesinoctanodes();
 }
 
@@ -618,30 +618,33 @@ void makeundo()                        // stores state of selected cubes before
     lastsel=sel;
     if(multiplayer(false)) return;
     undoblock u;
-    initundocube(u, sel);
+	initundocube(u, sel);
     addundo(u);
 }
 
 void swapundo(vector<undoblock> &a, vector<undoblock> &b, const char *s)
 {
     if(noedit() || multiplayer()) return;
-    if(a.empty()) { conoutf("nothing more to %s", s); return; }
-    undoblock u = a.pop();
-    if(u.b)
-    {
-        sel.o = u.b->o;
-        sel.s = u.b->s;
-        sel.grid = u.b->grid;
-        sel.orient = u.b->orient;
-    }
-    undoblock r;
-    if(u.g) initundocube(r, sel);
-    if(u.n) copyundoents(r, u);
-    b.add(r);
-    pasteundo(u);
-    if(u.b) changed(sel);
-    freeundo(u);
-    clearheightmap();
+    if(a.empty()) { conoutf("nothing more to %s", s); return; }	
+	int ts = a.last().ts;
+	while(!a.empty() && ts==a.last().ts)
+	{
+		undoblock u = a.pop();
+		if(u.b)
+		{
+			sel.o = u.b->o;
+			sel.s = u.b->s;
+			sel.grid = u.b->grid;
+			sel.orient = u.b->orient;
+		}
+		undoblock r;
+		if(u.g) initundocube(r, sel);
+		if(u.n) copyundoents(r, u);
+		b.add(r);
+		pasteundo(u);
+		if(u.b) changed(sel);
+		freeundo(u);
+	}    
     reorient();
     forcenextundo();
 }
@@ -649,7 +652,7 @@ void swapundo(vector<undoblock> &a, vector<undoblock> &b, const char *s)
 void editundo() { swapundo(undos, redos, "undo"); }
 void editredo() { swapundo(redos, undos, "redo"); }
 
-editinfo *localedit=NULL;
+editinfo *localedit = NULL;
 
 void freeeditinfo(editinfo *&e)
 {
@@ -682,7 +685,7 @@ void mppaste(editinfo *&e, selinfo &sel, bool local)
         int o = sel.orient;
         sel.orient = e->copy->orient;
         cube *s = e->copy->c();
-        loopselxyz(pastecube(*s++, c));
+        loopselxyz(if (!isempty(*s) || s->children) pastecube(*s, c); s++); // 'transparent'. old opaque by 'delcube; paste'
         sel.orient = o;
     }
 }
@@ -693,324 +696,392 @@ void copy()
     mpcopy(localedit, sel, true);
 }
 
-void paste(int *isdown)
+void pastehilite()
+{
+    if(!localedit) return;
+	sel.s = localedit->copy->s;
+    reorient();
+    havesel = true;
+}
+
+void paste()
 {
     if(noedit()) return;
-    if(*isdown!=0 && localedit && localedit->copy)
-    {
-        sel.s = localedit->copy->s;
-        havesel = true;
-    }
-    else if(havesel)
-        mppaste(localedit, sel, true);
+    mppaste(localedit, sel, true);
 }
 
 COMMAND(copy, "");
-COMMAND(paste, "D");
+COMMAND(pastehilite, "");
+COMMAND(paste, "");
 COMMANDN(undo, editundo, "");
 COMMANDN(redo, editredo, "");
 
 ///////////// height maps ////////////////
 
-void pushside(cube &c, int d, int x, int y, int z)
-{
-    ivec a;
-    getcubevector(c, d, x, y, z, a);
-    a[R[d]] = 8 - a[R[d]];
-    setcubevector(c, d, x, y, z, a);
-}
-
-void setheightmap(selinfo &b)
-{
-    int d = dimension(sel.orient);
-    int dc = dimcoord(sel.orient);
-    int w = sel.s[R[d]] + 1;
-    int h = (dc ? sel.s[D[d]] : 0) + (dc ? sel.o[D[d]] : hdr.worldsize - sel.o[D[d]]) / sel.grid;
-
-    makeundo();
-
-    b.o[R[d]]  = max(sel.o[R[d]], b.o[R[d]]-sel.grid);
-    b.o[C[d]]  = max(sel.o[C[d]], b.o[C[d]]-sel.grid);
-    int sx     = (b.o[R[d]]-sel.o[R[d]])/sel.grid;
-    int sy     = (b.o[C[d]]-sel.o[C[d]])/sel.grid;
-    b.s[R[d]]  = max(0, min(b.s[R[d]]-sx+2, sel.s[R[d]]-sx));
-    b.s[C[d]]  = max(0, min(b.s[C[d]]-sy+2, sel.s[C[d]]-sy));
-
-    loopxyz(b, b.grid,
-
-        if(c.children)
-            discardchildren(c);
-        solidfaces(c);
-
-        int e[2][2];
-
-        loopi(2) loopj(2)
-            e[i][j] = min(8, hmap[x+sx+i+(y+sy+j)*w] - (h-z-1)*8);
-
-        loopi(2) loopj(2)
-        {
-            int f = e[i][j];
-            if(f<0 || (f==0 && e[1-i][j]==0 && e[i][1-j]==0))
-            {
-                f=0;
-                pushside(c, d, i, j, 0);
-                pushside(c, d, i, j, 1);
-            }
-            edgeset(cubeedge(c, d, i, j), dc, dc ? f : 8-f);
-        }
-
-        c.texture[sel.orient] = (htex ? htex[x+sx+(y+sy)*w] : htexture);
-        optiface((uchar *)&c.faces[d], c);
-    );
-
-    changed(b);
-}
-
-void extendchangedhmap(selinfo &b, int x, int y, int d) 
-{
-    b.o[R[d]] = min(b.o[R[d]], x*sel.grid+sel.o[R[d]]);
-    b.o[C[d]] = min(b.o[C[d]], y*sel.grid+sel.o[C[d]]);
-    b.s[R[d]] = max(b.s[R[d]], x+1);
-    b.s[C[d]] = max(b.s[C[d]], y+1);
-}
-
-void cubifyheightmap(selinfo &b, bool downbias)     // pull up heighfields to where they don't cross cube boundaries
-{
-    int d = dimension(sel.orient);
-    int w = sel.s[R[d]] + 1;
-    int l = sel.s[C[d]] + 1;
-    for(;;)
-    {
-        bool changed = false;
-        loop(x, w-1)
-        {
-            loop(y, l-1)
-            {
-                int *o[4], best, par, q = 0;
-                loopi(2) loopj(2) o[i+j*2] = hmap+x+i+(y+j)*w;
-
-                // to avoid infinite loops, we must always ensure:
-                // a) all changes satisfies the reason it needed to be pulled
-                // b) all changes are done in the same direction
-                #define pullhmap(I, LT, GT, M, N, A) { \
-                    best = I; \
-                    loopi(4) if(*o[i] LT best) best = *o[q = i] - M; \
-                    par = (best&(~7)) + N; \
-                    /* dual layer for extra smoothness */ \
-                    if(*o[q^3] GT par && !(*o[q^1] LT par || *o[q^2] LT par)) { \
-                        if(*o[q^3] GT par A 8 || *o[q^1] != par || *o[q^2] != par) { \
-                            *o[q^3] = (*o[q^3] GT par A 8 ? par A 8 : *o[q^3]); \
-                            *o[q^1] = *o[q^2] = par; \
-                            changed = true; \
-                            extendchangedhmap(b, x, y, d); \
-                        } \
-                    /* single layer */ \
-                    } else { \
-                        loopj(4) if(*o[j] GT par) { \
-                            *o[j] = par; \
-                            changed = true; \
-                            extendchangedhmap(b, x, y, d); \
-                        } \
-                    } \
-                }
-                
-                if(downbias) {
-                    pullhmap(hdr.worldsize, <, >, 0, 8, +);
-                }
-                else
-                    pullhmap(0, >, <, 1, 0, -);
-            }
-        }
-        if(!changed) break;
-    }
-}
-
-void getlimits(int &d, int &dc, int &w, int &l, int &lo, int &hi, int &himax)
-{
-    d = dimension(sel.orient);
-    dc = dimcoord(sel.orient);
-    w = sel.s[R[d]] + 1;
-    l = sel.s[C[d]] + 1;
-    lo = 8 * (sel.o[D[d]] / sel.grid);
-    hi = 8 * (sel.s[D[d]]) + lo;
-    himax = 8 * hdr.worldsize / sel.grid;
-
-    if(!dc)
-    {
-        swap(int, hi, lo);
-        hi = himax - hi;
-        lo = himax - lo;
-    }
-}
-
-void createheightmap()
-{
-    int d, dc, w, l, lo, hi, himax;
-    getlimits(d, dc, w, l, lo, hi, himax);
-
-    clearheightmap();
-    hmap = new int[w*l];
-    htex = new ushort[w*l];
-    loop(x, w) loop(y, l)
-    {
-        hmap[x+y*w] = lo;
-        htex[x+y*w] = 0;
-    }
-
-    int h = hi / 8;
-    selinfo b(sel);
-    makeundo();
-    loopxyz(b, b.grid,
-        if(c.children) { solidfaces(c); discardchildren(c); }
-        if(!htex[x+y*w] && z == sel.s[D[d]]-1) htex[x+y*w] = c.texture[sel.orient];
-        if(isempty(c)) continue;
-        if(!htex[x+y*w]) htex[x+y*w] = c.texture[sel.orient];
-
-        loopi(2) loopj(2)
-        {
-            int a = x+i+(y+j)*w;
-            int e = edgeget(cubeedge(c, d, i, j), dc);
-            e = (h-z-1)*8 + (dc ? e : 8-e);
-            hmap[a] = max(hmap[a], e);// simply take the heighest points
-        }
-    );
-    cubifyheightmap(b, !dc);
-    setheightmap(b);
-}
-
-const int MAXBRUSH = 50;
+#define MAXBRUSH    64
+#define MAXBRUSHC   63
+#define MAXBRUSH2   32
 int brush[MAXBRUSH][MAXBRUSH];
-VAR(brushx, 0, 25, MAXBRUSH);
-VAR(brushy, 0, 25, MAXBRUSH);
-int brushmaxx = 0;
-int brushmaxy = 0;
+VAR(brushx, 0, MAXBRUSH2, MAXBRUSH);
+VAR(brushy, 0, MAXBRUSH2, MAXBRUSH);
+bool paintbrush = 0;
+int brushmaxx = 0, brushminx = MAXBRUSH;
+int brushmaxy = 0, brushminy = MAXBRUSH;
 
 void clearbrush()
 {
-    loopi(MAXBRUSH) loopj(MAXBRUSH)
-        brush[i][j] = 0;
-    brushmaxx = 0;
-    brushmaxy = 0;
+    memset(brush, 0, sizeof brush);
+    brushmaxx = brushmaxy = 0;
+    brushminx = brushminy = MAXBRUSH;
+    paintbrush = false;
 }
 
 void brushvert(int *x, int *y, int *v)
 {
+    *x += MAXBRUSH2 - brushx + 1; // +1 for automatic padding
+    *y += MAXBRUSH2 - brushy + 1;
     if(*x<0 || *y<0 || *x>=MAXBRUSH || *y>=MAXBRUSH) return;
-    brush[*x][*y] = *v;
-    brushmaxx = max(brushmaxx, *x+1);
-    brushmaxy = max(brushmaxy, *y+1);
+    brush[*x][*y] = clamp(*v, 0, 8);
+    paintbrush = paintbrush || (brush[*x][*y] > 0);
+    brushmaxx = min(MAXBRUSH-1, max(brushmaxx, *x+1));
+    brushmaxy = min(MAXBRUSH-1, max(brushmaxy, *y+1));
+    brushminx = max(0,          min(brushminx, *x-1));
+    brushminy = max(0,          min(brushminy, *y-1));
 }
 
-int getxcursor() { int d = dimension(sel.orient); return (cur[R[d]] - sel.o[R[d]]) / sel.grid + (sel.corner&1 ? 1 : 0); }
-int getycursor() { int d = dimension(sel.orient); return (cur[C[d]] - sel.o[C[d]]) / sel.grid + (sel.corner&2 ? 1 : 0); }
-
-void copybrush()
-{
-    if(hmap == NULL) return;
-    int d, dc, w, l, lo, hi, himax; getlimits(d, dc, w, l, lo, hi, himax);
-    if(w>=MAXBRUSH || l>=MAXBRUSH) return conoutf("Selection is too big to generate brush");
-    clearbrush();
-    loop(x, w) loop(y, l)
-        brush[x][y] = hi - hmap[x+y*w];
-    brushx = max(0, min(MAXBRUSH, getxcursor()));
-    brushy = max(0, min(MAXBRUSH, getycursor()));
-}
-
-void savebrush(const char *name)
-{
-    FILE *f = fopen("mybrushes.cfg", "a");
-    if(!f) return;
-    fprintf(f, "newbrush \"%s\" %d %d [\n", name, brushx, brushy);
-    int skipped = 0;
-    loop(y, MAXBRUSH)
-    {
-        int last = 0;
-        loop(x, MAXBRUSH) if(brush[x][y]!=0) last = x+1;
-        if(!last) { skipped++; continue; }
-        while(skipped) { fprintf(f, "\"\"\n"); skipped--; }
-        fprintf(f, "\"");
-        loop(x, last)
-            fprintf(f, "%d ", brush[x][y]);
-        fprintf(f, "\"\n");
-    }
-    fprintf(f, "]\n\n");
-    conoutf("Brush \"%s\" saved", name);
-    fclose(f);
-}
+vector<int> htextures;
 
 COMMAND(clearbrush, "");
 COMMAND(brushvert, "iii");
-COMMAND(copybrush, "");
-COMMAND(savebrush, "s");
+ICOMMAND(hmapcancel, "", (), htextures.setsizenodelete(0); );
+ICOMMAND(hmapselect, "", (), 
+    int t = lookupcube(cur.x, cur.y, cur.z).texture[orient];
+    int i = htextures.find(t);
+    if(i<0)
+        htextures.add(t);
+    else
+        htextures.remove(i);
+);
 
-void edithmap(int dir, int mode)
+bool ischildless(cube &c)
 {
-    if(multiplayer() || hmap == NULL) return;
-
-    int d, dc, w, l, lo, hi, himax;
-    getlimits(d, dc, w, l, lo, hi, himax);
-    int x = getxcursor();
-    int y = getycursor();
-
-    if(x<0 || y<0 || x>=w || y>=l) return;
-
-    int bx = x>brushx ? 0 : brushx - x;
-    int by = y>brushy ? 0 : brushy - y;
-    int sx = x>brushx ? brushx : x;
-    int sy = y>brushy ? brushy : y;
-    int ex = min(brushmaxx - brushx, w - x);
-    int ey = min(brushmaxy - brushy, l - y);
-
-    loopi(sx+ex) loopj(sy+ey)
+    if(!c.children)
+        return true;
+    loopi(8)
     {
-        int index = x+i-sx+(y+j-sy)*w;
-        hmap[index] -= brush[i+bx][j+by]*dir;
-        hmap[index] = max(0, min(hmap[index], himax));
-        while(hmap[index] > hi) { hi += 8; sel.s[D[d]] += 1; if(!dc) sel.o[D[d]] -= sel.grid; }
-        while(hmap[index] < lo) { lo -= 8; sel.s[D[d]] += 1; if(dc)  sel.o[D[d]] -= sel.grid; }
+        if(!ischildless(c.children[i]) || !isempty(c.children[i]))
+            return false;
+    }
+    emptyfaces(c);
+    discardchildren(c);    
+    return true;
+}
+
+inline bool ishtexture(int t)
+{    
+    loopv(htextures)
+        if(t == htextures[i])
+            return false;
+    return true;
+}
+
+VARP(bypassheightmapcheck, 0, 0, 1);    // temp
+
+inline bool isheightmap(int o, int d, bool empty, cube *c) 
+{
+    return bypassheightmapcheck || (ischildless(*c) && 
+           ( (empty && isempty(*c)) ||
+           (         
+            (c->faces[R[d]] & 0x77777777) == 0 &&
+            (c->faces[C[d]] & 0x77777777) == 0 &&
+            ishtexture(c->texture[o])
+           )));
+}
+
+namespace hmap 
+{
+#   define PAINTED     1
+#   define NOTHMAP     2
+#   define MAPPED      16
+    uchar  flags[MAXBRUSH][MAXBRUSH];
+    cube   *cmap[MAXBRUSHC][MAXBRUSHC][4];
+    int    mapz[MAXBRUSHC][MAXBRUSHC];    
+    int    map [MAXBRUSH][MAXBRUSH];
+    
+    selinfo changes;    
+    bool selecting;
+    int d, dc, dr, dcr, biasup, br, hws, fg;
+    int gx, gy, gz, mx, my, mz, nx, ny, nz, bmx, bmy, bnx, bny;
+    uint fs;
+    
+    cube *getcube(ivec t, int f) 
+    {
+        t[d] += dcr*f*gridsize;    
+        if(t[d] > nz || t[d] < mz) return NULL;
+        cube *c = &lookupcube(t.x, t.y, t.z, -gridsize);
+        if(!isheightmap(sel.orient, d, true, c)) return NULL;        
+        if(lusize > gridsize)
+            c = &lookupcube(t.x, t.y, t.z, gridsize);
+        discardchildren(*c);    
+        if     (t.x < changes.o.x) changes.o.x = t.x;
+        else if(t.x > changes.s.x) changes.s.x = t.x;
+        if     (t.y < changes.o.y) changes.o.y = t.y;
+        else if(t.y > changes.s.y) changes.s.y = t.y;
+        if     (t.z < changes.o.z) changes.o.z = t.z;
+        else if(t.z > changes.s.z) changes.s.z = t.z;
+        return c;
     }
 
-    selinfo b;
-    b.o[D[d]] = sel.o[D[d]];
-    b.s[D[d]] = sel.s[D[d]];
-    b.o[C[d]] = sel.o[C[d]]+(y-sy)*sel.grid;
-    b.o[R[d]] = sel.o[R[d]]+(x-sx)*sel.grid;
-    b.s[C[d]] = y+ey;
-    b.s[R[d]] = x+ex;
-    b.grid   = sel.grid;
-    b.orient = sel.orient;
-    cubifyheightmap(b, mode == dir>0);
-    setheightmap(b);
-}
+    uint getface(cube *c, int d)
+    {
+        return  0x0f0f0f0f & ((dc ? c->faces[d] : 0x88888888 - c->faces[d]) >> fs);
+    }
 
-void smoothmap()
-{    
-    if(multiplayer() || hmap == NULL) return;
-    int d = dimension(sel.orient);
-    int w = sel.s[R[d]] + 1;
-    int l = sel.s[C[d]] + 1;
-    for(int x=1; x<w-2; x++) 
-        for(int y=1; y<l-2; y++) 
-        {
-            int i = x+(y*w);
-            hmap[i] = (hmap[i+1] + hmap[i-1] + hmap[i] +hmap[i+w] + hmap[i-w]) / 5;    
+    void pushside(cube &c, int d, int x, int y, int z)
+    {
+        ivec a;
+        getcubevector(c, d, x, y, z, a);
+        a[R[d]] = 8 - a[R[d]];
+        setcubevector(c, d, x, y, z, a);
+    }
+
+    void addpoint(int x, int y, int z, int v)
+    {
+        if(!(flags[x][y] & MAPPED))
+          map[x][y] = v + (z*8);
+      flags[x][y] |= MAPPED;
+    }
+    
+    void select(int x, int y, int z)
+    {        
+        if((NOTHMAP & flags[x][y]) || (PAINTED & flags[x][y])) return;
+        ivec t(d, x+gx, y+gy, dc ? z : hws-z);
+        t.shl(gridpower);
+        
+        cube **c = cmap[x][y];     
+        loopk(4) c[k] = NULL;
+        c[1] = getcube(t, 0);
+        if(!c[1] || !isempty(*c[1])) 
+        {   // try up             
+            c[2] = c[1];
+            c[1] = getcube(t, 1);
+            if(!c[1] || isempty(*c[1])) {
+                c[0] = c[1], c[1] = c[2], c[2] = NULL;
+            }else
+                z++, t[d]+=fg;
         }
-    cubifyheightmap(sel, !dimcoord(sel.orient));
-    setheightmap(sel);
+        else // drop down
+        { 
+            z--; 
+            t[d]-= fg; 
+            c[0] = c[1];
+            c[1] = getcube(t, 0);            
+        }
+        
+        if(!c[1] || isempty(*c[1])) { flags[x][y] |= NOTHMAP; return; }
+
+        flags[x][y] |= PAINTED;
+        mapz [x][y]  = z;
+        
+        if(!c[0]) c[0] = getcube(t, 1);
+        if(!c[2]) c[2] = getcube(t, -1);
+        c[3] = getcube(t, -2);
+        c[2] = !c[2] || isempty(*c[2]) ? NULL : c[2];
+        c[3] = !c[3] || isempty(*c[3]) ? NULL : c[3];
+        
+        uint face = getface(c[1], d);          
+        if(face == 0x08080808 && (!c[0] || !isempty(*c[0]))) { flags[x][y] |= NOTHMAP; return; }              
+        if(c[1]->faces[R[d]] == F_SOLID)   // was single
+            face += 0x08080808;      
+        else                               // was pair
+            face += c[2] ? getface(c[2], d) : 0x08080808;
+        face += 0x08080808;                // c[3]        
+        uchar *f = (uchar*)&face;
+        addpoint(x,   y,   z, f[0]);
+        addpoint(x+1, y,   z, f[1]);
+        addpoint(x,   y+1, z, f[2]);
+        addpoint(x+1, y+1, z, f[3]);
+                
+        if(selecting) // continue to adjacent cubes
+        {        
+            if(x>bmx) select(x-1, y, z);
+            if(x<bnx) select(x+1, y, z);
+            if(y>bmy) select(x, y-1, z);
+            if(y<bny) select(x, y+1, z);
+        }
+    }       
+
+    void ripple(int x, int y, int z, bool force)
+    {
+        if(force) select(x, y, z);
+        if((NOTHMAP & flags[x][y]) || !(PAINTED & flags[x][y])) return;
+
+        bool changed = false;
+        int *o[4], best, par, q = 0;
+        loopi(2) loopj(2) o[i+j*2] = &map[x+i][y+j];
+        #define pullhmap(I, LT, GT, M, N, A) do { \
+            best = I; \
+            loopi(4) if(*o[i] LT best) best = *o[q = i] - M; \
+            par = (best&(~7)) + N; \
+            /* dual layer for extra smoothness */ \
+            if(*o[q^3] GT par && !(*o[q^1] LT par || *o[q^2] LT par)) { \
+                if(*o[q^3] GT par A 8 || *o[q^1] != par || *o[q^2] != par) { \
+                    *o[q^3] = (*o[q^3] GT par A 8 ? par A 8 : *o[q^3]); \
+                    *o[q^1] = *o[q^2] = par; \
+                    changed = true; \
+                } \
+            /* single layer */ \
+            } else { \
+                loopj(4) if(*o[j] GT par) { \
+                    *o[j] = par; \
+                    changed = true; \
+                } \
+            } \
+        } while(0)
+        
+        if(biasup)
+            pullhmap(0, >, <, 1, 0, -);
+        else
+            pullhmap(hdr.worldsize, <, >, 0, 8, +);     
+   
+        cube **c  = cmap[x][y];
+        int e[2][2];
+        int notempty = 0;
+
+        loopk(4) if(c[k]) {
+            loopi(2) loopj(2) {
+                e[i][j] = min(8, map[x+i][y+j] - (mapz[x][y]+3-k)*8);
+                notempty |= e[i][j] > 0;         
+            }
+            if(notempty) 
+            {
+                c[k]->texture[sel.orient] = c[1]->texture[sel.orient];
+                solidfaces(*c[k]);
+                loopi(2) loopj(2)
+                {
+                    int f = e[i][j];
+                    if(f<0 || (f==0 && e[1-i][j]==0 && e[i][1-j]==0))
+                    {
+                        f=0;
+                        pushside(*c[k], d, i, j, 0);
+                        pushside(*c[k], d, i, j, 1);
+                    }
+                    edgeset(cubeedge(*c[k], d, i, j), dc, dc ? f : 8-f);
+                }
+            }
+            else 
+                emptyfaces(*c[k]);
+        }
+
+        if(!changed) return;
+        if(x>mx) ripple(x-1, y, mapz[x][y], true);
+        if(x<nx) ripple(x+1, y, mapz[x][y], true);
+        if(y>my) ripple(x, y-1, mapz[x][y], true);
+        if(y<ny) ripple(x, y+1, mapz[x][y], true);    
+               
+#define DIAGONAL_RIPPLE(a,b,exp) if(exp) { \
+            if(flags[x a][ y] & PAINTED) \
+                ripple(x a, y b, mapz[x a][y], true); \
+            else if(flags[x][y b] & PAINTED) \
+                ripple(x a, y b, mapz[x][y b], true); \
+        }
+        
+        DIAGONAL_RIPPLE(-1, -1, (x>mx && y>my)); // do diagonals because adjacents
+        DIAGONAL_RIPPLE(-1, +1, (x>mx && y<ny)); //    won't unless changed
+        DIAGONAL_RIPPLE(+1, +1, (x<nx && y<ny));
+        DIAGONAL_RIPPLE(+1, -1, (x<nx && y>my));
+    }
+
+#define loopbrush(i) for(int x=bmx; x<=bnx+i; x++) for(int y=bmy; y<=bny+i; y++)
+
+    void paint()
+    {
+        loopbrush(1)
+            map[x][y] -= dr * brush[x][y];        
+    }
+
+    void smooth()
+    {
+        int sum, div;
+        loopbrush(-2)        
+        {
+            sum = 0;
+            div = 9;
+            loopi(3) loopj(3)
+                if(flags[x+i][y+j] & MAPPED)
+                    sum += map[x+i][y+j];
+                else div--;
+            if(div)
+                map[x+1][y+1] = sum / div;
+        }
+    }
+
+    void rippleandset()
+    {              
+        loopbrush(0)
+            ripple(x, y, gz, false);        
+    }
+
+    void run(int dir, int mode) 
+    {                 
+        d  = dimension(sel.orient);
+        dc = dimcoord(sel.orient);
+        dcr= dc ? 1 : -1;
+        dr = dir>0 ? 1 : -1;
+        br = dir>0 ? 0x08080808 : 0;
+     //   biasup = mode == dir<0;
+        biasup = dir<0;
+        int cx = (sel.corner&1 ? 0 : -1);
+        int cy = (sel.corner&2 ? 0 : -1);
+        hws= (hdr.worldsize>>gridpower);
+        gx = (cur[R[d]] >> gridpower) + cx - MAXBRUSH2;
+        gy = (cur[C[d]] >> gridpower) + cy - MAXBRUSH2;
+        gz = (cur[D[d]] >> gridpower);
+        fs = dc ? 4 : 0;  
+        fg = dc ? gridsize : -gridsize;
+        mx = max(0, -gx); // ripple range
+        my = max(0, -gy);
+        nx = min(MAXBRUSH-1, hws-gx) - 1; 
+        ny = min(MAXBRUSH-1, hws-gy) - 1; 
+        if(havesel)
+        {   // selection range
+            mx = max(mx, (sel.o[R[d]]>>gridpower)-gx);
+            my = max(my, (sel.o[C[d]]>>gridpower)-gy);
+            nx = min(nx, (sel.s[R[d]]+(sel.o[R[d]]>>gridpower))-gx-1);
+            ny = min(ny, (sel.s[C[d]]+(sel.o[C[d]]>>gridpower))-gy-1);
+        }
+        bmx = max(mx, brushminx); // brush range
+        bmy = max(my, brushminy);
+        bnx = min(nx, brushmaxx-1);
+        bny = min(ny, brushmaxy-1);   
+        nz = hdr.worldsize-gridsize;
+        mz = 0;
+                    
+        changes.grid = gridsize;
+        changes.s = changes.o = cur;
+        memset(map, 0, sizeof map);
+        memset(flags, 0, sizeof flags);
+        
+        selecting = true;
+        select(clamp(MAXBRUSH2-cx, bmx, bnx),
+               clamp(MAXBRUSH2-cy, bmy, bny),
+               dc ? gz : hws - gz);
+        selecting = false;
+        if(paintbrush)
+            paint();
+        else 
+            smooth();
+        rippleandset();                       // pull up points to cubify, and set
+        changes.s.sub(changes.o).shr(gridpower).add(1);
+        changed(changes);
+    }
 }
 
-void getheightmap()
-{
-    if(noedit() || multiplayer() || !havesel) return;
-    if(hmap == NULL)
-        createheightmap();
-    else 
-        smoothmap();
+void edithmap(int dir, int mode) {    
+    if(multiplayer() || !hmapsel || gridsize < 8) return;    
+    hmap::run(dir, mode);        
 }
 
-COMMAND(getheightmap, "");
-COMMAND(smoothmap, "");
-
 ///////////// main cube edit ////////////////
 
 int bounded(int n) { return n<0 ? 0 : (n>8 ? 8 : n); }
@@ -1131,10 +1202,10 @@ void mpeditface(int dir, int mode, selinfo &sel, bool local)
 void editface(int *dir, int *mode)
 {
     if(noedit(moving!=0)) return;
-    if(hmap)
-        edithmap(*dir, *mode);
-    else
-        mpeditface(*dir, *mode, sel, true);
+    if(hmapedit!=1)
+        mpeditface(*dir, *mode, sel, true);        
+    else 
+        edithmap(*dir, *mode);       
 }
 
 VAR(selectionsurf, 0, 0, 1);
@@ -1168,6 +1239,7 @@ COMMAND(delcube, "");
 
 int curtexindex = -1, lasttex = 0;
 int texpaneltimer = 0;
+vector<ushort> texmru;
 
 void tofronttex()                                       // maintain most recently used of the texture lists when applying texture
 {
@@ -1222,18 +1294,20 @@ void filltexlist()
     }
 }
 
-void edittex(int *dir)
+void edittex(int i)
+{
+    curtexindex = i = min(max(i, 0), curtexnum-1);
+    int t = lasttex = texmru[i];    
+    mpedittex(t, allfaces, sel, true);
+}
+
+void edittex_(int *dir)
 {
     if(noedit()) return;
     filltexlist();
     texpaneltimer = 5000;
     if(!(lastsel==sel)) tofronttex();
-    int i = curtexindex;
-    i = i<0 ? 0 : i+*dir;
-    curtexindex = i = min(max(i, 0), curtexnum-1);
-    int t = lasttex = htexture = texmru[i];
-    clearheighttexture();
-    mpedittex(t, allfaces, sel, true);
+    edittex(curtexindex<0 ? 0 : curtexindex+*dir);
 }
 
 void gettex()
@@ -1249,7 +1323,7 @@ void gettex()
     }
 }
 
-COMMAND(edittex, "i");
+COMMANDN(edittex, edittex_, "i");
 COMMAND(gettex, "");
 
 void replacetexcube(cube &c, int oldtex, int newtex)
@@ -1344,7 +1418,7 @@ void mpflip(selinfo &sel, bool local)
 
 void flip()
 {
-    if(noedit() || hmap!=NULL) return;
+    if(noedit()) return;
     mpflip(sel, true);
 }
 
@@ -1372,7 +1446,7 @@ void mprotate(int cw, selinfo &sel, bool local)
 
 void rotate(int *cw)
 {
-    if(noedit() || hmap!=NULL) return;
+    if(noedit()) return;
     mprotate(*cw, sel, true);
 }
 
@@ -1405,7 +1479,7 @@ COMMAND(editmat, "s");
 
 #define TEXTURE_WIDTH 10
 #define TEXTURE_HEIGHT 7
-extern int menudistance;
+extern int menudistance, menuautoclose;
 
 VAR(thumbtime, 0, 50, 1000);
 
@@ -1434,20 +1508,17 @@ struct texturegui : g3d_callback
                     int ti = (i*TEXTURE_HEIGHT+j)*TEXTURE_WIDTH+k;
                     if(ti<curtexnum) 
                     {
-                        Texture *tex = crosshair;
+                        Texture *tex = notexture;
                         Slot &slot = lookuptexture(texmru[ti], false);
                         if(slot.sts.empty()) continue;
                         else if(slot.loaded) tex = slot.sts[0].t;
                         else if(slot.thumbnail) tex = slot.thumbnail;
                         else if(lastmillis-lastthumbnail>=thumbtime) { tex = loadthumbnail(slot); lastthumbnail = lastmillis; }
-                        if(g.texture(tex, 1.0)&G3D_UP && (slot.loaded || tex!=crosshair)) 
-                        {
-                            curtexindex = ti;
-                            mpedittex(texmru[ti], allfaces, sel, true); 
-                        }
+                        if(g.texture(tex, 1.0)&G3D_UP && (slot.loaded || tex!=notexture)) 
+                            edittex(ti);
                     }
                     else
-                        g.texture(crosshair, 1.0); //create an empty space
+                        g.texture(notexture, 1.0); //create an empty space
                 }
                 g.poplist();
             }
@@ -1465,7 +1536,7 @@ struct texturegui : g3d_callback
     {   
         if(!menuon) return;
         filltexlist();
-        if(!editmode || camera1->o.dist(menupos) > menudistance*3) menuon = false;
+        if(!editmode || camera1->o.dist(menupos) > menuautoclose) menuon = false;
         else g3d_addgui(this, menupos); //follow?
     }
 } gui;
@@ -1475,7 +1546,11 @@ void g3d_texturemenu()
     gui.show(); 
 }
 
-void showtexgui(int *n) { gui.showtextures(((*n==0) ? !gui.menuon : (*n==1)) && editmode); }
+void showtexgui(int *n) 
+{ 
+    if(!editmode) { conoutf("operation only allowed in edit mode"); return; }
+    gui.showtextures(*n==0 ? !gui.menuon : *n==1); 
+}
 
 // 0/noargs = toggle, 1 = on, other = off - will autoclose if too far away or exit editmode
 COMMAND(showtexgui, "i");
