@@ -30,21 +30,25 @@ VARF(floatvtx, 0, 0, 1, allchanged());
             f->x = v.x; \
             f->y = v.y; \
             f->z = v.z; \
-            f->u = v.u; \
-            f->v = v.v; \
             body; \
             f++; \
         } \
     }
+#define GENVERTSUV(type, ptr, body) GENVERTS(type, ptr, { f->u = v.u; f->v = v.v; body; })
 
 void genverts(void *buf)
 {
     if(renderpath==R_FIXEDFUNCTION)
     {
-        if(floatvtx) { GENVERTS(fvertexff, buf, {}); }
-        else { GENVERTS(vertexff, buf, {}); }
+        if(nolights)
+        {
+            if(floatvtx) { GENVERTS(fvertexffc, buf, {}); }
+            else { GENVERTS(vertexffc, buf, {}); }
+        }
+        else if(floatvtx) { GENVERTSUV(fvertexff, buf, {}); }
+        else { GENVERTSUV(vertexff, buf, {}); }
     }
-    else if(floatvtx) { GENVERTS(fvertex, buf, { f->n = v.n; }); }
+    else if(floatvtx) { GENVERTSUV(fvertex, buf, { f->n = v.n; }); }
 }
 
 struct vboinfo
@@ -211,7 +215,7 @@ struct sortkey
 struct sortval
 {
      int unlit;
-     usvector dims[3];
+     usvector dims[6];
 
      sortval() : unlit(0) {}
 };
@@ -283,7 +287,7 @@ struct lodcollect
             LightMap &lm = lightmaps[t.unlit-LMID_RESERVED];
             short u = short((lm.unlitx + 0.5f) * SHRT_MAX/LM_PACKW), 
                   v = short((lm.unlity + 0.5f) * SHRT_MAX/LM_PACKH);
-            loopl(3) loopvj(t.dims[l])
+            loopl(6) loopvj(t.dims[l])
             {
                 vertex &vtx = verts[t.dims[l][j]];
                 if(!vtx.u && !vtx.v)
@@ -300,7 +304,7 @@ struct lodcollect
                 }
             }
             sortval *dst = indices.access(sortkey(k.tex, t.unlit, k.envmap));
-            if(dst) loopl(3) loopvj(t.dims[l]) dst->dims[l].add(t.dims[l][j]);
+            if(dst) loopl(6) loopvj(t.dims[l]) dst->dims[l].add(t.dims[l][j]);
         }
     }
                     
@@ -310,7 +314,7 @@ struct lodcollect
 
         texs.setsizenodelete(0);
         enumeratekt(indices, sortkey, k, sortval, t,
-            loopl(3) if(t.dims[l].length() && t.unlit<=0)
+            loopl(6) if(t.dims[l].length() && t.unlit<=0)
             {
                 if(k.lmid>=LMID_RESERVED && lightmaps[k.lmid-LMID_RESERVED].unlitx>=0)
                 {
@@ -381,14 +385,8 @@ struct lodcollect
 
             if(skybuf)
             {
-#if 1
                 if(offsetindices) loopi(lod.sky+lod.explicitsky) skybuf[i] += offsetindices;
                 lod.skybuf = (ushort *)addvbo(va, &lod==&va->l0 ? VBO_SKYBUF_L0 : VBO_SKYBUF_L1, skybuf, (lod.sky+lod.explicitsky)*sizeof(ushort));
-#else
-                glGenBuffers_(1, &lod.skybufGL);
-                glBindBuffer_(GL_ELEMENT_ARRAY_BUFFER_ARB, lod.skybufGL);
-                glBufferData_(GL_ELEMENT_ARRAY_BUFFER_ARB, (lod.sky+lod.explicitsky)*sizeof(ushort), skybuf, GL_STATIC_DRAW_ARB);
-#endif
                 delete[] skybuf;
             }
             else lod.skybufGL = 0;
@@ -404,25 +402,35 @@ struct lodcollect
             {
                 const sortkey &k = texs[i];
                 const sortval &t = indices[k];
-                lod.eslist[i].texture = k.tex;
-                lod.eslist[i].lmid = t.unlit>0 ? t.unlit : k.lmid;
-                lod.eslist[i].envmap = k.envmap;
-                loopl(3) if((lod.eslist[i].length[l] = t.dims[l].length()))
+                elementset &e = lod.eslist[i];
+                e.texture = k.tex;
+                e.lmid = t.unlit>0 ? t.unlit : k.lmid;
+                e.envmap = k.envmap;
+                ushort *startbuf = curbuf;
+                loopl(6) 
                 {
-                    memcpy(curbuf, t.dims[l].getbuf(), t.dims[l].length() * sizeof(ushort));
-                    curbuf += t.dims[l].length();
+                    e.minvert[l] = USHRT_MAX;
+                    e.maxvert[l] = 0;
+
+                    if(t.dims[l].length())
+                    {
+                        memcpy(curbuf, t.dims[l].getbuf(), t.dims[l].length() * sizeof(ushort));
+
+                        loopvj(t.dims[l])
+                        {
+                            curbuf[j] += offsetindices;
+                            e.minvert[l] = min(e.minvert[l], curbuf[j]);
+                            e.maxvert[l] = max(e.maxvert[l], curbuf[j]);
+                        }
+
+                        curbuf += t.dims[l].length();
+                    }
+                    e.length[l] = curbuf-startbuf;
                 }
             }
             if(hasVBO)
             {
-#if 1
-                if(offsetindices) loopi(3*curtris) ebuf[i] += offsetindices;
                 lod.ebuf = (ushort *)addvbo(va, &lod==&va->l0 ? VBO_EBUF_L0 : VBO_EBUF_L1, ebuf, 3*curtris*sizeof(ushort));
-#else
-                glGenBuffers_(1, &lod.ebufGL);
-                glBindBuffer_(GL_ELEMENT_ARRAY_BUFFER_ARB, lod.ebufGL);
-                glBufferData_(GL_ELEMENT_ARRAY_BUFFER_ARB, 3*curtris*sizeof(ushort), ebuf, GL_STATIC_DRAW_ARB);
-#endif
                 delete[] ebuf;
             }
         }
@@ -433,22 +441,23 @@ struct lodcollect
     }
 } l0, l1;
 
-int explicitsky = 0, skyarea = 0;
+int explicitsky = 0;
+double skyarea = 0;
 
 VARF(lodsize, 0, 32, 128, hdr.mapwlod = lodsize);
 VAR(loddistance, 0, 2000, 100000);
 
-int addtriindexes(usvector &v, int index[4])
+int addtriindexes(usvector &v, int index[4], int mask = 3)
 {
     int tris = 0;
-    if(index[0]!=index[1] && index[0]!=index[2] && index[1]!=index[2])
+    if(mask&1 && index[0]!=index[1] && index[0]!=index[2] && index[1]!=index[2])
     {
         tris++;
         v.add(index[0]);
         v.add(index[1]);
         v.add(index[2]);
     }
-    if(index[0]!=index[2] && index[0]!=index[3] && index[2]!=index[3])
+    if(mask&2 && index[0]!=index[2] && index[0]!=index[3] && index[2]!=index[3])
     {
         tris++;
         v.add(index[0]);
@@ -458,32 +467,75 @@ int addtriindexes(usvector &v, int index[4])
     return tris;
 }
 
+vvec shadowmapmin, shadowmapmax;
+
+int calcshadowmask(vvec *vv)
+{
+    extern vec shadowdir;
+    int mask = 0, used = 0;
+    if(vv[0]==vv[2]) return 0;
+    if(vv[0]!=vv[1] && vv[1]!=vv[2])
+    {
+        vvec e1(vv[1]), e2(vv[2]);
+        e1.sub(vv[0]);
+        e2.sub(vv[0]);
+        vec v;
+        v.cross(e1.tovec(), e2.tovec());
+        if(v.dot(shadowdir)>0) { mask |= 1; used |= 0x7; } 
+    }
+    if(vv[0]!=vv[3] && vv[2]!=vv[3])
+    {
+        vvec e1(vv[2]), e2(vv[3]);
+        e1.sub(vv[0]);
+        e2.sub(vv[0]);
+        vec v;
+        v.cross(e1.tovec(), e2.tovec());
+        if(v.dot(shadowdir)>0) { mask |= 2; used |= 0xD; }
+    }
+    if(used) loopi(4) if(used&(1<<i))
+    {
+        const vvec &v = vv[i];
+        loopk(3)
+        {
+            if(v[k]<shadowmapmin[k]) shadowmapmin[k] = v[k];
+            if(v[k]>shadowmapmax[k]) shadowmapmax[k] = v[k];
+        }
+    }
+    return mask;
+}
+
 void addcubeverts(int orient, int size, bool lodcube, vvec *vv, ushort texture, surfaceinfo *surface, surfacenormals *normals, ushort envmap = EMID_NONE)
 {
     int index[4];
+    int shadowmask = texture==DEFAULT_SKY || renderpath==R_FIXEDFUNCTION ? 0 : calcshadowmask(vv);
     loopk(4)
     {
         short u, v;
-        if(surface && surface->lmid >= LMID_RESERVED)
+        if(!nolights && surface && surface->lmid >= LMID_RESERVED)
         {
             u = short((surface->x + (surface->texcoords[k*2] / 255.0f) * (surface->w - 1) + 0.5f) * SHRT_MAX/LM_PACKW);
             v = short((surface->y + (surface->texcoords[k*2 + 1] / 255.0f) * (surface->h - 1) + 0.5f) * SHRT_MAX/LM_PACKH);
         }
         else u = v = 0;
-        index[k] = vh.access(vv[k], u, v, normals ? normals->normals[k] : bvec(128, 128, 128));
+        index[k] = vh.access(vv[k], u, v, renderpath!=R_FIXEDFUNCTION && normals ? normals->normals[k] : bvec(128, 128, 128));
     }
 
     extern vector<GLuint> lmtexids;
+    int dim = dimension(orient);
     sortkey key(texture, surface && lmtexids.inrange(surface->lmid) ? surface->lmid : LMID_AMBIENT, envmap);
     if(!lodcube)
     {
-        int tris = addtriindexes(texture == DEFAULT_SKY ? l0.explicitskyindices : l0.indices[key].dims[dimension(orient)], index);
-        if(texture == DEFAULT_SKY) explicitsky += tris;
-        else l0.curtris += tris;
+        if(texture == DEFAULT_SKY) explicitsky += addtriindexes(l0.explicitskyindices, index);
+        else
+        {
+            sortval &val = l0.indices[key];
+            l0.curtris += addtriindexes(val.dims[2*dim], index, shadowmask^3);
+            if(shadowmask) l0.curtris += addtriindexes(val.dims[2*dim+1], index, shadowmask); 
+        }
     }
     if(lodsize && size>=lodsize)
     {
-        int tris = addtriindexes(texture == DEFAULT_SKY ? l1.explicitskyindices : l1.indices[key].dims[dimension(orient)], index);
+        int tris = addtriindexes(texture == DEFAULT_SKY ? l1.explicitskyindices : l1.indices[key].dims[2*dim], index);
         if(texture != DEFAULT_SKY) l1.curtris += tris;
     }
 }
@@ -507,7 +559,11 @@ void gencubeverts(cube &c, int x, int y, int z, int size, int csi, bool lodcube)
         loopk(4) calcvert(c, x, y, z, size, vv[k], faceverts(c, i, k));
         ushort envmap = EMID_NONE;
         Slot &slot = lookuptexture(c.texture[i], false);
-        if(slot.shader && slot.shader->type & SHADER_ENVMAP) envmap = closestenvmap(i, x, y, z, size); 
+        if(slot.shader && slot.shader->type&SHADER_ENVMAP)
+        {
+            loopv(slot.sts) if(slot.sts[i].type==TEX_ENVMAP) { envmap = EMID_CUSTOM; break; }
+            if(envmap==EMID_NONE) envmap = closestenvmap(i, x, y, z, size); 
+        }
         addcubeverts(i, size, lodcube, vv, c.texture[i], e.surfaces ? &e.surfaces[i] : NULL, e.normals ? &e.normals[i] : NULL, envmap);
         if(!lodcube && slot.autograss && i!=O_BOTTOM) 
         {
@@ -517,6 +573,7 @@ void gencubeverts(cube &c, int x, int y, int z, int size, int csi, bool lodcube)
             g.texture = c.texture[i];
         }
     }
+    else if(touchingface(c, i) && visibleface(c, i, x, y, z, size, MAT_AIR, MAT_NOCLIP, lodcube)) ext(c).visible |= 1<<i;
 }
 
 bool skyoccluded(cube &c, int orient)
@@ -627,32 +684,8 @@ vtxarray *newva(int x, int y, int z, int size)
     int allocsize = sizeof(vtxarray) + l0.size() + l1.size();
     int bufsize = verts.length()*VTXSIZE;
     if(!hasVBO) allocsize += bufsize; // length of vertex buffer
-    vtxarray *va = (vtxarray *)new uchar[allocsize];
-    va->vbufGL = 0;
-    va->vbuf = 0; // Offset in VBO, or just null
-    if(hasVBO && verts.length())
-    {
-        void *vbuf;
-        if(VTXSIZE!=sizeof(vertex))
-        {
-            char *f = new char[bufsize];
-            genverts(f);
-            vbuf = (vertex *)addvbo(va, VBO_VBUF, f, bufsize); 
-            delete[] f;
-        }
-        else vbuf = (vertex *)addvbo(va, VBO_VBUF, verts.getbuf(), bufsize);
-        int offset = int(size_t(vbuf)) / VTXSIZE;
-        l0.offsetindices = offset;
-        l1.offsetindices = offset;
-    }
-    char *buf = l1.setup(va, va->l1, l0.setup(va, va->l0, (char *)(va+1)));
-    if(!hasVBO)
-    {
-        va->vbuf = (vertex *)buf;
-        if(VTXSIZE!=sizeof(vertex)) genverts(buf);
-        else memcpy(va->vbuf, verts.getbuf(), bufsize);
-    }
 
+    vtxarray *va = (vtxarray *)new uchar[allocsize];
     va->parent = NULL;
     va->children = new vector<vtxarray *>;
     va->allocsize = allocsize;
@@ -671,7 +704,38 @@ vtxarray *newva(int x, int y, int z, int size)
     }
     va->grasssamples = NULL;
     va->hasmerges = 0;
-    wverts += va->verts = verts.length();
+
+    va->verts = verts.length();
+    va->vbufGL = 0;
+    va->vbuf = 0; // Offset in VBO, or just null
+    va->minvert = 0;
+    va->maxvert = verts.length()-1;
+    if(hasVBO && verts.length())
+    {
+        void *vbuf;
+        if(VTXSIZE!=sizeof(vertex))
+        {
+            char *f = new char[bufsize];
+            genverts(f);
+            vbuf = (vertex *)addvbo(va, VBO_VBUF, f, bufsize); 
+            delete[] f;
+        }
+        else vbuf = (vertex *)addvbo(va, VBO_VBUF, verts.getbuf(), bufsize);
+        int offset = int(size_t(vbuf)) / VTXSIZE;
+        l0.offsetindices = offset;
+        l1.offsetindices = offset;
+        va->minvert += offset;
+        va->maxvert += offset;
+    }
+    char *buf = l1.setup(va, va->l1, l0.setup(va, va->l0, (char *)(va+1)));
+    if(!hasVBO)
+    {
+        va->vbuf = (vertex *)buf;
+        if(VTXSIZE!=sizeof(vertex)) genverts(buf);
+        else memcpy(va->vbuf, verts.getbuf(), bufsize);
+    }
+
+    wverts += va->verts;
     wtris  += va->l0.tris;
     allocva++;
     valist.add(va);
@@ -736,7 +800,7 @@ static vector<mergedface> vamerges[VVEC_INT];
 
 void genmergedfaces(cube &c, const ivec &co, int size, int minlevel = 0)
 {
-    if(!c.ext || !c.ext->merges) return;
+    if(!c.ext || !c.ext->merges || isempty(c)) return;
     int index = 0;
     loopi(6) if(c.ext->mergeorigin & (1<<i))
     {
@@ -746,7 +810,12 @@ void genmergedfaces(cube &c, const ivec &co, int size, int minlevel = 0)
         mf.orient = i;
         mf.tex = c.texture[i];
         mf.envmap = EMID_NONE;
-        if(lookupshader(mf.tex)->type & SHADER_ENVMAP) mf.envmap = closestenvmap(i, co.x, co.y, co.z, size);
+        Slot &slot = lookuptexture(mf.tex, false);
+        if(slot.shader && slot.shader->type&SHADER_ENVMAP)
+        {
+            loopv(slot.sts) if(slot.sts[i].type==TEX_ENVMAP) { mf.envmap = EMID_CUSTOM; break; }
+            if(mf.envmap==EMID_NONE) mf.envmap = closestenvmap(i, co.x, co.y, co.z, size);
+        }
         mf.surface = c.ext->surfaces ? &c.ext->surfaces[i] : NULL;
         mf.normals = c.ext->normals ? &c.ext->normals[i] : NULL;
         genmergedverts(c, i, co, size, m, mf.v);
@@ -867,7 +936,10 @@ void setva(cube &c, int cx, int cy, int cz, int size, int csi)
 {
     ASSERT(size <= VVEC_INT_MASK+1);
 
-    explicitsky = skyarea = 0;
+    explicitsky = 0;
+    skyarea = 0;
+    shadowmapmin = vvec(cx+size, cy+size, cz+size);
+    shadowmapmax = vvec(cx, cy, cz);
 
     rendercube(c, cx, cy, cz, size, csi);
 
@@ -881,13 +953,17 @@ void setva(cube &c, int cx, int cy, int cz, int size, int csi)
     ext(c).va = va;
     va->min = bbmin;
     va->max = bbmax;
+    va->shadowmapmin = shadowmapmin.toivec(va->x, va->y, va->z);
+    loopk(3) shadowmapmax[k] += (1<<VVEC_FRAC)-1;
+    va->shadowmapmax = shadowmapmax.toivec(va->x, va->y, va->z); 
     if(vamms.length()) va->mapmodels = new vector<octaentities *>(vamms);
     va->hasmerges = vahasmerges;
 
     verts.setsizenodelete(0);
     vamms.setsizenodelete(0);
     grasstris.setsizenodelete(0);
-    explicitsky = skyarea = 0;
+    explicitsky = 0;
+    skyarea = 0;
     vh.clear(); 
     l0.clear();
     l1.clear();
@@ -1005,9 +1081,11 @@ void allchanged(bool load)
     memset(cstats, 0, sizeof(cstat)*32);
     resetqueries();
     if(load) initenvmaps();
+    guessshadowdir();
     octarender();
     if(load) precacheall();
     setupmaterials();
+    invalidatereflections();
     if(load) 
     {
         entitiesinoctanodes();
