@@ -108,7 +108,7 @@ bool raycubeintersect(const cube &c, const vec &o, const vec &ray, float &dist)
 extern void entselectionbox(const entity &e, vec &eo, vec &es);
 extern int entselradius;
 float hitentdist;
-int  hitent, hitorient;
+int hitent, hitorient;
 
 static float disttoent(octaentities *oc, octaentities *last, const vec &o, const vec &ray, float radius, int mode, extentity *t)
 {
@@ -153,102 +153,119 @@ static float disttoent(octaentities *oc, octaentities *last, const vec &o, const
     return dist;
 }
 
-int rayent(const vec &o, const vec &ray, int &orient)
+// optimized shadow version
+static float shadowent(octaentities *oc, octaentities *last, const vec &o, const vec &ray, float radius, int mode, extentity *t)
 {
-    hitent = -1;
-    float d = raycube(o, ray, hitentdist = 1e20f, RAY_ENTS);
-    orient = hitorient;
-    if(hitentdist == d)
-        return hitent;
-    else
-        return -1;
-}
-
-float raycubepos(const vec &o, vec &ray, vec &hitpos, float radius, int mode, int size)
-{
-    ray.normalize();
-    hitpos = ray;
-    float dist = raycube(o, ray, radius, mode, size);
-    hitpos.mul(dist);
-    hitpos.add(o);
+    float dist = 1e16f, f = 0.0f;
+    if(oc == last || oc == NULL) return dist;
+    const vector<extentity *> &ents = et->getents();
+    loopv(oc->mapmodels) if(!last || last->mapmodels.find(oc->mapmodels[i])<0)
+    {
+        extentity &e = *ents[oc->mapmodels[i]];
+        if(!e.inoctanode || &e==t) continue;
+        if(e.attr3 && (e.triggerstate == TRIGGER_DISAPPEARED || !checktriggertype(e.attr3, TRIG_COLLIDE) || e.triggerstate == TRIGGERED)) continue;
+        if(!mmintersect(e, o, ray, radius, mode, f)) continue;
+        if(f>0 && f<dist) dist = f;
+    } 
     return dist;
 }
 
-bool raycubelos(vec &o, vec &dest, vec &hitpos)
+bool insideworld(const vec &o)
 {
-    vec ray(dest);
-    ray.sub(o);
-    float mag = ray.magnitude();
-    float distance = raycubepos(o, ray, hitpos, mag, RAY_CLIPMAT|RAY_POLY);
-    return distance >= mag; 
+    loopi(3) if(o[i]<0 || o[i]>=hdr.worldsize) return false;
+    return true;
 }
 
-float rayfloor(const vec &o, vec &floor, int mode)
-{
-    if(o.z<=0) return -1;
-    hitslope = vec(0, 0, 1);
-    float dist = raycube(o, vec(0, 0, -1), o.z, mode); 
-    if(dist>=o.z) return -1;
-    floor = hitslope;
-    return dist;
-}
+#define INITRAYCUBE \
+    octaentities *oclast = NULL; \
+    float dist = 0, dent = mode&RAY_BB ? 1e16f : 1e14f; \
+    cube *last = NULL; \
+    vec v(o), invray(ray.x ? 1/ray.x : 1e16f, ray.y ? 1/ray.y : 1e16f, ray.z ? 1/ray.z : 1e16f); \
+    static cube *levels[32]; \
+    levels[0] = worldroot; \
+    int l = 0, lsize = hdr.worldsize; \
+    ivec lo(0, 0, 0), lsizemask(invray.x>0 ? 0x7FFFFFFF : 0, invray.y>0 ? 0x7FFFFFFF : 0, invray.z>0 ? 0x7FFFFFFF : 0); \
+
+#define CHECKINSIDEWORLD \
+    if(!insideworld(v)) \
+    { \
+        float disttoworld = 1e16f, exitworld = 1e16f; \
+        loopi(3) \
+        { \
+            float c = v[i]; \
+            if(c>=0 && c<hdr.worldsize) \
+            { \
+                float d = ((invray[i]>0?hdr.worldsize:0)-c)*invray[i]; \
+                exitworld = min(exitworld, d); \
+            } \
+            else \
+            { \
+                float d = ((invray[i]>0?0:hdr.worldsize)-c)*invray[i]; \
+                if(d<0) return (radius>0?radius:-1); \
+                disttoworld = min(disttoworld, 0.1f + d); \
+            } \
+        } \
+        if(disttoworld > exitworld) return (radius>0?radius:-1); \
+        pushvec(v, ray, disttoworld); \
+        dist += disttoworld; \
+    }
+
+#define DOWNOCTREE(disttoent, earlyexit) \
+        cube *lc = levels[l]; \
+        for(;;) \
+        { \
+            lsize >>= 1; \
+            if(z>=lo.z+lsize) { lo.z += lsize; lc += 4; } \
+            if(y>=lo.y+lsize) { lo.y += lsize; lc += 2; } \
+            if(x>=lo.x+lsize) { lo.x += lsize; lc += 1; } \
+            if(lc->ext && lc->ext->ents && dent > 1e15f) \
+            { \
+                dent = disttoent(lc->ext->ents, oclast, o, ray, radius, mode, t); \
+                if(dent < 1e15f earlyexit) return min(dent, dist); \
+                oclast = lc->ext->ents; \
+            } \
+            if(lc->children==NULL) break; \
+            lc = lc->children; \
+            levels[++l] = lc; \
+        }
+
+#define FINDCLOSEST \
+        float dx = (lo.x+(lsize&lsizemask.x)-v.x)*invray.x, \
+              dy = (lo.y+(lsize&lsizemask.y)-v.y)*invray.y, \
+              dz = (lo.z+(lsize&lsizemask.z)-v.z)*invray.z; \
+        float disttonext = dx; \
+        disttonext = min(disttonext, dy); \
+        disttonext = min(disttonext, dz); \
+        disttonext += 0.1f; \
+        pushvec(v, ray, disttonext); \
+        dist += disttonext; \
+        last = &c;
+
+#define UPOCTREE \
+        x = int(v.x); \
+        y = int(v.y); \
+        z = int(v.z); \
+        for(;;) \
+        { \
+            lo.mask(~lsize); \
+            lsize <<= 1; \
+            uint lx = uint(x - lo.x), ly = uint(y - lo.y), lz = uint(z - lo.z); \
+            if(lx<uint(lsize) && ly<uint(lsize) && lz<uint(lsize)) break; \
+            if(!l) break; \
+            --l; \
+        }
 
 float raycube(const vec &o, const vec &ray, float radius, int mode, int size, extentity *t)
 {
-    octaentities *oclast = NULL;
-    float dist = 0, dent = 1e16f;
-    cube *last = NULL;
-    vec v = o;
-    if(ray.iszero()) return dist;
+    if(ray.iszero()) return 0;
 
-    static cube *levels[32];
-    levels[0] = worldroot;
-    int l = 0, lsize = hdr.worldsize;
-    ivec lo(0, 0, 0);
-
-    if(!insideworld(v))
-    {
-        float disttoworld = 1e16f, exitworld = 1e16f;
-        loopi(3) if(ray[i]!=0)
-        {
-            float c = v[i];
-            if(c>=0 && c<hdr.worldsize)
-            {
-                float d = (float(ray[i]>0?hdr.worldsize:0)-c)/ray[i];
-                exitworld = min(exitworld, d);
-            }
-            else
-            {
-                float d = (float(ray[i]>0?0:hdr.worldsize)-c)/ray[i];
-                if(d<0) return (radius>0?radius:-1);
-                disttoworld = min(disttoworld, 0.1f + d);
-            }
-        }
-        if(disttoworld > exitworld) return (radius>0?radius:-1);
-        pushvec(v, ray, disttoworld);
-        dist += disttoworld;
-    }
+    INITRAYCUBE
+    CHECKINSIDEWORLD
 
     int x = int(v.x), y = int(v.y), z = int(v.z);
     for(;;)
     {
-        cube *lc = levels[l];
-        for(;;)
-        {
-            lsize >>= 1;
-            if(z>=lo.z+lsize) { lo.z += lsize; lc += 4; }
-            if(y>=lo.y+lsize) { lo.y += lsize; lc += 2; }
-            if(x>=lo.x+lsize) { lo.x += lsize; lc += 1; }
-            if(lc->ext && lc->ext->ents && (mode&RAY_BB) && dent > 1e15f)
-            {
-                dent = disttoent(lc->ext->ents, oclast, o, ray, radius, mode, t);
-                if((mode&RAY_SHADOW) && dent < 1e15f) return min(dent, dist);
-                oclast = lc->ext->ents;
-            }
-            if(lc->children==NULL) break;
-            lc = lc->children;
-            levels[++l] = lc;
-        }
+        DOWNOCTREE(disttoent, && (mode&RAY_SHADOW));
 
         cube &c = *lc;
         if((dist>0 || !(mode&RAY_SKIPFIRST)) &&
@@ -256,10 +273,13 @@ float raycube(const vec &o, const vec &ray, float radius, int mode, int size, ex
             ((mode&RAY_EDITMAT) && c.ext && c.ext->material != MAT_AIR) ||
             (!(mode&RAY_PASS) && lsize==size && !isempty(c)) ||
             isentirelysolid(c) ||
-            dent < dist ||
-            last==&c))
+            dent < dist))
         {
-            if(last==&c && radius>0) dist = radius;
+            return min(dent, dist);
+        }
+        else if(last==&c)
+        {
+            if(radius>0) dist = radius;
             return min(dent, dist);
         }
         else if(!isempty(c))
@@ -270,38 +290,81 @@ float raycube(const vec &o, const vec &ray, float radius, int mode, int size, ex
                 return min(dent, dist+f);
         }
 
-        float disttonext = 1e16f;
-        loopi(3) if(ray[i]!=0)
-        {
-            float d = (float(lo[i]+(ray[i]>0?lsize:0))-v[i])/ray[i];
-            if(d >= 0) disttonext = min(disttonext, 0.1f + d);
-        }
-        pushvec(v, ray, disttonext);
-        dist += disttonext;
-        last = &c;
+        FINDCLOSEST
 
         if(radius>0 && dist>=radius) return min(dent, dist);
 
-        x = int(v.x);
-        y = int(v.y);
-        z = int(v.z);
-
-        for(;;)
-        {
-            lo.x &= ~lsize;
-            lo.y &= ~lsize;
-            lo.z &= ~lsize;
-            lsize <<= 1;
-            if(x<lo.x+lsize && y<lo.y+lsize && z<lo.z+lsize)
-            {
-                if(x>=lo.x && y>=lo.y && z>=lo.z) break;
-            }
-            if(!l) break;
-            --l;
-        }
+        UPOCTREE
     }
 }
 
+// optimized version for lightmap shadowing... every cycle here counts!!!
+float shadowray(const vec &o, const vec &ray, float radius, int mode, extentity *t)
+{
+    INITRAYCUBE
+    CHECKINSIDEWORLD 
+
+    int x = int(v.x), y = int(v.y), z = int(v.z);
+    for(;;)
+    {
+        DOWNOCTREE(shadowent, );
+
+        cube &c = *lc;
+        if(isentirelysolid(c)) return dist;
+        else if(last==&c) return radius;        
+        else if(!isempty(c))
+        {
+            float f = 0;
+            setcubeclip(c, lo.x, lo.y, lo.z, lsize);
+            if(raycubeintersect(c, v, ray, f)) return dist+f;
+        }
+
+        FINDCLOSEST
+
+        if(dist>=radius) return dist;
+
+        UPOCTREE
+    }
+}
+
+float rayent(const vec &o, vec &ray, vec &hitpos, float radius, int mode, int size, int &orient, int &ent)
+{
+    hitent = -1;
+    float d = raycubepos(o, ray, hitpos, hitentdist = radius, mode, size);
+    orient = hitorient;
+    ent = (hitentdist == d) ? hitent : -1;
+    return d;
+}
+
+float raycubepos(const vec &o, vec &ray, vec &hitpos, float radius, int mode, int size)
+{
+    ray.normalize();
+    hitpos = ray;
+    float dist = raycube(o, ray, radius, mode, size);
+    hitpos.mul(dist); 
+    hitpos.add(o);
+    return dist; 
+}
+
+bool raycubelos(vec &o, vec &dest, vec &hitpos)
+{
+    vec ray(dest);
+    ray.sub(o);
+    float mag = ray.magnitude();
+    float distance = raycubepos(o, ray, hitpos, mag, RAY_CLIPMAT|RAY_POLY);
+    return distance >= mag;
+}
+
+float rayfloor(const vec &o, vec &floor, int mode, float radius)
+{
+    if(o.z<=0) return -1;
+    hitslope = vec(0, 0, 1);
+    float dist = raycube(o, vec(0, 0, -1), radius, mode);
+    if(dist<0 || (radius>0 && dist>=radius)) return dist;
+    floor = hitslope;
+    return dist;
+}
+
 /////////////////////////  entity collision  ///////////////////////////////////////////////
 
 // info about collisions
@@ -317,30 +380,30 @@ const float JUMPVEL = 125.0f;
 const float GRAVITY = 200.0f;
 const float STEPSPEED = 1.0f;
 
-bool ellipsecollide(dynent *d, const vec &dir, dynent *e)
+bool ellipsecollide(physent *d, const vec &dir, const vec &o, float yaw, float xr, float yr,  float hi, float lo)
 {
-    float below = (e->o.z-e->eyeheight) - (d->o.z+d->aboveeye),
-          above = (d->o.z-d->eyeheight) - (e->o.z+e->aboveeye);
+    float below = (o.z-lo) - (d->o.z+d->aboveeye),
+          above = (d->o.z-d->eyeheight) - (o.z+hi);
     if(below>=0 || above>=0) return true;
-    float x = e->o.x - d->o.x, y = e->o.y - d->o.y;
-    float angle = atan2f(y, x), dangle = angle-(d->yaw+90)*RAD, eangle = angle-(e->yaw+90)*RAD;
+    float x = o.x - d->o.x, y = o.y - d->o.y;
+    float angle = atan2f(y, x), dangle = angle-(d->yaw+90)*RAD, eangle = angle-(yaw+90)*RAD;
     float dx = d->xradius*cosf(dangle), dy = d->yradius*sinf(dangle);
-    float ex = e->xradius*cosf(eangle), ey = e->yradius*sinf(eangle);
+    float ex = xr*cosf(eangle), ey = yr*sinf(eangle);
     float dist = sqrtf(x*x + y*y) - sqrtf(dx*dx + dy*dy) - sqrtf(ex*ex + ey*ey);
     if(dist < 0)
     {
-        if(dir.iszero() || (below >= -(d->eyeheight+d->aboveeye)/4.0f) && dir.z > 0)                             
-        {    
-            wall = vec(0, 0, -1); 
+        if(dir.iszero() || ((d->type>=ENT_INANIMATE || below >= -(d->eyeheight+d->aboveeye)/4.0f) && dir.z > 0))
+        {
+            wall = vec(0, 0, -1);
             return false;
         }
-        if(dir.iszero() || (above >= -(d->eyeheight+d->aboveeye)/2.0f && dir.z < 0))              
-        { 
-            wall = vec(0, 0, 1); 
+        if(dir.iszero() || ((d->type>=ENT_INANIMATE || above >= -(d->eyeheight+d->aboveeye)/3.0f) && dir.z < 0))
+        {
+            wall = vec(0, 0, 1);
             return false;
         }
         if(dir.iszero() || -x*dir.x + -y*dir.y < 0)
-        {    
+        {
             wall = vec(-x, -y, 0);
             wall.normalize();
             return false;
@@ -354,8 +417,9 @@ bool rectcollide(physent *d, const vec &dir, const vec &o, float xr, float yr,
     if(collideonly && !visible) return true;
     vec s(d->o);
     s.sub(o);
-    xr += d->radius;
-    yr += d->radius;
+    float dxr = d->collidetype==COLLIDE_ELLIPSE ? d->radius : d->xradius, dyr = d->collidetype==COLLIDE_ELLIPSE ? d->radius : d->yradius;
+    xr += dxr;
+    yr += dyr;
     walldistance = -1e10f;
     float zr = s.z>0 ? d->eyeheight+hi : d->aboveeye+lo;
     float ax = fabs(s.x)-xr;
@@ -365,14 +429,14 @@ bool rectcollide(physent *d, const vec &dir, const vec &o, float xr, float yr,
     wall.x = wall.y = wall.z = 0;
 #define TRYCOLLIDE(dim, ON, OP, N, P) \
     { \
-        if(s.dim<0) { if(visible&(1<<ON) && (dir.iszero() || (dir.dim>0 && (N)))) { walldistance = a ## dim; wall.dim = -1; return false; } } \
-        else if(visible&(1<<OP) && (dir.iszero() || (dir.dim<0 && (P)))) { walldistance = a ## dim; wall.dim = 1; return false; } \
+        if(s.dim<0) { if(visible&(1<<ON) && (dir.iszero() || (dir.dim>0 && (d->type>=ENT_INANIMATE || (N))))) { walldistance = a ## dim; wall.dim = -1; return false; } } \
+        else if(visible&(1<<OP) && (dir.iszero() || (dir.dim<0 && (d->type>=ENT_INANIMATE || (P))))) { walldistance = a ## dim; wall.dim = 1; return false; } \
     }
-    if(ax>ay && ax>az) TRYCOLLIDE(x, O_LEFT, O_RIGHT, ax > -d->radius, ax > -d->radius);
-    if(ay>az) TRYCOLLIDE(y, O_BACK, O_FRONT, ay > -d->radius, ay > -d->radius);
+    if(ax>ay && ax>az) TRYCOLLIDE(x, O_LEFT, O_RIGHT, ax > -dxr, ax > -dxr);
+    if(ay>az) TRYCOLLIDE(y, O_BACK, O_FRONT, ay > -dyr, ay > -dyr);
     TRYCOLLIDE(z, O_BOTTOM, O_TOP,
          az >= -(d->eyeheight+d->aboveeye)/4.0f,
-         az >= -(d->eyeheight+d->aboveeye)/2.0f);
+         az >= -(d->eyeheight+d->aboveeye)/3.0f);
     if(collideonly) inside = true;
     return collideonly;
 }
@@ -457,9 +521,19 @@ bool plcollide(physent *d, const vec &dir)    // collide with player or monster
         {
             physent *o = dynents[i];
             if(o==d || d->o.reject(o->o, d->radius+o->radius)) continue;
-            if(!ellipsecollide((dynent *)d, dir, (dynent *)o))
+            if(d->collidetype!=COLLIDE_ELLIPSE || o->collidetype!=COLLIDE_ELLIPSE)
+            { 
+                if(!rectcollide(d, dir, o->o, o->collidetype==COLLIDE_ELLIPSE ? o->radius : o->xradius, o->collidetype==COLLIDE_ELLIPSE ? o->radius : o->yradius, o->aboveeye, o->eyeheight))
+                {
+                    hitplayer = o;
+                    if((d->type==ENT_AI || d->type==ENT_INANIMATE) && wall.z>0) d->onplayer = o;
+                    return false;
+                }
+            }
+            else if(!ellipsecollide(d, dir, o->o, o->yaw, o->xradius, o->yradius, o->aboveeye, o->eyeheight))
             {
                 hitplayer = o;
+                if((d->type==ENT_AI || d->type==ENT_INANIMATE) && wall.z>0) d->onplayer = o;
                 return false;
             }
         }
@@ -475,7 +549,7 @@ void rotatebb(vec &center, vec &radius, int yaw)
     switch(yaw)
     {
         case 0: break;
-        case 180: 
+        case 180:
             center.x = -center.x;
             center.y = -center.y;
             break;
@@ -484,12 +558,12 @@ void rotatebb(vec &center, vec &radius, int yaw)
             swap(float, center.x, center.y);
             center.x = -center.x;
             break;
-        case 270: 
-            swap(float, radius.x, radius.y); 
+        case 270:
+            swap(float, radius.x, radius.y);
             swap(float, center.x, center.y);
             center.y = -center.y;
             break;
-        default: 
+        default:
             radius.x = radius.y = max(radius.x, radius.y) + max(fabs(center.x), fabs(center.y));
             center.x = center.y = 0.0f;
             break;
@@ -497,27 +571,29 @@ void rotatebb(vec &center, vec &radius, int yaw)
 }
 
 bool mmcollide(physent *d, const vec &dir, octaentities &oc)               // collide with a mapmodel
-{
+{   
     const vector<extentity *> &ents = et->getents();
     loopv(oc.mapmodels)
     {
         extentity &e = *ents[oc.mapmodels[i]];
-        if(e.attr3 && (e.triggerstate == TRIGGER_DISAPPEARED || !checktriggertype(e.attr3, TRIG_COLLIDE) || e.triggerstate == TRIGGERED)) continue;
+        if(e.attr3 && e.attr3!=15 && (e.triggerstate == TRIGGER_DISAPPEARED || !checktriggertype(e.attr3, TRIG_COLLIDE) || e.triggerstate == TRIGGERED)) continue;
         model *m = loadmodel(NULL, e.attr2);
         if(!m || !m->collide) continue;
         vec center, radius;
         m->collisionbox(0, center, radius);
-        center.z -= radius.z;
-        radius.z *= 2;
-        rotatebb(center, radius, e.attr1);
-        if(!rectcollide(d, dir, center.add(e.o), radius.x, radius.y, radius.z, 0)) return false;
+        if(!m->ellipsecollide || d->collidetype!=COLLIDE_ELLIPSE)
+        {
+            rotatebb(center, radius, e.attr1);
+            if(!rectcollide(d, dir, center.add(e.o), radius.x, radius.y, radius.z, radius.z)) return false;
+        }
+        else if(!ellipsecollide(d, dir, center.add(e.o), float((e.attr1+7)-(e.attr1+7)%15), radius.x, radius.y, radius.z, radius.z)) return false;
     }
     return true;
 }
 
 bool cubecollide(physent *d, const vec &dir, float cutoff, cube &c, int x, int y, int z, int size) // collide with cube geometry
 {
-    if(isentirelysolid(c) || (c.ext && (d->type<ENT_CAMERA || c.ext->material != MAT_CLIP) && isclipped(c.ext->material)))
+    if(isentirelysolid(c) || (c.ext && ((d->type==ENT_AI && c.ext->material==MAT_AICLIP) || ((d->type<ENT_CAMERA || c.ext->material != MAT_CLIP) && isclipped(c.ext->material)))))
     {
         int s2 = size>>1;
         vec o = vec(x+s2, y+s2, z+s2);
@@ -564,13 +640,12 @@ bool cubecollide(physent *d, const vec &dir, float cutoff, cube &c, int x, int y
                 if(!dir.iszero())
                 {
                     if(f.dot(dir) >= -cutoff) continue;
-                    if((dir.x*f.x < 0 || dir.y*f.y < 0) && dist <= -r) continue;
-                    if(dir.z*f.z < 0 && dist < -(d->eyeheight+d->aboveeye)/(dir.z < 0 ? 2.0f : 4.0f)) continue;
+                    if(d->type<ENT_CAMERA && dist < (dir.z*f.z < 0 ? -(d->eyeheight+d->aboveeye)/(dir.z < 0 ? 3.0f : 4.0f) : ((dir.x*f.x < 0 || dir.y*f.y < 0) ? -r : 0))) continue;
                 }
-                if(p.p[i].x) { if((p.p[i].x>0 ? o.x-p.o.x : p.o.x-o.x) + p.r.x - r < dist) continue; } 
-                if(p.p[i].y) { if((p.p[i].y>0 ? o.y-p.o.y : p.o.y-o.y) + p.r.y - r < dist) continue; }
-                if(p.p[i].z) { if((p.p[i].z>0 ? o.z-p.o.z : p.o.z-o.z) + p.r.z - zr < dist) continue; }
-                w = &p.p[i];
+                if(f.x && (f.x>0 ? o.x-p.o.x : p.o.x-o.x) + p.r.x - r < dist && f.dist(vec(p.o.x + (f.x>0 ? -p.r.x : p.r.x), o.y, o.z)) >= vec(0, f.y*r, f.z*zr).magnitude()) continue;
+                if(f.y && (f.y>0 ? o.y-p.o.y : p.o.y-o.y) + p.r.y - r < dist && f.dist(vec(o.x, p.o.y + (f.y>0 ? -p.r.y : p.r.y), o.z)) >= vec(f.x*r, 0, f.z*zr).magnitude()) continue;
+                if(f.z && (f.z>0 ? o.z-p.o.z : p.o.z-o.z) + p.r.z - zr < dist && f.dist(vec(o.x, o.y, p.o.z + (f.z>0 ? -p.r.z : p.r.z))) >= vec(f.x*r, f.y*r, 0).magnitude()) continue;
+                w = &f;
                 m = dist;
             }
         }
@@ -594,7 +669,7 @@ bool octacollide(physent *d, const vec &dir, float cutoff, const ivec &bo, const
         {
             if(!octacollide(d, dir, cutoff, bo, bs, c[i].children, o, size>>1)) return false;
         }
-        else if(c[i].ext && c[i].ext->material!=MAT_NOCLIP && (!isempty(c[i]) || ((d->type<ENT_CAMERA || c[i].ext->material != MAT_CLIP) && isclipped(c[i].ext->material))))
+        else if(c[i].ext && c[i].ext->material!=MAT_NOCLIP && (!isempty(c[i]) || (d->type==ENT_AI && c[i].ext->material==MAT_AICLIP) || ((d->type<ENT_CAMERA || c[i].ext->material != MAT_CLIP) && isclipped(c[i].ext->material))))
         {
             if(!cubecollide(d, dir, cutoff, c[i], o.x, o.y, o.z, size)) return false;
         }
@@ -603,7 +678,7 @@ bool octacollide(physent *d, const vec &dir, float cutoff, const ivec &bo, const
 }
 
 // all collision happens here
-bool collide(physent *d, const vec &dir, float cutoff)
+bool collide(physent *d, const vec &dir, float cutoff, bool playercol)
 {
     inside = false;
     hitplayer = NULL;
@@ -612,7 +687,7 @@ bool collide(physent *d, const vec &dir, float cutoff)
          bs(int(d->radius)*2, int(d->radius)*2, int(d->eyeheight+d->aboveeye));
     bs.add(2);  // guard space for rounding errors
     if(!octacollide(d, dir, cutoff, bo, bs, worldroot, ivec(0, 0, 0), hdr.worldsize>>1)) return false; // collide with world
-    return plcollide(d, dir);
+    return !playercol || plcollide(d, dir);
 }
 
 void slopegravity(float g, const vec &slope, vec &gvec)
@@ -817,7 +892,7 @@ bool move(physent *d, vec &dir)
 {
     vec old(d->o);
 #if 0
-    if(d->physstate == PHYS_STEP_DOWN && dir.z <= 0.0f && (d->move || d->strafe))
+    if(d->physstate == PHYS_STEP_DOWN && dir.z <= 0.0f && cl->allowmove(pl) && (d->move || d->strafe))
     {
         float step = dir.magnitude()*STEPSPEED;
         if(trystepdown(d, dir, step, 0.75f, 0.25f)) return true;
@@ -831,7 +906,7 @@ bool move(physent *d, vec &dir)
     bool collided = false;
     vec obstacle;
     d->o.add(dir);
-    if(!collide(d, d->type!=ENT_CAMERA ? dir : vec(0, 0, 0)) || (d->type==ENT_AI ? !collide(d) : (!dir.z && !collide(d, vec(0, 0, -1)))))
+    if(!collide(d, d->type!=ENT_CAMERA ? dir : vec(0, 0, 0)) || ((d->type==ENT_AI || d->type==ENT_INANIMATE) && !collide(d)))
     {
         d->o = old;
         if(d->type == ENT_CAMERA) return false;
@@ -853,7 +928,7 @@ bool move(physent *d, vec &dir)
     if(slide)
     {
         slideagainst(d, dir, obstacle);
-        if(d->type == ENT_AI) d->blocked = true;
+        if(d->type == ENT_AI || d->type == ENT_INANIMATE) d->blocked = true;
     }
     if(found)
     {
@@ -866,15 +941,15 @@ bool move(physent *d, vec &dir)
 
 bool bounce(physent *d, float secs, float elasticity, float waterfric)
 {
-    cube &c = lookupcube(int(d->o.x), int(d->o.y), int(d->o.z));
-    bool water = c.ext && c.ext->material == MAT_WATER;
+    int mat = lookupmaterial(d->o);
+    bool water = isliquid(mat);
     if(water)
     {
         if(d->vel.z > 0 && d->vel.z + d->gravity.z < 0) d->vel.z = 0.0f;
+        d->vel.mul(1.0f - secs/waterfric);
         d->gravity.z = -4.0f*GRAVITY*secs;
-    }
+    } 
     else d->gravity.z -= GRAVITY*secs;
-    if(water) d->vel.mul(1.0f - secs/waterfric);
     vec old(d->o);
     loopi(2)
     {
@@ -904,13 +979,13 @@ bool bounce(physent *d, float secs, float elasticity, float waterfric)
         d->vel.mul(k);
         d->vel.sub(wvel);
     }
-    if(d->physstate != PHYS_BOUNCE)
+    if(d->physstate!=PHYS_BOUNCE)
     {
         // make sure bouncers don't start inside geometry
-        if(d->o == old) return true;
+        if(d->o == old) return !hitplayer;
         d->physstate = PHYS_BOUNCE;
     }
-    return hitplayer != 0;
+    return hitplayer!=0;
 }
 
 void avoidcollision(physent *d, const vec &dir, physent *obstacle, float space)
@@ -995,6 +1070,7 @@ void vectoyawpitch(const vec &v, float &yaw, float &pitch)
 }
 
 VARP(maxroll, 0, 3, 20);
+VAR(floatspeed, 10, 100, 1000);
 
 void modifyvelocity(physent *pl, bool local, bool water, bool floating, int curtime)
 {
@@ -1053,7 +1129,7 @@ void modifyvelocity(physent *pl, bool local, bool water, bool floating, int curt
         }
     }
 
-    if(m.iszero() && (pl->move || pl->strafe))
+    if(m.iszero() && cl->allowmove(pl) && (pl->move || pl->strafe))
     {
         vecfromyawpitch(pl->yaw, floating || water || pl->type==ENT_CAMERA ? pl->pitch : 0, pl->move, pl->strafe, m);
 
@@ -1072,7 +1148,11 @@ void modifyvelocity(physent *pl, bool local, bool water, bool floating, int curt
 
     vec d(m);
     d.mul(pl->maxspeed);
-    if(!water && !floating) d.mul((pl->move && !pl->strafe ? 1.3f : 1.0f) * (pl->physstate < PHYS_SLOPE && pl->move>=0 ? 1.3f : 1.0f)); // EXPERIMENTAL
+    if(floating) 
+    {
+        if(pl==player) d.mul(floatspeed/100.0f);
+    }
+    else if(!water && cl->allowmove(pl)) d.mul((pl->move && !pl->strafe ? 1.3f : 1.0f) * (pl->physstate < PHYS_SLOPE ? 1.3f : 1.0f)); // EXPERIMENTAL
     float friction = water && !floating ? 20.0f : (pl->physstate >= PHYS_SLOPE || floating ? 6.0f : 30.0f);
     float fpsfric = friction/curtime*20.0f;
 
@@ -1100,24 +1180,24 @@ void modifygravity(physent *pl, bool water, float secs)
 
 bool moveplayer(physent *pl, int moveres, bool local, int curtime)
 {
-    cube &c = lookupcube(int(pl->o.x), int(pl->o.y), int(pl->o.z));
-    bool water = c.ext && c.ext->material == MAT_WATER;
+    int material = lookupmaterial(pl->o);
+    bool water = isliquid(material);
     bool floating = (editmode && local) || pl->state==CS_EDITING || pl->state==CS_SPECTATOR;
     float secs = curtime/1000.f;
 
     // apply any player generated changes in velocity
     modifyvelocity(pl, local, water, floating, curtime);
-
     // apply gravity
     if(!floating && pl->type!=ENT_CAMERA) modifygravity(pl, water, secs);
 
-    vec d(pl->vel);
+    vec d(pl->vel), oldpos(pl->o);
     if(!floating && pl->type!=ENT_CAMERA && water) d.mul(0.5f);
     d.add(pl->gravity);
     d.mul(secs);
 
     pl->blocked = false;
     pl->moving = true;
+    pl->onplayer = NULL;
 
     if(floating)                // just apply velocity
     {
@@ -1143,10 +1223,13 @@ bool moveplayer(physent *pl, int moveres, bool local, int curtime)
         }
     }
 
-    if(pl->state==CS_ALIVE) updatedynentcache(pl);
+    if(pl->type!=ENT_CAMERA && pl->state==CS_ALIVE) updatedynentcache(pl);
 
     if(!pl->timeinair && pl->physstate >= PHYS_FLOOR && pl->vel.squaredlen() < 1e-4f && pl->gravity.iszero()) pl->moving = false;
 
+    pl->lastmoveattempt = lastmillis;
+    if(pl->o!=oldpos) pl->lastmove = lastmillis;
+
     // automatically apply smooth roll when strafing
 
     if(pl->strafe==0)
@@ -1164,17 +1247,19 @@ bool moveplayer(physent *pl, int moveres, bool local, int curtime)
 
     if(pl->type!=ENT_CAMERA)
     {
-        cube &c = lookupcube((int)pl->o.x, (int)pl->o.y, (int)pl->o.z+1);
-        bool inwater = c.ext && c.ext->material == MAT_WATER;
+        int mat = lookupmaterial(vec(pl->o.x, pl->o.y, pl->o.z+1));
+        bool inwater = isliquid(mat);
         if(!pl->inwater && inwater) cl->physicstrigger(pl, local, 0, -1);
         else if(pl->inwater && !inwater) cl->physicstrigger(pl, local, 0, 1);
         pl->inwater = inwater;
+
+        if(pl->state==CS_ALIVE && material==MAT_LAVA) cl->suicide(pl);
     }
 
     return true;
 }
 
-VAR(minframetime, 5, 10, 20);
+VARP(minframetime, 5, 10, 20);
 
 int physicsfraction = 0, physicsrepeat = 0;
 
@@ -1188,7 +1273,7 @@ void physicsframe()          // optimally schedule physics frames inside the gra
     }
     else
     {
-        physicsrepeat = 1;
+        physicsrepeat = curtime>0 ? 1 : 0;
     }
     cleardynentcache();
 }
@@ -1196,7 +1281,7 @@ void physicsframe()          // optimally schedule physics frames inside the gra
 void moveplayer(physent *pl, int moveres, bool local)
 {
     loopi(physicsrepeat) moveplayer(pl, moveres, local, min(curtime, minframetime));
-    if(pl->o.z<0 && pl->state==CS_ALIVE) cl->worldhurts(pl, 400);
+    if(pl->o.z<0 && pl->state==CS_ALIVE) cl->suicide(pl);
 }
 
 void updatephysstate(physent *d)
@@ -1268,17 +1353,151 @@ bool intersect(physent *d, vec &from, vec &to)   // if lineseg hits entity bound
         && p->z >= d->o.z-d->eyeheight;
 }
 
-#define dir(name,v,d,s,os) ICOMMAND(name, "D", { player->s = args!=NULL; player->v = player->s ? d : (player->os ? -(d) : 0); });
+const float PLATFORMMARGIN = 0.2f;
+const float PLATFORMBORDER = 10.0f;
+
+struct platformcollision
+{
+    physent *d;
+    int next;
+
+    platformcollision() {}
+    platformcollision(physent *d, int next) : d(d), next(next) {}
+};
+
+bool platformcollide(physent *d, physent *o, const vec &dir, float margin = 0)
+{
+    if(d->collidetype!=COLLIDE_ELLIPSE || o->collidetype!=COLLIDE_ELLIPSE)
+    {
+        if(rectcollide(d, dir, o->o,
+            o->collidetype==COLLIDE_ELLIPSE ? o->radius : o->xradius,
+            o->collidetype==COLLIDE_ELLIPSE ? o->radius : o->yradius, o->aboveeye,
+            o->eyeheight + margin))
+            return true;
+    }
+    else if(ellipsecollide(d, dir, o->o, o->yaw, o->xradius, o->yradius, o->aboveeye, o->eyeheight + margin))
+        return true;
+    return false;
+}
+
+bool moveplatform(physent *p, const vec &dir)
+{   
+    vec oldpos(p->o);
+    p->o.add(dir);
+    if(!collide(p, dir, 0, dir.z<=0))
+    {
+        p->o = oldpos; 
+        return false;
+    }
+    p->o = oldpos; 
+
+    static vector<physent *> candidates;
+    candidates.setsizenodelete(0);
+    for(int x = int(max(p->o.x-p->radius-PLATFORMBORDER, 0))>>dynentsize, ex = int(min(p->o.x+p->radius+PLATFORMBORDER, hdr.worldsize-1))>>dynentsize; x <= ex; x++)
+    for(int y = int(max(p->o.y-p->radius-PLATFORMBORDER, 0))>>dynentsize, ey = int(min(p->o.y+p->radius+PLATFORMBORDER, hdr.worldsize-1))>>dynentsize; y <= ey; y++)
+    {
+        const vector<physent *> &dynents = checkdynentcache(x, y);
+        loopv(dynents)
+        {
+            physent *d = dynents[i];
+            if(p==d || d->o.z-d->eyeheight < p->o.z+p->aboveeye || p->o.reject(d->o, p->radius+PLATFORMBORDER+d->radius) || candidates.find(d) >= 0) continue;
+            candidates.add(d);
+            d->stacks = d->collisions = -1;
+        }
+    }
+    static vector<physent *> passengers, colliders;
+    passengers.setsizenodelete(0);
+    colliders.setsizenodelete(0);
+    static vector<platformcollision> collisions;
+    collisions.setsizenodelete(0);
+    // build up collision DAG of colliders to be pushed off, and DAG of stacked passengers
+    loopv(candidates)
+    {
+        physent *d = candidates[i];
+        // check if the dynent is on top of the platform
+        if(!platformcollide(p, d, vec(0, 0, 1), PLATFORMMARGIN)) passengers.add(d);
+        vec doldpos(d->o);
+        d->o.add(dir);
+        if(!collide(d, dir, 0, false)) colliders.add(d);
+        d->o = doldpos;
+        loopvj(candidates)
+        {
+            physent *o = candidates[j];
+            if(!platformcollide(d, o, dir))
+            {
+                collisions.add(platformcollision(d, o->collisions));
+                o->collisions = collisions.length() - 1;
+            }
+            if(d->o.z < o->o.z && !platformcollide(d, o, vec(0, 0, 1), PLATFORMMARGIN))
+            {
+                collisions.add(platformcollision(o, d->stacks));
+                d->stacks = collisions.length() - 1;
+            }
+        }
+    }
+    loopv(colliders) // propagate collisions
+    {
+        physent *d = colliders[i];
+        for(int n = d->collisions; n>=0; n = collisions[n].next)
+        {
+            physent *o = collisions[n].d;
+            if(colliders.find(o)<0) colliders.add(o);
+        }
+    }
+    if(dir.z>0)
+    {
+        loopv(passengers) // if any stacked passengers collide, stop the platform
+        {
+            physent *d = passengers[i];
+            if(colliders.find(d)>=0) return false;
+            for(int n = d->stacks; n>=0; n = collisions[n].next)
+            {
+                physent *o = collisions[n].d;
+                if(passengers.find(o)<0) passengers.add(o);
+            }
+        }
+        loopv(passengers)
+        {
+            physent *d = passengers[i];
+            d->o.add(dir);
+            d->lastmove = lastmillis;
+            if(dir.x || dir.y) updatedynentcache(d);
+        }
+    }
+    else loopv(passengers) // move any stacked passengers who aren't colliding with non-passengers
+    {
+        physent *d = passengers[i];
+        if(colliders.find(d)>=0) continue;
+
+        d->o.add(dir);
+        d->lastmove = lastmillis;
+        if(dir.x || dir.y) updatedynentcache(d);
+
+        for(int n = d->stacks; n>=0; n = collisions[n].next)
+        {
+            physent *o = collisions[n].d;
+            if(passengers.find(o)<0) passengers.add(o);
+        }
+    }
+
+    p->o.add(dir);
+    p->lastmove = lastmillis;
+    if(dir.x || dir.y) updatedynentcache(p);
+
+    return true;
+}
+
+#define dir(name,v,d,s,os) ICOMMAND(name, "D", (int *down), { player->s = *down!=0; player->v = player->s ? d : (player->os ? -(d) : 0); });
 
 dir(backward, move,   -1, k_down,  k_up);
 dir(forward,  move,    1, k_up,    k_down);
 dir(left,     strafe,  1, k_left,  k_right);
 dir(right,    strafe, -1, k_right, k_left);
 
-ICOMMAND(jump,   "D", { if(cl->canjump()) player->jumpnext = args!=NULL; });
-ICOMMAND(attack, "D", { cl->doattack(args!=NULL); });
+ICOMMAND(jump,   "D", (int *down), { if(cl->canjump()) player->jumpnext = *down!=0; });
+ICOMMAND(attack, "D", (int *down), { cl->doattack(*down!=0); });
 
-VARP(sensitivity, 0, 10, 1000);
+VARP(sensitivity, 0, 3, 1000);
 VARP(sensitivityscale, 1, 1, 100);
 VARP(invmouse, 0, 0, 1);
 
