@@ -3,7 +3,7 @@
 #include "pch.h"
 #include "engine.h"
 
-bool hasVBO = false, hasOQ = false, hasTR = false, hasFBO = false, hasCM = false, hasTC = false, hasstencil = false;
+bool hasVBO = false, hasDRE = false, hasOQ = false, hasTR = false, hasFBO = false, hasDS = false, hasTF = false, hasBE = false, hasCM = false, hasNP2 = false, hasTC = false, hasTE = false, hasMT = false, hasD3, hasstencil = false, hasAF = false;
 int renderpath;
 
 // GL_ARB_vertex_buffer_object
@@ -23,8 +23,11 @@ PFNGLMULTITEXCOORD3FARBPROC     glMultiTexCoord3f_     = NULL;
  
 // GL_ARB_vertex_program, GL_ARB_fragment_program
 PFNGLGENPROGRAMSARBPROC            glGenPrograms_            = NULL;
+PFNGLDELETEPROGRAMSARBPROC         glDeletePrograms_         = NULL;
 PFNGLBINDPROGRAMARBPROC            glBindProgram_            = NULL;
 PFNGLPROGRAMSTRINGARBPROC          glProgramString_          = NULL;
+PFNGLGETPROGRAMIVARBPROC           glGetProgramiv_           = NULL;
+
 PFNGLPROGRAMENVPARAMETER4FARBPROC  glProgramEnvParameter4f_  = NULL;
 PFNGLPROGRAMENVPARAMETER4FVARBPROC glProgramEnvParameter4fv_ = NULL;
 
@@ -48,6 +51,7 @@ PFNGLDELETEFRAMEBUFFERSEXTPROC      glDeleteFramebuffers_      = NULL;
 PFNGLGENFRAMEBUFFERSEXTPROC         glGenFramebuffers_         = NULL;
 PFNGLFRAMEBUFFERTEXTURE2DEXTPROC    glFramebufferTexture2D_    = NULL;
 PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC glFramebufferRenderbuffer_ = NULL;
+PFNGLGENERATEMIPMAPEXTPROC          glGenerateMipmap_          = NULL;
 
 // GL_ARB_shading_language_100, GL_ARB_shader_objects, GL_ARB_fragment_shader, GL_ARB_vertex_shader
 PFNGLCREATEPROGRAMOBJECTARBPROC       glCreateProgramObject_      = NULL;
@@ -64,6 +68,12 @@ PFNGLGETUNIFORMLOCATIONARBPROC        glGetUniformLocation_       = NULL;
 PFNGLUNIFORM4FVARBPROC                glUniform4fv_               = NULL;
 PFNGLUNIFORM1IARBPROC                 glUniform1i_                = NULL;
 
+// GL_EXT_draw_range_elements
+PFNGLDRAWRANGEELEMENTSEXTPROC glDrawRangeElements_ = NULL;
+
+// GL_EXT_blend_minmax
+PFNGLBLENDEQUATIONEXTPROC glBlendEquation_ = NULL;
+
 void *getprocaddress(const char *name)
 {
     return SDL_GL_GetProcAddress(name);
@@ -74,6 +84,12 @@ VAR(ati_texgen_bug, 0, 0, 1);
 VAR(ati_oq_bug, 0, 0, 1);
 VAR(nvidia_texgen_bug, 0, 0, 1);
 VAR(apple_glsldepth_bug, 0, 0, 1);
+VAR(apple_minmax_bug, 0, 0, 1);
+VAR(apple_ff_bug, 0, 0, 1);
+VAR(intel_quadric_bug, 0, 0, 1);
+VAR(mesa_dre_bug, 0, 0, 1);
+VAR(minimizetcusage, 1, 0, 0);
+VAR(emulatefog, 1, 0, 0);
 
 void gl_init(int w, int h, int bpp, int depth, int fsaa)
 {
@@ -97,7 +113,6 @@ void gl_init(int w, int h, int bpp, int depth, int fsaa)
 
     glEnable(GL_LINE_SMOOTH);
     glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
-    glEnable(GL_POLYGON_OFFSET_LINE);
     glPolygonOffset(-3.0f, -3.0f);
 
     glCullFace(GL_FRONT);
@@ -110,25 +125,32 @@ void gl_init(int w, int h, int bpp, int depth, int fsaa)
     conoutf("Renderer: %s (%s)", renderer, vendor);
     conoutf("Driver: %s", version);
     
-    extern int shaderprecision;
+#ifdef __APPLE__
+    extern int mac_osversion();
+    int osversion = mac_osversion();  /* 0x1050 = 10.5 (Leopard) */
+#endif
+    
+    //extern int shaderprecision;
     // default to low precision shaders on certain cards, can be overridden with -f3
     // char *weakcards[] = { "GeForce FX", "Quadro FX", "6200", "9500", "9550", "9600", "9700", "9800", "X300", "X600", "FireGL", "Intel", "Chrome", NULL } 
     // if(shaderprecision==2) for(char **wc = weakcards; *wc; wc++) if(strstr(renderer, *wc)) shaderprecision = 1;
-    
-    if(!strstr(exts, "GL_EXT_texture_env_combine") && !strstr(exts, "GL_ARB_texture_env_combine")) 
-        fatal("No texture_env_combine extension! (your video card is WAY too old)");
+  
+    if(strstr(exts, "GL_EXT_texture_env_combine") || strstr(exts, "GL_ARB_texture_env_combine")) hasTE = true;
+    else conoutf("WARNING: No texture_env_combine extension! (your video card is WAY too old)");
 
-    if(!strstr(exts, "GL_ARB_multitexture")) fatal("No multitexture extension!");
-    glActiveTexture_       = (PFNGLACTIVETEXTUREARBPROC)      getprocaddress("glActiveTextureARB");
-    glClientActiveTexture_ = (PFNGLCLIENTACTIVETEXTUREARBPROC)getprocaddress("glClientActiveTextureARB");
-    glMultiTexCoord2f_     = (PFNGLMULTITEXCOORD2FARBPROC)    getprocaddress("glMultiTexCoord2fARB");
-    glMultiTexCoord3f_     = (PFNGLMULTITEXCOORD3FARBPROC)    getprocaddress("glMultiTexCoord3fARB");
+    if(strstr(exts, "GL_EXT_texture_env_dot3") || strstr(exts, "GL_ARB_texture_env_dot3")) hasD3 = true;
 
-    if(!strstr(exts, "GL_ARB_vertex_buffer_object"))
+    if(strstr(exts, "GL_ARB_multitexture"))
     {
-        conoutf("WARNING: No vertex_buffer_object extension! (geometry heavy maps will be SLOW)");
+        glActiveTexture_       = (PFNGLACTIVETEXTUREARBPROC)      getprocaddress("glActiveTextureARB");
+        glClientActiveTexture_ = (PFNGLCLIENTACTIVETEXTUREARBPROC)getprocaddress("glClientActiveTextureARB");
+        glMultiTexCoord2f_     = (PFNGLMULTITEXCOORD2FARBPROC)    getprocaddress("glMultiTexCoord2fARB");
+        glMultiTexCoord3f_     = (PFNGLMULTITEXCOORD3FARBPROC)    getprocaddress("glMultiTexCoord3fARB");
+        hasMT = true;
     }
-    else
+    else conoutf("WARNING: No multitexture extension!");
+
+    if(strstr(exts, "GL_ARB_vertex_buffer_object"))
     {
         glGenBuffers_    = (PFNGLGENBUFFERSARBPROC)   getprocaddress("glGenBuffersARB");
         glBindBuffer_    = (PFNGLBINDBUFFERARBPROC)   getprocaddress("glBindBufferARB");
@@ -140,23 +162,83 @@ void gl_init(int w, int h, int bpp, int depth, int fsaa)
         hasVBO = true;
         //conoutf("Using GL_ARB_vertex_buffer_object extension.");
     }
+    else conoutf("WARNING: No vertex_buffer_object extension! (geometry heavy maps will be SLOW)");
 
+    if(strstr(exts, "GL_EXT_draw_range_elements"))
+    {
+        glDrawRangeElements_ = (PFNGLDRAWRANGEELEMENTSEXTPROC)getprocaddress("glDrawRangeElementsEXT");
+        hasDRE = true;
+    }
+
+    if(strstr(exts, "GL_ARB_occlusion_query"))
+    {
+        GLint bits;
+        glGetQueryiv_ = (PFNGLGETQUERYIVARBPROC)getprocaddress("glGetQueryivARB");
+        glGetQueryiv_(GL_SAMPLES_PASSED_ARB, GL_QUERY_COUNTER_BITS_ARB, &bits);
+        if(bits)
+        {
+            glGenQueries_ =        (PFNGLGENQUERIESARBPROC)       getprocaddress("glGenQueriesARB");
+            glDeleteQueries_ =     (PFNGLDELETEQUERIESARBPROC)    getprocaddress("glDeleteQueriesARB");
+            glBeginQuery_ =        (PFNGLBEGINQUERYARBPROC)       getprocaddress("glBeginQueryARB");
+            glEndQuery_ =          (PFNGLENDQUERYARBPROC)         getprocaddress("glEndQueryARB");
+            glGetQueryObjectiv_ =  (PFNGLGETQUERYOBJECTIVARBPROC) getprocaddress("glGetQueryObjectivARB");
+            glGetQueryObjectuiv_ = (PFNGLGETQUERYOBJECTUIVARBPROC)getprocaddress("glGetQueryObjectuivARB");
+            hasOQ = true;
+            //conoutf("Using GL_ARB_occlusion_query extension.");
+#if defined(__APPLE__) && SDL_BYTEORDER == SDL_BIG_ENDIAN
+            if(strstr(vendor, "ATI") && (osversion<0x1050)) ati_oq_bug = 1;
+#endif
+            if(ati_oq_bug) conoutf("WARNING: Using ATI occlusion query bug workaround. (use \"/ati_oq_bug 0\" to disable if unnecessary)");
+        }
+    }
+    if(!hasOQ)
+    {
+        conoutf("WARNING: No occlusion query support! (large maps may be SLOW)");
+        if(renderpath==R_FIXEDFUNCTION) zpass = 0;
+    }
+
+    bool avoidshaders = false;
     if(strstr(vendor, "ATI"))
     {
         floatvtx = 1;
         conoutf("WARNING: ATI cards may show garbage in skybox. (use \"/ati_skybox_bug 1\" to fix)");
+
+        extern int reservedynlighttc, reserveshadowmaptc;
+        reservedynlighttc = 2;
+        reserveshadowmaptc = 3;
+        minimizetcusage = 1;
+        emulatefog = 1;
     }
     else if(strstr(vendor, "Tungsten"))
     {
+        avoidshaders = true;
         floatvtx = 1;
-    }
-    if(floatvtx) conoutf("WARNING: Using floating point vertexes. (use \"/floatvtx 0\" to disable)");
 
-    if(!shaderprecision || !strstr(exts, "GL_ARB_vertex_program") || !strstr(exts, "GL_ARB_fragment_program"))
+        extern int maxtexsize;
+        maxtexsize = 256;
+
+        if(!hasOQ) waterreflect = waterrefract = 0;
+        if(hasDRE) mesa_dre_bug = 1;
+    }
+    else if(strstr(vendor, "Intel"))
     {
-        conoutf("WARNING: No shader support! Using fixed function fallback. (no fancy visuals for you)");
+        avoidshaders = true;
+        intel_quadric_bug = 1;
+
+        extern int maxtexsize;
+        maxtexsize = 256;
+
+        if(!hasOQ) waterreflect = waterrefract = 0;
+    } 
+    //if(floatvtx) conoutf("WARNING: Using floating point vertexes. (use \"/floatvtx 0\" to disable)");
+
+    extern int useshaders;
+    if(!useshaders || (useshaders<0 && avoidshaders) || !hasMT || !strstr(exts, "GL_ARB_vertex_program") || !strstr(exts, "GL_ARB_fragment_program"))
+    {
+        if(!hasMT || !strstr(exts, "GL_ARB_vertex_program") || !strstr(exts, "GL_ARB_fragment_program")) conoutf("WARNING: No shader support! Using fixed function fallback. (no fancy visuals for you)");
+        else if(useshaders<0 && !hasTF) conoutf("WARNING: Disabling shaders for extra performance. (use \"/shaders 1\" to enable shaders if desired)");
         renderpath = R_FIXEDFUNCTION;
-        if(strstr(vendor, "ATI") && !shaderprecision) ati_texgen_bug = 1;
+        if(strstr(vendor, "ATI") && !useshaders) ati_texgen_bug = 1;
         else if(strstr(vendor, "NVIDIA")) nvidia_texgen_bug = 1;
         if(ati_texgen_bug) conoutf("WARNING: Using ATI texgen bug workaround. (use \"/ati_texgen_bug 0\" to disable if unnecessary)");
         if(nvidia_texgen_bug) conoutf("WARNING: Using NVIDIA texgen bug workaround. (use \"/nvidia_texgen_bug 0\" to disable if unnecessary)");
@@ -164,8 +246,10 @@ void gl_init(int w, int h, int bpp, int depth, int fsaa)
     else
     {
         glGenPrograms_ =            (PFNGLGENPROGRAMSARBPROC)           getprocaddress("glGenProgramsARB");
+        glDeletePrograms_ =         (PFNGLDELETEPROGRAMSARBPROC)        getprocaddress("glDeleteProgramsARB");
         glBindProgram_ =            (PFNGLBINDPROGRAMARBPROC)           getprocaddress("glBindProgramARB");
         glProgramString_ =          (PFNGLPROGRAMSTRINGARBPROC)         getprocaddress("glProgramStringARB");
+        glGetProgramiv_ =           (PFNGLGETPROGRAMIVARBPROC)          getprocaddress("glGetProgramivARB");
         glProgramEnvParameter4f_ =  (PFNGLPROGRAMENVPARAMETER4FARBPROC) getprocaddress("glProgramEnvParameter4fARB");
         glProgramEnvParameter4fv_ = (PFNGLPROGRAMENVPARAMETER4FVARBPROC)getprocaddress("glProgramEnvParameter4fvARB");
         renderpath = R_ASMSHADER;
@@ -192,43 +276,13 @@ void gl_init(int w, int h, int bpp, int depth, int fsaa)
                 renderpath = R_GLSLANG;
                 conoutf("Rendering using the OpenGL 1.5 GLSL shader path.");
 #ifdef __APPLE__
+                //if(osversion<0x1050) ??
                 apple_glsldepth_bug = 1;
 #endif
                 if(apple_glsldepth_bug) conoutf("WARNING: Using Apple GLSL depth bug workaround. (use \"/apple_glsldepth_bug 0\" to disable if unnecessary");
             }
         }
         if(renderpath==R_ASMSHADER) conoutf("Rendering using the OpenGL 1.5 assembly shader path.");
-
-        glEnable(GL_VERTEX_PROGRAM_ARB);
-        glEnable(GL_FRAGMENT_PROGRAM_ARB);
-    }
-
-    if(strstr(exts, "GL_ARB_occlusion_query"))
-    {
-        GLint bits;
-        glGetQueryiv_ = (PFNGLGETQUERYIVARBPROC)getprocaddress("glGetQueryivARB");
-        glGetQueryiv_(GL_SAMPLES_PASSED_ARB, GL_QUERY_COUNTER_BITS_ARB, &bits);
-        if(bits)
-        {
-            glGenQueries_ =        (PFNGLGENQUERIESARBPROC)       getprocaddress("glGenQueriesARB");
-            glDeleteQueries_ =     (PFNGLDELETEQUERIESARBPROC)    getprocaddress("glDeleteQueriesARB");
-            glBeginQuery_ =        (PFNGLBEGINQUERYARBPROC)       getprocaddress("glBeginQueryARB");
-            glEndQuery_ =          (PFNGLENDQUERYARBPROC)         getprocaddress("glEndQueryARB");
-            glGetQueryObjectiv_ =  (PFNGLGETQUERYOBJECTIVARBPROC) getprocaddress("glGetQueryObjectivARB");
-            glGetQueryObjectuiv_ = (PFNGLGETQUERYOBJECTUIVARBPROC)getprocaddress("glGetQueryObjectuivARB");
-            hasOQ = true;
-            //conoutf("Using GL_ARB_occlusion_query extension.");
-#if defined(__APPLE__) && SDL_BYTEORDER == SDL_BIG_ENDIAN
-            if(strstr(vendor, "ATI")) ati_oq_bug = 1; 
-#endif            
-            if(ati_oq_bug) conoutf("WARNING: Using ATI occlusion query bug workaround. (use \"/ati_oq_bug 0\" to disable if unnecessary)");
-        }
-    }
-    if(!hasOQ)
-    {
-        conoutf("WARNING: No occlusion query support! (large maps may be SLOW)");
-        extern int zpass;
-        if(renderpath==R_FIXEDFUNCTION) zpass = 0;
     }
 
     if(renderpath!=R_FIXEDFUNCTION)
@@ -253,40 +307,82 @@ void gl_init(int w, int h, int bpp, int depth, int fsaa)
         glGenFramebuffers_         = (PFNGLGENFRAMEBUFFERSEXTPROC)        getprocaddress("glGenFramebuffersEXT");
         glFramebufferTexture2D_    = (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC)   getprocaddress("glFramebufferTexture2DEXT");
         glFramebufferRenderbuffer_ = (PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC)getprocaddress("glFramebufferRenderbufferEXT");
+        glGenerateMipmap_          = (PFNGLGENERATEMIPMAPEXTPROC)         getprocaddress("glGenerateMipmapEXT");
         hasFBO = true;
         //conoutf("Using GL_EXT_framebuffer_object extension.");
     }
     else conoutf("WARNING: No framebuffer object support. (reflective water may be slow)");
 
+    if(strstr(exts, "GL_EXT_packed_depth_stencil") || strstr(exts, "GL_NV_packed_depth_stencil"))
+    {
+        hasDS = true;
+        //conoutf("Using GL_EXT_packed_depth_stencil extension.");
+    }
+
+    if(strstr(exts, "GL_EXT_blend_minmax"))
+    {
+        glBlendEquation_ = (PFNGLBLENDEQUATIONEXTPROC) getprocaddress("glBlendEquationEXT");
+        hasBE = true;
+#ifdef __APPLE__
+        if(osversion<0x1050) apple_minmax_bug = 1;
+#endif
+#ifndef WIN32
+        if(strstr(vendor, "ATI")) apple_minmax_bug = 1;
+#endif
+        //conoutf("Using GL_EXT_blend_minmax extension.");
+    }
+
+    if(strstr(exts, "GL_ARB_texture_float") || strstr(exts, "GL_ATI_texture_float"))
+    {
+        hasTF = true;
+        //conoutf("Using GL_ARB_texture_float extension");
+        shadowmap = 1;
+    }
+
     if(strstr(exts, "GL_ARB_texture_cube_map"))
     {
+        glGetIntegerv(GL_MAX_CUBE_MAP_TEXTURE_SIZE_ARB, (GLint *)&hwcubetexsize);
         hasCM = true;
         //conoutf("Using GL_ARB_texture_cube_map extension.");
     }
     else conoutf("WARNING: No cube map texture support. (no reflective glass)");
 
-    if(!strstr(exts, "GL_ARB_texture_non_power_of_two")) conoutf("WARNING: Non-power-of-two textures not supported!");
-
+    if(strstr(exts, "GL_ARB_texture_non_power_of_two")) 
+    {
+        hasNP2 = true;
+        //conoutf("Using GL_ARB_texture_non_power_of_two extension.");
+    }
+    else conoutf("WARNING: Non-power-of-two textures not supported!");
+     
     if(strstr(exts, "GL_EXT_texture_compression_s3tc"))
     {
         hasTC = true;
         //conoutf("Using GL_EXT_texture_compression_s3tc extension.");
     }
+    
+#ifdef __APPLE__
+     if((renderpath!=R_FIXEDFUNCTION) && (osversion>=0x1050)) 
+     {
+        apple_ff_bug = 1;
+        conoutf("WARNING: Using leopard OPTION ARB_position_invariant workaround (use \"/apple_ff_bug 0\" to disable if unnecessary)");
+    }
+#endif
 
-    extern int maxtexsize;
-    glGetIntegerv(GL_MAX_TEXTURE_SIZE, (GLint *)&maxtexsize);
+    if(strstr(exts, "GL_EXT_texture_filter_anisotropic"))
+    {
+       hasAF = true;
+       glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, (GLint *)&hwmaxaniso);
+       //conoutf("Using GL_EXT_texture_filter_anisotropic extension.");
+    }
 
     if(fsaa) glEnable(GL_MULTISAMPLE);
 
-    
-    exec("data/stdshader.cfg");
-    defaultshader = lookupshaderbyname("default");
-    notextureshader = lookupshaderbyname("notexture");
-    nocolorshader = lookupshaderbyname("nocolor");
-    defaultshader->set();
+    glGetIntegerv(GL_MAX_TEXTURE_SIZE, (GLint *)&hwtexsize);
+
+    inittmus();
 }
 
-VARF(wireframe, 0, 0, 1, if(noedit(true)) wireframe = 0);
+VAR(wireframe, 0, 0, 1);
 
 vec worldpos, camright, camup;
 
@@ -331,8 +427,10 @@ bool isthirdperson() { return player!=camera1 || player->state==CS_DEAD || (refl
 
 void recomputecamera()
 {
+    cl->setupcamera();
     if(deathcam && player->state!=CS_DEAD) deathcam = false;
-    if((editmode || !thirdperson) && player->state!=CS_DEAD)
+    extern int testanims;
+    if(((editmode && !testanims) || !thirdperson) && player->state!=CS_DEAD)
     {
         //if(camera1->state==CS_DEAD) camera1->o.z -= camera1->eyeheight-0.8f;
         camera1 = player;
@@ -359,10 +457,11 @@ void recomputecamera()
     }
 }
 
-void project(float fovy, float aspect, int farplane)
+void project(float fovy, float aspect, int farplane, bool flipx = false, bool flipy = false)
 {
     glMatrixMode(GL_PROJECTION);
     glLoadIdentity();
+    if(flipx || flipy) glScalef(flipx ? -1 : 1, flipy ? -1 : 1, 1);
     gluPerspective(fovy, aspect, 0.54f, farplane);
     glMatrixMode(GL_MODELVIEW);
 }
@@ -403,23 +502,40 @@ void undoclipmatrix()
 VAR(reflectclip, 0, 6, 64);
 VARP(reflectmms, 0, 1, 1);
 
-void setfogplane(float scale, float z)
+void setfogplane(const plane &p, bool flush)
 {
-    float fogplane[4] = {1, 0, 0, 0};
-    if(scale)
+    static float fogselect[4] = {0, 0, 0, 0};
+    setenvparamfv("fogselect", SHPARAM_VERTEX, 8, fogselect);
+    setenvparamfv("fogplane", SHPARAM_VERTEX, 9, p.v);
+    if(flush)
     {
-        fogplane[0] = 0;
+        flushenvparam(SHPARAM_VERTEX, 8);
+        flushenvparam(SHPARAM_VERTEX, 9);
+    }
+}
+
+void setfogplane(float scale, float z, bool flush)
+{
+    float fogselect[4] = {1, 1, 1, 1}, fogplane[4] = {0, 0, 0, 0};
+    if(scale || z)
+    {
+        loopk(4) fogselect[k] = 0;
+
         fogplane[2] = scale;
         fogplane[3] = -z;
     }  
+    setenvparamfv("fogselect", SHPARAM_VERTEX, 8, fogselect);
     setenvparamfv("fogplane", SHPARAM_VERTEX, 9, fogplane);
-//    flushenvparam(SHPARAM_VERTEX, 9);
+    if(flush)
+    {
+        flushenvparam(SHPARAM_VERTEX, 8);
+        flushenvparam(SHPARAM_VERTEX, 9);
+    }
 }
 
-extern void rendercaustics(float z, bool refract);
-
 void drawreflection(float z, bool refract, bool clear)
 {
+    uchar wcol[3];
     getwatercolour(wcol);
     float fogc[4] = { wcol[0]/256.0f, wcol[1]/256.0f, wcol[2]/256.0f, 1.0f };
     
@@ -434,8 +550,8 @@ void drawreflection(float z, bool refract, bool clear)
     if(refract) refracting = z;
 
     float oldfogstart, oldfogend, oldfogcolor[4];
-
-    if((renderpath!=R_FIXEDFUNCTION && refract) || camera1->o.z < z)
+    if(renderpath==R_FIXEDFUNCTION && refract && refractfog) glDisable(GL_FOG);
+    else if((renderpath!=R_FIXEDFUNCTION && refract) || camera1->o.z < z)
     {
         glGetFloatv(GL_FOG_START, &oldfogstart);
         glGetFloatv(GL_FOG_END, &oldfogend);
@@ -475,7 +591,7 @@ void drawreflection(float z, bool refract, bool clear)
     //if(!refract && explicitsky) drawskybox(farplane, true, z);
 
     GLfloat clipmatrix[16];
-    if(reflectclip) 
+    if(reflectclip)
     {
         float zoffset = reflectclip/4.0f, zclip;
         if(refract)
@@ -494,12 +610,10 @@ void drawreflection(float z, bool refract, bool clear)
 
     //if(!refract && explicitsky) drawskylimits(true, z);
 
-    extern void renderreflectedgeom(float z, bool refract);
-    renderreflectedgeom(z, refract);
+    renderreflectedgeom(z, refract, caustics && refract, (renderpath!=R_FIXEDFUNCTION || refractfog) && refract);
 
-    if(refract) rendercaustics(z, refract);
+    if(refract && renderpath!=R_FIXEDFUNCTION) glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_FALSE);
 
-    extern void renderreflectedmapmodels(float z, bool refract);
     if(reflectmms) renderreflectedmapmodels(z, refract);
     cl->rendergame();
 
@@ -511,14 +625,14 @@ void drawreflection(float z, bool refract, bool clear)
         if(reflectclip) setclipmatrix(clipmatrix);
     }
 
+    setfogplane(1, z);
+    if(refract) rendergrass();
     rendermaterials(z, refract);
+    render_particles(0);
+
+    if(refract && renderpath!=R_FIXEDFUNCTION) glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
 
     setfogplane();
-    defaultshader->set();
-
-    glDisable(GL_FOG);
-    render_particles(0);
-    glEnable(GL_FOG);
 
     if(reflectclip) undoclipmatrix();
 
@@ -529,7 +643,8 @@ void drawreflection(float z, bool refract, bool clear)
         glCullFace(GL_FRONT);
     }
 
-    if((renderpath!=R_FIXEDFUNCTION && refract) || camera1->o.z < z)
+    if(renderpath==R_FIXEDFUNCTION && refract && refractfog) glEnable(GL_FOG);
+    else if((renderpath!=R_FIXEDFUNCTION && refract) || camera1->o.z < z)
     {
         glFogf(GL_FOG_START, oldfogstart);
         glFogf(GL_FOG_END, oldfogend);
@@ -540,7 +655,7 @@ void drawreflection(float z, bool refract, bool clear)
     reflecting = 0;
 }
 
-static void setfog(bool underwater)
+static void setfog(int fogmat)
 {
     glFogi(GL_FOG_START, (fog+64)/8);
     glFogi(GL_FOG_END, fog);
@@ -548,20 +663,26 @@ static void setfog(bool underwater)
     glFogfv(GL_FOG_COLOR, fogc);
     glClearColor(fogc[0], fogc[1], fogc[2], 1.0f);
     
-    if(underwater)
+    if(fogmat==MAT_WATER || fogmat==MAT_LAVA)
     {
-        getwatercolour(wcol);
-        float fogwc[4] = { wcol[0]/256.0f, wcol[1]/256.0f, wcol[2]/256.0f, 1.0f };
+        uchar col[3];
+        if(fogmat==MAT_WATER) getwatercolour(col);
+        else getlavacolour(col);
+        float fogwc[4] = { col[0]/256.0f, col[1]/256.0f, col[2]/256.0f, 1 };
         glFogfv(GL_FOG_COLOR, fogwc); 
         glFogi(GL_FOG_START, 0);
-        glFogi(GL_FOG_END, min(fog, max(waterfog*4, 32)));//(fog+96)/8);
+        glFogi(GL_FOG_END, min(fog, max((fogmat==MAT_WATER ? waterfog : lavafog)*4, 32)));//(fog+96)/8);
     }
 
     if(renderpath!=R_FIXEDFUNCTION) setfogplane();
 }
 
+bool envmapping = false;
+
 void drawcubemap(int size, const vec &o, float yaw, float pitch)
 {
+    envmapping = true;
+
     physent *oldcamera = camera1;
     static physent cmcamera;
     cmcamera = *player;
@@ -575,16 +696,16 @@ void drawcubemap(int size, const vec &o, float yaw, float pitch)
    
     defaultshader->set();
 
-    cube &c = lookupcube(int(o.x), int(o.y), int(o.z));
-    bool underwater = c.ext && c.ext->material == MAT_WATER;
+    int fogmat = lookupmaterial(o);
+    if(fogmat!=MAT_WATER && fogmat!=MAT_LAVA) fogmat = MAT_AIR;
 
-    setfog(underwater);    
+    setfog(fogmat);
 
     glClear(GL_DEPTH_BUFFER_BIT);
 
     int farplane = max(max(fog*2, 384), hdr.worldsize*2);
 
-    project(90.0f, 1.0f, farplane);
+    project(90.0f, 1.0f, farplane, true, true);
 
     transplayer();
 
@@ -616,23 +737,29 @@ void drawcubemap(int size, const vec &o, float yaw, float pitch)
     glDisable(GL_TEXTURE_2D);
 
     camera1 = oldcamera;
+    envmapping = false;
 }
 
 VAR(hudgunfov, 10, 65, 150);
 
-void gl_drawframe(int w, int h, float curfps)
+void gl_drawhud(int w, int h, int fogmat);
+
+void gl_drawframe(int w, int h)
 {
     defaultshader->set();
 
     recomputecamera();
-    
+   
+    cleardynlights();
+    cl->adddynlights();
+
     float fovy = (float)fov*h/w;
     float aspect = w/(float)h;
-    cube &c = lookupcube((int)camera1->o.x, (int)camera1->o.y, int(camera1->o.z + camera1->aboveeye*0.5f));
-    bool underwater = c.ext && c.ext->material == MAT_WATER;
-    
-    setfog(underwater);
-    if(underwater)
+    int fogmat = lookupmaterial(vec(camera1->o.x, camera1->o.y, camera1->o.z + camera1->aboveeye*0.5f));
+    if(fogmat!=MAT_WATER && fogmat!=MAT_LAVA) fogmat = MAT_AIR;
+
+    setfog(fogmat);
+    if(fogmat!=MAT_AIR)
     {
         fovy += (float)sin(lastmillis/1000.0)*2.0f;
         aspect += (float)sin(lastmillis/1000.0+PI)*0.1f;
@@ -646,19 +773,23 @@ void gl_drawframe(int w, int h, float curfps)
 
     glEnable(GL_TEXTURE_2D);
 
-    glPolygonMode(GL_FRONT_AND_BACK, wireframe ? GL_LINE : GL_FILL);
+    glPolygonMode(GL_FRONT_AND_BACK, wireframe && editmode ? GL_LINE : GL_FILL);
     
     xtravertsva = xtraverts = glde = 0;
 
     if(!hasFBO) drawreflections();
 
-    glClear(GL_DEPTH_BUFFER_BIT|(wireframe ? GL_COLOR_BUFFER_BIT : 0)|(hasstencil ? GL_STENCIL_BUFFER_BIT : 0));
-
     visiblecubes(worldroot, hdr.worldsize/2, 0, 0, 0, w, h, fov);
     
+    extern GLuint shadowmapfb;
+    if(shadowmap && !shadowmapfb) rendershadowmap();
+
+    glClear(GL_DEPTH_BUFFER_BIT|(wireframe && editmode ? GL_COLOR_BUFFER_BIT : 0)|(hasstencil ? GL_STENCIL_BUFFER_BIT : 0));
+
     if(limitsky()) drawskybox(farplane, true);
 
-    rendergeom();
+    bool causticspass = caustics && fogmat==MAT_WATER && lookupmaterial(vec(camera1->o.x, camera1->o.y, camera1->o.z + camera1->aboveeye*1.25f));
+    rendergeom(causticspass);
 
     queryreflections();
 
@@ -666,14 +797,10 @@ void gl_drawframe(int w, int h, float curfps)
 
     rendermapmodels();
 
-    defaultshader->set();
-
-    cl->rendergame();
-
-    if(underwater) 
+    if(!waterrefract) 
     {
-        cube &s = lookupcube((int)camera1->o.x, (int)camera1->o.y, int(camera1->o.z + camera1->aboveeye*1.25f));
-        if(s.ext && s.ext->material==MAT_WATER) rendercaustics(0, false);
+        defaultshader->set();
+        cl->rendergame();
     }
 
     defaultshader->set();
@@ -682,11 +809,19 @@ void gl_drawframe(int w, int h, float curfps)
 
     if(hasFBO) drawreflections();
 
+    if(waterrefract)
+    {
+        defaultshader->set();
+        cl->rendergame();
+    }
+
     renderwater();
     rendergrass();
-    rendermaterials();
 
-    if(!isthirdperson()) 
+    rendermaterials();
+    render_particles(curtime);
+
+    if(!isthirdperson())
     {
         project(hudgunfov, aspect, farplane);
         cl->drawhudgun();
@@ -694,20 +829,17 @@ void gl_drawframe(int w, int h, float curfps)
     }
 
     glDisable(GL_FOG);
-    defaultshader->set();
-    
-    render_particles(curtime);
-
-    g3d_render();
-
     glDisable(GL_CULL_FACE);
 
     renderfullscreenshader(w, h);
-    
+   
+    defaultshader->set();
+    g3d_render();
+
     glDisable(GL_TEXTURE_2D);
     notextureshader->set();
 
-    gl_drawhud(w, h, (int)curfps, 0, verts.length(), underwater);
+    gl_drawhud(w, h, fogmat);
 
     glEnable(GL_CULL_FACE);
     glEnable(GL_FOG);
@@ -722,21 +854,42 @@ void damageblend(int n) { dblend += n; }
 
 VARP(hidestats, 0, 0, 1);
 VARP(hidehud, 0, 0, 1);
-VARP(crosshairfx, 0, 0, 1);
+VARP(crosshairfx, 0, 1, 1);
+
+static Texture *crosshair = NULL;
+
+void loadcrosshair(const char *name)
+{
+    crosshair = textureload(name, 3, true);
+    if(crosshair==notexture) crosshair = textureload("data/crosshair.png", 3, true);
+}
+
+COMMAND(loadcrosshair, "s");
+
+void writecrosshairs(FILE *f)
+{
+    fprintf(f, "loadcrosshair \"%s\"\n", crosshair->name);
+    fprintf(f, "\n");
+}
 
 void drawcrosshair(int w, int h)
 {
     bool windowhit = g3d_windowhit(true, false);
-    if(!windowhit && (hidehud || player->state==CS_SPECTATOR)) return;;
+    if(!windowhit && (hidehud || player->state==CS_SPECTATOR)) return;
 
     static Texture *cursor = NULL;
-    if(!cursor) cursor = textureload("data/guicursor.png", 3, false);
+    if(!cursor) cursor = textureload("data/guicursor.png", 3, true);
+    if(!crosshair) crosshair = textureload("data/crosshair.png", 3, true);
     if((windowhit ? cursor : crosshair)->bpp==32) glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
     else glBlendFunc(GL_ONE, GL_ONE);
-    glColor3f(1, 1, 1);
+    float r = 1, g = 1, b = 1;
+    if(!windowhit && crosshairfx) cl->crosshaircolor(r, g, b);
+    glColor3f(r, g, b);
     float chsize = (windowhit ? cursorsize : crosshairsize)*w/300.0f;
-    float x = w*1.5f - (windowhit ? 0 : chsize/2.0f);
-    float y = h*1.5f - (windowhit ? 0 : chsize/2.0f);
+    float cx = 0.5f, cy = 0.5f;
+    if(windowhit) g3d_cursorpos(cx, cy);
+    float x = cx*w*3.0f - (windowhit ? 0 : chsize/2.0f);
+    float y = cy*h*3.0f - (windowhit ? 0 : chsize/2.0f);
     glBindTexture(GL_TEXTURE_2D, (windowhit ? cursor : crosshair)->gl);
     glBegin(GL_QUADS);
     glTexCoord2d(0.0, 0.0); glVertex2f(x,          y);
@@ -746,7 +899,9 @@ void drawcrosshair(int w, int h)
     glEnd();
 }
 
-void gl_drawhud(int w, int h, int curfps, int nquads, int curvert, bool underwater)
+VARP(showfpsrange, 0, 0, 1);
+
+void gl_drawhud(int w, int h, int fogmat)
 {
     if(editmode && !hidehud)
     {
@@ -768,9 +923,18 @@ void gl_drawhud(int w, int h, int curfps, int nquads, int curvert, bool underwat
     glLoadIdentity();
     glOrtho(0, w, h, 0, -1, 1);
 
+    glColor3f(1, 1, 1);
+
+    extern int debugsm;
+    if(debugsm)
+    {
+        extern void viewshadowmap();
+        viewshadowmap();
+    }
+
     glEnable(GL_BLEND);
 
-    if(dblend || underwater)
+    if(dblend || fogmat==MAT_WATER || fogmat==MAT_LAVA)
     {
         glDepthMask(GL_FALSE);
         glBlendFunc(GL_ZERO, GL_SRC_COLOR);
@@ -778,11 +942,13 @@ void gl_drawhud(int w, int h, int curfps, int nquads, int curvert, bool underwat
         if(dblend) glColor3f(1.0f, 0.1f, 0.1f);
         else
         {
-            getwatercolour(wcol);
-            float maxc = max(wcol[0], max(wcol[1], wcol[2]));
-            float wblend[3];
-            loopi(3) wblend[i] = wcol[i] / min(32 + maxc*7/8, 255);
-            glColor3fv(wblend);
+            uchar col[3];
+            if(fogmat==MAT_WATER) getwatercolour(col);
+            else getlavacolour(col);
+            float maxc = max(col[0], max(col[1], col[2]));
+            float blend[3];
+            loopi(3) blend[i] = col[i] / min(32 + maxc*7/8, 255);
+            glColor3fv(blend);
             //glColor3f(0.1f, 0.5f, 1.0f);
         }
         glVertex2i(0, 0);
@@ -816,7 +982,11 @@ void gl_drawhud(int w, int h, int curfps, int nquads, int curvert, bool underwat
 
         if(!hidestats)
         {
-            draw_textf("fps %d", w*3-5*FONTH, h*3-100, curfps);
+            extern void getfps(int &fps, int &bestdiff, int &worstdiff);
+            int fps, bestdiff, worstdiff;
+            getfps(fps, bestdiff, worstdiff);
+            if(showfpsrange) draw_textf("fps %d+%d-%d", w*3-7*FONTH, h*3-100, fps, bestdiff, worstdiff);
+            else draw_textf("fps %d", w*3-5*FONTH, h*3-100, fps);
 
             if(editmode)
             {
@@ -826,7 +996,15 @@ void gl_drawhud(int w, int h, int curfps, int nquads, int curvert, bool underwat
             }
         }
 
-        if(editmode) draw_text(executeret("if (enthavesel) [ result ( concatword (entget) \" : \" (enthavesel) \" selected\" ) ] [ result ]"), FONTH/2, abovegameplayhud);
+        if(editmode) 
+        {
+            char *editinfo = executeret("edithud");
+            if(editinfo)
+            {
+                draw_text(editinfo, FONTH/2, abovegameplayhud);
+                DELETEA(editinfo);
+            }
+        }
 
         cl->gameplayhud(w, h);
         render_texture_panel(w, h);
