@@ -2,13 +2,65 @@
 // runs dedicated or as client coroutine
 
 #include "pch.h"
-#include "engine.h" 
 
 #ifdef STANDALONE
+#include "cube.h"
+#include "iengine.h"
+#include "igame.h"
 void localservertoclient(int chan, uchar *buf, int len) {}
-void fatal(char *s, char *o) { cleanupserver(); printf("servererror: %s\n", s); exit(1); }
+void fatal(const char *s, ...) 
+{ 
+    void cleanupserver(); 
+    cleanupserver(); 
+    s_sprintfdlv(msg,s,s);
+    printf("servererror: %s\n", msg); 
+    exit(EXIT_FAILURE); 
+}
+#else
+#include "engine.h"
 #endif
 
+igameclient     *cl = NULL;
+igameserver     *sv = NULL;
+iclientcom      *cc = NULL;
+icliententities *et = NULL;
+
+hashtable<const char *, igame *> *gamereg = NULL;
+
+vector<char *> gameargs;
+
+void registergame(const char *name, igame *ig)
+{
+    if(!gamereg) gamereg = new hashtable<const char *, igame *>;
+    (*gamereg)[name] = ig;
+}
+
+void initgame(const char *game)
+{
+    igame **ig = gamereg->access(game);
+    if(!ig) fatal("cannot start game module: ", game);
+    sv = (*ig)->newserver();
+    cl = (*ig)->newclient();
+    if(cl)
+    {
+        cc = cl->getcom();
+        et = cl->getents();
+        cl->initclient();
+    }
+    loopv(gameargs)
+    {
+        if(!cl || !cl->clientoption(gameargs[i]))
+        {
+            if(!sv->serveroption(gameargs[i])) 
+#ifdef STANDALONE
+                printf("unknown command-line option: %s\n", gameargs[i]);
+#else
+                conoutf("unknown command-line option: %s", gameargs[i]);
+#endif
+        }
+    }
+}
+
 // all network traffic is in 32bit ints, which are then compressed using the following simple scheme (assumes that most values are small).
 
 void putint(ucharbuf &p, int n)
@@ -81,6 +133,23 @@ void getstring(char *text, ucharbuf &p, int len)
     while(*t++);
 }
 
+void filtertext(char *dst, const char *src, bool whitespace, int len)
+{
+    for(int c = *src; c; c = *++src)
+    {
+        switch(c)
+        {
+        case '\f': ++src; continue;
+        }
+        if(isspace(c) ? whitespace : isprint(c))
+        {
+            *dst++ = c;
+            if(!--len) break;
+        }
+    }
+    *dst = '\0';
+}
+
 enum { ST_EMPTY, ST_LOCAL, ST_TCPIP };
 
 struct client                   // server side version of "dynent" type
@@ -99,7 +168,12 @@ size_t bsend = 0, brec = 0;
 int laststatus = 0; 
 ENetSocket pongsock = ENET_SOCKET_NULL;
 
-void sendfile(int cn, int chan, FILE *file)
+void cleanupserver()
+{
+    if(serverhost) enet_host_destroy(serverhost);
+}
+
+void sendfile(int cn, int chan, FILE *file, const char *format, ...)
 {
 #ifndef STANDALONE
     extern ENetHost *clienthost;
@@ -115,9 +189,30 @@ void sendfile(int cn, int chan, FILE *file)
 
     fseek(file, 0, SEEK_END);
     int len = ftell(file);
-    ENetPacket *packet = enet_packet_create(NULL, len, ENET_PACKET_FLAG_RELIABLE);
+    bool reliable = false;
+    if(*format=='r') { reliable = true; ++format; }
+    ENetPacket *packet = enet_packet_create(NULL, MAXTRANS+len, ENET_PACKET_FLAG_RELIABLE);
     rewind(file);
-    fread(packet->data, 1, len, file);
+
+    ucharbuf p(packet->data, packet->dataLength);
+    va_list args;
+    va_start(args, format);
+    while(*format) switch(*format++)
+    {
+        case 'i':
+        {
+            int n = isdigit(*format) ? *format++-'0' : 1;
+            loopi(n) putint(p, va_arg(args, int));
+            break;
+        }
+        case 's': sendstring(va_arg(args, const char *), p); break;
+        case 'l': putint(p, len); break;
+    }
+    va_end(args);
+    enet_packet_resize(packet, p.length()+len);
+
+    fread(&packet->data[p.length()], 1, len, file);
+    enet_packet_resize(packet, p.length()+len);
 
     if(cn >= 0) enet_peer_send(clients[cn]->peer, chan, packet);
 #ifndef STANDALONE
@@ -134,11 +229,12 @@ void *getinfo(int i)    { return !clients.inrange(i) || clients[i]->type==ST_EMP
 int getnumclients()     { return clients.length(); }
 uint getclientip(int n) { return clients.inrange(n) && clients[n]->type==ST_TCPIP ? clients[n]->peer->address.host : 0; }
 
-void sendpacket(int n, int chan, ENetPacket *packet)
+void sendpacket(int n, int chan, ENetPacket *packet, int exclude)
 {
     if(n<0)
     {
-        loopv(clients) sendpacket(i, chan, packet);
+        sv->recordpacket(chan, packet->data, packet->dataLength);
+        loopv(clients) if(i!=exclude) sendpacket(i, chan, packet);
         return;
     }
     switch(clients[n]->type)
@@ -158,6 +254,7 @@ void sendpacket(int n, int chan, ENetPacket *packet)
 
 void sendf(int cn, int chan, const char *format, ...)
 {
+    int exclude = -1;
     bool reliable = false;
     if(*format=='r') { reliable = true; ++format; }
     ENetPacket *packet = enet_packet_create(NULL, MAXTRANS, reliable ? ENET_PACKET_FLAG_RELIABLE : 0);
@@ -166,6 +263,18 @@ void sendf(int cn, int chan, const char *format, ...)
     va_start(args, format);
     while(*format) switch(*format++)
     {
+        case 'x':
+            exclude = va_arg(args, int);
+            break;
+
+        case 'v':
+        {
+            int n = va_arg(args, int);
+            int *v = va_arg(args, int *);
+            loopi(n) putint(p, v[i]);
+            break;
+        }
+
         case 'i': 
         {
             int n = isdigit(*format) ? *format++-'0' : 1;
@@ -173,14 +282,23 @@ void sendf(int cn, int chan, const char *format, ...)
             break;
         }
         case 's': sendstring(va_arg(args, const char *), p); break;
+        case 'm':
+        {
+            int n = va_arg(args, int);
+            enet_packet_resize(packet, packet->dataLength+n);
+            p.buf = packet->data;
+            p.maxlen += n;
+            p.put(va_arg(args, uchar *), n);
+            break;
+        }
     }
     va_end(args);
     enet_packet_resize(packet, p.length());
-    sendpacket(cn, chan, packet);
+    sendpacket(cn, chan, packet, exclude);
     if(packet->referenceCount==0) enet_packet_destroy(packet);
 }
 
-char *disc_reasons[] = { "normal", "end of packet", "client num", "kicked/banned", "tag type", "ip is banned", "server is in private mode", "server FULL (maxclients)" };
+const char *disc_reasons[] = { "normal", "end of packet", "client num", "kicked/banned", "tag type", "ip is banned", "server is in private mode", "server FULL (maxclients)" };
 
 void disconnect_client(int n, int reason)
 {
@@ -191,6 +309,8 @@ void disconnect_client(int n, int reason)
     sv->clientdisconnect(n);
     clients[n]->type = ST_EMPTY;
     clients[n]->peer->data = NULL;
+    sv->deleteinfo(clients[n]->info);
+    clients[n]->info = NULL;
     sv->sendservmsg(s);
 }
 
@@ -221,7 +341,7 @@ client &addclient()
 {
     loopv(clients) if(clients[i]->type==ST_EMPTY)
     {
-        sv->resetinfo(clients[i]->info);
+        clients[i]->info = sv->newinfo();
         return *clients[i];
     }
     client *c = new client;
@@ -231,9 +351,10 @@ client &addclient()
     return *c;
 }
 
-int nonlocalclients = 0;
+int localclients = 0, nonlocalclients = 0;
 
 bool hasnonlocalclients() { return nonlocalclients!=0; }
+bool haslocalclients() { return localclients!=0; }
 
 void sendpongs()        // reply all server info requests
 {
@@ -260,56 +381,60 @@ bool resolverwait(const char *name, ENetAddress *address)
 {
     return enet_address_set_host(address, name) >= 0;
 }
+
+int connectwithtimeout(ENetSocket sock, const char *hostname, ENetAddress &remoteaddress)
+{
+    int result = enet_socket_connect(sock, &remoteaddress);
+    if(result<0) enet_socket_destroy(sock);
+    return result;
+}
 #endif
 
-ENetSocket mssock = ENET_SOCKET_NULL;
-
-void httpgetsend(ENetAddress &ad, char *hostname, char *req, char *ref, char *agent)
+ENetSocket httpgetsend(ENetAddress &remoteaddress, const char *hostname, const char *req, const char *ref, const char *agent, ENetAddress *localaddress = NULL)
 {
-    if(mssock!=ENET_SOCKET_NULL)
-    {
-        enet_socket_destroy(mssock);
-        mssock = ENET_SOCKET_NULL;
-    }
-    if(ad.host==ENET_HOST_ANY)
+    if(remoteaddress.host==ENET_HOST_ANY)
     {
+#ifdef STANDALONE
         printf("looking up %s...\n", hostname);
-        if(!resolverwait(hostname, &ad)) return;
+#endif
+        if(!resolverwait(hostname, &remoteaddress)) return ENET_SOCKET_NULL;
     }
-    mssock = enet_socket_create(ENET_SOCKET_TYPE_STREAM, NULL);
-    if(mssock==ENET_SOCKET_NULL) { printf("could not open socket\n"); return; }
-    if(enet_socket_connect(mssock, &ad)<0) 
+    ENetSocket sock = enet_socket_create(ENET_SOCKET_TYPE_STREAM, localaddress);
+    if(sock==ENET_SOCKET_NULL || connectwithtimeout(sock, hostname, remoteaddress)<0) 
     { 
-        printf("could not connect\n"); 
-        enet_socket_destroy(mssock);
-        mssock = ENET_SOCKET_NULL;
-        return; 
+#ifdef STANDALONE
+        printf(sock==ENET_SOCKET_NULL ? "could not open socket\n" : "could not connect\n"); 
+#endif
+        return ENET_SOCKET_NULL; 
     }
     ENetBuffer buf;
     s_sprintfd(httpget)("GET %s HTTP/1.0\nHost: %s\nReferer: %s\nUser-Agent: %s\n\n", req, hostname, ref, agent);
     buf.data = httpget;
     buf.dataLength = strlen((char *)buf.data);
+#ifdef STANDALONE
     printf("sending request to %s...\n", hostname);
-    enet_socket_send(mssock, NULL, &buf, 1);
+#endif
+    enet_socket_send(sock, NULL, &buf, 1);
+    return sock;
 }  
 
-void httpgetreceive(ENetBuffer &buf)
+bool httpgetreceive(ENetSocket sock, ENetBuffer &buf, int timeout = 0)
 {
-    if(mssock==ENET_SOCKET_NULL) return;
+    if(sock==ENET_SOCKET_NULL) return false;
     enet_uint32 events = ENET_SOCKET_WAIT_RECEIVE;
-    if(enet_socket_wait(mssock, &events, 0) >= 0 && events)
+    if(enet_socket_wait(sock, &events, timeout) >= 0 && events)
     {
-        int len = enet_socket_receive(mssock, NULL, &buf, 1);
+        int len = enet_socket_receive(sock, NULL, &buf, 1);
         if(len<=0)
         {
-            enet_socket_destroy(mssock);
-            mssock = ENET_SOCKET_NULL;
-            return;
+            enet_socket_destroy(sock);
+            return false;
         }
         buf.data = ((char *)buf.data)+len;
         ((char*)buf.data)[0] = 0;
         buf.dataLength -= len;
     }
+    return true;
 }  
 
 uchar *stripheader(uchar *b)
@@ -319,8 +444,10 @@ uchar *stripheader(uchar *b)
     return s ? (uchar *)s : b;
 }
 
+ENetSocket mssock = ENET_SOCKET_NULL;
+ENetAddress msaddress = { ENET_HOST_ANY, ENET_PORT_ANY };
 ENetAddress masterserver = { ENET_HOST_ANY, 80 };
-int updmaster = 0;
+int lastupdatemaster = 0;
 string masterbase;
 string masterpath;
 uchar masterrep[MAXTRANS];
@@ -329,7 +456,8 @@ ENetBuffer masterb;
 void updatemasterserver()
 {
     s_sprintfd(path)("%sregister.do?action=add", masterpath);
-    httpgetsend(masterserver, masterbase, path, sv->servername(), sv->servername());
+    if(mssock!=ENET_SOCKET_NULL) enet_socket_destroy(mssock);
+    mssock = httpgetsend(masterserver, masterbase, path, sv->servername(), sv->servername(), &msaddress);
     masterrep[0] = 0;
     masterb.data = masterrep;
     masterb.dataLength = MAXTRANS-1;
@@ -337,109 +465,170 @@ void updatemasterserver()
 
 void checkmasterreply()
 {
-    bool busy = mssock!=ENET_SOCKET_NULL;
-    httpgetreceive(masterb);
-    if(busy && mssock==ENET_SOCKET_NULL) printf("masterserver reply: %s\n", stripheader(masterrep));
+    if(mssock!=ENET_SOCKET_NULL && !httpgetreceive(mssock, masterb))
+    {
+        mssock = ENET_SOCKET_NULL;
+        printf("masterserver reply: %s\n", stripheader(masterrep));
+    }
 } 
 
+#ifndef STANDALONE
+
+#define RETRIEVELIMIT 20000
+
 uchar *retrieveservers(uchar *buf, int buflen)
 {
+    buf[0] = '\0';
+
     s_sprintfd(path)("%sretrieve.do?item=list", masterpath);
-    httpgetsend(masterserver, masterbase, path, sv->servername(), sv->servername());
+    ENetAddress address = masterserver;
+    ENetSocket sock = httpgetsend(address, masterbase, path, sv->servername(), sv->servername());
+    if(sock==ENET_SOCKET_NULL) return buf;
+    /* only cache this if connection succeeds */
+    masterserver = address;
+
+    s_sprintfd(text)("retrieving servers from %s... (esc to abort)", masterbase);
+    show_out_of_renderloop_progress(0, text);
+
     ENetBuffer eb;
-    buf[0] = 0;
     eb.data = buf;
     eb.dataLength = buflen-1;
-    while(mssock!=ENET_SOCKET_NULL) httpgetreceive(eb);
+   
+    int starttime = SDL_GetTicks(), timeout = 0;
+    while(httpgetreceive(sock, eb, 250))
+    {
+        timeout = SDL_GetTicks() - starttime;
+        show_out_of_renderloop_progress(min(float(timeout)/RETRIEVELIMIT, 1), text);
+        SDL_Event event;
+        while(SDL_PollEvent(&event))
+        {
+            if(event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_ESCAPE) timeout = RETRIEVELIMIT + 1;
+        }
+        if(timeout > RETRIEVELIMIT)
+        {
+            buf[0] = '\0';
+            enet_socket_destroy(sock);
+            return buf;
+        }
+    }
+
     return stripheader(buf);
 }
+#endif
 
 #define DEFAULTCLIENTS 6
 
 int uprate = 0, maxclients = DEFAULTCLIENTS;
-char *sdesc = "", *ip = "", *master = NULL, *adminpass = NULL;
-char *game = "fps";
-bool pubserv = false;
+const char *ip = "", *master = NULL;
+const char *game = "fps";
 
-void serverslice(int seconds, uint timeout)   // main server update, called from main loop in sp, or from below in dedicated server
+#ifdef STANDALONE
+int lastmillis = 0, totalmillis = 0;
+#endif
+
+void serverslice(uint timeout)   // main server update, called from main loop in sp, or from below in dedicated server
 {
-    sv->serverupdate(seconds);
+    localclients = nonlocalclients = 0;
+    loopv(clients) switch(clients[i]->type)
+    {
+        case ST_LOCAL: localclients++; break;
+        case ST_TCPIP: nonlocalclients++; break;
+    }
 
-    if(!serverhost) return;     // below is network only
+    if(!serverhost) 
+    {
+        sv->serverupdate(lastmillis, totalmillis);
+        return;
+    }
+       
+    // below is network only
+
+    lastmillis = totalmillis = (int)enet_time_get();
+    sv->serverupdate(lastmillis, totalmillis);
 
     sendpongs();
     
     if(*masterpath) checkmasterreply();
 
-    if(seconds>updmaster && *masterpath)       // send alive signal to masterserver every hour of uptime
+    if(totalmillis-lastupdatemaster>60*60*1000 && *masterpath)       // send alive signal to masterserver every hour of uptime
     {
         updatemasterserver();
-        updmaster = seconds+60*60;
+        lastupdatemaster = totalmillis;
     }
     
-    nonlocalclients = 0;
-    loopv(clients) if(clients[i]->type==ST_TCPIP) nonlocalclients++;
-
-    if(seconds-laststatus>60)   // display bandwidth stats, useful for server ops
+    if(totalmillis-laststatus>60*1000)   // display bandwidth stats, useful for server ops
     {
-        laststatus = seconds;     
+        laststatus = totalmillis;     
         if(nonlocalclients || bsend || brec) printf("status: %d remote clients, %.1f send, %.1f rec (K/sec)\n", nonlocalclients, bsend/60.0f/1024, brec/60.0f/1024);
         bsend = brec = 0;
     }
 
     ENetEvent event;
-    if(enet_host_service(serverhost, &event, timeout) > 0)
-    switch(event.type)
+    bool serviced = false;
+    while(!serviced)
     {
-        case ENET_EVENT_TYPE_CONNECT:
+        if(enet_host_check_events(serverhost, &event) <= 0)
         {
-            client &c = addclient();
-            c.type = ST_TCPIP;
-            c.peer = event.peer;
-            c.peer->data = &c;
-            char hn[1024];
-            s_strcpy(c.hostname, (enet_address_get_host_ip(&c.peer->address, hn, sizeof(hn))==0) ? hn : "unknown");
-            printf("client connected (%s)\n", c.hostname);
-            int reason = DISC_MAXCLIENTS;
-            if(nonlocalclients<maxclients && !(reason = sv->clientconnect(c.num, c.peer->address.host))) send_welcome(c.num);
-            else disconnect_client(c.num, reason);
-            break;
+            if(enet_host_service(serverhost, &event, timeout) <= 0) break;
+            serviced = true;
         }
-        case ENET_EVENT_TYPE_RECEIVE:
+        switch(event.type)
         {
-            brec += event.packet->dataLength;
-            client *c = (client *)event.peer->data;
-            if(c) process(event.packet, c->num, event.channelID);
-            if(event.packet->referenceCount==0) enet_packet_destroy(event.packet);
-            break;
+            case ENET_EVENT_TYPE_CONNECT:
+            {
+                client &c = addclient();
+                c.type = ST_TCPIP;
+                c.peer = event.peer;
+                c.peer->data = &c;
+                char hn[1024];
+                s_strcpy(c.hostname, (enet_address_get_host_ip(&c.peer->address, hn, sizeof(hn))==0) ? hn : "unknown");
+                printf("client connected (%s)\n", c.hostname);
+                int reason = DISC_MAXCLIENTS;
+                if(nonlocalclients<maxclients && !(reason = sv->clientconnect(c.num, c.peer->address.host))) 
+                {
+                    nonlocalclients++;
+                    send_welcome(c.num);
+                }
+                else disconnect_client(c.num, reason);
+                break;
+            }
+            case ENET_EVENT_TYPE_RECEIVE:
+            {
+                brec += event.packet->dataLength;
+                client *c = (client *)event.peer->data;
+                if(c) process(event.packet, c->num, event.channelID);
+                if(event.packet->referenceCount==0) enet_packet_destroy(event.packet);
+                break;
+            }
+            case ENET_EVENT_TYPE_DISCONNECT: 
+            {
+                client *c = (client *)event.peer->data;
+                if(!c) break;
+                printf("disconnected client (%s)\n", c->hostname);
+                sv->clientdisconnect(c->num);
+                nonlocalclients--;
+                c->type = ST_EMPTY;
+                event.peer->data = NULL;
+                sv->deleteinfo(c->info);
+                c->info = NULL;
+                break;
+            }
+            default:
+                break;
         }
-        case ENET_EVENT_TYPE_DISCONNECT: 
-        {
-            client *c = (client *)event.peer->data;
-            if(!c) break;
-            printf("disconnected client (%s)\n", c->hostname);
-            sv->clientdisconnect(c->num);
-            c->type = ST_EMPTY;
-            event.peer->data = NULL;
-            break;
-        }
-        default:
-            break;
     }
     if(sv->sendpackets()) enet_host_flush(serverhost);
 }
 
-void cleanupserver()
-{
-    if(serverhost) enet_host_destroy(serverhost);
-}
-
 void localdisconnect()
 {
     loopv(clients) if(clients[i]->type==ST_LOCAL) 
     {
         sv->localdisconnect(i);
+        localclients--;
         clients[i]->type = ST_EMPTY;
+        sv->deleteinfo(clients[i]->info);
+        clients[i]->info = NULL;
     }
 }
 
@@ -448,43 +637,17 @@ void localconnect()
     client &c = addclient();
     c.type = ST_LOCAL;
     s_strcpy(c.hostname, "local");
+    localclients++;
     sv->localconnect(c.num);
     send_welcome(c.num); 
 }
 
-hashtable<char *, igame *> *gamereg = NULL;
-
-void registergame(char *name, igame *ig)
-{
-    if(!gamereg) gamereg = new hashtable<char *, igame *>;
-    (*gamereg)[name] = ig;
-}
-
-igameclient     *cl = NULL;
-igameserver     *sv = NULL;
-iclientcom      *cc = NULL;
-icliententities *et = NULL;
-
-void initgame(char *game)
-{
-    igame **ig = gamereg->access(game);
-    if(!ig) fatal("cannot start game module: ", game);
-    sv = (*ig)->newserver();
-    cl = (*ig)->newclient();
-    if(cl)
-    {
-        cc = cl->getcom();
-        et = cl->getents();
-        cl->initclient();
-    }
-}
-
 void initserver(bool dedicated)
 {
     initgame(game);
 
     if(!master) master = sv->getdefaultmaster();
-    char *mid = strstr(master, "/");
+    const char *mid = strstr(master, "/");
     if(!mid) mid = master;
     s_strcpy(masterpath, mid);
     s_strncpy(masterbase, master, mid-master+1);
@@ -492,16 +655,21 @@ void initserver(bool dedicated)
     if(dedicated)
     {
         ENetAddress address = { ENET_HOST_ANY, sv->serverport() };
-        if(*ip && enet_address_set_host(&address, ip)<0) printf("WARNING: server ip not resolved");
+        if(*ip)
+        {
+            if(enet_address_set_host(&address, ip)<0) printf("WARNING: server ip not resolved");
+            else msaddress.host = address.host;
+        }
         serverhost = enet_host_create(&address, maxclients+1, 0, uprate);
-        if(!serverhost) fatal("could not create server host\n");
+        if(!serverhost) fatal("could not create server host");
         loopi(maxclients) serverhost->peers[i].data = NULL;
         address.port = sv->serverinfoport();
         pongsock = enet_socket_create(ENET_SOCKET_TYPE_DATAGRAM, &address);
-        if(pongsock == ENET_SOCKET_NULL) fatal("could not create server info socket\n");
+        if(pongsock == ENET_SOCKET_NULL) fatal("could not create server info socket");
+        else enet_socket_set_option(pongsock, ENET_SOCKOPT_NONBLOCK, 1);
     }
 
-    sv->serverinit(sdesc, adminpass, pubserv);
+    sv->serverinit();
 
     if(dedicated)       // do not return, this becomes main loop
     {
@@ -509,9 +677,11 @@ void initserver(bool dedicated)
         SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);
         #endif
         printf("dedicated server started, waiting for clients...\nCtrl-C to exit\n\n");
-        atexit(cleanupserver);
         atexit(enet_deinitialize);
-        for(;;) serverslice(time(NULL), 5);
+        atexit(cleanupserver);
+        enet_time_set(0);
+        if(*masterpath) updatemasterserver();
+        for(;;) serverslice(5);
     }
 }
 
@@ -527,12 +697,9 @@ bool serveroption(char *opt)
             else maxclients = DEFAULTCLIENTS;
             return true;
         }
-        case 'n': sdesc = opt+2; return true;
         case 'i': ip = opt+2; return true;
         case 'm': master = opt+2; return true;
         case 'g': game = opt+2; return true;
-        case 'p': adminpass = opt+2; return true;
-        case 'o': pubserv = atoi(opt+2)!=0; return true;
         default: return false;
     }
 }
@@ -540,7 +707,7 @@ bool serveroption(char *opt)
 #ifdef STANDALONE
 int main(int argc, char* argv[])
 {   
-    for(int i = 1; i<argc; i++) if(argv[i][0]!='-' || !serveroption(argv[i])) printf("WARNING: unknown commandline option\n");
+    for(int i = 1; i<argc; i++) if(argv[i][0]!='-' || !serveroption(argv[i])) gameargs.add(argv[i]);
     if(enet_initialize()<0) fatal("Unable to initialise network module");
     initserver(true);
     return 0;
