@@ -9,7 +9,7 @@ SDL_Surface *texrotate(SDL_Surface *s, int numrots, int type)
     if(numrots<1 || numrots>5) return s; 
     SDL_Surface *d = SDL_CreateRGBSurface(SDL_SWSURFACE, (numrots&5)==1 ? s->h : s->w, (numrots&5)==1 ? s->w : s->h, s->format->BitsPerPixel, s->format->Rmask, s->format->Gmask, s->format->Bmask, s->format->Amask);
     if(!d) fatal("create surface");    
-    int depth = s->format->BitsPerPixel/8;
+    int depth = s->format->BytesPerPixel;
     loop(y, s->h) loop(x, s->w)
     {
         uchar *src = (uchar *)s->pixels+(y*s->w+x)*depth;
@@ -39,7 +39,7 @@ SDL_Surface *texoffset(SDL_Surface *s, int xoffset, int yoffset)
     if(!xoffset && !yoffset) return s;
     SDL_Surface *d = SDL_CreateRGBSurface(SDL_SWSURFACE, s->w, s->h, s->format->BitsPerPixel, s->format->Rmask, s->format->Gmask, s->format->Bmask, s->format->Amask);
     if(!d) fatal("create surface");
-    int depth = s->format->BitsPerPixel/8;
+    int depth = s->format->BytesPerPixel;
     uchar *src = (uchar *)s->pixels;
     loop(y, s->h)
     {
@@ -52,18 +52,98 @@ SDL_Surface *texoffset(SDL_Surface *s, int xoffset, int yoffset)
     return d;
 }
 
-void texmad(SDL_Surface *s, float mul, float add)
+void texmad(SDL_Surface *s, const vec &mul, const vec &add)
 {
-    int depth = s->format->BitsPerPixel/8;
+    int maxk = min(s->format->BytesPerPixel, 3);
     uchar *src = (uchar *)s->pixels;
-    loopi(s->h*s->w*depth)
+    loopi(s->h*s->w) 
     {
-        float val = *src*mul + 255*add;
-        *src++ = uchar(min(max(val, 0), 255));
+        loopk(maxk)
+        {
+            float val = src[k]*mul[k] + 255*add[k];
+            src[k] = uchar(min(max(val, 0), 255));
+        }
+        src += s->format->BytesPerPixel;
     }
 }
 
-VARP(mintexcompresssize, 0, 1<<10, 1<<12);
+SDL_Surface *texffmask(SDL_Surface *s, int minval)
+{
+    if(nomasks || s->format->BytesPerPixel<3) { SDL_FreeSurface(s); return NULL; }
+    bool glow = false, envmap = true;
+    uchar *src = (uchar *)s->pixels;
+    loopi(s->h*s->w)
+    {
+        if(src[1]>minval) glow = true;
+        if(src[2]>minval) { glow = envmap = true; break; }
+        src += s->format->BytesPerPixel;
+    }
+    if(!glow && !envmap) { SDL_FreeSurface(s); return NULL; }
+    SDL_Surface *m = SDL_CreateRGBSurface(SDL_SWSURFACE, s->w, s->h, envmap ? 16 : 8, 0, 0, 0, 0);
+    if(!m) fatal("create surface");
+    uchar *dst = (uchar *)m->pixels;
+    src = (uchar *)s->pixels;
+    loopi(s->h*s->w)
+    {
+        *dst++ = src[1];
+        if(envmap) *dst++ = src[2];
+        src += s->format->BytesPerPixel;
+    }
+    SDL_FreeSurface(s);
+    return m;
+}
+
+void texdup(SDL_Surface *s, int srcchan, int dstchan)
+{
+    if(srcchan==dstchan || max(srcchan, dstchan) >= s->format->BytesPerPixel) return;
+    uchar *src = (uchar *)s->pixels;
+    loopi(s->h*s->w)
+    {
+        src[dstchan] = src[srcchan];
+        src += s->format->BytesPerPixel;
+    }
+}
+
+VAR(hwtexsize, 1, 0, 0);
+VAR(hwcubetexsize, 1, 0, 0);
+VAR(hwmaxaniso, 1, 0, 0);
+VARP(maxtexsize, 0, 0, 1<<12);
+VARP(texreduce, 0, 0, 12);
+VARP(texcompress, 0, 1<<10, 1<<12);
+VARP(trilinear, 0, 1, 1);
+VARP(bilinear, 0, 1, 1);
+VARP(aniso, 0, 0, 16);
+
+GLenum compressedformat(GLenum format, int w, int h)
+{
+#ifdef __APPLE__
+#undef GL_COMPRESSED_RGB_S3TC_DXT1_EXT
+#undef GL_COMPRESSED_RGBA_S3TC_DXT5_EXT
+#define GL_COMPRESSED_RGB_S3TC_DXT1_EXT GL_COMPRESSED_RGB_ARB
+#define GL_COMPRESSED_RGBA_S3TC_DXT5_EXT GL_COMPRESSED_RGBA_ARB
+#endif
+    if(hasTC && texcompress && max(w, h) >= texcompress) switch(format)
+    {
+        case GL_RGB5:
+        case GL_RGB8:
+        case GL_RGB: return GL_COMPRESSED_RGB_S3TC_DXT1_EXT;
+        case GL_RGBA: return GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
+    }
+    return format;
+}
+
+int formatsize(GLenum format)
+{
+    switch(format)
+    {
+        case GL_LUMINANCE:
+        case GL_ALPHA: return 1;
+        case GL_LUMINANCE_ALPHA: return 2;
+        case GL_RGB: return 3;
+        case GL_RGBA: return 4;
+        default: return 4;
+    }
+}
 
 void createtexture(int tnum, int w, int h, void *pixels, int clamp, bool mipit, GLenum component, GLenum subtarget)
 {
@@ -84,89 +164,154 @@ void createtexture(int tnum, int w, int h, void *pixels, int clamp, bool mipit,
         glBindTexture(target, tnum);
         glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
         glTexParameteri(target, GL_TEXTURE_WRAP_S, clamp&1 ? GL_CLAMP_TO_EDGE : GL_REPEAT);
-        glTexParameteri(target, GL_TEXTURE_WRAP_T, clamp&2 ? GL_CLAMP_TO_EDGE : GL_REPEAT);
-        glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
-        glTexParameteri(target, GL_TEXTURE_MIN_FILTER, mipit ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
-        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
+        if(target!=GL_TEXTURE_1D) glTexParameteri(target, GL_TEXTURE_WRAP_T, clamp&2 ? GL_CLAMP_TO_EDGE : GL_REPEAT);
+        if(target==GL_TEXTURE_2D && hasAF && min(aniso, hwmaxaniso) > 0) glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, min(aniso, hwmaxaniso));
+        glTexParameteri(target, GL_TEXTURE_MAG_FILTER, bilinear ? GL_LINEAR : GL_NEAREST);
+        glTexParameteri(target, GL_TEXTURE_MIN_FILTER, 
+            mipit ?
+                (trilinear ? 
+                    (bilinear ? GL_LINEAR_MIPMAP_LINEAR : GL_NEAREST_MIPMAP_LINEAR) : 
+                    (bilinear ? GL_LINEAR_MIPMAP_NEAREST : GL_NEAREST_MIPMAP_NEAREST)) :
+                (bilinear ? GL_LINEAR : GL_NEAREST));
     }
-#ifdef __APPLE__
-#undef GL_COMPRESSED_RGB_S3TC_DXT1_EXT
-#undef GL_COMPRESSED_RGBA_S3TC_DXT5_EXT
-#define GL_COMPRESSED_RGB_S3TC_DXT1_EXT GL_COMPRESSED_RGB_ARB
-#define GL_COMPRESSED_RGBA_S3TC_DXT5_EXT GL_COMPRESSED_RGBA_ARB
-#endif
-    GLenum format = component, type = GL_UNSIGNED_BYTE, compressed = component;
+    GLenum format = component, type = GL_UNSIGNED_BYTE;
     switch(component)
     {
+        case GL_RGB16F_ARB:
+        case GL_RGB32F_ARB:
+            format = GL_RGB;
+            type = GL_FLOAT;
+            break;
+
+        case GL_RGBA16F_ARB:
+        case GL_RGBA32F_ARB:
+            format = GL_RGBA;
+            type = GL_FLOAT;
+            break;
+
         case GL_DEPTH_COMPONENT:
             type = GL_FLOAT;
             break;
 
-        case GL_RGB8:
         case GL_RGB5:
+        case GL_RGB8:
+        case GL_RGB16:
             format = GL_RGB;
-            if(mipit && hasTC && mintexcompresssize && max(w, h) >= mintexcompresssize) compressed = GL_COMPRESSED_RGB_S3TC_DXT1_EXT;
             break;
 
-        case GL_RGB:
-            if(mipit && hasTC && mintexcompresssize && max(w, h) >= mintexcompresssize) compressed = GL_COMPRESSED_RGB_S3TC_DXT1_EXT;
-            break;
-
-        case GL_RGBA:
-            if(mipit && hasTC && mintexcompresssize && max(w, h) >= mintexcompresssize) compressed = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
+        case GL_RGBA8:
+        case GL_RGBA16:
+            format = GL_RGBA;
             break;
     }
-    //component = format == GL_RGB ? GL_COMPRESSED_RGB_S3TC_DXT1_EXT : GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
-    if(mipit)
+    uchar *scaled = NULL;
+    int hwlimit = target==GL_TEXTURE_CUBE_MAP_ARB ? hwcubetexsize : hwtexsize,
+        sizelimit = mipit && maxtexsize ? min(maxtexsize, hwlimit) : hwlimit;
+    if(pixels && max(w, h) > sizelimit && (!mipit || sizelimit < hwlimit))
     {
-        if(gluBuild2DMipmaps(subtarget, compressed, w, h, format, type, pixels))
+        int oldw = w, oldh = h;
+        while(w > sizelimit || h > sizelimit) { w /= 2; h /= 2; }
+        scaled = new uchar[w*h*formatsize(format)];
+        gluScaleImage(format, oldw, oldh, type, pixels, w, h, type, scaled);
+        pixels = scaled;
+    }
+    if(mipit && pixels)
+    {
+        GLenum compressed = compressedformat(component, w, h);
+        if(target==GL_TEXTURE_1D)
+        {
+            if(gluBuild1DMipmaps(subtarget, compressed, w, format, type, pixels))
+            {
+                if(compressed==component || gluBuild1DMipmaps(subtarget, component, w, format, type, pixels)) conoutf("could not build mipmaps");
+            }
+        }
+        else if(gluBuild2DMipmaps(subtarget, compressed, w, h, format, type, pixels))
         {
             if(compressed==component || gluBuild2DMipmaps(subtarget, component, w, h, format, type, pixels)) conoutf("could not build mipmaps");
         }
     }
+    else if(target==GL_TEXTURE_1D) glTexImage1D(subtarget, 0, component, w, 0, format, type, pixels);
     else glTexImage2D(subtarget, 0, component, w, h, 0, format, type, pixels);
+    if(scaled) delete[] scaled;
 }
 
 hashtable<char *, Texture> textures;
 
-Texture *crosshair = NULL; // used as default, ensured to be loaded
-
-VAR(maxtexsize, 0, 0, 1<<12);
+Texture *notexture = NULL; // used as default, ensured to be loaded
 
 static GLenum texformat(int bpp)
 {
     switch(bpp)
     {
         case 8: return GL_LUMINANCE;
+        case 16: return GL_LUMINANCE_ALPHA;
         case 24: return GL_RGB;
         case 32: return GL_RGBA;
         default: return 0;
     }
 }
 
-static Texture *newtexture(const char *rname, SDL_Surface *s, int clamp = 0, bool mipit = true)
+static void resizetexture(int &w, int &h, bool mipit = true, GLenum format = GL_RGB, GLenum target = GL_TEXTURE_2D)
+{
+    if(mipit) return;
+    int hwlimit = target==GL_TEXTURE_CUBE_MAP_ARB ? hwcubetexsize : hwtexsize,
+        sizelimit = mipit && maxtexsize ? min(maxtexsize, hwlimit) : hwlimit;
+    int w2 = w, h2 = h;
+    if(!hasNP2 && (w&(w-1) || h&(h-1)))
+    {
+        w2 = h2 = 1;
+        while(w2 < w) w2 *= 2;
+        while(h2 < h) h2 *= 2;
+        if(w2 > sizelimit || (w - w2)/2 < (w2 - w)/2) w2 /= 2;
+        if(h2 > sizelimit || (h - h2)/2 < (h2 - h)/2) h2 /= 2;
+    }
+    while(w2 > sizelimit || h2 > sizelimit)
+    {
+        w2 /= 2;
+        h2 /= 2;
+    }
+    w = w2;
+    h = h2;
+}
+
+static Texture *newtexture(const char *rname, SDL_Surface *s, int clamp = 0, bool mipit = true, bool canreduce = false)
 {
     char *key = newstring(rname);
     Texture *t = &textures[key];
     t->name = key;
     t->bpp = s->format->BitsPerPixel;
-    t->w = t->xs = s->w;
-    t->h = t->ys = s->h;
+    int w = t->xs = s->w;
+    int h = t->ys = s->h;
     glGenTextures(1, &t->gl);
-    if(maxtexsize && (t->w > maxtexsize || t->h > maxtexsize))
+    if(canreduce) loopi(texreduce)
     {
-        do { t->w /= 2; t->h /= 2; } while(t->w > maxtexsize || t->h > maxtexsize);
-        if(gluScaleImage(texformat(t->bpp), t->xs, t->ys, GL_UNSIGNED_BYTE, s->pixels, t->w, t->h, GL_UNSIGNED_BYTE, s->pixels))
-        {
-            t->w = t->xs;
-            t->h = t->ys;
-        }
+        if(w > 1) w /= 2;
+        if(h > 1) h /= 2;
     }
-    createtexture(t->gl, t->w, t->h, s->pixels, clamp, mipit, texformat(t->bpp));
+    GLenum format = texformat(t->bpp);
+    resizetexture(w, h, mipit, format);
+    uchar *pixels = (uchar *)s->pixels;
+    if(w != t->xs || h != t->ys)
+    {
+        if(w*h > t->xs*t->ys) pixels = new uchar[formatsize(format)*w*h];
+        gluScaleImage(format, t->xs, t->ys, GL_UNSIGNED_BYTE, s->pixels, w, h, GL_UNSIGNED_BYTE, pixels);
+    }
+    createtexture(t->gl, w, h, pixels, clamp, mipit, format);
+    if(pixels!=s->pixels) delete[] pixels;
     SDL_FreeSurface(s);
     return t;
 }
 
+static vec parsevec(const char *arg)
+{
+    vec v(0, 0, 0);
+    int i = 0;
+    for(; arg[0] && arg[0]!='>' && i<3; arg += strcspn(i ? arg+1 : arg, "/>"), i++)
+        v[i] = atof(i ? arg+1 : arg);
+    if(i==1) v.y = v.z = v.x;
+    return v;
+}
+
 static SDL_Surface *texturedata(const char *tname, Slot::Tex *tex = NULL, bool msg = true)
 {
     if(tex && !tname)
@@ -185,9 +330,9 @@ static SDL_Surface *texturedata(const char *tname, Slot::Tex *tex = NULL, bool m
         file++;
     }
     
-    show_out_of_renderloop_progress(0, file);
+    if(msg) show_out_of_renderloop_progress(0, file);
 
-    SDL_Surface *s = IMG_Load(file);
+    SDL_Surface *s = IMG_Load(findfile(file, "rb"));
     if(!s) { if(msg) conoutf("could not load texture %s", tname); return NULL; }
     int bpp = s->format->BitsPerPixel;
     if(!texformat(bpp)) { SDL_FreeSurface(s); conoutf("texture must be 8, 24, or 32 bpp: %s", tname); return NULL; }
@@ -200,7 +345,9 @@ static SDL_Surface *texturedata(const char *tname, Slot::Tex *tex = NULL, bool m
     {
         const char *cmd = &tname[1], *arg1 = strchr(cmd, ':'), *arg2 = arg1 ? strchr(arg1, ',') : NULL;
         if(!arg1) arg1 = strchr(cmd, '>');
-        if(!strncmp(cmd, "mad", arg1-cmd)) texmad(s, atof(arg1+1), arg2 ? atof(arg2+1) : 0);
+        if(!strncmp(cmd, "mad", arg1-cmd)) texmad(s, parsevec(arg1+1), arg2 ? parsevec(arg2+1) : vec(0, 0, 0)); 
+        else if(!strncmp(cmd, "ffmask", arg1-cmd)) s = texffmask(s, atoi(arg1+1));
+        else if(!strncmp(cmd, "dup", arg1-cmd)) texdup(s, atoi(arg1+1), atoi(arg2+1));
     }
     return s;
 }
@@ -208,7 +355,7 @@ static SDL_Surface *texturedata(const char *tname, Slot::Tex *tex = NULL, bool m
 void loadalphamask(Texture *t)
 {
     if(t->alphamask || t->bpp!=32) return;
-    SDL_Surface *s = IMG_Load(t->name);
+    SDL_Surface *s = IMG_Load(findfile(t->name, "rb"));
     if(!s || !s->format->Amask) { if(s) SDL_FreeSurface(s); return; }
     uint alpha = s->format->Amask;
     t->alphamask = new uchar[s->h * ((s->w+7)/8)];
@@ -235,7 +382,7 @@ Texture *textureload(const char *name, int clamp, bool mipit, bool msg)
     Texture *t = textures.access(path(tname));
     if(t) return t;
     SDL_Surface *s = texturedata(tname, NULL, msg); 
-    return s ? newtexture(tname, s, clamp, mipit) : crosshair;
+    return s ? newtexture(tname, s, clamp, mipit) : notexture;
 }
 
 void cleangl()
@@ -244,9 +391,9 @@ void cleangl()
     textures.clear();
 }
 
-void settexture(const char *name)
+void settexture(const char *name, bool clamp)
 {
-    glBindTexture(GL_TEXTURE_2D, textureload(name)->gl);
+    glBindTexture(GL_TEXTURE_2D, textureload(name, clamp, true, false)->gl);
 }
 
 vector<Slot> slots;
@@ -281,6 +428,7 @@ void texture(char *type, char *name, int *rot, int *xoffset, int *yoffset, float
         {"g", TEX_GLOW},
         {"s", TEX_SPEC},
         {"z", TEX_DEPTH},
+        {"e", TEX_ENVMAP}
     };
     int tnum = -1, matslot = findmaterial(type);
     loopi(sizeof(types)/sizeof(types[0])) if(!strcmp(types[i].name, type)) { tnum = i; break; }
@@ -332,15 +480,27 @@ static int findtextype(Slot &s, int type, int last = -1)
         loop(y, t->h) loop(x, t->w) \
         { \
             body; \
-            dst += t->format->BitsPerPixel/8; \
+            dst += t->format->BytesPerPixel; \
         } \
     }
 
-#define sourcetex(s) uchar *src = &((uchar *)s->pixels)[(s->format->BitsPerPixel/8)*(y*s->w + x)];
+#define sourcetex(s) uchar *src = &((uchar *)s->pixels)[s->format->BytesPerPixel*(y*s->w + x)];
+
+static void scaleglow(SDL_Surface *g, Slot &s)
+{
+    ShaderParam *cparam = findshaderparam(s, "glowscale", SHPARAM_PIXEL, 0);
+    if(!cparam) cparam = findshaderparam(s, "glowscale", SHPARAM_VERTEX, 0);
+    float color[3] = {1, 1, 1};
+    if(cparam) memcpy(color, cparam->val, sizeof(color));
+    writetex(g,
+        loopk(3) dst[k] = min(255, int(dst[k] * color[k]));
+    );
+}
 
 static void addglow(SDL_Surface *c, SDL_Surface *g, Slot &s)
 {
     ShaderParam *cparam = findshaderparam(s, "glowscale", SHPARAM_PIXEL, 0);
+    if(!cparam) cparam = findshaderparam(s, "glowscale", SHPARAM_VERTEX, 0);
     float color[3] = {1, 1, 1};
     if(cparam) memcpy(color, cparam->val, sizeof(color));
     writetex(c,
@@ -397,13 +557,15 @@ static void addname(vector<char> &key, Slot &slot, Slot::Tex &t)
         s_sprintfd(toffset)("+%d,%d", t.xoffset, t.yoffset);
         for(const char *s = toffset; *s; key.add(*s++));
     }
-    switch(t.type)
+    if(t.combined>=0 || renderpath==R_FIXEDFUNCTION) switch(t.type)
     {
         case TEX_GLOW:
         {
             ShaderParam *cparam = findshaderparam(slot, "glowscale", SHPARAM_PIXEL, 0);
+            if(!cparam) cparam = findshaderparam(slot, "glowscale", SHPARAM_VERTEX, 0);
             s_sprintfd(suffix)("?%.2f,%.2f,%.2f", cparam ? cparam->val[0] : 1.0f, cparam ? cparam->val[1] : 1.0f, cparam ? cparam->val[2] : 1.0f);
             for(const char *s = suffix; *s; key.add(*s++));
+            break;
         }
     }
 }
@@ -440,15 +602,15 @@ SDL_Surface *scalesurface(SDL_Surface *os, int w, int h)
 
 static void texcombine(Slot &s, int index, Slot::Tex &t, bool forceload = false)
 {
-    vector<char> key;
+    if(renderpath==R_FIXEDFUNCTION && t.type!=TEX_DIFFUSE && t.type!=TEX_GLOW && !forceload) { t.t = notexture; return; }
+    vector<char> key; 
     addname(key, s, t);
-    if(renderpath==R_FIXEDFUNCTION && t.type!=TEX_DIFFUSE && !forceload) { t.t = crosshair; return; }
     switch(t.type)
     {
         case TEX_DIFFUSE:
             if(renderpath==R_FIXEDFUNCTION)
             {
-                for(int i = -1; (i = findtextype(s, (1<<TEX_DECAL)|(1<<TEX_GLOW)|(1<<TEX_NORMAL), i))>=0;)
+                for(int i = -1; (i = findtextype(s, (1<<TEX_DECAL)|(1<<TEX_NORMAL), i))>=0;)
                 {
                     s.sts[i].combined = index;
                     addname(key, s, s.sts[i]);
@@ -470,9 +632,13 @@ static void texcombine(Slot &s, int index, Slot::Tex &t, bool forceload = false)
     t.t = textures.access(key.getbuf());
     if(t.t) return;
     SDL_Surface *ts = texturedata(NULL, &t);
-    if(!ts) { t.t = crosshair; return; }
+    if(!ts) { t.t = notexture; return; }
     switch(t.type)
     {
+        case TEX_GLOW:
+            if(renderpath==R_FIXEDFUNCTION) scaleglow(ts, s);  
+            break;
+
         case TEX_DIFFUSE:
             if(renderpath==R_FIXEDFUNCTION)
             {
@@ -513,7 +679,7 @@ static void texcombine(Slot &s, int index, Slot::Tex &t, bool forceload = false)
             }
             break;
     }
-    t.t = newtexture(key.getbuf(), ts);
+    t.t = newtexture(key.getbuf(), ts, 0, true, true);
 }
 
 Slot &lookuptexture(int slot, bool load)
@@ -524,7 +690,17 @@ Slot &lookuptexture(int slot, bool load)
     loopv(s.sts)
     {
         Slot::Tex &t = s.sts[i];
-        if(t.combined<0) texcombine(s, i, t, slot<0 && slot>-MAT_EDIT);
+        if(t.combined>=0) continue;
+        switch(t.type)
+        {
+            case TEX_ENVMAP:
+                if(hasCM && (renderpath!=R_FIXEDFUNCTION || (slot<0 && slot>-MAT_EDIT))) t.t = cubemapload(t.name);
+                break;
+
+            default:
+                texcombine(s, i, t, slot<0 && slot>-MAT_EDIT);
+                break;
+        }
     }
     s.loaded = true;
     return s;
@@ -544,7 +720,7 @@ Texture *loadthumbnail(Slot &slot)
     else
     {
         SDL_Surface *s = texturedata(NULL, &slot.sts[0], false);
-        if(!s) slot.thumbnail = crosshair;
+        if(!s) slot.thumbnail = notexture;
         else
         {
             int xs = s->w, ys = s->h;
@@ -573,23 +749,35 @@ GLuint cubemapfromsky(int size)
 {
     extern Texture *sky[6];
     if(!sky[0]) return 0;
-    GLuint tex;
-    glGenTextures(1, &tex);
-    uchar *scaled = new uchar[3*size*size];
+    
+    int tsize = 0, cmw, cmh;
+    GLint tw[6], th[6]; 
     loopi(6)
     {
-        uchar *pixels = new uchar[3*sky[i]->w*sky[i]->h];
         glBindTexture(GL_TEXTURE_2D, sky[i]->gl);
-        glGetTexImage(GL_TEXTURE_2D, 0, GL_RGB, GL_UNSIGNED_BYTE, pixels);
-        gluScaleImage(GL_RGB, sky[i]->w, sky[i]->h, GL_UNSIGNED_BYTE, pixels, size, size, GL_UNSIGNED_BYTE, scaled);
-        createtexture(!i ? tex : 0, size, size, scaled, 3, true, GL_RGB5, cubemapsides[i].target);
-        delete[] pixels;
+        glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &tw[i]);
+        glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &th[i]);
+        tsize = max(tsize, max(tw[i], th[i]));
     }
-    delete[] scaled;
+    cmw = cmh = min(tsize, size);
+    resizetexture(cmw, cmh, true, GL_RGB5, GL_TEXTURE_CUBE_MAP_ARB);
+    
+    GLuint tex;
+    glGenTextures(1, &tex);
+    uchar *pixels = new uchar[3*max(cmw, tsize)*max(cmh, tsize)];
+    loopi(6)
+    {
+        glBindTexture(GL_TEXTURE_2D, sky[i]->gl);
+        glPixelStorei(GL_PACK_ALIGNMENT, 1);
+        glGetTexImage(GL_TEXTURE_2D, 0, GL_RGB, GL_UNSIGNED_BYTE, pixels);
+        if(tw[i]!=cmw || th[i]!=cmh) gluScaleImage(GL_RGB, tw[i], th[i], GL_UNSIGNED_BYTE, pixels, cmw, cmh, GL_UNSIGNED_BYTE, pixels);
+        createtexture(!i ? tex : 0, cmw, cmh, pixels, 3, true, GL_RGB5, cubemapsides[i].target);
+    }
+    delete[] pixels;
     return tex;
-}
+}   
 
-Texture *cubemapload(const char *name, bool mipit, bool msg)
+Texture *cubemaploadwildcard(const char *name, bool mipit, bool msg)
 {
     if(!hasCM) return NULL;
     string tname;
@@ -600,6 +788,8 @@ Texture *cubemapload(const char *name, bool mipit, bool msg)
     SDL_Surface *surface[6];
     string sname;
     if(!wildcard) s_strcpy(sname, tname);
+    GLenum format = 0;
+    int tsize = 0;
     loopi(6)
     {
         if(wildcard)
@@ -614,19 +804,57 @@ Texture *cubemapload(const char *name, bool mipit, bool msg)
             loopj(i) SDL_FreeSurface(surface[j]);
             return NULL;
         }
+        if(!format) format = texformat(surface[i]->format->BitsPerPixel);
+        else if(texformat(surface[i]->format->BitsPerPixel)!=format)
+        {
+            if(surface[i] && msg) conoutf("cubemap texture %s doesn't match other sides' format", sname);
+            loopj(i) SDL_FreeSurface(surface[j]);
+            return NULL;
+        }
+        tsize = max(tsize, max(surface[i]->w, surface[i]->h));
     }
-    t = &textures[newstring(tname)];
-    s_strcpy(t->name, tname);
+    char *key = newstring(tname);
+    t = &textures[key];
+    t->name = key;
     t->bpp = surface[0]->format->BitsPerPixel;
-    t->xs = surface[0]->w;
-    t->ys = surface[0]->h;
+    int w = t->xs = tsize;
+    int h = t->ys = tsize;
+    resizetexture(w, h, mipit, format, GL_TEXTURE_CUBE_MAP_ARB);
     glGenTextures(1, &t->gl);
+    uchar *pixels = NULL;
     loopi(6)
     {
         SDL_Surface *s = surface[i];
-        createtexture(!i ? t->gl : 0, s->w, s->h, s->pixels, 3, mipit, texformat(s->format->BitsPerPixel), cubemapsides[i].target);
+        if(s->w != w || s->h != h)
+        {
+            if(!pixels) pixels = new uchar[formatsize(format)*w*h];
+            gluScaleImage(format, s->w, s->h, GL_UNSIGNED_BYTE, s->pixels, w, h, GL_UNSIGNED_BYTE, pixels);
+        }
+        createtexture(!i ? t->gl : 0, w, h, s->w != w || s->h != h ? pixels : s->pixels, 3, mipit, format, cubemapsides[i].target);
         SDL_FreeSurface(s);
     }
+    if(pixels) delete[] pixels;
+    return t;
+}
+
+Texture *cubemapload(const char *name, bool mipit, bool msg)
+{
+    if(!hasCM) return NULL;
+    s_sprintfd(pname)("packages/%s", name);
+    path(pname);
+    Texture *t = NULL;
+    if(!strchr(pname, '*'))
+    {
+        s_sprintfd(jpgname)("%s_*.jpg", pname);
+        t = cubemaploadwildcard(jpgname, mipit, false);
+        if(!t)
+        {
+            s_sprintfd(pngname)("%s_*.png", pname);
+            t = cubemaploadwildcard(pngname, mipit, false);
+            if(!t && msg) conoutf("could not load envmap %s", name);
+        }
+    }
+    else t = cubemaploadwildcard(pname, mipit, msg);
     return t;
 }
 
@@ -654,18 +882,19 @@ void clearenvmaps()
     envmaps.setsize(0);
 }
 
-VAR(aaenvmap, 0, 1, 1);
+VAR(aaenvmap, 0, 2, 4);
 
 GLuint genenvmap(const vec &o, int envmapsize)
 {
-    int rendersize = 1;
-    while(rendersize < scr_w && rendersize < scr_h) rendersize *= 2;
-    if(rendersize > scr_w || rendersize > scr_h) rendersize /= 2;
-    if(!aaenvmap && rendersize > 1<<envmapsize) rendersize = 1<<envmapsize;
-    int texsize = rendersize < 1<<envmapsize ? rendersize : 1<<envmapsize;
+    int rendersize = 1<<(envmapsize+aaenvmap), sizelimit = min(hwcubetexsize, min(screen->w, screen->h));
+    if(maxtexsize) sizelimit = min(sizelimit, maxtexsize);
+    while(rendersize > sizelimit) rendersize /= 2;
+    int texsize = min(rendersize, 1<<envmapsize);
+    if(!aaenvmap) rendersize = texsize;
     GLuint tex;
     glGenTextures(1, &tex);
     glViewport(0, 0, rendersize, rendersize);
+    glPixelStorei(GL_PACK_ALIGNMENT, 1);
     float yaw = 0, pitch = 0;
     uchar *pixels = new uchar[3*rendersize*rendersize];
     loopi(6)
@@ -688,18 +917,12 @@ GLuint genenvmap(const vec &o, int envmapsize)
         }
         drawcubemap(rendersize, o, yaw, pitch);
         glReadPixels(0, 0, rendersize, rendersize, GL_RGB, GL_UNSIGNED_BYTE, pixels);
-        uchar *src = pixels, *dst = &pixels[3*rendersize*rendersize-3];
-        loop(y, rendersize/2) loop(x, rendersize)
-        {
-            loopk(3) swap(uchar, src[k], dst[k]);
-            src += 3;
-            dst -= 3;
-        }
         if(texsize<rendersize) gluScaleImage(GL_RGB, rendersize, rendersize, GL_UNSIGNED_BYTE, pixels, texsize, texsize, GL_UNSIGNED_BYTE, pixels);
         createtexture(tex, texsize, texsize, pixels, 3, true, GL_RGB5, side.target);
     }
     delete[] pixels;
-    glViewport(0, 0, scr_w, scr_h);
+    glViewport(0, 0, screen->w, screen->h);
+    clientkeepalive();
     return tex;
 }
 
@@ -761,7 +984,7 @@ ushort closestenvmap(int orient, int x, int y, int z, int size)
 
 GLuint lookupenvmap(ushort emid)
 {
-    if(emid==EMID_SKY) return skyenvmap;
+    if(emid==EMID_SKY || emid==EMID_CUSTOM) return skyenvmap;
     if(emid==EMID_NONE || !envmaps.inrange(emid-EMID_RESERVED)) return 0;
     GLuint tex = envmaps[emid-EMID_RESERVED].tex;
     return tex ? tex : skyenvmap;
@@ -769,7 +992,7 @@ GLuint lookupenvmap(ushort emid)
 
 void writetgaheader(FILE *f, SDL_Surface *s, int bits)
 {
-    fwrite("\0\0\x02\0\0\0\0 \0\0\0\0", 1, 12, f);
+    fwrite("\0\0\x02\0\0\0\0\0\0\0\0\0", 1, 12, f);
     ushort dim[] = { s->w, s->h };
     endianswap(dim, sizeof(ushort), 2);
     fwrite(dim, sizeof(short), 2, f);
@@ -779,9 +1002,9 @@ void writetgaheader(FILE *f, SDL_Surface *s, int bits)
 
 void flipnormalmapy(char *destfile, char *normalfile)           // RGB (jpg/png) -> BGR (tga)
 {
-    SDL_Surface *ns = IMG_Load(normalfile);
+    SDL_Surface *ns = IMG_Load(findfile(path(normalfile), "rb"));
     if(!ns) return;
-    FILE *f = fopen(destfile, "wb");
+    FILE *f = openfile(path(destfile), "wb");
     if(f)
     {
         writetgaheader(f, ns, 24);
@@ -795,25 +1018,25 @@ void flipnormalmapy(char *destfile, char *normalfile)           // RGB (jpg/png)
         fclose(f);
     }
     if(ns) SDL_FreeSurface(ns);
-}  
+}
 
 void mergenormalmaps(char *heightfile, char *normalfile)    // BGR (tga) -> BGR (tga) (SDL loads TGA as BGR!)
 {
-    SDL_Surface *hs = IMG_Load(heightfile);
-    SDL_Surface *ns = IMG_Load(normalfile);
+    SDL_Surface *hs = IMG_Load(findfile(path(heightfile), "rb"));
+    SDL_Surface *ns = IMG_Load(findfile(path(normalfile), "rb"));
     if(hs && ns)
     {
         uchar def_n[] = { 255, 128, 128 };
-        FILE *f = fopen(normalfile, "wb");
+        FILE *f = openfile(normalfile, "wb");
         if(f)
         {
-            writetgaheader(f, ns, 24);
+            writetgaheader(f, ns, 24); 
             for(int y = ns->h-1; y>=0; y--) loop(x, ns->w)
             {
                 int off = (x+y*ns->w)*3;
                 uchar *hd = hs ? (uchar *)hs->pixels+off : def_n;
                 uchar *nd = ns ? (uchar *)ns->pixels+off : def_n;
-                #define S(x) x/255.0f*2-1
+                #define S(x) x/255.0f*2-1 
                 vec n(S(nd[0]), S(nd[1]), S(nd[2]));
                 vec h(S(hd[0]), S(hd[1]), S(hd[2]));
                 n.mul(2).add(h).normalize().add(1).div(2).mul(255);
