@@ -3,6 +3,7 @@
 
 VARFP(waterreflect, 0, 1, 1, cleanreflections());
 VARFP(waterrefract, 0, 1, 1, cleanreflections());
+VARP(refractfog, 0, 1, 1);
 
 /* vertex water */
 VARP(watersubdiv, 0, 2, 3);
@@ -15,23 +16,57 @@ float wcol[4];
     inline void vertw(float v1, float v2, float v3, float t) \
     { \
         float angle = (v1-wx1)/wsize*(v2-wy1)/wsize*(v1-wx2)*(v2-wy2)*59/23+t; \
-        float s = sinf(angle), h = 0.8f*s-1.1f; \
+        float s = sinf(angle), h = WATER_AMPLITUDE*s-WATER_OFFSET; \
+        body; \
+        glVertex3f(v1, v2, v3+h); \
+    }
+#define VERTWN(vertw, body) \
+    inline void vertw(float v1, float v2, float v3) \
+    { \
+        float h = -WATER_OFFSET; \
         body; \
         glVertex3f(v1, v2, v3+h); \
     }
 #define VERTWT(vertwt, body) VERTW(vertwt, { float v = cosf(angle); float duv = 0.5f*v; body; })
+VERTW(vertwt, {
+    glTexCoord2f(v1/8.0f, v2/8.0f);
+})
+VERTWN(vertwtn, {
+    glTexCoord2f(v1/8.0f, v2/8.0f);
+})
 VERTW(vertwc, {
-    glColor4f(wcol[0], wcol[1], wcol[2], max(wcol[3], 0.6f) + fabs(s)*0.1f);
+    glColor4f(wcol[0], wcol[1], wcol[2], wcol[3] + fabs(s)*0.1f);
+})
+VERTWN(vertwcn, {
+    glColor4f(wcol[0], wcol[1], wcol[2], wcol[3]);
 })
 VERTWT(vertwtc, {
     glColor4f(1, 1, 1, 0.2f + fabs(s)*0.1f);
     glTexCoord3f(v1+duv, v2+duv, v3+h);
 })
+VERTWN(vertwtcn, {
+    glColor4f(1, 1, 1, 0.2f);
+    glTexCoord3f(v1, v2, v3+h);
+})
 VERTWT(vertwmtc, {
     glColor4f(1, 1, 1, 0.2f + fabs(s)*0.1f);
-    glMultiTexCoord3f_(GL_TEXTURE0_ARB, v1+duv, v2+duv, v3+h);
+    glMultiTexCoord3f_(GL_TEXTURE0_ARB, v1-duv, v2+duv, v3+h);
     glMultiTexCoord3f_(GL_TEXTURE1_ARB, v1+duv, v2+duv, v3+h);
 })
+VERTWN(vertwmtcn, {
+    glColor4f(1, 1, 1, 0.2f);
+    glMultiTexCoord3f_(GL_TEXTURE0_ARB, v1, v2, v3+h);
+    glMultiTexCoord3f_(GL_TEXTURE1_ARB, v1, v2, v3+h);
+})
+
+static float lavaxk = 1.0f, lavayk = 1.0f, lavascroll = 0.0f;
+
+VERTW(vertl, {
+    glTexCoord2f(lavaxk*(v1+lavascroll), lavayk*(v2+lavascroll));
+})
+VERTWN(vertln, {
+    glTexCoord2f(lavaxk*(v1+lavascroll), lavayk*(v2+lavascroll));
+})
 
 #define renderwaterstrips(vertw, z, t) \
     for(int x = wx1; x<wx2; x += subdiv) \
@@ -50,10 +85,8 @@ VERTWT(vertwmtc, {
         xtraverts += n; \
     }
 
-void rendervertwater(uint subdiv, int xo, int yo, int z, uint size)
+void rendervertwater(uint subdiv, int xo, int yo, int z, uint size, uchar mat = MAT_WATER)
 {   
-    float t = lastmillis/300.0f;
-
     wx1 = xo;
     wy1 = yo;
     wx2 = wx1 + size,
@@ -63,9 +96,25 @@ void rendervertwater(uint subdiv, int xo, int yo, int z, uint size)
     ASSERT((wx1 & (subdiv - 1)) == 0);
     ASSERT((wy1 & (subdiv - 1)) == 0);
 
-    if(waterrefract) { renderwaterstrips(vertwmtc, z, t); }
-    else if(waterreflect) { renderwaterstrips(vertwtc, z, t); }
-    else { renderwaterstrips(vertwc, z, t); }
+    switch(mat)
+    {
+        case MAT_WATER:
+        {
+            float t = lastmillis/(renderpath!=R_FIXEDFUNCTION ? 600.0f : 300.0f);
+            if(renderpath!=R_FIXEDFUNCTION) { renderwaterstrips(vertwt, z, t); }
+            else if(nowater || (!waterrefract && !waterreflect)) { renderwaterstrips(vertwc, z, t); }
+            else if(waterrefract) { renderwaterstrips(vertwmtc, z, t); }
+            else { renderwaterstrips(vertwtc, z, t); }
+            break;
+        }
+
+        case MAT_LAVA:
+        {
+            float t = lastmillis/2000.0f;
+            renderwaterstrips(vertl, z, t);
+            break;
+        }
+    }
 }
 
 uint calcwatersubdiv(int x, int y, int z, uint size)
@@ -87,12 +136,12 @@ uint calcwatersubdiv(int x, int y, int z, uint size)
     return subdiv;
 }
 
-uint renderwaterlod(int x, int y, int z, uint size)
+uint renderwaterlod(int x, int y, int z, uint size, uchar mat = MAT_WATER)
 {
     if(size <= (uint)(32 << waterlod))
     {
         uint subdiv = calcwatersubdiv(x, y, z, size);
-        if(subdiv < size * 2) rendervertwater(min(subdiv, size), x, y, z, size);
+        if(subdiv < size * 2) rendervertwater(min(subdiv, size), x, y, z, size, mat);
         return subdiv;
     }
     else
@@ -100,33 +149,74 @@ uint renderwaterlod(int x, int y, int z, uint size)
         uint subdiv = calcwatersubdiv(x, y, z, size);
         if(subdiv >= size)
         {
-            if(subdiv < size * 2) rendervertwater(size, x, y, z, size);
+            if(subdiv < size * 2) rendervertwater(size, x, y, z, size, mat);
             return subdiv;
         }
         uint childsize = size / 2,
-             subdiv1 = renderwaterlod(x, y, z, childsize),
-             subdiv2 = renderwaterlod(x + childsize, y, z, childsize),
-             subdiv3 = renderwaterlod(x + childsize, y + childsize, z, childsize),
-             subdiv4 = renderwaterlod(x, y + childsize, z, childsize),
+             subdiv1 = renderwaterlod(x, y, z, childsize, mat),
+             subdiv2 = renderwaterlod(x + childsize, y, z, childsize, mat),
+             subdiv3 = renderwaterlod(x + childsize, y + childsize, z, childsize, mat),
+             subdiv4 = renderwaterlod(x, y + childsize, z, childsize, mat),
              minsubdiv = subdiv1;
         minsubdiv = min(minsubdiv, subdiv2);
         minsubdiv = min(minsubdiv, subdiv3);
         minsubdiv = min(minsubdiv, subdiv4);
         if(minsubdiv < size * 2)
         {
-            if(minsubdiv >= size) rendervertwater(size, x, y, z, size);
+            if(minsubdiv >= size) rendervertwater(size, x, y, z, size, mat);
             else
             {
-                if(subdiv1 >= size) rendervertwater(childsize, x, y, z, childsize);
-                if(subdiv2 >= size) rendervertwater(childsize, x + childsize, y, z, childsize);
-                if(subdiv3 >= size) rendervertwater(childsize, x + childsize, y + childsize, z, childsize);
-                if(subdiv4 >= size) rendervertwater(childsize, x, y + childsize, z, childsize);
+                if(subdiv1 >= size) rendervertwater(childsize, x, y, z, childsize, mat);
+                if(subdiv2 >= size) rendervertwater(childsize, x + childsize, y, z, childsize, mat);
+                if(subdiv3 >= size) rendervertwater(childsize, x + childsize, y + childsize, z, childsize, mat);
+                if(subdiv4 >= size) rendervertwater(childsize, x, y + childsize, z, childsize, mat);
             } 
         }
         return minsubdiv;
     }
 }
 
+#define renderwaterquad(vertwn, z) \
+    { \
+        vertwn(x, y, z); \
+        vertwn(x+rsize, y, z); \
+        vertwn(x+rsize, y+csize, z); \
+        vertwn(x, y+csize, z); \
+        xtraverts += 4; \
+    }
+
+void renderflatwater(int x, int y, int z, uint rsize, uint csize, uchar mat = MAT_WATER)
+{
+    switch(mat)
+    {
+        case MAT_WATER:
+            if(renderpath!=R_FIXEDFUNCTION) { renderwaterquad(vertwtn, z); }
+            else if(nowater || (!waterrefract && !waterreflect)) { renderwaterquad(vertwcn, z); }
+            else if(waterrefract) { renderwaterquad(vertwmtcn, z); }
+            else { renderwaterquad(vertwtcn, z); }
+            break;
+
+        case MAT_LAVA:
+            renderwaterquad(vertln, z);
+            break;
+    }
+}
+
+VARFP(vertwater, 0, 1, 1, allchanged());
+
+void renderlava(materialsurface &m, Texture *tex, float scale)
+{
+    lavaxk = 8.0f/(tex->xs*scale);
+    lavayk = 8.0f/(tex->ys*scale); 
+    lavascroll = lastmillis/1000.0f;
+    if(vertwater)
+    {
+        if(renderwaterlod(m.o.x, m.o.y, m.o.z, m.csize, MAT_LAVA) >= (uint)m.csize * 2)
+            rendervertwater(m.csize, m.o.x, m.o.y, m.o.z, m.csize, MAT_LAVA);
+    }
+    else renderflatwater(m.o.x, m.o.y, m.o.z, m.rsize, m.csize, MAT_LAVA);
+}
+
 /* reflective/refractive water */
 
 #define MAXREFLECTIONS 16
@@ -148,6 +238,13 @@ Reflection *findreflection(int height);
 VARP(reflectdist, 0, 2000, 10000);
 VAR(waterfog, 0, 150, 10000);
 
+void getwatercolour(uchar *wcol)
+{
+    static const uchar defaultwcol[3] = { 20, 70, 80};
+    if(hdr.watercolour[0] || hdr.watercolour[1] || hdr.watercolour[2]) memcpy(wcol, hdr.watercolour, 3);
+    else memcpy(wcol, defaultwcol, 3);
+}
+
 void watercolour(int *r, int *g, int *b)
 {
     hdr.watercolour[0] = *r;
@@ -157,7 +254,25 @@ void watercolour(int *r, int *g, int *b)
 
 COMMAND(watercolour, "iii");
 
-Shader *watershader = NULL, *waterreflectshader = NULL, *waterrefractshader = NULL;
+VAR(lavafog, 0, 50, 10000);
+
+void getlavacolour(uchar *lcol)
+{
+    static const uchar defaultlcol[3] = { 255, 64, 0 };
+    if(hdr.lavacolour[0] || hdr.lavacolour[1] || hdr.lavacolour[2]) memcpy(lcol, hdr.lavacolour, 3);
+    else memcpy(lcol, defaultlcol, 3);
+}
+
+void lavacolour(int *r, int *g, int *b)
+{
+    hdr.lavacolour[0] = *r;
+    hdr.lavacolour[1] = *g;
+    hdr.lavacolour[2] = *b;
+}
+
+COMMAND(lavacolour, "iii");
+
+Shader *watershader = NULL, *waterreflectshader = NULL, *waterrefractshader = NULL, *waterfadeshader = NULL;
 
 void setprojtexmatrix(Reflection &ref, bool init = true)
 {
@@ -168,8 +283,8 @@ void setprojtexmatrix(Reflection &ref, bool init = true)
         glGetFloatv(GL_MODELVIEW_MATRIX, mm);
 
         glLoadIdentity();
-        glTranslatef(0.5f, 0.5f, 0.5f);
-        glScalef(0.5f, 0.5f, 0.5f);
+        glTranslatef(0.5f, 0.5f, 0);
+        glScalef(0.5f, 0.5f, 1);
         glMultMatrixf(pm);
         glMultMatrixf(mm);
 
@@ -180,79 +295,42 @@ void setprojtexmatrix(Reflection &ref, bool init = true)
 
 void setuprefractTMUs()
 {
-    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_INTERPOLATE_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_CONSTANT_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
-    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_TEXTURE);
-    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
-    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB,  GL_CONSTANT_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB_ARB, GL_SRC_ALPHA);
-    
+    if(!refractfog) setuptmu(0, "K , T @ Ka");
+
     glActiveTexture_(GL_TEXTURE1_ARB);
     glEnable(GL_TEXTURE_2D);
 
-    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_INTERPOLATE_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_PREVIOUS_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
-    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_TEXTURE);
-    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
-    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB,  GL_PRIMARY_COLOR_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB_ARB, GL_ONE_MINUS_SRC_ALPHA);
-   
+    setuptmu(1, "P , T @ C~a");
+
     glActiveTexture_(GL_TEXTURE0_ARB);
-    glMatrixMode(GL_TEXTURE);
 }
 
 void setupreflectTMUs()
 {
-    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB,  GL_INTERPOLATE_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB,  GL_TEXTURE);
-    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
-    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB,  GL_CONSTANT_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
-    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB,  GL_PRIMARY_COLOR_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB_ARB, GL_SRC_ALPHA);
+    setuptmu(0, "T , K @ Ca", "Ka * P~a");
 
-    glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB,  GL_MODULATE);
-    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB,  GL_CONSTANT_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_SRC_ALPHA);
-    glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_ALPHA_ARB,  GL_PREVIOUS_ARB);
-    glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_ALPHA_ARB, GL_ONE_MINUS_SRC_ALPHA);
-    
     glDepthMask(GL_FALSE);
     glEnable(GL_BLEND);
     glBlendFunc(GL_ONE, GL_SRC_ALPHA);
-
-    glMatrixMode(GL_TEXTURE);
 }
 
 void cleanupwaterTMUs(bool refract)
 {
-    extern void setupTMU();
-    setupTMU();
+    resettmu(0);
 
     if(refract)
     {
         glActiveTexture_(GL_TEXTURE1_ARB);
-        setupTMU();
+        resettmu(1);
         glLoadIdentity();
         glDisable(GL_TEXTURE_2D);
         glActiveTexture_(GL_TEXTURE0_ARB);
     }
     else
     {
-        glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_TEXTURE);
-        glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_ALPHA_ARB, GL_SRC_ALPHA);
-
         glDisable(GL_BLEND);
         glDepthMask(GL_TRUE);
     }
-
-    glLoadIdentity();
-    glMatrixMode(GL_MODELVIEW);
 }
 
 VAR(waterspec, 0, 150, 1000);
@@ -264,12 +342,17 @@ VAR(oqwater, 0, 1, 1);
 
 extern int oqfrags;
 
-void rendervertwater()
+void renderwaterff()
 {
     glDisable(GL_CULL_FACE);
     
-    if(waterrefract) setuprefractTMUs();
-    else if(waterreflect) setupreflectTMUs();
+    if(!nowater && (waterreflect || waterrefract))
+    {
+        if(waterrefract) setuprefractTMUs();
+        else setupreflectTMUs();
+
+        glMatrixMode(GL_TEXTURE);
+    }
     else
     {
         glDisable(GL_TEXTURE_2D);
@@ -279,8 +362,9 @@ void rendervertwater()
     }
 
     int lastdepth = -1;
-    float offset = -1.1f;
+    float offset = -WATER_OFFSET;
 
+    uchar wcolub[3];
     getwatercolour(wcolub);
     loopi(3) wcol[i] = wcolub[i]/255.0f;
 
@@ -290,25 +374,26 @@ void rendervertwater()
         Reflection &ref = reflections[i];
         if(ref.height<0 || ref.lastused<totalmillis || ref.matsurfs.empty()) continue;
 
-        if(waterreflect || waterrefract)
+        if(!nowater && (waterreflect || waterrefract))
         {
             if(hasOQ && oqfrags && oqwater && ref.query && ref.query->owner==&ref && checkquery(ref.query)) continue;
             if(waterrefract) glActiveTexture_(GL_TEXTURE1_ARB);
             glBindTexture(GL_TEXTURE_2D, ref.tex);
             setprojtexmatrix(ref);
-        }
 
-        if(waterrefract)
-        {
-            glActiveTexture_(GL_TEXTURE0_ARB);
-            glBindTexture(GL_TEXTURE_2D, camera1->o.z>=ref.height+offset ? ref.refracttex : ref.tex);
-            setprojtexmatrix(ref, !waterreflect);
-        }
-        else if(waterreflect)
-        {
-            if(camera1->o.z < ref.height+offset) { if(blended) { glDepthMask(GL_TRUE); glDisable(GL_BLEND); blended = false; } }
-            else if(!blended) { glDepthMask(GL_FALSE); glEnable(GL_BLEND); blended = true; }
+            if(waterrefract)
+            {
+                glActiveTexture_(GL_TEXTURE0_ARB);
+                glBindTexture(GL_TEXTURE_2D, camera1->o.z>=ref.height+offset ? ref.refracttex : ref.tex);
+                setprojtexmatrix(ref, !waterreflect);
+            }
+            else
+            {
+                if(camera1->o.z < ref.height+offset) { if(blended) { glDepthMask(GL_TRUE); glDisable(GL_BLEND); blended = false; } }
+                else if(!blended) { glDepthMask(GL_FALSE); glEnable(GL_BLEND); blended = true; }
+            }
         }
+        bool begin = false;
         loopvj(ref.matsurfs)
         {
             materialsurface &m = *ref.matsurfs[j];
@@ -316,22 +401,35 @@ void rendervertwater()
             if(m.depth!=lastdepth)
             {
                 float depth = !waterfog ? 1.0f : min(0.75f*m.depth/waterfog, 0.95f);
+                if(nowater || !waterrefract) depth = max(depth, nowater || !waterreflect ? 0.6f : 0.3f);
                 wcol[3] = depth;
-                if(waterreflect || waterrefract)
+                if(!nowater && (waterreflect || waterrefract))
                 {
+                    if(begin) { glEnd(); begin = false; }
                     float ec[4] = { wcol[0], wcol[1], wcol[2], depth };
                     if(!waterrefract) { loopk(3) ec[k] *= depth; ec[3] = 1-ec[3]; }
-                    glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, ec);
+                    colortmu(0, ec[0], ec[1], ec[2], ec[3]);
                 }
                 lastdepth = m.depth;
             }
 
-            if(renderwaterlod(m.o.x, m.o.y, m.o.z, m.csize) >= (uint)m.csize * 2)
+            if(!vertwater)
+            {
+                if(!begin) { glBegin(GL_QUADS); begin = true; }
+                renderflatwater(m.o.x, m.o.y, m.o.z, m.rsize, m.csize);
+            }
+            else if(renderwaterlod(m.o.x, m.o.y, m.o.z, m.csize) >= (uint)m.csize * 2)
                 rendervertwater(m.csize, m.o.x, m.o.y, m.o.z, m.csize);
         }
+        if(begin) glEnd();
     }
 
-    if(waterrefract || waterreflect) cleanupwaterTMUs(waterrefract!=0);
+    if(!nowater && (waterreflect || waterrefract))
+    {
+        cleanupwaterTMUs(waterrefract!=0);
+        glLoadIdentity();
+        glMatrixMode(GL_MODELVIEW);
+    }
     else
     {
         glEnable(GL_TEXTURE_2D);
@@ -342,12 +440,14 @@ void rendervertwater()
     glEnable(GL_CULL_FACE);
 }
 
+VARFP(waterfade, 0, 1, 1, cleanreflections());
+
 void renderwater()
 {
     if(editmode && showmat) return;
     if(!rplanes) return;
 
-    if(renderpath==R_FIXEDFUNCTION) { rendervertwater(); return; }
+    if(renderpath==R_FIXEDFUNCTION) { renderwaterff(); return; }
 
     glDisable(GL_CULL_FACE);
 
@@ -367,6 +467,11 @@ void renderwater()
     {
         glActiveTexture_(GL_TEXTURE3_ARB);
         glEnable(GL_TEXTURE_2D);
+        if(hasFBO && renderpath!=R_FIXEDFUNCTION && waterfade)
+        {
+            glEnable(GL_BLEND);
+            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+        }
     }
     else
     {
@@ -382,14 +487,18 @@ void renderwater()
     if(!watershader) watershader = lookupshaderbyname("water");
     if(!waterreflectshader) waterreflectshader = lookupshaderbyname("waterreflect");
     if(!waterrefractshader) waterrefractshader = lookupshaderbyname("waterrefract");
+    if(!waterfadeshader) waterfadeshader = lookupshaderbyname("waterfade");
     
-    (waterrefract ? waterrefractshader : (waterreflect ? waterreflectshader : watershader))->set();
+    (waterrefract ? (waterfade && hasFBO ? waterfadeshader : waterrefractshader) : (waterreflect ? waterreflectshader : watershader))->set();
 
     if(waterreflect || waterrefract) glMatrixMode(GL_TEXTURE);
 
+    GLfloat oldfogc[4];
+    glGetFloatv(GL_FOG_COLOR, oldfogc);
+    vec ambient(max(hdr.skylight[0], hdr.ambient), max(hdr.skylight[1], hdr.ambient), max(hdr.skylight[2], hdr.ambient));
     entity *lastlight = (entity *)-1;
     int lastdepth = -1;
-    float offset = -1.1f;
+    float offset = -WATER_OFFSET;
     bool blended = true;
     loopi(MAXREFLECTIONS)
     {
@@ -408,11 +517,18 @@ void renderwater()
             glActiveTexture_(GL_TEXTURE3_ARB);
             glBindTexture(GL_TEXTURE_2D, camera1->o.z>=ref.height+offset ? ref.refracttex : ref.tex);
             glActiveTexture_(GL_TEXTURE0_ARB);
+            if(waterfade) setlocalparamf("waterheight", SHPARAM_VERTEX, 7, ref.height+offset+2, ref.height+offset+2, ref.height+offset+2);
         }
         else if(waterreflect)
         {
-            if(camera1->o.z < ref.height+offset) { if(blended) { glDepthMask(GL_TRUE); glDisable(GL_BLEND); blended = false; } }
+            GLfloat fogc[4] = { 0, 0, 0, 1 };
+            if(camera1->o.z < ref.height+offset) 
+            { 
+                if(blended) { glDepthMask(GL_TRUE); glDisable(GL_BLEND); blended = false; } 
+                loopk(3) fogc[k] = oldfogc[k];
+            }
             else if(!blended) { glDepthMask(GL_FALSE); glEnable(GL_BLEND); blended = true; }
+            setlocalparamfv("rgbafog", SHPARAM_PIXEL, 6, fogc);
         }
 
         bool begin = false;
@@ -426,7 +542,7 @@ void renderwater()
                 if(begin) { glEnd(); begin = false; }
                 const vec &lightpos = light ? light->o : vec(hdr.worldsize/2, hdr.worldsize/2, hdr.worldsize);
                 float lightrad = light && light->attr1 ? light->attr1 : hdr.worldsize*8.0f;
-                const vec &lightcol = (light ? vec(light->attr2, light->attr3, light->attr4) : vec(hdr.ambient, hdr.ambient, hdr.ambient)).div(255.0f).mul(waterspec/100.0f);
+                const vec &lightcol = (light ? vec(light->attr2, light->attr3, light->attr4) : vec(ambient)).div(255.0f).mul(waterspec/100.0f);
                 setlocalparamf("lightpos", SHPARAM_VERTEX, 2, lightpos.x, lightpos.y, lightpos.z);
                 setlocalparamf("lightcolor", SHPARAM_PIXEL, 3, lightcol.x, lightcol.y, lightcol.z);
                 setlocalparamf("lightradius", SHPARAM_PIXEL, 4, lightrad, lightrad, lightrad);
@@ -437,12 +553,18 @@ void renderwater()
             {
                 if(begin) { glEnd(); begin = false; }
                 float depth = !waterfog ? 1.0f : min(0.75f*m.depth/waterfog, 0.95f);
+                depth = max(depth, waterreflect ? 0.3f : 0.6f);
                 setlocalparamf("depth", SHPARAM_PIXEL, 5, depth, 1.0f-depth);
                 lastdepth = m.depth;
             }
 
-            if(!begin) { glBegin(GL_QUADS); begin = true; }
-            drawmaterial(m.orient, m.o.x, m.o.y, m.o.z, m.csize, m.rsize, 1.1f, true);
+            if(!vertwater)
+            {
+                if(!begin) { glBegin(GL_QUADS); begin = true; }
+                renderflatwater(m.o.x, m.o.y, m.o.z, m.rsize, m.csize);
+            }
+            else if(renderwaterlod(m.o.x, m.o.y, m.o.z, m.csize) >= (uint)m.csize * 2)
+                rendervertwater(m.csize, m.o.x, m.o.y, m.o.z, m.csize);
         }
         if(begin) glEnd();
     }
@@ -457,6 +579,7 @@ void renderwater()
     {
         glActiveTexture_(GL_TEXTURE3_ARB);
         glDisable(GL_TEXTURE_2D);
+        if(hasFBO && renderpath!=R_FIXEDFUNCTION && waterfade) glDisable(GL_BLEND);
     }
     else
     {
@@ -520,6 +643,12 @@ void cleanreflections()
 
 VARFP(reflectsize, 6, 8, 10, cleanreflections());
 
+void invalidatereflections()
+{
+    if(hasFBO) return;
+    loopi(MAXREFLECTIONS) reflections[i].matsurfs.setsizenodelete(0);
+}
+
 void addreflection(materialsurface &m)
 {
     int height = m.o.z;
@@ -545,88 +674,114 @@ void addreflection(materialsurface &m)
         if(!oldest || oldest->lastused==totalmillis) return;
         ref = oldest;
     }
-    static GLenum fboFormat = GL_RGB, dbFormat = GL_DEPTH_COMPONENT;
-    char *buf = NULL;
-    int size = 1<<reflectsize;
-    if(!hasFBO) while(size>scr_w || size>scr_h) size /= 2;
-    if((waterreflect || waterrefract) && !ref->tex)
-    {
-        glGenTextures(1, &ref->tex);
-        buf = new char[size*size*3];
-        memset(buf, 0, size*size*3);
-        createtexture(ref->tex, size, size, buf, 3, false, fboFormat);
-    }
-    if((waterreflect || waterrefract) && hasFBO && !ref->fb)
-    {
-        glGenFramebuffers_(1, &ref->fb);
-        glBindFramebuffer_(GL_FRAMEBUFFER_EXT, ref->fb);
-        glFramebufferTexture2D_(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, ref->tex, 0);
-        if(fboFormat==GL_RGB && glCheckFramebufferStatus_(GL_FRAMEBUFFER_EXT)!=GL_FRAMEBUFFER_COMPLETE_EXT)
-        {
-            fboFormat = GL_RGB8;
-            createtexture(ref->tex, size, size, buf, 3, false, fboFormat);
-            glFramebufferTexture2D_(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, ref->tex, 0);
-        }
-
-        if(!reflectiondb)
-        {
-            glGenRenderbuffers_(1, &reflectiondb);
-            glBindRenderbuffer_(GL_RENDERBUFFER_EXT, reflectiondb);
-            glRenderbufferStorage_(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, size, size);
-        }
-        glFramebufferRenderbuffer_(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, reflectiondb);
-        if(dbFormat==GL_DEPTH_COMPONENT && glCheckFramebufferStatus_(GL_FRAMEBUFFER_EXT)!=GL_FRAMEBUFFER_COMPLETE_EXT)
-        {
-            GLenum alts[] = { GL_DEPTH_COMPONENT16, GL_DEPTH_COMPONENT24, GL_DEPTH_COMPONENT32 };
-            loopi(sizeof(alts)/sizeof(alts[0]))
-            {
-                dbFormat = alts[i];
-                glBindRenderbuffer_(GL_RENDERBUFFER_EXT, reflectiondb);
-                glRenderbufferStorage_(GL_RENDERBUFFER_EXT, dbFormat, size, size);
-                glFramebufferRenderbuffer_(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, reflectiondb);
-                if(glCheckFramebufferStatus_(GL_FRAMEBUFFER_EXT)==GL_FRAMEBUFFER_COMPLETE_EXT) break;
-            }
-        }
-
-        glBindFramebuffer_(GL_FRAMEBUFFER_EXT, 0);
-    }
-    if(waterrefract && !ref->refracttex)
-    {
-        glGenTextures(1, &ref->refracttex);
-        createtexture(ref->refracttex, size, size, buf, 3, false, fboFormat);
-    }
-    if(waterrefract && hasFBO && !ref->refractfb)
-    {
-        glGenFramebuffers_(1, &ref->refractfb);
-        glBindFramebuffer_(GL_FRAMEBUFFER_EXT, ref->refractfb);
-        glFramebufferTexture2D_(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, ref->refracttex, 0);
-        glFramebufferRenderbuffer_(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, reflectiondb);
-            
-        glBindFramebuffer_(GL_FRAMEBUFFER_EXT, 0);
-    }
-    if(buf) delete[] buf;
     if(ref->height!=height) ref->height = height;
     rplanes++;
     ref->lastused = totalmillis;
     ref->matsurfs.setsizenodelete(0);
     ref->matsurfs.add(&m);
+    if(nowater) return;
+
+    static const GLenum colorfmts[] = { GL_RGBA, GL_RGBA8, GL_RGB, GL_RGB8, GL_FALSE },
+                        depthfmts[] = { GL_DEPTH_STENCIL_EXT, GL_DEPTH24_STENCIL8_EXT, GL_DEPTH_COMPONENT24, GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT16, GL_DEPTH_COMPONENT32, GL_FALSE };
+    const int stencilfmts = 2;
+    static GLenum reflectfmt = GL_FALSE, refractfmt = GL_FALSE, depthfmt = GL_FALSE, stencilfmt = GL_FALSE;
+    char *buf = NULL;
+    int size = 1<<reflectsize;
+    if(!hasFBO) while(size>screen->w || size>screen->h) size /= 2;
+    while(size>hwtexsize) size /= 2;
+    if((waterreflect || waterrefract) && !ref->tex)
+    {
+        glGenTextures(1, &ref->tex);
+        buf = new char[size*size*4];
+        memset(buf, 0, size*size*4);
+        int find = 2;
+        if(hasFBO)
+        {
+            if(!ref->fb) glGenFramebuffers_(1, &ref->fb);
+            glBindFramebuffer_(GL_FRAMEBUFFER_EXT, ref->fb);
+        }
+        do
+        {
+            createtexture(ref->tex, size, size, buf, 3, false, reflectfmt ? reflectfmt : colorfmts[find]);
+            if(!hasFBO) break;
+            else
+            {
+                glFramebufferTexture2D_(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, ref->tex, 0);
+                if(glCheckFramebufferStatus_(GL_FRAMEBUFFER_EXT)==GL_FRAMEBUFFER_COMPLETE_EXT) break;
+            }
+        }
+        while(!reflectfmt && colorfmts[++find]);
+        if(!reflectfmt) reflectfmt = colorfmts[find];
+
+        if(hasFBO)
+        {
+            if(!reflectiondb) { glGenRenderbuffers_(1, &reflectiondb); depthfmt = stencilfmt = GL_FALSE; }
+            if(!depthfmt) glBindRenderbuffer_(GL_RENDERBUFFER_EXT, reflectiondb);
+            find = hasstencil && hasDS ? 0 : stencilfmts;
+            do
+            {
+                if(!depthfmt) glRenderbufferStorage_(GL_RENDERBUFFER_EXT, depthfmts[find], size, size);
+                glFramebufferRenderbuffer_(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, reflectiondb);
+                if(depthfmt ? stencilfmt : find<stencilfmts) glFramebufferRenderbuffer_(GL_FRAMEBUFFER_EXT, GL_STENCIL_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, reflectiondb);
+                if(glCheckFramebufferStatus_(GL_FRAMEBUFFER_EXT)==GL_FRAMEBUFFER_COMPLETE_EXT) break;
+            }
+            while(!depthfmt && depthfmts[++find]);
+            if(!depthfmt)
+            {
+                glBindRenderbuffer_(GL_RENDERBUFFER_EXT, 0);
+                depthfmt = depthfmts[find];
+                stencilfmt = find<stencilfmts ? depthfmt : GL_FALSE;
+            }
+
+            glBindFramebuffer_(GL_FRAMEBUFFER_EXT, 0);
+        }
+    }
+    if(waterrefract && !ref->refracttex)
+    {
+        glGenTextures(1, &ref->refracttex);
+
+        int find = hasFBO && renderpath!=R_FIXEDFUNCTION && waterrefract && waterfade ? 0 : 2;
+        if(hasFBO)
+        {
+            if(!ref->refractfb) glGenFramebuffers_(1, &ref->refractfb);
+            glBindFramebuffer_(GL_FRAMEBUFFER_EXT, ref->refractfb);
+        }
+        do
+        {
+            createtexture(ref->refracttex, size, size, buf, 3, false, refractfmt ? refractfmt : colorfmts[find]);
+            if(!hasFBO) break;
+            else
+            {
+                glFramebufferTexture2D_(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, ref->refracttex, 0);
+                if(glCheckFramebufferStatus_(GL_FRAMEBUFFER_EXT)==GL_FRAMEBUFFER_COMPLETE_EXT) break;
+            }
+        }
+        while(!refractfmt && colorfmts[++find]);
+        if(!refractfmt) refractfmt = colorfmts[find];
+
+        if(hasFBO)
+        {
+            glFramebufferRenderbuffer_(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, reflectiondb);
+            if(stencilfmt) glFramebufferRenderbuffer_(GL_FRAMEBUFFER_EXT, GL_STENCIL_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, reflectiondb);
+            glBindFramebuffer_(GL_FRAMEBUFFER_EXT, 0);
+        }
+    }
+    if(buf) delete[] buf;
 }
 
 extern vtxarray *visibleva;
 extern void drawreflection(float z, bool refract, bool clear);
 
-VARP(reflectfps, 1, 30, 200);
-
 int rplanes = 0;
 
-static int lastreflectframe = 0, lastquery = 0;
+static int lastquery = 0;
 
 void queryreflections()
 {
     rplanes = 0;
 
     static int lastsize = 0, size = 1<<reflectsize;
-    if(!hasFBO) while(size>scr_w || size>scr_h) size /= 2;
+    if(!hasFBO) while(size>screen->w || size>screen->h) size /= 2;
     if(size!=lastsize) { if(lastsize) cleanreflections(); lastsize = size; }
 
     for(vtxarray *va = visibleva; va; va = va->next)
@@ -641,14 +796,15 @@ void queryreflections()
     }
    
     lastquery = totalmillis;
- 
-    if((editmode && showmat) || !hasOQ || !oqfrags || !oqwater || (!waterreflect && !waterrefract)) return;
 
-    int refs = 0, minmillis = 1000/reflectfps;
+    if((editmode && showmat) || !hasOQ || !oqfrags || !oqwater || nowater || (!waterreflect && !waterrefract)) return;
+
+    float offset = vertwater ? 0.1f : WATER_OFFSET;
+    int refs = 0;
     loopi(MAXREFLECTIONS)
     {
         Reflection &ref = reflections[i];
-        if(ref.height<0 || ref.lastused<totalmillis || totalmillis-lastreflectframe<minmillis || ref.matsurfs.empty())
+        if(ref.height<0 || ref.lastused<totalmillis || ref.matsurfs.empty())
         {
             ref.query = NULL;
             continue;
@@ -669,7 +825,7 @@ void queryreflections()
         loopvj(ref.matsurfs)
         {
             materialsurface &m = *ref.matsurfs[j];
-            drawmaterial(m.orient, m.o.x, m.o.y, m.o.z, m.csize, m.rsize, 1.1f);
+            drawmaterial(m.orient, m.o.x, m.o.y, m.o.z, m.csize, m.rsize, offset);
         }
         glEnd();
         endquery(ref.query);
@@ -692,13 +848,13 @@ VAR(maskreflect, 0, 2, 16);
 
 void maskreflection(Reflection &ref, float offset, bool reflect)
 {
-    if(!maskreflect || (reflect && camera1->o.z<ref.height+offset+4.0f))
+    if(!maskreflect)
     {
-        glClear(GL_DEPTH_BUFFER_BIT);
+        glClear(GL_DEPTH_BUFFER_BIT | (hasstencil && hasDS ? GL_STENCIL_BUFFER_BIT : 0));
         return;
     }
     glClearDepth(0);
-    glClear(GL_DEPTH_BUFFER_BIT);
+    glClear(GL_DEPTH_BUFFER_BIT | (hasstencil && hasDS ? GL_STENCIL_BUFFER_BIT : 0));
     glClearDepth(1);
     glDepthRange(1, 1);
     glDepthFunc(GL_ALWAYS);
@@ -709,7 +865,7 @@ void maskreflection(Reflection &ref, float offset, bool reflect)
     if(reflect)
     {
         glPushMatrix();
-        glTranslatef(0, 0, 2*ref.height);
+        glTranslatef(0, 0, 2*(ref.height+offset));
         glScalef(1, 1, -1);
     }
     int border = maskreflect;
@@ -735,14 +891,11 @@ void maskreflection(Reflection &ref, float offset, bool reflect)
 void drawreflections()
 {
     if(editmode && showmat) return;
-    if(!waterreflect && !waterrefract) return;
-    int minmillis = 1000/reflectfps;
-    if(lastquery-lastreflectframe<minmillis) return;
-    lastreflectframe = lastquery-(lastquery%minmillis);
+    if(nowater || (!waterreflect && !waterrefract)) return;
 
     static int lastdrawn = 0;
     int refs = 0, n = lastdrawn;
-    float offset = -1.1f;
+    float offset = -WATER_OFFSET;
     loopi(MAXREFLECTIONS)
     {
         Reflection &ref = reflections[++n%MAXREFLECTIONS];
@@ -757,9 +910,9 @@ void drawreflections()
         }
 
         int size = 1<<reflectsize;
-        if(!hasFBO) while(size>scr_w || size>scr_h) size /= 2;
+        if(!hasFBO) while(size>screen->w || size>screen->h) size /= 2;
 
-        if(!refs) glViewport(hasFBO ? 0 : scr_w-size, hasFBO ? 0 : scr_h-size, size, size);
+        if(!refs) glViewport(hasFBO ? 0 : screen->w-size, hasFBO ? 0 : screen->h-size, size, size);
 
         refs++;
         ref.lastupdate = totalmillis;
@@ -773,7 +926,7 @@ void drawreflections()
             if(!ref.fb)
             {
                 glBindTexture(GL_TEXTURE_2D, ref.tex);
-                glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, scr_w-size, scr_h-size, size, size);
+                glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, screen->w-size, screen->h-size, size, size);
             }
         }
 
@@ -785,7 +938,7 @@ void drawreflections()
             if(!ref.refractfb)
             {
                 glBindTexture(GL_TEXTURE_2D, ref.refracttex);
-                glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, scr_w-size, scr_h-size, size, size);
+                glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, screen->w-size, screen->h-size, size, size);
             }   
         }    
 
@@ -793,7 +946,7 @@ void drawreflections()
     }
     
     if(!refs) return;
-    glViewport(0, 0, scr_w, scr_h);
+    glViewport(0, 0, screen->w, screen->h);
     if(hasFBO) glBindFramebuffer_(GL_FRAMEBUFFER_EXT, 0);
 
     defaultshader->set();
