@@ -4,69 +4,10 @@
 #include "engine.h"
 
 header hdr;
-vector<ushort> texmru;
 
 VAR(octaentsize, 0, 128, 1024);
 VAR(entselradius, 0, 2, 10);
 
-void modifyoctaentity(bool add, int id, cube *c, const ivec &cor, int size, const ivec &bo, const ivec &br, vtxarray *lastva = NULL)
-{
-    loopoctabox(cor, size, bo, br)
-    {
-        ivec o(i, cor.x, cor.y, cor.z, size);
-        vtxarray *va = c[i].ext && c[i].ext->va ? c[i].ext->va : lastva;
-        if(c[i].children != NULL && size > octaentsize)
-            modifyoctaentity(add, id, c[i].children, o, size>>1, bo, br, va);
-        else if(add)
-        {
-            if(!c[i].ext || !c[i].ext->ents) ext(c[i]).ents = new octaentities(o, size);
-            switch(et->getents()[id]->type)
-            {
-                case ET_MAPMODEL:
-                    if(loadmodel(NULL, et->getents()[id]->attr2))
-                    {
-                        if(va && c[i].ext->ents->mapmodels.empty()) 
-                        {
-                            if(!va->mapmodels) va->mapmodels = new vector<octaentities *>;
-                            va->mapmodels->add(c[i].ext->ents);
-                        }
-                        c[i].ext->ents->mapmodels.add(id);
-                        break;
-                    }
-                    // invisible mapmodel
-                default:
-                    c[i].ext->ents->other.add(id);
-                    break;
-            }
-
-        }
-        else if(c[i].ext && c[i].ext->ents)
-        {
-            switch(et->getents()[id]->type)
-            {
-                case ET_MAPMODEL:
-                    if(loadmodel(NULL, et->getents()[id]->attr2))
-                    {
-                        c[i].ext->ents->mapmodels.removeobj(id);
-                        if(va && va->mapmodels && c[i].ext->ents->mapmodels.empty())
-                        {
-                            va->mapmodels->removeobj(c[i].ext->ents);
-                            if(va->mapmodels->empty()) DELETEP(va->mapmodels);
-                        }
-                        break;
-                    }
-                    // invisible mapmodel
-                default:
-                    c[i].ext->ents->other.removeobj(id);
-                    break;
-            }
-            if(c[i].ext->ents->mapmodels.empty() && c[i].ext->ents->other.empty()) 
-                freeoctaentities(c[i]);
-        }
-        if(c[i].ext && c[i].ext->ents) c[i].ext->ents->query = NULL;
-    }
-}
-
 bool getentboundingbox(extentity &e, ivec &o, ivec &r)
 {
     switch(e.type)
@@ -81,7 +22,7 @@ bool getentboundingbox(extentity &e, ivec &o, ivec &r)
                 vec center, radius;
                 m->boundbox(0, center, radius);
                 rotatebb(center, radius, e.attr1);
-                
+
                 o = e.o;
                 o.add(center);
                 r = radius;
@@ -101,10 +42,94 @@ bool getentboundingbox(extentity &e, ivec &o, ivec &r)
     return true;
 }
 
+void modifyoctaentity(bool add, int id, cube *c, const ivec &cor, int size, const ivec &bo, const ivec &br, vtxarray *lastva = NULL)
+{
+    loopoctabox(cor, size, bo, br)
+    {
+        ivec o(i, cor.x, cor.y, cor.z, size);
+        vtxarray *va = c[i].ext && c[i].ext->va ? c[i].ext->va : lastva;
+        if(c[i].children != NULL && size > octaentsize)
+            modifyoctaentity(add, id, c[i].children, o, size>>1, bo, br, va);
+        else if(add)
+        {
+            if(!c[i].ext || !c[i].ext->ents) ext(c[i]).ents = new octaentities(o, size);
+            octaentities &oe = *c[i].ext->ents;
+            switch(et->getents()[id]->type)
+            {
+                case ET_MAPMODEL:
+                    if(loadmodel(NULL, et->getents()[id]->attr2))
+                    {
+                        if(va && oe.mapmodels.empty()) 
+                        {
+                            if(!va->mapmodels) va->mapmodels = new vector<octaentities *>;
+                            va->mapmodels->add(&oe);
+                        }
+                        oe.mapmodels.add(id);
+                        loopk(3)
+                        {
+                            oe.bbmin[k] = min(oe.bbmin[k], max(oe.o[k], bo[k]));
+                            oe.bbmax[k] = max(oe.bbmax[k], min(oe.o[k]+size, bo[k]+br[k]));
+                        }
+                        break;
+                    }
+                    // invisible mapmodel
+                default:
+                    oe.other.add(id);
+                    break;
+            }
+
+        }
+        else if(c[i].ext && c[i].ext->ents)
+        {
+            octaentities &oe = *c[i].ext->ents;
+            switch(et->getents()[id]->type)
+            {
+                case ET_MAPMODEL:
+                    if(loadmodel(NULL, et->getents()[id]->attr2))
+                    {
+                        oe.mapmodels.removeobj(id);
+                        if(va && va->mapmodels && oe.mapmodels.empty())
+                        {
+                            va->mapmodels->removeobj(&oe);
+                            if(va->mapmodels->empty()) DELETEP(va->mapmodels);
+                        }
+                        oe.bbmin = oe.bbmax = oe.o;
+                        oe.bbmin.add(oe.size);
+                        loopvj(oe.mapmodels)
+                        {
+                            extentity &e = *et->getents()[oe.mapmodels[j]];
+                            ivec eo, er;
+                            if(getentboundingbox(e, eo, er)) loopk(3)
+                            {
+                                oe.bbmin[k] = min(oe.bbmin[k], eo[k]);
+                                oe.bbmax[k] = max(oe.bbmax[k], eo[k]+er[k]);
+                            }
+                        }
+                        loopk(3)
+                        {
+                            oe.bbmin[k] = max(oe.bbmin[k], oe.o[k]);
+                            oe.bbmax[k] = min(oe.bbmax[k], oe.o[k]+size);
+                        }
+                        break;
+                    }
+                    // invisible mapmodel
+                default:
+                    oe.other.removeobj(id);
+                    break;
+            }
+            if(oe.mapmodels.empty() && oe.other.empty()) 
+                freeoctaentities(c[i]);
+        }
+        if(c[i].ext && c[i].ext->ents) c[i].ext->ents->query = NULL;
+    }
+}
+
 static void modifyoctaent(bool add, int id)
 {
+    vector<extentity *> &ents = et->getents();
+    if(!ents.inrange(id)) return;
     ivec o, r;
-    extentity &e = *et->getents()[id];
+    extentity &e = *ents[id];
     if((e.inoctanode!=0)==add || !getentboundingbox(e, o, r)) return;
     e.inoctanode = add;
     modifyoctaentity(add, id, worldroot, ivec(0, 0, 0), hdr.worldsize>>1, o, r);
@@ -118,8 +143,11 @@ static inline void removeentity(int id) { modifyoctaent(false, id); }
 void freeoctaentities(cube &c)
 {
     if(!c.ext) return;
-    while(c.ext->ents && !c.ext->ents->mapmodels.empty()) removeentity(c.ext->ents->mapmodels.pop());
-    while(c.ext->ents && !c.ext->ents->other.empty())     removeentity(c.ext->ents->other.pop());
+    if(et->getents().length())
+    {
+        while(c.ext->ents && !c.ext->ents->mapmodels.empty()) removeentity(c.ext->ents->mapmodels.pop());
+        while(c.ext->ents && !c.ext->ents->other.empty())     removeentity(c.ext->ents->other.pop());
+    }
     if(c.ext->ents)
     {
         delete c.ext->ents;
@@ -147,10 +175,18 @@ char *entname(entity &e)
 }
 
 extern selinfo sel;
-extern int orient;
 extern bool havesel, selectcorners;
 int entlooplevel = 0;
-int efocus = -1, enthover, entorient = -1;
+int efocus = -1, enthover = -1, entorient = -1, oldhover = -1;
+bool undonext = true;
+
+VAR(entediting, 0, 0, 1);
+
+bool noentedit()
+{
+    if(!editmode) { conoutf("operation only allowed in edit mode"); return true; }
+    return !entediting;
+}
 
 bool pointinsel(selinfo &sel, vec &o)
 {
@@ -169,7 +205,16 @@ bool haveselent()
     return entgroup.length() > 0;
 }
 
-extern void entcancel();
+void entcancel()
+{
+    entgroup.setsize(0);
+}
+
+void entadd(int id)
+{
+    undonext = true;
+    entgroup.add(id);
+}
 
 void initundoent(undoblock &u)
 {
@@ -188,18 +233,88 @@ void initundoent(undoblock &u)
 
 void makeundoent()
 {
+    if(!undonext) return;
+    undonext = false;
+    oldhover = enthover;
     undoblock u;
     initundoent(u);
     if(u.n) addundo(u);
 }
 
+void detachentity(extentity &e)
+{
+    if(!e.attached) return;
+    e.attached->attached = NULL;
+    e.attached = NULL;
+}
+
+VAR(attachradius, 1, 100, 1000);
+
+void attachentity(extentity &e)
+{
+    switch(e.type)
+    {
+        case ET_SPOTLIGHT:
+            break;
+
+        default:
+            if(e.type<ET_GAMESPECIFIC || !et->mayattach(e)) return;
+            break;
+    }
+
+    detachentity(e);
+
+    vector<extentity *> &ents = et->getents();
+    int closest = -1;
+    float closedist = 1e10f;
+    loopv(ents)
+    {
+        extentity *a = ents[i];
+        if(a->attached) continue;
+        switch(e.type)
+        {
+            case ET_SPOTLIGHT: 
+                if(a->type!=ET_LIGHT) continue; 
+                break;
+
+            default:
+                if(e.type<ET_GAMESPECIFIC || !et->attachent(e, *a)) continue;
+                break;
+        }
+        float dist = e.o.dist(a->o);
+        if(dist < closedist)
+        {
+            closest = i;
+            closedist = dist;
+        }
+    }
+    if(closedist>attachradius) return;
+    e.attached = ents[closest];
+    ents[closest]->attached = &e;
+}
+
+void attachentities()
+{
+    vector<extentity *> &ents = et->getents();
+    loopv(ents) attachentity(*ents[i]);
+}
+
 // convenience macros implicitly define:
 // e         entity, currently edited ent
 // n         int,    index to currently edited ent
-#define addimplicit(f)  { if(entgroup.empty() && enthover>=0) { entgroup.add(enthover); f; entgroup.drop(); } else f; }
-#define entfocus(i, f)  { int n = efocus = (i); if(n>=0) { entity &e = *et->getents()[n]; f; } }
-#define entedit(i, f)   { entfocus(i, removeentity(n); f; if(e.type != ET_EMPTY) addentity(n); et->editent(n)); }
-#define addgroup(exp)   { loopv(et->getents()) entfocus(i, if(exp) entgroup.add(n)); }
+#define addimplicit(f)  { if(entgroup.empty() && enthover>=0) { entadd(enthover); undonext = (enthover != oldhover); f; entgroup.drop(); } else f; }
+#define entfocus(i, f)  { int n = efocus = (i); if(n>=0) { extentity &e = *et->getents()[n]; f; } }
+#define entedit(i, f) \
+{ \
+    entfocus(i, \
+    int oldtype = e.type; \
+    removeentity(n);  \
+    f; \
+    if(oldtype!=e.type) detachentity(e); \
+    if(e.type!=ET_EMPTY) { addentity(n); if(oldtype!=e.type) attachentity(e); } \
+    et->editent(n)); \
+}
+#define addgroup(exp)   { loopv(et->getents()) entfocus(i, if(exp) entadd(n)); }
 #define setgroup(exp)   { entcancel(); addgroup(exp); }
 #define groupeditloop(f){ entlooplevel++; int _ = efocus; loopv(entgroup) entedit(entgroup[i], f); efocus = _; entlooplevel--; }
 #define groupeditpure(f){ if(entlooplevel>0) { entedit(efocus, f); } else groupeditloop(f); }
@@ -210,19 +325,21 @@ void copyundoents(undoblock &d, undoblock &s)
 {
     entcancel();
     loopi(s.n)
-        entgroup.add(s.e[i].i);
+        entadd(s.e[i].i);
     initundoent(d);
+   	loopi(s.n) if(s.e[i].e.type==ET_EMPTY)
+		entgroup.remove(s.e[i].i);
 }
 
 void pasteundoents(undoblock &u)
 {
     loopi(u.n)
-        entedit(u.e[i].i, e = u.e[i].e);
+        entedit(u.e[i].i, (entity &)e = u.e[i].e);
 }
 
 void entflip()
 {
-    if(noedit(true)) return;
+    if(noentedit()) return;
     int d = dimension(sel.orient);
     float mid = sel.s[d]*sel.grid/2+sel.o[d];
     groupeditundo(e.o[d] -= (e.o[d]-mid)*2);
@@ -230,7 +347,7 @@ void entflip()
 
 void entrotate(int *cw)
 {
-    if(noedit(true)) return;
+    if(noentedit()) return;
     int d = dimension(sel.orient);
     int dd = *cw<0 == dimcoord(sel.orient) ? R[d] : C[d];
     float mid = sel.s[dd]*sel.grid/2+sel.o[dd];
@@ -266,10 +383,8 @@ void entselectionbox(const entity &e, vec &eo, vec &es)
 }
 
 VAR(entselsnap, 0, 0, 1);
-VAR(passthroughent, 0, 0, 1);
 VAR(entmovingshadow, 0, 1, 1);
 
-extern int rayent(const vec &o, const vec &ray, int &orient);
 extern void boxs(int orient, vec o, const vec &s);
 extern void boxs3D(const vec &o, vec s, int g);
 extern void editmoveplane(const vec &o, const vec &ray, int d, float off, vec &handle, vec &dest, bool first);
@@ -278,7 +393,7 @@ bool initentdragging = true;
 
 void entdrag(const vec &ray)
 {
-    if(!haveselent()) return;
+    if(noentedit() || !haveselent()) return;
 
     float r = 0, c = 0;
     static vec v, handle;
@@ -305,10 +420,157 @@ void entdrag(const vec &ray)
     initentdragging = false;
 }
 
+VAR(showentradius, 0, 1, 1);
+
+void renderentradius(extentity &e)
+{
+    if(!showentradius) return;
+    float radius = 0.0f, angle = 0.0f, ring = 0.0f;
+    vec dir(0, 0, 0);
+    float color[3] = {0, 1, 1};
+    switch(e.type)
+    {
+        case ET_LIGHT:
+            radius = e.attr1;
+            color[0] = e.attr2/255.0f;
+            color[1] = e.attr3/255.0f;
+            color[2] = e.attr4/255.0f;
+            break;
+
+        case ET_SPOTLIGHT:
+            if(e.attached)
+            {
+                radius = e.attached->attr1;
+                if(!radius) radius = 2*e.o.dist(e.attached->o);
+                dir = vec(e.o).sub(e.attached->o).normalize();
+                angle = max(1, min(90, e.attr1));
+            }
+            break;
+
+        case ET_SOUND:
+            radius = e.attr2;
+            break;
+
+        case ET_ENVMAP:
+        {
+            extern int envmapradius;
+            radius = e.attr1 ? max(0, min(10000, e.attr1)) : envmapradius;
+            break;
+        }
+
+        case ET_MAPMODEL:
+        case ET_PLAYERSTART:
+            radius = 4;
+            if(e.type==ET_MAPMODEL && e.attr3) ring = checktriggertype(e.attr3, TRIG_COLLIDE) ? 20 : 12;
+            vecfromyawpitch(e.attr1, 0, 1, 0, dir);
+            break;
+
+        default:
+            if(e.type>=ET_GAMESPECIFIC) et->entradius(e, radius, angle, dir);
+            break;
+    }
+    if(radius<=0) return;
+    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
+    loopj(2)
+    {
+        if(!j)
+        {
+            glDepthFunc(GL_GREATER);
+            glColor3f(0.25f, 0.25f, 0.25f);
+        }
+        else 
+        {
+            glDepthFunc(GL_LESS);
+            glColor3fv(color);
+        }
+        if(e.attached)
+        {
+            glBegin(GL_LINES);
+            glVertex3fv(e.o.v);
+            glVertex3fv(e.attached->o.v);
+            glEnd();
+        }
+        if(ring)
+        {
+            glBegin(GL_LINE_LOOP);
+            loopi(16)
+            {
+                vec p(e.o);
+                p.x += ring*cosf(2*M_PI*i/16.0f);
+                p.y += ring*sinf(2*M_PI*i/16.0f);
+                glVertex3fv(p.v);
+            }
+            glEnd();
+        }
+        if(dir.iszero()) loopk(3)
+        {
+            glBegin(GL_LINE_LOOP);
+            loopi(16)
+            {
+                vec p(e.o);
+                p[k>=2 ? 1 : 0] += radius*cosf(2*M_PI*i/16.0f);
+                p[k>=1 ? 2 : 1] += radius*sinf(2*M_PI*i/16.0f);
+                glVertex3fv(p.v);
+            }
+            glEnd();
+        }
+        else if(!angle)
+        {
+            float arrowsize = min(radius/8, 0.5f);
+            vec target(vec(dir).mul(radius).add(e.o)), arrowbase(vec(dir).mul(radius - arrowsize).add(e.o)), spoke;
+            spoke.orthogonal(dir);
+            spoke.normalize();
+            spoke.mul(arrowsize);
+            glBegin(GL_LINES);
+            glVertex3fv(e.o.v);
+            glVertex3fv(target.v);
+            glEnd();
+            glBegin(GL_TRIANGLE_FAN);
+            glVertex3fv(target.v);
+            loopi(5)
+            {
+                vec p(spoke);
+                p.rotate(2*M_PI*i/4.0f, dir);
+                p.add(arrowbase);
+                glVertex3fv(p.v);
+            }
+            glEnd();
+        }
+        else
+        {
+            vec spot(vec(dir).mul(radius*cosf(angle*RAD)).add(e.attached->o)), spoke;
+            spoke.orthogonal(dir);
+            spoke.normalize();
+            spoke.mul(radius*sinf(angle*RAD));
+            glBegin(GL_LINES);
+            loopi(8)
+            {
+                vec p(spoke);
+                p.rotate(2*M_PI*i/8.0f, dir);
+                p.add(spot);
+                glVertex3fv(e.attached->o.v);
+                glVertex3fv(p.v);
+            }
+            glEnd();
+            glBegin(GL_LINE_LOOP);
+            loopi(8)
+            {
+                vec p(spoke);
+                p.rotate(2*M_PI*i/8.0f, dir);
+                p.add(spot);
+                glVertex3fv(p.v);
+            }
+            glEnd();
+        }
+    }
+    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
+}
+
 void renderentselection(const vec &o, const vec &ray, bool entmoving)
 {   
+    if(noentedit()) return;
     vec eo, es;
- 
+
     glColor3ub(0, 40, 0);
     loopv(entgroup) entfocus(entgroup[i],     
         entselectionbox(e, eo, es);
@@ -332,16 +594,14 @@ void renderentselection(const vec &o, const vec &ray, bool entmoving)
         boxs(entorient, eo, es);
         glLineWidth(1);
     }
-}
 
-void entadd(int id)
-{
-    entgroup.removeobj(id);
-    entgroup.add(id);
+    loopv(entgroup) entfocus(entgroup[i], renderentradius(e));
+    if(enthover>=0) entfocus(enthover, renderentradius(e));
 }
 
 bool enttoggle(int id)
 {
+    undonext = true;
     int i = entgroup.find(id);
     if(i < 0)
         entadd(id);
@@ -350,11 +610,12 @@ bool enttoggle(int id)
     return i < 0;
 }
 
-bool hoveringonent(const vec &o, const vec &ray)
+bool hoveringonent(int ent, int orient)
 {
-    if(!passthroughent)          
-        if((efocus = enthover = rayent(o, ray, entorient)) >= 0)
-            return true;
+    if(noentedit()) return false;
+    entorient = orient;
+    if((efocus = enthover = ent) >= 0)
+        return true;
     efocus   = entgroup.empty() ? -1 : entgroup.last();
     enthover = -1;
     return false;
@@ -362,11 +623,11 @@ bool hoveringonent(const vec &o, const vec &ray)
 
 VAR(entitysurf, 0, 0, 1);
 VARF(entmoving, 0, 0, 2,
-    if(enthover < 0)
+    if(enthover < 0 || noentedit())
         entmoving = 0;
     else if(entmoving == 1)
         entmoving = enttoggle(enthover);
-    else if(entmoving == 2)
+    else if(entmoving == 2 && entgroup.find(enthover) < 0)
         entadd(enthover);
     if(entmoving > 0)
         initentdragging = true;
@@ -374,7 +635,7 @@ VARF(entmoving, 0, 0, 2,
 
 void entpush(int *dir)
 {
-    if(noedit(true)) return;
+    if(noentedit()) return;
     int d = dimension(entorient);
     int s = dimcoord(entorient) ? -*dir : *dir;
     if(entmoving) 
@@ -412,7 +673,7 @@ COMMAND(entpush, "i");
 
 void delent()
 {
-    if(noedit(true)) return;
+    if(noentedit()) return;
     groupedit(e.type = ET_EMPTY;);
     entcancel();
 }
@@ -446,7 +707,7 @@ bool dropentity(entity &e, int drop = -1)
     switch(drop)
     {
     case 1:
-        if(e.type != ET_LIGHT)
+        if(e.type != ET_LIGHT && e.type != ET_SPOTLIGHT)
             dropenttofloor(&e);
         break;
     case 2:
@@ -475,10 +736,18 @@ bool dropentity(entity &e, int drop = -1)
 
 void dropent()
 {
-    if(noedit(true)) return;
+    if(noentedit()) return;
     groupedit(dropentity(e));
 }
 
+void attachent()
+{
+    if(noentedit()) return;
+    groupedit(attachentity(e));
+}
+
+COMMAND(attachent, "");
+
 extentity *newentity(bool local, const vec &o, int type, int v1, int v2, int v3, int v4)
 {
     extentity &e = *et->newentity();
@@ -524,7 +793,7 @@ void newentity(int type, int a1, int a2, int a3, int a4)
 
 void newent(char *what, int *a1, int *a2, int *a3, int *a4)
 {
-    if(noedit(true)) return;
+    if(noentedit()) return;
     int type = findtype(what);
     if(type != ET_EMPTY)
         newentity(type, *a1, *a2, *a3, *a4);
@@ -535,6 +804,7 @@ vector<entity> entcopybuf;
 
 void entcopy()
 {
+    if(noentedit()) return;
     entcopygrid = sel.grid;
     entcopybuf.setsize(0);
     loopv(entgroup) 
@@ -543,6 +813,7 @@ void entcopy()
 
 void entpaste()
 {
+    if(noentedit()) return;
     if(entcopybuf.length()==0) return;
     entcancel();
     int last = et->getents().length()-1;
@@ -554,10 +825,10 @@ void entpaste()
         o.mul(m).add(sel.o.v);
         extentity *e = newentity(true, o, ET_EMPTY, c.attr1, c.attr2, c.attr3, c.attr4);
         et->getents().add(e);
-        entgroup.add(++last);
+        entadd(++last);
     }
     int j = 0;
-    groupeditundo(e.type = entcopybuf[j++].type);
+    groupeditundo(e.type = entcopybuf[j++].type;);
 }
 
 COMMAND(newent, "siiii");
@@ -568,7 +839,7 @@ COMMAND(entpaste, "");
 
 void entset(char *what, int *a1, int *a2, int *a3, int *a4)
 {
-    if(noedit(true)) return;
+    if(noentedit()) return;
     int type = findtype(what);
     groupedit(e.type=type;
               e.attr1=*a1;
@@ -577,11 +848,11 @@ void entset(char *what, int *a1, int *a2, int *a3, int *a4)
               e.attr4=*a4;);
 }
 
-ICOMMAND(enthavesel,"",  addimplicit(intret(entgroup.length())));
-ICOMMAND(entselect, "s", addgroup(e.type != ET_EMPTY && entgroup.find(n)<0 && execute(args[0])>0));
-ICOMMAND(entloop,   "s", addimplicit(groupeditloop(((void)e, execute(args[0])))));
-ICOMMAND(insel,     "",  entfocus(efocus, intret(pointinsel(sel, e.o))));
-ICOMMAND(entget,    "",  entfocus(efocus, s_sprintfd(s)("%s %d %d %d %d", et->entname(e.type), e.attr1, e.attr2, e.attr3, e.attr4);  result(s)));
+ICOMMAND(enthavesel,"",  (), addimplicit(intret(entgroup.length())));
+ICOMMAND(entselect, "s", (char *body), if(!noentedit()) addgroup(e.type != ET_EMPTY && entgroup.find(n)<0 && execute(body)>0));
+ICOMMAND(entloop,   "s", (char *body), if(!noentedit()) addimplicit(groupeditloop(((void)e, execute(body)))));
+ICOMMAND(insel,     "",  (), entfocus(efocus, intret(pointinsel(sel, e.o))));
+ICOMMAND(entget,    "",  (), entfocus(efocus, s_sprintfd(s)("%s %d %d %d %d", et->entname(e.type), e.attr1, e.attr2, e.attr3, e.attr4);  result(s)));
 COMMAND(entset, "siiii");
 
 
@@ -593,7 +864,7 @@ int findentity(int type, int index)
     return -1;
 }
 
-int spawncycle = -1, fixspawn = 2;
+int spawncycle = -1, fixspawn = 4;
 
 void findplayerspawn(dynent *d, int forceent)   // place at random spawn. also used by monsters!
 {
@@ -646,13 +917,15 @@ void resetmap()
     clearmapsounds();
     cleanreflections();
     resetlightmaps();
+    clearparticles();
     clearsleep();
     cancelsel();
     pruneundos();
 
     setvar("gamespeed", 100);
     setvar("paused", 0);
-    setvar("wireframe", 0);
+
+    et->getents().deletecontentsp();
 }
 
 void startmap(const char *name)
@@ -682,11 +955,10 @@ bool emptymap(int scale, bool force)    // main empty world creation routine
     hdr.mapprec = 32;
     hdr.mapllod = 0;
     hdr.lightmaps = 0;
-    memset(hdr.reserved1, 0, sizeof(hdr.reserved1));
-    memset(hdr.reserved2, 0, sizeof(hdr.reserved2));
+    memset(hdr.skylight, 0, sizeof(hdr.skylight));
+    memset(hdr.reserved, 0, sizeof(hdr.reserved));
     texmru.setsize(0);
     freeocta(worldroot);
-    et->getents().setsize(0);
     worldroot = newcubes(F_EMPTY);
     loopi(4) solidfaces(worldroot[i]);
 
@@ -732,6 +1004,13 @@ void mapenlarge() { if(enlargemap(false)) cl->newmap(-1); }
 COMMAND(newmap, "i");
 COMMAND(mapenlarge, "");
 
+void mapname()
+{
+    result(cl->getclientmap());
+}
+
+COMMAND(mapname, "");
+
 void mpeditent(int i, const vec &o, int type, int attr1, int attr2, int attr3, int attr4, bool local)
 {
     if(et->getents().length()<=i)
@@ -740,27 +1019,25 @@ void mpeditent(int i, const vec &o, int type, int attr1, int attr2, int attr3, i
         extentity *e = newentity(local, o, type, attr1, attr2, attr3, attr4);
         et->getents().add(e);
         addentity(i);
+        attachentity(*e);
     }
     else
     {
         extentity &e = *et->getents()[i];
         removeentity(i);
+        int oldtype = e.type;
+        if(oldtype!=type) detachentity(e);
         e.type = type;
         e.o = o;
         e.attr1 = attr1; e.attr2 = attr2; e.attr3 = attr3; e.attr4 = attr4;
         addentity(i);
+        if(oldtype!=type) attachentity(e);
     }
 }
 
 int getworldsize() { return hdr.worldsize; }
 int getmapversion() { return hdr.version; }
 
-bool insideworld(const vec &o)
-{
-    loopi(3) if(o.v[i]<0 || o.v[i]>=hdr.worldsize) return false;
-    return true;
-}
-
 int triggertypes[NUMTRIGGERTYPES] =
 {
     0,
@@ -777,7 +1054,8 @@ int triggertypes[NUMTRIGGERTYPES] =
     TRIG_COLLIDE | TRIG_TOGGLE | TRIG_LOCKED | TRIG_RUMBLE,
     TRIG_DISAPPEAR,
     TRIG_DISAPPEAR | TRIG_RUMBLE,
-    0, 0 /* reserved */
+    TRIG_DISAPPEAR | TRIG_COLLIDE | TRIG_LOCKED,
+    0 /* reserved */
 };
 
 void resettriggers()
