@@ -5,19 +5,28 @@
 
 void backup(char *name, char *backupname)
 {
-    remove(backupname);
-    rename(name, backupname);
+    string backupfile;
+    s_strcpy(backupfile, findfile(backupname, "wb"));
+    remove(backupfile);
+    rename(findfile(name, "wb"), backupfile);
 }
 
-string cgzname, bakname, pcfname, mcfname;
+string cgzname, bakname, pcfname, mcfname, picname;
 
 VARP(savebak, 0, 2, 2);
 
+void cutogz(char *s)
+{
+    char *ogzp = strstr(s, ".ogz");
+    if(ogzp) *ogzp = '\0';
+}
+
 void setnames(const char *fname, const char *cname = 0)
 {
     if(!cname) cname = fname;
     string name, pakname, mapname, cfgname;
     s_strncpy(name, cname, 100);
+    cutogz(name);
     char *slash = strpbrk(name, "/\\");
     if(slash)
     {
@@ -31,15 +40,18 @@ void setnames(const char *fname, const char *cname = 0)
     }
     if(strpbrk(fname, "/\\")) s_strcpy(mapname, fname);
     else s_sprintf(mapname)("base/%s", fname);
+    cutogz(mapname);
 
     s_sprintf(cgzname)("packages/%s.ogz", mapname);
     if(savebak==1) s_sprintf(bakname)("packages/%s.BAK", mapname);
     else s_sprintf(bakname)("packages/%s_%d.BAK", mapname, lastmillis);
     s_sprintf(pcfname)("packages/%s/package.cfg", pakname);
     s_sprintf(mcfname)("packages/%s/%s.cfg",      pakname, cfgname);
+    s_sprintf(picname)("packages/%s.jpg", mapname);
 
     path(cgzname);
     path(bakname);
+    path(picname);
 }
 
 ushort readushort(gzFile f)
@@ -234,9 +246,9 @@ cube *loadchildren(gzFile f)
 void save_world(char *mname, bool nolms)
 {
     if(!*mname) mname = cl->getclientmap();
-    setnames(mname);
+    setnames(*mname ? mname : "untitled");
     if(savebak) backup(cgzname, bakname);
-    gzFile f = gzopen(cgzname, "wb9");
+    gzFile f = opengzfile(cgzname, "wb9");
     if(!f) { conoutf("could not write map to %s", cgzname); return; }
     hdr.version = MAPVERSION;
     hdr.numents = 0;
@@ -310,7 +322,7 @@ void load_world(const char *mname, const char *cname)        // still supports a
 {
     int loadingstart = SDL_GetTicks();
     setnames(mname, cname);
-    gzFile f = gzopen(cgzname, "rb9");
+    gzFile f = opengzfile(cgzname, "rb9");
     if(!f) { conoutf("could not read map %s", cgzname); return; }
     header newhdr;
     gzread(f, &newhdr, sizeof(header));
@@ -319,7 +331,8 @@ void load_world(const char *mname, const char *cname)        // still supports a
     if(newhdr.version>MAPVERSION) { conoutf("map %s requires a newer version of cube 2", cgzname); gzclose(f); return; }
     hdr = newhdr;
     resetmap();
-    computescreen(mname);
+    Texture *mapshot = textureload(picname, 0, true, false);
+    computescreen(mname, mapshot!=notexture ? mapshot : NULL);
     if(hdr.version<=20) conoutf("loading older / less efficient map format, may benefit from \"calclight 2\", then \"savecurrentmap\"");
     if(!hdr.ambient) hdr.ambient = 25;
     if(!hdr.lerpsubdivsize)
@@ -381,9 +394,8 @@ void load_world(const char *mname, const char *cname)        // still supports a
     worldroot = NULL;
 
     show_out_of_renderloop_progress(0, "loading entities...");
-    vector<extentity *> &ents = et->getents();
-    ents.deletecontentsp();
 
+    vector<extentity *> &ents = et->getents();
     char *ebuf = new char[et->extraentinfosize()];
     loopi(hdr.numents)
     {
@@ -411,6 +423,7 @@ void load_world(const char *mname, const char *cname)        // still supports a
         if(hdr.version <= 20 && e.type >= ET_ENVMAP) e.type++;
         if(hdr.version <= 21 && e.type >= ET_PARTICLES) e.type++;
         if(hdr.version <= 22 && e.type >= ET_SOUND) e.type++;
+        if(hdr.version <= 23 && e.type >= ET_SPOTLIGHT) e.type++;
         if(!samegame)
         {
             if(e.type>=ET_GAMESPECIFIC || hdr.version<=14)
@@ -421,7 +434,7 @@ void load_world(const char *mname, const char *cname)        // still supports a
         }
         if(!insideworld(e.o))
         {
-            if(e.type != ET_LIGHT)
+            if(e.type != ET_LIGHT && e.type != ET_SPOTLIGHT)
             {
                 conoutf("warning: ent outside of world: enttype[%s] index %d (%f, %f, %f)", et->entname(e.type), i, e.o.x, e.o.y, e.o.z);
             }
@@ -447,34 +460,26 @@ void load_world(const char *mname, const char *cname)        // still supports a
     show_out_of_renderloop_progress(0, "validating...");
     validatec(worldroot, hdr.worldsize>>1);
 
-    if(hdr.version < 7 || !hdr.lightmaps) clearlights();
-    else
+    if(hdr.version >= 7) loopi(hdr.lightmaps)
     {
-        loopi(hdr.lightmaps)
+        show_out_of_renderloop_progress(i/(float)hdr.lightmaps, "loading lightmaps...");
+        LightMap &lm = lightmaps.add();
+        if(hdr.version >= 17)
         {
-            show_out_of_renderloop_progress(i/(float)hdr.lightmaps, "loading lightmaps...");
-            LightMap &lm = lightmaps.add();
-            if(hdr.version >= 17)
+            int type = gzgetc(f);
+            lm.type = type&0xF;
+            if(hdr.version >= 20 && type&0x80)
             {
-                int type = gzgetc(f);
-                lm.type = type&0xF;
-                if(hdr.version >= 20 && type&0x80)
-                {
-                    lm.unlitx = readushort(f);
-                    lm.unlity = readushort(f);
-                }
+                lm.unlitx = readushort(f);
+                lm.unlity = readushort(f);
             }
-            gzread(f, lm.data, 3 * LM_PACKW * LM_PACKH);
-            lm.finalize();
         }
-        initlights();
+        gzread(f, lm.data, 3 * LM_PACKW * LM_PACKH);
+        lm.finalize();
     }
 
     gzclose(f);
 
-// FIXME: need to load texture slots before VAs can be properly sorted    
-//    allchanged();
-
     conoutf("read map %s (%.1f seconds)", cgzname, (SDL_GetTicks()-loadingstart)/1000.0f);
     conoutf("%s", hdr.maptitle);
 
@@ -484,8 +489,6 @@ void load_world(const char *mname, const char *cname)        // still supports a
     execfile(mcfname);
     overrideidents = false;
 
-    allchanged(true);
-
     loopv(ents)
     {
         extentity &e = *ents[i];
@@ -493,10 +496,18 @@ void load_world(const char *mname, const char *cname)        // still supports a
         {
             mapmodelinfo &mmi = getmminfo(e.attr2);
             if(!&mmi) conoutf("could not find map model: %d", e.attr2);
-            else if(!loadmodel(mmi.name)) conoutf("could not load model: %s", mmi.name);
+            else if(!loadmodel(NULL, e.attr2, true)) conoutf("could not load model: %s", mmi.name);
         }
     }
 
+    cl->preload();
+
+    initlights();
+    allchanged(true);
+
+    computescreen(mname, mapshot!=notexture ? mapshot : NULL);
+    attachentities();
+
     startmap(cname ? cname : mname);
 }
 
@@ -512,25 +523,34 @@ void writeobj(char *name)
     hasVBO = false;
     allchanged();
     s_sprintfd(fname)("%s.obj", name);
-    FILE *f = fopen(fname, "w"); 
+    FILE *f = openfile(path(fname), "w"); 
     if(!f) return;
     fprintf(f, "# obj file of sauerbraten level\n");
     extern vector<vtxarray *> valist;
     loopv(valist)
     {
-        vtxarray &v = *valist[i];
-        vertex *verts = v.vbuf;
-        if(verts)
+        vtxarray &va = *valist[i];
+        uchar *verts = (uchar *)va.vbuf;
+        if(!verts) continue;
+        int vtxsize = VTXSIZE;
+        loopj(va.verts) 
         {
-            loopj(v.verts) fprintf(f, "v %d %d %d\n", verts[j].x, verts[j].y, verts[j].z);
-            ushort *ebuf = v.l0.ebuf;
-            loopi(v.l0.texs) loopl(3) loopj(v.l0.eslist[i].length[l]/3)
-            {
-                fprintf(f, "f");
-                for(int k = 0; k<3; k++) fprintf(f, " %d", ebuf[k]-v.verts);
-                ebuf += 3;
-                fprintf(f, "\n");
-            }
+            vvec vv;
+            if(floatvtx) { vec &f = *(vec *)verts; loopk(3) vv[k] = short(f[k]); }
+            else vv = *(vvec *)verts;
+            vec v = vv.tovec(va.x, va.y, va.z);
+            if(vv.x&((1<<VVEC_FRAC)-1)) fprintf(f, "v %.3f ", v.x); else fprintf(f, "v %d ", int(v.x));
+            if(vv.y&((1<<VVEC_FRAC)-1)) fprintf(f, "%.3f ", v.y); else fprintf(f, "%d ", int(v.y));
+            if(vv.z&((1<<VVEC_FRAC)-1)) fprintf(f, "%.3f\n", v.z); else fprintf(f, "%d\n", int(v.z));
+            verts += vtxsize;
+        }
+        ushort *ebuf = va.l0.ebuf;
+        loopi(va.l0.tris)
+        {
+            fprintf(f, "f");
+            for(int k = 0; k<3; k++) fprintf(f, " %d", ebuf[k]-va.verts);
+            ebuf += 3;
+            fprintf(f, "\n");
         }
     }
     fclose(f);
