@@ -8,21 +8,24 @@ struct capturestate
     static const int OCCUPYLIMIT = 100;
     static const int CAPTURESCORE = 1;
     static const int SCORESECS = 10;
-    static const int AMMOSECS = 20;
+    static const int AMMOSECS = 15;
+    static const int REGENSECS = 1;
+    static const int REGENHEALTH = 10;
+    static const int REGENARMOUR = 10;
+    static const int REGENAMMO = 20;
     static const int MAXAMMO = 5;
     static const int REPAMMODIST = 32;
     static const int RESPAWNSECS = 10;        
-    static const int RADARRADIUS = 1024;
 
     struct baseinfo
     {
         vec o;
         string owner, enemy;
 #ifndef CAPTURESERV
-        string info;
+        string name, info;
         extentity *ent;
 #endif
-        int ammotype, ammo, enemies, converted, capturetime;
+        int ammotype, ammo, owners, enemies, converted, capturetime;
 
         baseinfo() { reset(); }
 
@@ -40,14 +43,23 @@ struct capturestate
             capturetime = -1;
             ammotype = 0;
             ammo = 0;
+            owners = 0;
         }
 
         bool enter(const char *team)
         {
-            if(!enemy[0])
+            if(!strcmp(owner, team))
             {
-                if(!strcmp(owner, team)) return false;
-                s_strcpy(enemy, team);
+                owners++;
+                return false;
+            }
+            if(!enemies)
+            {
+                if(strcmp(enemy, team))
+                {
+                    converted = 0;
+                    s_strcpy(enemy, team);
+                }
                 enemies++;
                 return true;
             }
@@ -58,14 +70,18 @@ struct capturestate
 
         bool steal(const char *team)
         {
-            return !enemy[0] && strcmp(owner, team);
+            return !enemies && strcmp(owner, team);
         }
             
         bool leave(const char *team)
         {
+            if(!strcmp(owner, team))
+            {
+                owners--;
+                return false;
+            }
             if(strcmp(enemy, team)) return false;
             enemies--;
-            if(!enemies) noenemy();
             return !enemies;
         }
 
@@ -73,9 +89,14 @@ struct capturestate
         {
             if(strcmp(enemy, team)) return -1;
             converted += units;
-            if(converted<(owner[0] ? 2 : 1)*OCCUPYLIMIT) return -1;
+            if(units<0)
+            {
+                if(converted<=0) noenemy();
+                return -1;
+            }
+            else if(converted<(owner[0] ? 2 : 1)*OCCUPYLIMIT) return -1;
             if(owner[0]) { owner[0] = '\0'; converted = 0; s_strcpy(enemy, team); return 0; }
-            else { s_strcpy(owner, team); ammo = 0; capturetime = 0; noenemy(); return 1; }
+            else { s_strcpy(owner, team); ammo = 0; capturetime = 0; owners = enemies; noenemy(); return 1; }
         }
 
         bool addammo(int i)
@@ -179,18 +200,21 @@ struct capturestate
 struct captureclient : capturestate
 {
     fpsclient &cl;
+    float radarscale;
 
-    captureclient(fpsclient &cl) : cl(cl)
+    captureclient(fpsclient &cl) : cl(cl), radarscale(0)
     {
-        CCOMMAND(captureclient, repammo, "", self->replenishammo()); 
+        CCOMMAND(repammo, "", (captureclient *self), self->replenishammo());
     }
     
     void replenishammo()
     {
+        int gamemode = cl.gamemode;
+        if(m_noitems) return;
         loopv(bases)
         {
             baseinfo &b = bases[i];
-            if(b.ammotype>0 && b.ammotype<=I_CARTRIDGES-I_SHELLS+1 && insidebase(b, cl.player1->o) && cl.et.hasmaxammo(cl.player1, b.ammotype-1+I_SHELLS)) return;
+            if(b.ammotype>0 && b.ammotype<=I_CARTRIDGES-I_SHELLS+1 && insidebase(b, cl.player1->o) && cl.player1->hasmaxammo(b.ammotype-1+I_SHELLS)) return;
         }
         cl.cc.addmsg(SV_REPAMMO, "r");
     }
@@ -204,19 +228,20 @@ struct captureclient : capturestate
 
     void renderbases()
     {
+        int gamemode = cl.gamemode;
         loopv(bases)
         {
             baseinfo &b = bases[i];
             const char *flagname = b.owner[0] ? (strcmp(b.owner, cl.player1->team) ? "flags/red" : "flags/blue") : "flags/neutral";
-            rendermodel(b.ent->color, b.ent->dir, flagname, ANIM_MAPMODEL|ANIM_LOOP, 0, 0, b.o.x, b.o.y, b.o.z, 0, 0, 10.0f, 0, NULL, MDL_SHADOW | MDL_CULL_VFC | MDL_CULL_OCCLUDED);
-            if(b.ammotype>0 && b.ammotype<=I_CARTRIDGES-I_SHELLS+1) loopi(b.ammo)
+            rendermodel(b.ent->color, b.ent->dir, flagname, ANIM_MAPMODEL|ANIM_LOOP, 0, 0, b.o, 0, 0, 0, 0, NULL, MDL_SHADOW | MDL_CULL_VFC | MDL_CULL_OCCLUDED);
+            if(b.ammotype>0 && b.ammotype<=I_CARTRIDGES-I_SHELLS+1) loopi(m_noitemsrail ? 0 : (m_noitems ? 1 : b.ammo))
             {
                 float angle = 2*M_PI*(cl.lastmillis/4000.0f + i/float(MAXAMMO));
                 vec p(b.o);
                 p.x += 10*cosf(angle);
                 p.y += 10*sinf(angle);
                 p.z += 4;
-                rendermodel(b.ent->color, b.ent->dir, cl.et.entmdlname(I_SHELLS+b.ammotype-1), ANIM_MAPMODEL|ANIM_LOOP, 0, 0, p.x, p.y, p.z, 0, 0, 10.0f, 0, NULL, MDL_SHADOW | MDL_CULL_VFC | MDL_CULL_OCCLUDED);
+                rendermodel(b.ent->color, b.ent->dir, cl.et.entmdlname(I_SHELLS+b.ammotype-1), ANIM_MAPMODEL|ANIM_LOOP, 0, 0, p, 0, 0, 0, 0, NULL, MDL_SHADOW | MDL_CULL_VFC | MDL_CULL_OCCLUDED);
             }
             int ttype = 11, mtype = -1;
             if(b.owner[0])
@@ -255,12 +280,13 @@ struct captureclient : capturestate
         glTexCoord2f(1.0f, 1.0f); glVertex2f(x+s, y+s);
         glTexCoord2f(0.0f, 1.0f); glVertex2f(x,   y+s);
     }
-    
+   
     void drawblips(int x, int y, int s, int type, bool skipenemy = false)
     {
         const char *textures[3] = {"data/blip_red.png", "data/blip_grey.png", "data/blip_blue.png"};
         settexture(textures[max(type+1, 0)]);
         glBegin(GL_QUADS);
+        float scale = radarscale<=0 || radarscale>cl.maxradarscale() ? cl.maxradarscale() : radarscale;
         loopv(bases)
         {
             baseinfo &b = bases[i];
@@ -276,24 +302,30 @@ struct captureclient : capturestate
             dir.sub(cl.player1->o);
             dir.z = 0.0f;
             float dist = dir.magnitude();
-            if(dist >= RADARRADIUS) dir.mul(RADARRADIUS/dist);
+            if(dist >= scale) dir.mul(scale/dist);
             dir.rotate_around_z(-cl.player1->yaw*RAD);
-            drawradar(x + s*0.5f*0.95f*(1.0f+dir.x/RADARRADIUS), y + s*0.5f*0.95f*(1.0f+dir.y/RADARRADIUS), 0.05f*s);
+            drawradar(x + s*0.5f*0.95f*(1.0f+dir.x/scale), y + s*0.5f*0.95f*(1.0f+dir.y/scale), 0.05f*s);
         }
         glEnd();
     }
-    
+   
+    int respawnwait()
+    {
+        int gamemode = cl.gamemode;
+        if(m_regencapture) return -1;
+        return max(0, (m_noitemsrail ? RESPAWNSECS/2 : RESPAWNSECS)-(cl.lastmillis-cl.player1->lastpain)/1000);
+    }
+
     void capturehud(int w, int h)
     {
         glEnable(GL_BLEND);
-        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
+        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
         int x = 1800*w/h*34/40, y = 1800*1/40, s = 1800*w/h*5/40;
-        glColor4f(1, 1, 1, 0.5f);
+        glColor3f(1, 1, 1);
         settexture("data/radar.png");
         glBegin(GL_QUADS);
         drawradar(float(x), float(y), float(s));
         glEnd();
-        glBlendFunc(GL_ONE, GL_ONE);
         bool showenemies = cl.lastmillis%1000 >= 500;
         drawblips(x, y, s, 1, showenemies);
         drawblips(x, y, s, 0, showenemies);
@@ -301,12 +333,15 @@ struct captureclient : capturestate
         if(showenemies) drawblips(x, y, s, -2);
         if(cl.player1->state == CS_DEAD)
         {
-            glPushMatrix();
-            glLoadIdentity();
-            glOrtho(0, w*900/h, 900, 0, -1, 1);
-            int wait = max(0, RESPAWNSECS-(cl.lastmillis-cl.player1->lastaction)/1000);
-            draw_textf("%d", (x+s/2)/2-16, (y+s/2)/2-32, wait);
-            glPopMatrix();
+            int wait = respawnwait();
+            if(wait>=0)
+            {
+                glPushMatrix();
+                glLoadIdentity();
+                glOrtho(0, w*900/h, 900, 0, -1, 1);
+                draw_textf("%d", (x+s/2)/2-(wait>=10 ? 28 : 16), (y+s/2)/2-32, wait);
+                glPopMatrix();
+            }
         }
         glDisable(GL_BLEND);
     }
@@ -321,8 +356,16 @@ struct captureclient : capturestate
             baseinfo &b = bases.add();
             b.o = e->o;
             b.ammotype = e->attr1;
+            s_sprintfd(alias)("base_%d", e->attr2);
+            const char *name = getalias(alias);
+            if(name[0]) s_strcpy(b.name, name); else s_sprintf(b.name)("base %d", bases.length());
             b.ent = e;
         }
+        vec center(0, 0, 0);
+        loopv(bases) center.add(bases[i].o);
+        center.div(bases.length());
+        radarscale = 0;
+        loopv(bases) radarscale = max(radarscale, 2*center.dist(bases[i].o));
     }
             
     void sendbases(ucharbuf &p)
@@ -345,9 +388,17 @@ struct captureclient : capturestate
         baseinfo &b = bases[i];
         if(owner[0])
         {
-            if(strcmp(b.owner, owner)) { conoutf("\f2%s captured base %d", owner, i); playsound(S_V_BASECAP); }
+            if(strcmp(b.owner, owner)) 
+            { 
+                conoutf("\f2%s captured %s", owner, b.name); 
+                if(!strcmp(owner, cl.player1->team)) playsound(S_V_BASECAP); 
+            }
+        }
+        else if(b.owner[0]) 
+        { 
+            conoutf("\f2%s lost %s", b.owner, b.name); 
+            if(!strcmp(b.owner, cl.player1->team)) playsound(S_V_BASELOST); 
         }
-        else if(b.owner[0]) { conoutf("\f2%s lost base %d", b.owner, i); playsound(S_V_BASELOST); }
         s_strcpy(b.owner, owner);
         s_strcpy(b.enemy, enemy);
         b.converted = converted;
@@ -358,11 +409,12 @@ struct captureclient : capturestate
     void setscore(const char *team, int total)
     {
         findscore(team).total = total;
+        if(total>=10000) conoutf("team %s captured all bases", team);
     }
 
-    int closesttoenemy(const char *team, bool noattacked = false)
+    int closesttoenemy(const char *team, bool noattacked = false, bool farthest = false)
     {
-        float bestdist = 1e10f;
+        float bestdist = farthest ? -1e10f : 1e10f;
         int best = -1;
         int attackers = INT_MAX, attacked = -1;
         loopv(bases)
@@ -371,7 +423,7 @@ struct captureclient : capturestate
             if(!b.owner[0] || strcmp(b.owner, team)) continue;
             if(noattacked && b.enemy[0]) continue;
             float dist = disttoenemy(b);
-            if(dist < bestdist)
+            if(farthest ? dist > bestdist : dist < bestdist)
             {
                 best = i;
                 bestdist = dist;
@@ -388,13 +440,13 @@ struct captureclient : capturestate
 
     int pickspawn(const char *team)
     {
-        int closest = closesttoenemy(team, true);
-        if(closest < 0) closest = closesttoenemy(team, false);
+        int gamemode = cl.gamemode, closest = closesttoenemy(team, true, m_regencapture);
+        if(!m_regencapture && closest < 0) closest = closesttoenemy(team, false);
         if(closest < 0) return -1;
         baseinfo &b = bases[closest];
 
-        float bestdist = 1e10f;
-        int best = -1;
+        float bestdist = 1e10f, altdist = 1e10f;
+        int best = -1, alt = -1;
         loopv(cl.et.ents)
         {
             extentity *e = cl.et.ents[i];
@@ -402,27 +454,35 @@ struct captureclient : capturestate
             float dist = e->o.dist(b.o);
             if(dist < bestdist)
             {
+                alt = best;
+                altdist = bestdist;
                 best = i;
                 bestdist = dist;
             }
+            else if(dist < altdist)
+            {
+                alt = i;
+                altdist = dist;
+            }
         }
-        return best;
+        return rnd(2) ? best : alt;
     }
 };
 
 #else
 
-struct captureserv : capturestate
+struct captureservmode : capturestate, servmode
 {
-    fpsserver &sv;
     int scoresec;
-    
-    captureserv(fpsserver &sv) : sv(sv), scoresec(0) {}
+    bool notgotbases;
+ 
+    captureservmode(fpsserver &sv) : servmode(sv), scoresec(0), notgotbases(false) {}
 
-    void reset()
+    void reset(bool empty)
     {
         capturestate::reset();
         scoresec = 0;
+        notgotbases = !empty;
     }
 
     void stealbase(int n, const char *team)
@@ -431,22 +491,24 @@ struct captureserv : capturestate
         loopv(sv.clients)
         {
             fpsserver::clientinfo *ci = sv.clients[i];
-            if(!ci->spectator && ci->state==CS_ALIVE && ci->team[0] && !strcmp(ci->team, team) && insidebase(b, ci->o))
+            if(!ci->spectator && ci->state.state==CS_ALIVE && ci->team[0] && !strcmp(ci->team, team) && insidebase(b, ci->state.o))
                 b.enter(ci->team);
         }
         sendbaseinfo(n);
     }
 
-    void replenishammo(int client, const char *team, const vec &o)
+    void replenishammo(clientinfo *ci)
     {
-        if(!team[0]) return;
+        int gamemode = sv.gamemode;
+        if(m_noitems || notgotbases || ci->state.state!=CS_ALIVE || !ci->team[0]) return;
         loopv(bases)
         {
             baseinfo &b = bases[i];
-            if(insidebase(b, o) && b.takeammo(team))
+            if(b.ammotype>0 && b.ammotype<=I_CARTRIDGES-I_SHELLS+1 && insidebase(b, ci->state.o) && !ci->state.hasmaxammo(b.ammotype-1+I_SHELLS) && b.takeammo(ci->team))
             {
                 sendbaseinfo(i);
-                sendf(client, 1, "rii", SV_REPAMMO, b.ammotype);
+                sendf(ci->clientnum, 1, "rii", SV_REPAMMO, b.ammotype);
+                ci->state.addammo(b.ammotype);
                 break;
             }
         }
@@ -476,12 +538,6 @@ struct captureserv : capturestate
         movebases(team, vec(-1e10f, -1e10f, -1e10f), o);
     }
     
-    void changeteam(const char *oldteam, const char *newteam, const vec &o)
-    {
-        leavebases(oldteam, o);
-        enterbases(newteam, o);
-    }
-
     void addscore(const char *team, int n)
     {
         if(!n) return;
@@ -490,27 +546,72 @@ struct captureserv : capturestate
         sendf(-1, 1, "risi", SV_TEAMSCORE, team, cs.total);
     }
 
-    void updatescores(int secs)
+    void regenowners(baseinfo &b, int ticks)
+    {
+        loopv(sv.clients)
+        {
+            fpsserver::clientinfo *ci = sv.clients[i];
+            if(!ci->spectator && ci->state.state==CS_ALIVE && ci->team[0] && !strcmp(ci->team, b.owner) && insidebase(b, ci->state.o))
+            {
+                bool notify = false;
+                if(ci->state.health < ci->state.maxhealth) 
+                {
+                    ci->state.health = min(ci->state.health + ticks*REGENHEALTH, ci->state.maxhealth);
+                    notify = true;
+                }
+                if(ci->state.armour < itemstats[I_GREENARMOUR-I_SHELLS].max) 
+                {
+                    ci->state.armour = min(ci->state.armour + ticks*REGENARMOUR, itemstats[I_GREENARMOUR-I_SHELLS].max);
+                    notify = true;
+                }
+                if(b.ammotype>0)
+                {
+                    int ammotype = b.ammotype-1+I_SHELLS;
+                    if(ammotype<=I_CARTRIDGES && !ci->state.hasmaxammo(ammotype)) 
+                    {
+                        ci->state.addammo(b.ammotype, ticks*REGENAMMO, 100);
+                        notify = true;
+                    }
+                }
+                if(notify)
+                    sendf(ci->clientnum, 1, "ri5", SV_BASEREGEN, ci->state.health, ci->state.armour, b.ammotype, b.ammotype>0 ? ci->state.ammo[b.ammotype] : 0);
+            }
+        }
+    }
+
+    void update()
     {
         if(sv.minremain<0) return;
         endcheck();
-        int t = secs-sv.lastsec;
+        int t = sv.gamemillis/1000 - (sv.gamemillis-sv.curtime)/1000;
         if(t<1) return;
+        int gamemode = sv.gamemode;
         loopv(bases)
         {
             baseinfo &b = bases[i];
             if(b.enemy[0])
             {
-                if(b.occupy(b.enemy, OCCUPYPOINTS*b.enemies*t)==1) addscore(b.owner, CAPTURESCORE);
+                if((!b.owners || !b.enemies) && b.occupy(b.enemy, (m_noitemsrail ? OCCUPYPOINTS*2 : OCCUPYPOINTS)*(b.enemies ? b.enemies : -(1+b.owners))*t)==1) addscore(b.owner, CAPTURESCORE);
                 sendbaseinfo(i);
             }
             else if(b.owner[0])
             {
                 b.capturetime += t;
-                int score = b.capturetime/SCORESECS - (b.capturetime-t)/SCORESECS,
-                    ammo = b.capturetime/AMMOSECS - (b.capturetime-t)/AMMOSECS;
+                int score = b.capturetime/SCORESECS - (b.capturetime-t)/SCORESECS;
                 if(score) addscore(b.owner, score);
-                if(b.addammo(ammo)) sendbaseinfo(i);
+                if(m_noitems)
+                {
+                    if(!m_noitemsrail)
+                    {
+                        int regen = b.capturetime/REGENSECS - (b.capturetime-t)/REGENSECS;
+                        if(regen) regenowners(b, regen);
+                    }
+                }
+                else 
+                {
+                    int ammo = b.capturetime/AMMOSECS - (b.capturetime-t)/AMMOSECS;
+                    if(ammo && b.addammo(ammo)) sendbaseinfo(i);
+                }
             }
         }
     }
@@ -525,13 +626,13 @@ struct captureserv : capturestate
     {
         ENetPacket *packet = enet_packet_create(NULL, MAXTRANS, ENET_PACKET_FLAG_RELIABLE);
         ucharbuf p(packet->data, packet->dataLength);
-        initclient(p, false);
+        initclient(NULL, p, false);
         enet_packet_resize(packet, p.length());
         sendpacket(-1, 1, packet);
         if(!packet->referenceCount) enet_packet_destroy(packet);
     }
 
-    void initclient(ucharbuf &p, bool connecting)
+    void initclient(clientinfo *ci, ucharbuf &p, bool connecting)
     {
         if(connecting) loopv(scores)
         {
@@ -577,10 +678,66 @@ struct captureserv : capturestate
         }
 
         if(!lastteam) return;
-        s_sprintfd(msg)("team %s captured all bases", lastteam); 
-        sv.sendservmsg(msg);
+        findscore(lastteam).total = 10000;
+        sendf(-1, 1, "risi", SV_TEAMSCORE, lastteam, 10000);
         sv.startintermission(); 
     }
+
+    void entergame(clientinfo *ci) 
+    {
+        if(notgotbases || ci->state.state!=CS_ALIVE) return;
+        enterbases(ci->team, ci->state.o);
+    }        
+
+    void spawned(clientinfo *ci)
+    {
+        if(notgotbases) return;
+        enterbases(ci->team, ci->state.o);
+    }
+
+    void leavegame(clientinfo *ci, bool disconnecting = false)
+    {
+        if(notgotbases || ci->state.state!=CS_ALIVE) return;
+        leavebases(ci->team, ci->state.o);
+    }
+
+    void died(clientinfo *ci, clientinfo *actor)
+    {
+        if(notgotbases) return;
+        leavebases(ci->team, ci->state.o);
+    }
+
+    void moved(clientinfo *ci, const vec &oldpos, const vec &newpos)
+    {
+        if(notgotbases) return;
+        movebases(ci->team, oldpos, newpos);
+    }
+
+    void changeteam(clientinfo *ci, const char *oldteam, const char *newteam)
+    {
+        if(notgotbases) return;
+        leavebases(oldteam, ci->state.o);
+        enterbases(newteam, ci->state.o);
+    }
+
+    void parsebases(ucharbuf &p)
+    {
+        int ammotype;
+        while((ammotype = getint(p))>=0)
+        {
+            vec o;
+            o.x = getint(p)/DMF;
+            o.y = getint(p)/DMF;
+            o.z = getint(p)/DMF;
+            if(notgotbases) addbase(ammotype>=GUN_SG && ammotype<=GUN_PISTOL ? ammotype : 0, o);
+        }
+        if(notgotbases)
+        {
+            notgotbases = false;
+            sendbases();
+            loopv(sv.clients) if(sv.clients[i]->state.state==CS_ALIVE) entergame(sv.clients[i]);
+        }
+    }
 };
 
 #endif
