@@ -7,26 +7,37 @@ struct clientcom : iclientcom
     bool senditemstoserver;     // after a map change, since server doesn't have map data
     int lastping;
 
-    bool connected, remote;
+    bool connected, remote, demoplayback;
 
-    int currentmaster;
     bool spectator;
 
     fpsent *player1;
 
-    clientcom(fpsclient &_cl) : cl(_cl), c2sinit(false), senditemstoserver(false), lastping(0), connected(false), remote(false), currentmaster(-1), spectator(false), player1(_cl.player1)
+    clientcom(fpsclient &_cl) : cl(_cl), c2sinit(false), senditemstoserver(false), lastping(0), connected(false), remote(false), demoplayback(false), spectator(false), player1(_cl.player1)
     {
-        CCOMMAND(clientcom, say, "C", self->toserver(args[0]));
-        CCOMMAND(clientcom, name, "s", self->switchname(args[0]));
-        CCOMMAND(clientcom, team, "s", self->switchteam(args[0]));
-        CCOMMAND(clientcom, map, "s", self->changemap(args[0]));
-        CCOMMAND(clientcom, kick, "s", self->kick(args[0]));
-        CCOMMAND(clientcom, spectator, "ss", self->togglespectator(args[0], args[1]));
-        CCOMMAND(clientcom, mastermode, "s", if(self->remote) self->addmsg(SV_MASTERMODE, "ri", atoi(args[0])));
-        CCOMMAND(clientcom, setmaster, "s", self->setmaster(args[0]));
-        CCOMMAND(clientcom, setteam, "s", self->setteam(args[0], args[1]));
-        CCOMMAND(clientcom, getmap, "", self->getmap());
-        CCOMMAND(clientcom, sendmap, "", self->sendmap());
+        CCOMMAND(say, "C", (clientcom *self, char *s), self->toserver(s));
+        CCOMMAND(name, "s", (clientcom *self, char *s), self->switchname(s));
+        CCOMMAND(team, "s", (clientcom *self, char *s), self->switchteam(s));
+        CCOMMAND(map, "s", (clientcom *self, char *s), self->changemap(s));
+        CCOMMAND(clearbans, "", (clientcom *self, char *s), self->clearbans());
+        CCOMMAND(kick, "s", (clientcom *self, char *s), self->kick(s));
+        CCOMMAND(goto, "s", (clientcom *self, char *s), self->gotoplayer(s));
+        CCOMMAND(spectator, "is", (clientcom *self, int *val, char *who), self->togglespectator(*val, who));
+        CCOMMAND(mastermode, "i", (clientcom *self, int *val), if(self->remote) self->addmsg(SV_MASTERMODE, "ri", *val));
+        CCOMMAND(setmaster, "s", (clientcom *self, char *s), self->setmaster(s));
+        CCOMMAND(approve, "s", (clientcom *self, char *s), self->approvemaster(s));
+        CCOMMAND(setteam, "ss", (clientcom *self, char *who, char *team), self->setteam(who, team));
+        CCOMMAND(getmap, "", (clientcom *self), self->getmap());
+        CCOMMAND(sendmap, "", (clientcom *self), self->sendmap());
+        CCOMMAND(listdemos, "", (clientcom *self), self->listdemos());
+        CCOMMAND(getdemo, "i", (clientcom *self, int *val), self->getdemo(*val));
+        CCOMMAND(recorddemo, "i", (clientcom *self, int *val), self->recorddemo(*val));
+        CCOMMAND(stopdemo, "", (clientcom *self), self->stopdemo());
+        CCOMMAND(cleardemos, "i", (clientcom *self, int *val), self->cleardemos(*val));
+
+        extern void result(const char *s);
+        CCOMMAND(getname, "", (clientcom *self), result(self->player1->name));
+        CCOMMAND(getteam, "", (clientcom *self), result(self->player1->team));
     }
 
     void switchname(const char *name)
@@ -36,7 +47,7 @@ struct clientcom : iclientcom
             c2sinit = false; 
             filtertext(player1->name, name, false, MAXNAMELEN);
         }
-        else conoutf("your name is: %s", player1->name);
+        else conoutf("your name is: %s", cl.colorname(player1));
     }
 
     void switchteam(const char *team)
@@ -51,26 +62,7 @@ struct clientcom : iclientcom
 
     int numchannels() { return 3; }
 
-    static void filtertext(char *dst, const char *src, bool whitespace = true, int len = sizeof(string)-1)
-    {
-        for(int c = *src; c; c = *++src)
-        {
-            switch(c)
-            {
-            case '\f':
-                if(src[1]>='0' && src[1]<='3') ++src;
-                continue;
-            }
-            if(isprint(c) || (whitespace && isspace(c)))
-            {
-                *dst++ = c;
-                if(!--len) break;
-            }
-        }
-        *dst = '\0';
-    }
-
-    void mapstart() { if(!spectator || currentmaster==player1->clientnum) senditemstoserver = true; }
+    void mapstart() { if(!spectator || player1->privilege) senditemstoserver = true; }
 
     void initclientnet()
     {
@@ -90,12 +82,14 @@ struct clientcom : iclientcom
 
     void gamedisconnect()
     {
+        if(remote) cl.stopfollowing();
         connected = false;
         player1->clientnum = -1;
         c2sinit = false;
         player1->lifesequence = 0;
-        currentmaster = -1;
+        player1->privilege = PRIV_NONE;
         spectator = false;
+        removetrackedparticles();
         loopv(cl.players) DELETEP(cl.players[i]);
         cleardynentcache();
     }
@@ -108,20 +102,38 @@ struct clientcom : iclientcom
         return allow;
     }
 
+    void edittoggled(bool on)
+    {
+        addmsg(SV_EDITMODE, "ri", on ? 1 : 0);
+    }
+
     int parseplayer(const char *arg)
     {
         char *end;
         int n = strtol(arg, &end, 10);
         if(!cl.players.inrange(n)) return -1;
         if(*arg && !*end) return n;
+        // try case sensitive first
         loopi(cl.numdynents())
         {
             fpsent *o = (fpsent *)cl.iterdynents(i);
             if(o && !strcmp(arg, o->name)) return o->clientnum;
         }
+        // nothing found, try case insensitive
+        loopi(cl.numdynents())
+        {
+            fpsent *o = (fpsent *)cl.iterdynents(i);
+            if(o && !strcasecmp(arg, o->name)) return o->clientnum;
+        }
         return -1;
     }
 
+    void clearbans()
+    {
+        if(!remote) return;
+        addmsg(SV_CLEARBANS, "r");
+    }
+
     void kick(const char *arg)
     {
         if(!remote) return;
@@ -146,11 +158,17 @@ struct clientcom : iclientcom
         addmsg(SV_SETMASTER, "ris", val, passwd);
     }
 
-    void togglespectator(const char *arg1, const char *arg2)
+    void approvemaster(const char *who)
     {
         if(!remote) return;
-        int i = arg2[0] ? parseplayer(arg2) : player1->clientnum,
-            val = atoi(arg1);
+        int i = parseplayer(who);
+        if(i>=0) addmsg(SV_APPROVEMASTER, "ri", i);
+    }
+
+    void togglespectator(int val, const char *who)
+    {
+        if(!remote) return;
+        int i = who[0] ? parseplayer(who) : player1->clientnum;
         if(i>=0) addmsg(SV_SPECTATOR, "rii", i, val);
     }
 
@@ -159,7 +177,7 @@ struct clientcom : iclientcom
 
     void addmsg(int type, const char *fmt = NULL, ...)
     {
-        if(spectator && (currentmaster!=player1->clientnum || type<SV_MASTERMODE))
+        if(remote && spectator && (!player1->privilege || type<SV_MASTERMODE))
         {
             static int spectypes[] = { SV_MAPVOTE, SV_GETMAP, SV_TEXT, SV_SETMASTER };
             bool allowed = false;
@@ -182,6 +200,15 @@ struct clientcom : iclientcom
             while(*fmt) switch(*fmt++)
             {
                 case 'r': reliable = true; break;
+                case 'v':
+                {
+                    int n = va_arg(args, int);
+                    int *v = va_arg(args, int *);
+                    loopi(n) putint(p, v[i]);
+                    numi += n;
+                    break;
+                }
+
                 case 'i': 
                 {
                     int n = isdigit(*fmt) ? *fmt++-'0' : 1;
@@ -193,19 +220,19 @@ struct clientcom : iclientcom
             }
             va_end(args);
         } 
-        int num = nums?0:numi;
-        if(num!=fpsserver::msgsizelookup(type)) { s_sprintfd(s)("inconsistant msg size for %d (%d != %d)", type, num, fpsserver::msgsizelookup(type)); fatal(s); }
+        int num = nums?0:numi, msgsize = msgsizelookup(type);
+        if(msgsize && num!=msgsize) { s_sprintfd(s)("inconsistent msg size for %d (%d != %d)", type, num, msgsize); fatal(s); }
         int len = p.length();
         messages.add(len&0xFF);
         messages.add((len>>8)|(reliable ? 0x80 : 0));
         loopi(len) messages.add(buf[i]);
     }
 
-    void toserver(char *text) { conoutf("%s:\f0 %s", player1->name, text); addmsg(SV_TEXT, "rs", text); }
+    void toserver(char *text) { conoutf("%s:\f0 %s", cl.colorname(player1), text); addmsg(SV_TEXT, "rs", text); }
 
     int sendpacketclient(ucharbuf &p, bool &reliable, dynent *d)
     {
-        if(!spectator || !c2sinit || messages.length())
+        if(d->state==CS_ALIVE || d->state==CS_EDITING)
         {
             // send position updates separately so as to not stall out aiming
             ENetPacket *packet = enet_packet_create(NULL, 100, 0);
@@ -221,15 +248,16 @@ struct clientcom : iclientcom
             putint(q, (int)(d->vel.x*DVELF));          // quantize to itself, almost always 1 byte
             putint(q, (int)(d->vel.y*DVELF));
             putint(q, (int)(d->vel.z*DVELF));
-            putint(q, d->physstate | (d->gravity.x || d->gravity.y ? 0x20 : 0) | (d->gravity.z ? 0x10 : 0));
+            putuint(q, d->physstate | (d->gravity.x || d->gravity.y ? 0x20 : 0) | (d->gravity.z ? 0x10 : 0) | ((((fpsent *)d)->lifesequence&1)<<6));
             if(d->gravity.x || d->gravity.y)
             {
                 putint(q, (int)(d->gravity.x*DVELF));      // quantize to itself, almost always 1 byte
                 putint(q, (int)(d->gravity.y*DVELF));
             }
             if(d->gravity.z) putint(q, (int)(d->gravity.z*DVELF));
-            // pack rest in 1 byte: strafe:2, move:2, state:3, reserved:1
-            putint(q, (d->strafe&3) | ((d->move&3)<<2) | ((editmode ? CS_EDITING : d->state)<<4) );
+            // pack rest in almost always 1 byte: strafe:2, move:2, garmour: 1, yarmour: 1, quad: 1
+            uint flags = (d->strafe&3) | ((d->move&3)<<2);
+            putuint(q, flags);
             enet_packet_resize(packet, q.length());
             sendpackettoserv(packet, 0);
         }
@@ -250,7 +278,6 @@ struct clientcom : iclientcom
             putint(p, SV_INITC2S);
             sendstring(player1->name, p);
             sendstring(player1->team, p);
-            putint(p, player1->lifesequence);
         }
         int i = 0;
         while(i < messages.length()) // send messages collected during the previous frames
@@ -291,7 +318,7 @@ struct clientcom : iclientcom
         int lagtime = cl.lastmillis-d->lastupdate;
         if(lagtime)
         {
-            d->plag = (d->plag*5+lagtime)/6;
+            if(d->state!=CS_SPAWNING && d->lastupdate) d->plag = (d->plag*5+lagtime)/6;
             d->lastupdate = cl.lastmillis;
         }
     }
@@ -316,7 +343,7 @@ struct clientcom : iclientcom
                 vel.x = getint(p)/DVELF;
                 vel.y = getint(p)/DVELF;
                 vel.z = getint(p)/DVELF;
-                physstate = getint(p);
+                physstate = getuint(p);
                 gravity = vec(0, 0, 0);
                 if(physstate&0x20)
                 {
@@ -324,24 +351,33 @@ struct clientcom : iclientcom
                     gravity.y = getint(p)/DVELF;
                 }
                 if(physstate&0x10) gravity.z = getint(p)/DVELF;
-                f = getint(p);
+                int seqcolor = (physstate>>6)&1;
+                f = getuint(p);
                 fpsent *d = cl.getclient(cn);
-                if(!d) continue;
-                d->o = o;
+                if(!d || seqcolor!=(d->lifesequence&1)) continue;
                 d->yaw = yaw;
                 d->pitch = pitch;
                 d->roll = roll;
-                d->vel = vel;
-                d->physstate = physstate & 0x0F;
-                d->gravity = gravity;       
                 d->strafe = (f&3)==3 ? -1 : f&3;
                 f >>= 2;
                 d->move = (f&3)==3 ? -1 : f&3;
-                int state = (f>>2)&7;
-                if(state==CS_DEAD && d->state!=CS_DEAD) d->lastaction = cl.lastmillis;
-                d->state = state;
-                updatephysstate(d);
-                updatepos(d);
+                f >>= 2;
+                if(f&1) { d->armourtype = A_GREEN; d->armour = 1; }
+                else if(f&2) { d->armourtype = A_YELLOW; d->armour = 1; }
+                else { d->armourtype = A_BLUE; d->armour = 0; }
+                d->quadmillis = f&4 ? 1 : 0;
+                f >>= 3;
+                d->maxhealth = 100 + f*itemstats[I_BOOST-I_SHELLS].add;
+                if(cl.allowmove(d))
+                {
+                    d->o = o;
+                    d->vel = vel;
+                    d->physstate = physstate & 0x0F;
+                    d->gravity = gravity;
+                    updatephysstate(d);
+                    updatepos(d);
+                }
+                if(d->state==CS_LAGGED || d->state==CS_SPAWNING) d->state = CS_ALIVE;
                 break;
             }
 
@@ -374,13 +410,13 @@ struct clientcom : iclientcom
         int gamemode = cl.gamemode;
         static char text[MAXTRANS];
         int type;
-        bool mapchanged = false, inited = false, joining = false;
+        bool mapchanged = false;
 
         while(p.remaining()) switch(type = getint(p))
         {
             case SV_INITS2C:                    // welcome messsage from the server
             {
-                int mycn = getint(p), prot = getint(p);
+                int mycn = getint(p), prot = getint(p), hasmap = getint(p);
                 if(prot!=PROTOCOL_VERSION)
                 {
                     conoutf("you are using a different game protocol (you: %d, server: %d)", PROTOCOL_VERSION, prot);
@@ -388,8 +424,7 @@ struct clientcom : iclientcom
                     return;
                 }
                 player1->clientnum = mycn;      // we are now fully connected
-                if(getint(p)) joining = true;
-                else if(cl.gamemode==1 || cl.getclientmap()[0]) changemap(cl.getclientmap()); // we are the first client on this server, set map
+                if(!hasmap && (cl.gamemode==1 || cl.getclientmap()[0])) changemap(cl.getclientmap()); // we are the first client on this server, set map
                 break;
             }
 
@@ -416,21 +451,42 @@ struct clientcom : iclientcom
                     s_sprintfd(ds)("@%s", &text);
                     particle_text(d->abovehead(), ds, 9);
                 }
-                conoutf("%s:\f0 %s", d->name, &text);
+                conoutf("%s:\f0 %s", cl.colorname(d), &text);
                 break;
             }
 
             case SV_MAPCHANGE:
                 getstring(text, p);
                 changemapserv(text, getint(p));
-                if(joining && m_arena)
-                {
-                    player1->state = CS_DEAD;
-                    cl.sb.showscores(true);
-                }
                 mapchanged = true;
                 break;
 
+            case SV_ARENAWIN:
+            {
+                int acn = getint(p);
+                fpsent *alive = acn<0 ? NULL : (acn==player1->clientnum ? player1 : cl.getclient(acn));
+                conoutf("arena round is over! next round in 5 seconds...");
+                if(!alive) conoutf("everyone died!");
+                else if(m_teammode) conoutf("team %s has won the round", alive->team);
+                else if(alive==player1) conoutf("you are the last man standing!");
+                else conoutf("%s is the last man standing", cl.colorname(alive));
+                break;
+            }
+
+            case SV_FORCEDEATH:
+            {
+                int cn = getint(p);
+                fpsent *d = cn==player1->clientnum ? player1 : cl.newclient(cn);
+                if(!d) break;
+                if(d==player1)
+                {
+                    if(editmode) toggleedit();
+                    cl.sb.showscores(true);
+                }
+                d->state = CS_DEAD;
+                break;
+            }
+
             case SV_ITEMLIST:
             {
                 int n;
@@ -454,142 +510,129 @@ struct clientcom : iclientcom
             case SV_INITC2S:            // another client either connected or changed name/team
             {
                 d = cl.newclient(cn);
+                if(!d)
+                {
+                    getstring(text, p);
+                    getstring(text, p);
+                    break;
+                }
                 getstring(text, p);
                 filtertext(text, text, false, MAXNAMELEN);
                 if(!text[0]) s_strcpy(text, "unnamed");
                 if(d->name[0])          // already connected
                 {
                     if(strcmp(d->name, text))
-                        conoutf("%s is now known as %s", d->name, &text);
+                    {
+                        string oldname, newname;
+                        s_strcpy(oldname, cl.colorname(d));
+                        s_strcpy(newname, cl.colorname(d, text));
+                        conoutf("%s is now known as %s", oldname, newname);
+                    }
                 }
                 else                    // new client
                 {
                     c2sinit = false;    // send new players my info again
-                    conoutf("connected: %s", &text);
+                    conoutf("connected: %s", cl.colorname(d, text));
                     loopv(cl.players)   // clear copies since new player doesn't have them
                         if(cl.players[i]) freeeditinfo(cl.players[i]->edit);
                     extern editinfo *localedit;
                     freeeditinfo(localedit);
                 }
-                s_strcpy(d->name, text);
+                s_strncpy(d->name, text, MAXNAMELEN+1);
                 getstring(text, p);
                 filtertext(d->team, text, false, MAXTEAMLEN);
-                d->lifesequence = getint(p);
-                inited = true;
                 break;
             }
 
             case SV_CDIS:
+                cl.clientdisconnected(getint(p));
+                break;
+
+            case SV_SPAWN:
             {
-                int cn = getint(p);
-                fpsent *d = cl.getclient(cn);
+                int ls = getint(p), gunselect = getint(p);
                 if(!d) break;
-                if(d->name[0]) conoutf("player %s disconnected", d->name);
-                DELETEP(cl.players[cn]);
-                cleardynentcache();
-                if(currentmaster==cn) currentmaster = -1;
+                d->lifesequence = ls;
+                d->gunselect = gunselect;
+                d->state = CS_SPAWNING;
                 break;
             }
 
-            case SV_SHOT:
+            case SV_SPAWNSTATE:
             {
-                if(!d) return;
-                int gun = getint(p);
-                vec s, e;
-                s.x = getint(p)/DMF;
-                s.y = getint(p)/DMF;
-                s.z = getint(p)/DMF;
-                e.x = getint(p)/DMF;
-                e.y = getint(p)/DMF;
-                e.z = getint(p)/DMF;
-                if(gun==GUN_SG) cl.ws.createrays(s, e);
-                d->gunselect = max(gun, 0);
-                d->gunwait = 0;
-                d->lastaction = cl.lastmillis;
-                d->lastattackgun = d->gunselect;
-                cl.ws.shootv(gun, s, e, d, false);
+                if(editmode) toggleedit();
+                player1->respawn();
+                player1->lifesequence = getint(p); 
+                player1->health = getint(p);
+                player1->maxhealth = getint(p);
+                player1->armour = getint(p);
+                player1->armourtype = getint(p);
+                player1->gunselect = getint(p);
+                loopi(GUN_PISTOL-GUN_SG+1) player1->ammo[GUN_SG+i] = getint(p);
+                player1->state = CS_ALIVE;
+                findplayerspawn(player1, m_capture ? cl.cpc.pickspawn(player1->team) : -1);
+                cl.sb.showscores(false);
+                if(m_arena) conoutf("new round starting... fight!");
+                addmsg(SV_SPAWN, "rii", player1->lifesequence, player1->gunselect);
+                break;
+            }
+
+            case SV_SHOTFX:
+            {
+                int scn = getint(p), gun = getint(p);
+                vec from, to;
+                loopk(3) from[k] = getint(p)/DMF;
+                loopk(3) to[k] = getint(p)/DMF;
+                fpsent *s = cl.getclient(scn);
+                if(!s) break;
+                if(gun==GUN_SG) cl.ws.createrays(from, to);
+                s->gunselect = max(gun, 0);
+                s->gunwait = 0;
+                s->lastaction = cl.lastmillis;
+                s->lastattackgun = s->gunselect;
+                cl.ws.shootv(gun, from, to, s, false);
                 break;
             }
 
             case SV_DAMAGE:
             {
-                if(!d) return;
-                int target = getint(p);
-                int damage = getint(p);
-                int ls = getint(p);
+                int tcn = getint(p), 
+                    acn = getint(p),
+                    damage = getint(p), 
+                    armour = getint(p),
+                    health = getint(p);
+                fpsent *target = tcn==player1->clientnum ? player1 : cl.getclient(tcn),
+                       *actor = acn==player1->clientnum ? player1 : cl.getclient(acn);
+                if(!target || !actor) break;
+                target->armour = armour;
+                target->health = health;
+                cl.damaged(damage, target, actor, false);
+                break;
+            }
+
+            case SV_HITPUSH:
+            {
+                int gun = getint(p), damage = getint(p);
                 vec dir;
-                dir.x = getint(p)/DVELF;
-                dir.y = getint(p)/DVELF;
-                dir.z = getint(p)/DVELF;
-                if(damage<=0 || damage>1000) break;
-                if(target==player1->clientnum)
-                {
-                    if(ls==player1->lifesequence) { cl.selfdamage(damage, cn, d); player1->vel.add(dir); }
-                }
-                else
-                {
-                    fpsent *victim = cl.getclient(target);
-                    if(victim)
-                    {
-                        victim->lastpain = cl.lastmillis;
-                        playsound(S_PAIN1+rnd(5), &victim->o);
-                        cl.ws.damageeffect(damage, victim);
-                    }
-                }
+                loopk(3) dir[k] = getint(p)/DNF;
+                player1->hitpush(damage, dir, NULL, gun);
                 break;
             }
 
             case SV_DIED:
             {
-                if(!d) return;
-                int actor = getint(p);
-                getint(p); // damage
-                d->superdamage = getint(p);
-                if(actor==cn)
+                int vcn = getint(p), acn = getint(p), frags = getint(p);
+                fpsent *victim = vcn==player1->clientnum ? player1 : cl.getclient(vcn),
+                       *actor = acn==player1->clientnum ? player1 : cl.getclient(acn);
+                if(!actor) break;
+                actor->frags = frags;
+                if(actor!=player1)
                 {
-                    conoutf("\f2%s suicided", d->name);
+                    s_sprintfd(ds)("@%d", actor->frags);
+                    particle_text(actor->abovehead(), ds, 9);
                 }
-                else if(actor==player1->clientnum)
-                {
-                    int frags;
-                    if(isteam(player1->team, d->team))
-                    {
-                        frags = -1;
-                        conoutf("\f2you fragged a teammate (%s)", d->name);
-                    }
-                    else
-                    {
-                        frags = 1;
-                        conoutf("\f2you fragged %s", d->name);
-                    }
-                    addmsg(SV_FRAGS, "ri", player1->frags += frags);
-                }
-                else
-                {
-                    fpsent *a = cl.getclient(actor);
-                    if(a)
-                    {
-                        if(isteam(a->team, d->team))
-                        {
-                            conoutf("\f2%s fragged his teammate (%s)", a->name, d->name);
-                        }
-                        else
-                        {
-                            conoutf("\f2%s fragged %s", a->name, d->name);
-                        }
-                    }
-                }
-                cl.ws.superdamageeffect(d->vel, d); 
-                playsound(S_DIE1+rnd(2), &d->o);
-                if(!inited) d->lifesequence++;
-                break;
-            }
-
-            case SV_FRAGS:
-            {
-                if(!d) return;
-                s_sprintfd(ds)("@%d", d->frags = getint(p));
-                particle_text(d->abovehead(), ds, 9);
+                if(!victim) break;
+                cl.killed(victim, actor);
                 break;
             }
 
@@ -602,28 +645,32 @@ struct clientcom : iclientcom
                 break;
             }
 
-            case SV_RESUME:
+            case SV_TAUNT:
             {
-                int cn = getint(p);
-                fpsent *d = (cn == player1->clientnum ? player1 : cl.newclient(cn));
-                if(d)
-                {
-                    d->maxhealth = getint(p);
-                    d->frags = getint(p);
-                }
+                if(!d) return;
+                d->lasttaunt = cl.lastmillis;
                 break;
             }
 
-            case SV_ITEMPICKUP:
+            case SV_RESUME:
             {
-                if(!d) return;
-                int i = getint(p);
-                getint(p);
-                if(!cl.et.ents.inrange(i)) break;
-                cl.et.setspawn(i, false);
-                char *name = cl.et.itemname(i);
-                if(name) particle_text(d->abovehead(), name, 15);
-                if(cl.et.ents[i]->type==I_BOOST && !inited) d->maxhealth += 10;
+                for(;;)
+                {
+                    int cn = getint(p);
+                    if(cn<0) break;
+                    int state = getint(p), lifesequence = getint(p), gunselect = getint(p), maxhealth = getint(p), frags = getint(p);
+                    fpsent *d = (cn == player1->clientnum ? player1 : cl.newclient(cn));
+                    if(!d) continue;
+                    if(d!=player1) 
+                    {
+                        d->state = state;
+                        d->gunselect = gunselect;
+                    }
+                    d->lifesequence = lifesequence;
+                    if(d->state==CS_ALIVE && d->health==d->maxhealth) d->health = maxhealth;
+                    d->maxhealth = maxhealth;
+                    d->frags = frags;
+                }
                 break;
             }
 
@@ -633,14 +680,18 @@ struct clientcom : iclientcom
                 if(!cl.et.ents.inrange(i)) break;
                 cl.et.setspawn(i, true);
                 playsound(S_ITEMSPAWN, &cl.et.ents[i]->o);
-                char *name = cl.et.itemname(i);
+                const char *name = cl.et.itemname(i);
                 if(name) particle_text(cl.et.ents[i]->o, name, 9);
                 break;
             }
 
             case SV_ITEMACC:            // server acknowledges that I picked up this item
-                cl.et.realpickup(getint(p), player1);
+            {
+                int i = getint(p), cn = getint(p);
+                fpsent *d = cn==player1->clientnum ? player1 : cl.getclient(cn);
+                cl.et.pickupeffects(i, d);
                 break;
+            }
 
             case SV_EDITF:              // coop editing messages
             case SV_EDITT:
@@ -675,6 +726,13 @@ struct clientcom : iclientcom
                 }
                 break;
             }
+            case SV_REMIP:
+            {
+                if(!d) return;
+                conoutf("%s remipped", cl.colorname(d));
+                mpremip(false);
+                break;
+            }
             case SV_EDITENT:            // coop edit of ent
             {
                 if(!d) return;
@@ -705,9 +763,48 @@ struct clientcom : iclientcom
                 conoutf("%s", text);
                 break;
 
-            case SV_CURRENTMASTER:
-                currentmaster = getint(p);
+            case SV_SENDDEMOLIST:
+            {
+                int demos = getint(p);
+                if(!demos) conoutf("no demos available");
+                else loopi(demos)
+                {
+                    getstring(text, p);
+                    conoutf("%d. %s", i+1, text);
+                }
                 break;
+            }
+
+            case SV_DEMOPLAYBACK:
+            {
+                int on = getint(p);
+                if(on) player1->state = CS_SPECTATOR;
+                else stopdemo();
+                demoplayback = on!=0;
+                break;
+            }
+
+            case SV_CURRENTMASTER:
+            {
+                int mn = getint(p), priv = getint(p);
+                player1->privilege = PRIV_NONE;
+                loopv(cl.players) if(cl.players[i]) cl.players[i]->privilege = PRIV_NONE;
+                if(mn>=0)
+                {
+                    fpsent *m = mn==player1->clientnum ? player1 : cl.newclient(mn);
+                    m->privilege = priv;
+                }
+                break;
+            }
+
+            case SV_EDITMODE:
+            {
+                int val = getint(p);
+                if(!d) break;
+                if(val) d->state = CS_EDITING;
+                else d->state = CS_ALIVE;
+                break;
+            }
 
             case SV_SPECTATOR:
             {
@@ -718,14 +815,18 @@ struct clientcom : iclientcom
                     spectator = val!=0;
                     s = player1;
                 }
-                else s = cl.getclient(sn);
+                else s = cl.newclient(sn);
                 if(!s) return;
                 if(val)
                 {
                     if(s==player1 && editmode) toggleedit();
                     s->state = CS_SPECTATOR;
                 }
-                else if(s->state==CS_SPECTATOR) s->state = CS_ALIVE;
+                else if(s->state==CS_SPECTATOR) 
+                {
+                    s->state = CS_DEAD;
+                    if(s==player1) cl.sb.showscores(true);
+                }
                 break;
             }
 
@@ -753,6 +854,18 @@ struct clientcom : iclientcom
                 break;
             }
 
+            case SV_BASEREGEN:
+            {
+                int health = getint(p), armour = getint(p), ammotype = getint(p), ammo = getint(p);
+                if(m_capture)
+                {
+                    player1->health = health;
+                    player1->armour = armour;
+                    if(ammotype>=GUN_SG && ammotype<=GUN_PISTOL) player1->ammo[ammotype] = ammo;
+                }
+                break;
+            }
+
             case SV_BASES:
             {
                 int base = 0, ammotype;
@@ -793,11 +906,69 @@ struct clientcom : iclientcom
                 break;
             }
 
+            case SV_CLEARTARGETS:
+                if(m_assassin) cl.asc.targets.setsize(0);
+                break;
+
+            case SV_CLEARHUNTERS:
+                if(m_assassin) cl.asc.hunters.setsize(0);
+                break;
+
+            case SV_ADDTARGET:
+            {
+                int tcn = getint(p);
+                if(m_assassin) 
+                {
+                    fpsent *t = cl.newclient(tcn);
+                    if(cl.asc.targets.find(t)<0) cl.asc.targets.add(t);
+                }
+                break;
+            }
+
+            case SV_REMOVETARGET:
+            {
+                int tcn = getint(p);
+                if(m_assassin)
+                {
+                    fpsent *t = cl.getclient(tcn);
+                    if(t) cl.asc.targets.removeobj(t);
+                }
+                break;
+            }
+
+            case SV_ADDHUNTER:
+            {
+                int hcn = getint(p);
+                if(m_assassin)
+                {
+                    fpsent *h = cl.newclient(hcn);
+                    if(cl.asc.hunters.find(h)<0) cl.asc.hunters.add(h);
+                }
+                break;
+            }
+
+            case SV_REMOVEHUNTER:
+            {
+                int hcn = getint(p);
+                if(m_assassin)
+                {
+                    fpsent *h = cl.getclient(hcn);
+                    if(h) cl.asc.hunters.removeobj(h);
+                }
+                break;
+            }    
+
             case SV_NEWMAP:
             {
                 int size = getint(p);
                 if(size>=0) emptymap(size, true);
                 else enlargemap(true);
+                if(d && d!=player1)
+                {
+                    int newsize = 0;
+                    while(1<<newsize < getworldsize()) newsize++;
+                    conoutf(size>=0 ? "%s started a new map of size %d" : "%s enlarged the map to size %d", cl.colorname(d), newsize);
+                }
                 break;
             }
 
@@ -814,30 +985,56 @@ struct clientcom : iclientcom
         cl.nextmode = gamemode;
         cl.minremain = -1;
         if(editmode && !allowedittoggle()) toggleedit();
+        if(m_demo) return;
         if(gamemode==1 && !name[0]) emptymap(0, true);
         else load_world(name);
         if(m_capture) cl.cpc.setupbases();
+        else if(m_assassin) cl.asc.reset();
+        if(editmode) edittoggled(editmode);
     }
 
     void changemap(const char *name) // request map change, server may ignore
     {
-        if(!spectator || currentmaster==player1->clientnum) addmsg(SV_MAPVOTE, "rsi", name, cl.nextmode);
+        if(!spectator || player1->privilege) addmsg(SV_MAPVOTE, "rsi", name, cl.nextmode);
     }
         
     void receivefile(uchar *data, int len)
     {
-        if(cl.gamemode!=1) return;
-        string oldname;
-        s_strcpy(oldname, cl.getclientmap());
-        s_sprintfd(mname)("getmap_%d", cl.lastmillis);
-        s_sprintfd(fname)("packages/base/%s.ogz", mname);
-        FILE *map = fopen(fname, "wb");
-        if(!map) return;
-        conoutf("received map");
-        fwrite(data, 1, len, map);
-        fclose(map);
-        load_world(mname, oldname[0] ? oldname : NULL);
-        remove(fname);
+        ucharbuf p(data, len);
+        int type = getint(p);
+        data += p.length();
+        len -= p.length();
+        switch(type)
+        {
+            case SV_SENDDEMO:
+            {
+                s_sprintfd(fname)("%d.dmo", cl.lastmillis);
+                FILE *demo = openfile(fname, "wb");
+                if(!demo) return;
+                conoutf("received demo \"%s\"", fname);
+                fwrite(data, 1, len, demo);
+                fclose(demo);
+                break;
+            }
+
+            case SV_SENDMAP:
+            {
+                if(cl.gamemode!=1) return;
+                string oldname;
+                s_strcpy(oldname, cl.getclientmap());
+                s_sprintfd(mname)("getmap_%d", cl.lastmillis);
+                s_sprintfd(fname)("packages/base/%s.ogz", mname);
+                const char *file = findfile(fname, "wb");
+                FILE *map = fopen(file, "wb");
+                if(!map) return;
+                conoutf("received map");
+                fwrite(data, 1, len, map);
+                fclose(map);
+                load_world(mname, oldname[0] ? oldname : NULL);
+                remove(file);
+                break;
+            }
+        }
     }
 
     void getmap()
@@ -847,14 +1044,56 @@ struct clientcom : iclientcom
         addmsg(SV_GETMAP, "r");
     }
 
+    void stopdemo()
+    {
+        if(remote)
+        {
+            if(player1->privilege<PRIV_ADMIN) return;
+            addmsg(SV_STOPDEMO, "r");
+        }
+        else
+        {
+            loopv(cl.players) if(cl.players[i]) cl.clientdisconnected(i);
+
+            extern igameserver *sv;
+            ((fpsserver *)sv)->enddemoplayback();
+        }
+    }
+
+    void recorddemo(int val)
+    {
+        if(player1->privilege<PRIV_ADMIN) return;
+        addmsg(SV_RECORDDEMO, "ri", val);
+    }
+
+    void cleardemos(int val)
+    {
+        if(player1->privilege<PRIV_ADMIN) return;
+        addmsg(SV_CLEARDEMOS, "ri", val);
+    }
+
+    void getdemo(int i)
+    {
+        if(i<=0) conoutf("getting demo...");
+        else conoutf("getting demo %d...", i);
+        addmsg(SV_GETDEMO, "ri", i);
+    }
+
+    void listdemos()
+    {
+        conoutf("listing demos...");
+        addmsg(SV_LISTDEMOS, "r");
+    }
+
     void sendmap()
     {
-        if(cl.gamemode!=1 || (spectator && currentmaster!=player1->clientnum)) { conoutf("\"sendmap\" only works in coopedit mode"); return; }
+        if(cl.gamemode!=1 || (spectator && !player1->privilege)) { conoutf("\"sendmap\" only works in coopedit mode"); return; }
         conoutf("sending map...");
         s_sprintfd(mname)("sendmap_%d", cl.lastmillis);
         save_world(mname, true);
         s_sprintfd(fname)("packages/base/%s.ogz", mname);
-        FILE *map = fopen(fname, "rb");
+        const char *file = findfile(fname, "rb");
+        FILE *map = fopen(file, "rb");
         if(map)
         {
             fseek(map, 0, SEEK_END);
@@ -863,6 +1102,21 @@ struct clientcom : iclientcom
             fclose(map);
         }
         else conoutf("could not read map");
-        remove(fname);
+        remove(file);
+    }
+
+    void gotoplayer(const char *arg)
+    {
+        if(player1->state!=CS_SPECTATOR && player1->state!=CS_EDITING) return;
+        int i = parseplayer(arg);
+        if(i>=0 && i!=player1->clientnum) 
+        {
+            fpsent *d = cl.getclient(i);
+            if(!d) return;
+            player1->o = d->o;
+            vec dir;
+            vecfromyawpitch(player1->yaw, player1->pitch, 1, 0, dir);
+            player1->o.add(dir.mul(-32));
+        }
     }
 };
