@@ -16,19 +16,28 @@ struct fpsclient : igameclient
     // these define classes local to fpsclient
     #include "weapon.h"
     #include "monster.h"
+    #include "movable.h"
     #include "scoreboard.h"
     #include "fpsrender.h"
     #include "entities.h"
     #include "client.h"
     #include "capture.h"
+    #include "assassin.h"
 
     int nextmode, gamemode;         // nextmode becomes gamemode after next map load
     bool intermission;
     int lastmillis;
     string clientmap;
-    int arenarespawnwait, arenadetectwait;
     int maptime, minremain;
     int respawnent;
+    int swaymillis;
+    vec swaydir;
+    int respawned, suicided;
+    int lastslowmohealth, slowmorealtimestart;
+
+    int following;
+    IVARP(followdist, 10, 50, 1000);
+    IVARP(followorient, 0, 1, 1);
 
     fpsent *player1;                // our client
     vector<fpsent *> players;       // other clients
@@ -36,20 +45,30 @@ struct fpsclient : igameclient
 
     weaponstate ws;
     monsterset  ms;
+    movableset  mo;
     scoreboard  sb;
     fpsrender   fr;
     entities    et;
     clientcom   cc;
+
+    IVARP(maxradarscale, 0, 1024, 10000);
+
     captureclient cpc;
+    assassinclient asc;
 
     fpsclient()
         : nextmode(0), gamemode(0), intermission(false), lastmillis(0),
-          arenarespawnwait(0), arenadetectwait(0), maptime(0), minremain(0), respawnent(-1),
+          maptime(0), minremain(0), respawnent(-1), 
+          swaymillis(0), swaydir(0, 0, 0),
+          respawned(-1), suicided(-1), 
+          following(-1),
           player1(spawnstate(new fpsent())),
-          ws(*this), ms(*this), sb(*this), et(*this), cc(*this), cpc(*this)
+          ws(*this), ms(*this), mo(*this), sb(*this), fr(*this), et(*this), cc(*this), cpc(*this), asc(*this)
     {
-        CCOMMAND(fpsclient, mode, "s", { self->setmode(atoi(args[0])); });
-        CCOMMAND(fpsclient, kill, "",  { self->selfdamage(self->player1->health+self->player1->armour, -1, self->player1); });
+        CCOMMAND(mode, "i", (fpsclient *self, int *val), { self->setmode(*val); });
+        CCOMMAND(kill, "",  (fpsclient *self), { self->suicide(self->player1); });
+        CCOMMAND(taunt, "", (fpsclient *self), { self->taunt(); });
+        CCOMMAND(follow, "s", (fpsclient *self, char *s), { self->follow(s); });
     }
 
     iclientcom      *getcom()  { return &cc; }
@@ -61,15 +80,30 @@ struct fpsclient : igameclient
         nextmode = mode;
     }
 
+    void taunt()
+    {
+        if(player1->state!=CS_ALIVE || player1->physstate<PHYS_SLOPE) return;
+        if(lastmillis-player1->lasttaunt<1000) return;
+        player1->lasttaunt = lastmillis;
+        cc.addmsg(SV_TAUNT, "r");
+    }
+
+	void follow(char *arg)
+    {
+        if(player1->state!=CS_SPECTATOR && arg[0]) return;
+        following = arg[0] ? cc.parseplayer(arg) : -1;
+        conoutf("follow %s", following>=0 ? "on" : "off");
+	}
+
     char *getclientmap() { return clientmap; }
 
-    void rendergame() { fr.rendergame(*this, gamemode); }
+    void rendergame() { fr.rendergame(gamemode); }
 
     void resetgamestate()
     {
-        player1->health = player1->maxhealth;
         if(m_classicsp) 
         {
+            mo.clear(gamemode);
             ms.monsterclear(gamemode);                 // all monsters back at their spawns for editing
             resettriggers();
         }
@@ -79,49 +113,25 @@ struct fpsclient : igameclient
     fpsent *spawnstate(fpsent *d)              // reset player state not persistent accross spawns
     {
         d->respawn();
-        if(m_noitems || m_capture)
-        {
-            d->gunselect = GUN_RIFLE;
-            d->armour = 0;
-            if(m_noitemsrail)
-            {
-                d->health = 1;
-                d->ammo[GUN_RIFLE] = 100;
-            }
-            else
-            {
-                //d->health = 100;
-                d->armour = 100;
-                d->armourtype = A_GREEN;
-                if(m_tarena || m_capture)
-                {
-                    d->ammo[GUN_PISTOL] = 80;
-                    int spawngun1 = rnd(5)+1, spawngun2;
-                    et.baseammo(d->gunselect = spawngun1, m_capture ? 1 : 2);
-                    do spawngun2 = rnd(5)+1; while(spawngun1==spawngun2);
-                    et.baseammo(spawngun2, m_capture ? 1 : 2);
-                    if(!m_capture) d->ammo[GUN_GL] += 1;
-                }
-                else // efficiency
-                {
-                    loopi(5) et.baseammo(i+1);
-                    d->gunselect = GUN_CG;
-                }
-                d->ammo[GUN_CG] /= 2;
-            }
-        }
-        else
-        {
-            d->ammo[GUN_PISTOL] = m_sp ? 80 : 40;
-            d->ammo[GUN_GL] = 1;
-        }
+        d->spawnstate(gamemode);
         return d;
     }
 
     void respawnself()
     {
-        spawnplayer(player1);
-        sb.showscores(false);
+        if(m_mp(gamemode)) 
+        {
+            if(respawned!=player1->lifesequence)
+            {
+                cc.addmsg(SV_TRYSPAWN, "r");
+                respawned = player1->lifesequence;
+            }
+        }
+        else
+        {
+            spawnplayer(player1);
+            sb.showscores(false);
+        }
     }
 
     fpsent *pointatplayer()
@@ -135,51 +145,38 @@ struct fpsclient : igameclient
         return NULL;
     }
 
-    void arenacount(fpsent *d, int &alive, int &dead, char *&lastteam, bool &oneteam)
+    void followplayer(fpsent *target)
     {
-        if(d->state==CS_SPECTATOR) return;
-        if(d->state!=CS_DEAD)
+		if(followorient() && target->state!=CS_DEAD) interpolateorientation(target, player1->yaw, player1->pitch);
+
+        physent followcam;
+        followcam.o = target->o;
+        followcam.yaw = player1->yaw;
+        followcam.pitch = player1->pitch;
+        followcam.type = ENT_CAMERA;
+        followcam.move = -1;
+        followcam.eyeheight = 2;
+        loopi(10)
         {
-            alive++;
-            if(lastteam && strcmp(lastteam, d->team)) oneteam = false;
-            lastteam = d->team;
-        }
-        else
-        {
-            dead++;
+            if(!moveplayer(&followcam, 10, true, followdist())) break;
         }
+
+        player1->o = followcam.o;
     }
 
-    void arenarespawn()
+    void stopfollowing()
     {
-        if(arenarespawnwait)
-        {
-            if(arenarespawnwait<lastmillis)
-            {
-                arenarespawnwait = 0;
-                conoutf("\f2new round starting... fight!");
-                playsound(S_V_FIGHT);
-                if(!cc.spectator) respawnself();
-            }
-        }
-        else if(arenadetectwait==0 || arenadetectwait<lastmillis)
-        {
-            arenadetectwait = 0;
-            int alive = 0, dead = 0;
-            char *lastteam = NULL;
-            bool oneteam = true;
-            loopv(players) if(players[i]) arenacount(players[i], alive, dead, lastteam, oneteam);
-            arenacount(player1, alive, dead, lastteam, oneteam);
-            if(dead>0 && (alive<=1 || (m_teammode && oneteam)))
-            {
-                conoutf("\f2arena round is over! next round in 5 seconds...");
-                if(alive) conoutf("\f2team %s is last man standing", lastteam);
-                else conoutf("\f2everyone died!");
-                arenarespawnwait = lastmillis+5000;
-                arenadetectwait  = lastmillis+10000;
-                player1->roll = 0;
-            }
-        }
+        if(following<0) return;
+        following = -1;
+        conoutf("follow off");
+    }
+
+    void setupcamera()
+    {
+        if(player1->state!=CS_SPECTATOR || following<0) return;
+        fpsent *target = getclient(following);
+        if(!target || target->state!=CS_ALIVE) return;
+        followplayer(target);
     }
 
     void otherplayers()
@@ -192,36 +189,59 @@ struct fpsclient : igameclient
                 players[i]->state = CS_LAGGED;
                 continue;
             }
-            if(lagtime && players[i]->state==CS_ALIVE && !intermission) moveplayer(players[i], 2, false);   // use physics to extrapolate player position
+            if(lagtime && (players[i]->state==CS_ALIVE || (players[i]->state==CS_DEAD && lastmillis-players[i]->lastpain<2000)) && !intermission) moveplayer(players[i], 2, false);   // use physics to extrapolate player position
         }
     }
 
     void updateworld(vec &pos, int curtime, int lm)        // main game update loop
     {
+        if(!maptime)
+        {
+            maptime = lm + curtime;
+            extern int totalmillis;
+            slowmorealtimestart = totalmillis;
+            return;
+        }
         lastmillis = lm;
-        maptime += curtime;
+
+        if(m_slowmo)
+        {
+            setvar("gamespeed", intermission ? 100 : player1->health);
+            if(player1->health<player1->maxhealth && lastmillis-lastslowmohealth>player1->health*player1->health/2)
+            {
+                lastslowmohealth = lastmillis;
+                player1->health++;
+            }
+        }
+
         if(!curtime) return;
+
         physicsframe();
-        et.checkquad(curtime);
-        if(m_arena) arenarespawn();
+        et.checkquad(curtime, player1);
         ws.moveprojectiles(curtime);
-        ws.bounceupdate(curtime);
-        if(player1->clientnum>=0 && player1->state==CS_ALIVE) ws.shoot(player1, pos);     // only shoot when connected to server
+        if(player1->clientnum>=0 && player1->state==CS_ALIVE) ws.shoot(player1, pos); // only shoot when connected to server
+        ws.bounceupdate(curtime); // need to do this after the player shoots so grenades don't end up inside player's BB next frame
         gets2c();           // do this first, so we have most accurate information when our player moves
         otherplayers();
+        mo.update(curtime);
         ms.monsterthink(curtime, gamemode);
         if(player1->state==CS_DEAD)
         {
-            if(lastmillis-player1->lastaction<2000)
+            if(lastmillis-player1->lastpain<2000)
             {
                 player1->move = player1->strafe = 0;
                 moveplayer(player1, 10, false);
             }
+            if(m_assassin && asc.respawnwait()<=0) respawnself();
         }
         else if(!intermission)
         {
             moveplayer(player1, 20, true);
-            et.checkitems();
+            if(player1->physstate>=PHYS_SLOPE) swaymillis += curtime;
+            float k = pow(0.7f, curtime/10.0f);
+            swaydir.mul(k); 
+            swaydir.add(vec(player1->vel).mul((1-k)/(15*max(player1->vel.magnitude(), player1->maxspeed))));
+            et.checkitems(player1);
             if(m_classicsp) checktriggers();
         }
         if(player1->clientnum>=0) c2sinfo(player1);   // do this last, to reduce the effective frame lag
@@ -231,7 +251,7 @@ struct fpsclient : igameclient
     {
         findplayerspawn(d, m_capture ? cpc.pickspawn(d->team) : (respawnent>=0 ? respawnent : -1));
         spawnstate(d);
-        d->state = cc.spectator ? CS_SPECTATOR : (editmode ? CS_EDITING : CS_ALIVE);
+        d->state = cc.spectator ? CS_SPECTATOR : (d==player1 && editmode ? CS_EDITING : CS_ALIVE);
     }
 
     void respawn()
@@ -239,10 +259,14 @@ struct fpsclient : igameclient
         if(player1->state==CS_DEAD)
         {
             player1->attacking = false;
-            if(m_capture && lastmillis-player1->lastaction<cpc.RESPAWNSECS*1000)
+            if(m_capture)
             {
-                conoutf("\f2you must wait %d seconds before respawn!", cpc.RESPAWNSECS-(lastmillis-player1->lastaction)/1000);
-                return;
+                int wait = cpc.respawnwait();
+                if(wait>0)
+                {
+                    conoutf("\f2you must wait %d second%s before respawn!", wait, wait!=1 ? "s" : "");
+                    return;
+                }
             }
             if(m_arena) { conoutf("\f2waiting for new round to start..."); return; }
             if(m_dmsp) { nextmode = gamemode; cc.changemap(clientmap); return; }    // if we die in SP we try the same map again
@@ -250,8 +274,7 @@ struct fpsclient : igameclient
             {
                 respawnself();
                 conoutf("\f2You wasted another life! The monsters stole your armour and some ammo...");
-                loopi(NUMGUNS) if((player1->ammo[i] = lastplayerstate.ammo[i])>5) player1->ammo[i] = max(player1->ammo[i]/3, 5); 
-                player1->ammo[GUN_PISTOL] = 80;
+                loopi(NUMGUNS) if(i!=GUN_PISTOL && (player1->ammo[i] = lastplayerstate.ammo[i])>5) player1->ammo[i] = max(player1->ammo[i]/3, 5); 
                 return;
             }
             respawnself();
@@ -272,64 +295,85 @@ struct fpsclient : igameclient
         return player1->state!=CS_DEAD && !intermission; 
     }
 
-    // damage arriving from the network, monsters, yourself, all ends up here.
-
-    void selfdamage(int damage, int actor, fpsent *act)
+    bool allowmove(physent *d)
     {
-        if(player1->state!=CS_ALIVE || editmode || intermission) return;
-        damageblend(damage);
-        int ad = damage*(player1->armourtype+1)*25/100;     // let armour absorb when possible
-        if(ad>player1->armour) ad = player1->armour;
-        player1->armour -= ad;
-        damage -= ad;
-        float droll = damage/0.5f;
-        player1->roll += player1->roll>0 ? droll : (player1->roll<0 ? -droll : (rnd(2) ? droll : -droll));  // give player a kick depending on amount of damage
-        if((player1->health -= damage)<=0)
+        if(d->type!=ENT_PLAYER) return true;
+        return !((fpsent *)d)->lasttaunt || lastmillis-((fpsent *)d)->lasttaunt>=1000;
+    }
+
+    void damaged(int damage, fpsent *d, fpsent *actor, bool local = true)
+    {
+        if(d->state!=CS_ALIVE || intermission) return;
+
+        if(local) damage = d->dodamage(damage);
+        else if(actor==player1) return;
+
+        if(d==player1)
         {
-            if(actor==-2)
-            {
-                conoutf("\f2you got killed by %s!", &act->name);
+            damageblend(damage);
+            d->damageroll(damage);
+            if(m_slowmo && player1->health<1) player1->health = 1;
+        }
+        ws.damageeffect(damage, d);
+
+        if(d->health<=0) { if(local) killed(d, actor); }
+        else if(d==player1) playsound(S_PAIN6);
+        else playsound(S_PAIN1+rnd(5), &d->o);
+    }
+
+    void killed(fpsent *d, fpsent *actor)
+    {
+        if(d->state!=CS_ALIVE || intermission) return;
+
+        string dname, aname;
+        s_strcpy(dname, d==player1 ? "you" : colorname(d));
+        s_strcpy(aname, actor==player1 ? "you" : (actor->type!=ENT_INANIMATE ? colorname(actor) : ""));
+        if(actor->type==ENT_AI)
+            conoutf("\f2%s got killed by %s!", dname, aname);
+        else if(d==actor || actor->type==ENT_INANIMATE)
+            conoutf("\f2%s suicided%s", dname, d==player1 ? "!" : "");
+        else if(isteam(d->team, actor->team))
+        {
+            if(d==player1) conoutf("\f2you got fragged by a teammate (%s)", aname);
+            else conoutf("\f2%s fragged a teammate (%s)", aname, dname);
+        }
+        else if(m_assassin && (d==player1 || actor==player1))
+        {
+            if(d==player1) 
+            {   
+                conoutf("\f2you got fragged by %s (%s)", aname, asc.hunters.find(actor)>=0 ? "assassin" : (asc.targets.find(actor)>=0 ? "target" : "friend"));
+                if(asc.hunters.find(actor)>=0) asc.hunters.removeobj(actor);
             }
-            else if(actor==-1)
+            else 
             {
-                actor = player1->clientnum;
-                conoutf("\f2you suicided!");
-                cc.addmsg(SV_FRAGS, "ri", --player1->frags);
+                conoutf("\f2you fragged %s (%s)", dname, asc.targets.find(d)>=0 ? "target +1" : (asc.hunters.find(d)>=0 ? "assassin +0" : "friend -1")); 
+                if(asc.targets.find(d)>=0) asc.targets.removeobj(d);
             }
-            else
-            {
-                fpsent *a = getclient(actor);
-                if(a)
-                {
-                    if(isteam(a->team, player1->team))
-                    {
-                        conoutf("\f2you got fragged by a teammate (%s)", a->name);
-                    }
-                    else
-                    {
-                        conoutf("\f2you got fragged by %s", a->name);
-                    }
-                }
-            }
-            sb.showscores(true);
-            player1->superdamage = -player1->health;
-            cc.addmsg(SV_DIED, "ri3", actor, damage+ad, player1->superdamage);
-            lastplayerstate = *player1;
-            player1->lifesequence++;
-            player1->attacking = false;
-            player1->state = CS_DEAD;
-            player1->deaths++;
-            player1->pitch = 0;
-            player1->roll = 0;
-            playsound(S_DIE1+rnd(2));
-            vec vel = player1->vel;
-            spawnstate(player1);
-            player1->vel = vel;
-            player1->lastaction = lastmillis;
         }
         else
         {
-            playsound(S_PAIN6);
+            if(d==player1) conoutf("\f2you got fragged by %s", aname);
+            else conoutf("\f2%s fragged %s", aname, dname);
+        }
+
+        d->state = CS_DEAD;
+        d->lastpain = lastmillis;
+        d->superdamage = max(-d->health, 0);
+        if(d==player1)
+        {
+            sb.showscores(true);
+            lastplayerstate = *player1;
+            d->attacking = false;
+            d->deaths++;
+            d->pitch = 0;
+            d->roll = 0;
+            playsound(S_DIE1+rnd(2));
+        }
+        else
+        {
+            d->move = d->strafe = 0;
+            playsound(S_DIE1+rnd(2), &d->o);
+            ws.superdamageeffect(d->vel, d);
         }
     }
 
@@ -349,7 +393,8 @@ struct fpsclient : igameclient
             {
                 conoutf("\f2--- single player time score: ---");
                 int pen, score = 0;
-                pen = maptime/1000;       score += pen; if(pen) conoutf("\f2time taken: %d seconds", pen); 
+                extern int totalmillis;
+                pen = (totalmillis-slowmorealtimestart)/1000; score += pen; if(pen) conoutf("\f2time taken: %d seconds (%d simulated seconds)", pen, (lastmillis-maptime)/1000); 
                 pen = player1->deaths*60; score += pen; if(pen) conoutf("\f2time penalty for %d deaths (1 minute each): %d seconds", player1->deaths, pen);
                 pen = ms.remain*10;       score += pen; if(pen) conoutf("\f2time penalty for %d monsters remaining (10 seconds each): %d seconds", ms.remain, pen);
                 pen = (10-ms.skill())*20; score += pen; if(pen) conoutf("\f2time penalty for lower skill level (20 seconds each): %d seconds", pen);
@@ -392,17 +437,64 @@ struct fpsclient : igameclient
         return players.inrange(cn) ? players[cn] : NULL;
     }
 
+    void clientdisconnected(int cn)
+    {
+        if(!players.inrange(cn)) return;
+        if(following==cn) stopfollowing();
+        fpsent *d = players[cn];
+        if(!d) return; 
+        if(d->name[0]) conoutf("player %s disconnected", colorname(d));
+        ws.removebouncers(d);
+        ws.removeprojectiles(d);
+        removetrackedparticles(d);
+        if(m_assassin) asc.removeplayer(d);
+        DELETEP(players[cn]);
+        cleardynentcache();
+    }
+
     void initclient()
     {
         clientmap[0] = 0;
         cc.initclientnet();
     }
 
+    void preloadcharacters()
+    {
+        loadmodel(fr.ogro() ? "monster/ogro" : "ironsnout", -1, true);
+        loadmodel(fr.ogro() ? "monster/ogro/blue" : "ironsnout/blue", -1, true);
+        loadmodel(fr.ogro() ? "monster/ogro/red" : "ironsnout/red", -1, true);
+    }
+
+    void preloadweapons()
+    {
+        loopi(NUMGUNS)
+        {
+            const char *file = guns[i].file;
+            if(!file) continue;
+            s_sprintfd(mdl)("hudguns/%s", file);
+            loadmodel(mdl, -1, true);
+            s_sprintf(mdl)("hudguns/%s/blue", file);
+            loadmodel(mdl, -1, true);
+            s_sprintf(mdl)("vwep/%s", file);
+            loadmodel(mdl, -1, true);
+        }
+    }
+
+    void preload()
+    {
+        preloadweapons();
+        preloadcharacters();
+        et.preloadentities();
+        if(m_sp) ms.preloadmonsters();
+    }
+
     void startmap(const char *name)   // called just after a map load
     {
+        respawned = suicided = -1;
         respawnent = -1;
         if(multiplayer(false) && m_sp) { gamemode = 0; conoutf("coop sp not supported yet"); }
         cc.mapstart();
+        mo.clear(gamemode);
         ms.monsterclear(gamemode);
         ws.projreset();
 
@@ -421,7 +513,8 @@ struct fpsclient : igameclient
             players[i]->maxhealth = 100;
         }
 
-        spawnplayer(player1);
+        if(!m_mp(gamemode)) spawnplayer(player1);
+        else findplayerspawn(player1, -1);
         et.resetspawns();
         s_strcpy(clientmap, name);
         sb.showscores(false);
@@ -433,20 +526,30 @@ struct fpsclient : igameclient
             s_sprintfd(aname)("bestscore_%s", getclientmap());
             const char *best = getalias(aname);
             if(*best) conoutf("\f2try to beat your best score so far: %s", best);
+            lastslowmohealth = lastmillis;
         }
     }
 
     void physicstrigger(physent *d, bool local, int floorlevel, int waterlevel)
     {
+        if(d->type==ENT_INANIMATE) return;
         if     (waterlevel>0) playsound(S_SPLASH1, d==player1 ? NULL : &d->o);
         else if(waterlevel<0) playsound(S_SPLASH2, d==player1 ? NULL : &d->o);
-        if     (floorlevel>0) { if(local) playsoundc(S_JUMP); else if(d->type==ENT_AI) playsound(S_JUMP, &d->o); }
-        else if(floorlevel<0) { if(local) playsoundc(S_LAND); else if(d->type==ENT_AI) playsound(S_LAND, &d->o); }
+        if     (floorlevel>0) { if(local) playsoundc(S_JUMP, (fpsent *)d); else if(d->type==ENT_AI) playsound(S_JUMP, &d->o); }
+        else if(floorlevel<0) { if(local) playsoundc(S_LAND, (fpsent *)d); else if(d->type==ENT_AI) playsound(S_LAND, &d->o); }
     }
 
-    void playsoundc(int n) { cc.addmsg(SV_SOUND, "i", n); playsound(n); }
+    void playsoundc(int n, fpsent *d = NULL) 
+    { 
+        if(!d || d==player1)
+        {
+            cc.addmsg(SV_SOUND, "i", n); 
+            playsound(n); 
+        }
+        else playsound(n, &d->o);
+    }
 
-    int numdynents() { return 1+players.length()+ms.monsters.length(); }
+    int numdynents() { return 1+players.length()+ms.monsters.length()+mo.movables.length(); }
 
     dynent *iterdynents(int i)
     {
@@ -455,24 +558,78 @@ struct fpsclient : igameclient
         if(i<players.length()) return players[i];
         i -= players.length();
         if(i<ms.monsters.length()) return ms.monsters[i];
+        i -= ms.monsters.length(); 
+        if(i<mo.movables.length()) return mo.movables[i];
         return NULL;
     }
 
-    void worldhurts(physent *d, int damage)
+    bool duplicatename(fpsent *d, char *name = NULL)
     {
-        if(d==player1) selfdamage(damage, -1, player1);
-        else if(d->type==ENT_AI) ((monsterset::monster *)d)->monsterpain(damage, player1);
+        if(!name) name = d->name;
+        if(d!=player1 && !strcmp(name, player1->name)) return true;
+        loopv(players) if(players[i] && d!=players[i] && !strcmp(name, players[i]->name)) return true;
+        return false;
     }
 
-    IVAR(hudgun, 0, 1, 1);
+    char *colorname(fpsent *d, char *name = NULL, const char *prefix = "")
+    {
+        if(!name) name = d->name;
+        if(name[0] && !duplicatename(d, name)) return name;
+        static string cname;
+        s_sprintf(cname)("%s%s \fs\f5(%d)\fr", prefix, name, d->clientnum);
+        return cname;
+    }
 
-    void drawhudmodel(int anim, float speed, int base)
+    void suicide(physent *d)
+    {
+        if(d==player1)
+        {
+            if(d->state!=CS_ALIVE) return;
+            if(!m_mp(gamemode)) killed(player1, player1);
+            else if(suicided!=player1->lifesequence)
+            {
+                cc.addmsg(SV_SUICIDE, "r");
+                suicided = player1->lifesequence;
+            }
+        }
+        else if(d->type==ENT_AI) ((monsterset::monster *)d)->monsterpain(400, player1);
+        else if(d->type==ENT_INANIMATE) ((movableset::movable *)d)->suicide();
+    }
+
+    IVARP(hudgun, 0, 1, 1);
+    IVARP(hudgunsway, 0, 1, 1);
+    IVARP(teamhudguns, 0, 1, 1);
+   
+    void drawhudmodel(int anim, float speed = 0, int base = 0)
     {
-        static char *hudgunnames[] = { "hudguns/fist", "hudguns/shotg", "hudguns/chaing", "hudguns/rocket", "hudguns/rifle", "hudguns/gl", "hudguns/pistol" };
         if(player1->gunselect>GUN_PISTOL) return;
-        vec color, dir;
-        lightreaching(player1->o, color, dir);
-        rendermodel(color, dir, hudgunnames[player1->gunselect], anim, 0, 0, player1->o.x, player1->o.y, player1->o.z, player1->yaw+90, player1->pitch, speed, base, NULL, 0);
+
+        vec sway, color, dir;
+        vecfromyawpitch(player1->yaw, player1->pitch, 1, 0, sway);
+        float swayspeed = min(4.0f, player1->vel.magnitude());
+        sway.mul(swayspeed);
+        float swayxy = sinf(swaymillis/115.0f)/100.0f,
+              swayz = cosf(swaymillis/115.0f)/100.0f;
+        swap(float, sway.x, sway.y);
+        sway.x *= -swayxy;
+        sway.y *= swayxy;
+        sway.z = -fabs(swayspeed*swayz);
+        sway.add(swaydir).add(player1->o);
+        if(!hudgunsway()) sway = player1->o;
+        lightreaching(sway, color, dir);
+        dynlightreaching(sway, color, dir);
+
+#if 0
+        if(player1->state!=CS_DEAD && player1->quadmillis)
+        {
+            float t = 0.5f + 0.5f*sinf(2*M_PI*lastmillis/1000.0f);
+            color.y = color.y*(1-t) + t;
+        }
+#endif
+
+        s_sprintfd(gunname)("hudguns/%s", guns[player1->gunselect].file);
+        if((m_teamskins || fr.teamskins()) && teamhudguns()) s_strcat(gunname, "/blue");
+        rendermodel(color, dir, gunname, anim, 0, 0, sway, player1->yaw+90, player1->pitch, speed, base, NULL, 0);
     }
 
     void drawhudgun()
@@ -480,13 +637,13 @@ struct fpsclient : igameclient
         if(!hudgun() || editmode || player1->state==CS_SPECTATOR) return;
 
         int rtime = ws.reloadtime(player1->gunselect);
-        if(player1->lastattackgun==player1->gunselect && lastmillis-player1->lastaction<rtime)
+        if(player1->lastaction && player1->lastattackgun==player1->gunselect && lastmillis-player1->lastaction<rtime)
         {
             drawhudmodel(ANIM_GUNSHOOT, rtime/17.0f, player1->lastaction);
         }
         else
         {
-            drawhudmodel(ANIM_GUNIDLE|ANIM_LOOP, 100, 0);
+            drawhudmodel(ANIM_GUNIDLE|ANIM_LOOP);
         }
     }
 
@@ -511,6 +668,12 @@ struct fpsclient : igameclient
         if(player1->state==CS_SPECTATOR)
         {
             draw_text("SPECTATOR", 10, 827);
+            if(m_capture)
+            {
+                glLoadIdentity();
+                glOrtho(0, w*1800/h, 1800, 0, -1, 1);
+                cpc.capturehud(w, h);
+            }
             return;
         }
         draw_textf("%d",  90, 822, player1->state==CS_DEAD ? 0 : player1->health);
@@ -535,6 +698,40 @@ struct fpsclient : igameclient
             drawicon((float)(g*64), (float)r, 1220, 1650);
         }
         if(m_capture) cpc.capturehud(w, h);
+        else if(m_assassin) asc.drawhud(w, h);
+    }
+
+    void crosshaircolor(float &r, float &g, float &b)
+    {
+        if(player1->state!=CS_ALIVE) return;
+        if(player1->gunwait) r = g = b = 0.5f;
+        else if(!editmode && !m_noitemsrail)
+        {
+            if(player1->health<=25) { r = 1.0f; g = b = 0; }
+            else if(player1->health<=50) { r = 1.0f; g = 0.5f; b = 0; }
+        }
+    }
+
+    void lighteffects(dynent *e, vec &color, vec &dir)
+    {
+#if 0
+        fpsent *d = (fpsent *)e;
+        if(d->state!=CS_DEAD && d->quadmillis)
+        {
+            float t = 0.5f + 0.5f*sinf(2*M_PI*lastmillis/1000.0f);
+            color.y = color.y*(1-t) + t;
+        }
+#endif
+    }
+
+    void particletrack(physent *owner, vec &o, vec &d)
+    {
+        if(owner->type!=ENT_PLAYER && owner->type!=ENT_AI) return;
+        float dist = o.dist(d);
+        vecfromyawpitch(owner->yaw, owner->pitch, 1, 0, d);
+        float newdist = raycube(owner->o, d, dist, RAY_CLIPMAT|RAY_POLY);
+        d.mul(min(newdist, dist)).add(owner->o);
+        o = ws.hudgunorigin(GUN_PISTOL, owner->o, d, (fpsent *)owner);
     }
 
     void newmap(int size)
@@ -544,7 +741,7 @@ struct fpsclient : igameclient
 
     void edittrigger(const selinfo &sel, int op, int arg1, int arg2, int arg3)
     {
-        switch(op)
+        if(gamemode==1) switch(op)
         {
             case EDIT_FLIP:
             case EDIT_COPY:
@@ -575,16 +772,26 @@ struct fpsclient : igameclient
                    arg1, arg2);
                 break;
             }
+            case EDIT_REMIP:
+            {
+                cc.addmsg(SV_EDITF + op, "r");
+                break;
+            }
         }
     }
-    
+   
     void g3d_gamemenus() { sb.show(); }
 
     // any data written into this vector will get saved with the map data. Must take care to do own versioning, and endianess if applicable. Will not get called when loading maps from other games, so provide defaults.
     void writegamedata(vector<char> &extras) {}
     void readgamedata(vector<char> &extras) {}
 
-    char *gameident() { return "fps"; }
+    const char *gameident() { return "fps"; }
+    const char *defaultmap() { return "metl4"; }
+    const char *savedconfig() { return "config.cfg"; }
+    const char *defaultconfig() { return "data/defaults.cfg"; }
+    const char *autoexec() { return "autoexec.cfg"; }
+    const char *savedservers() { return "servers.cfg"; }
 };
 
 REGISTERGAME(fpsgame, "fps", new fpsclient(), new fpsserver());
