@@ -5,49 +5,59 @@ struct weaponstate
     fpsclient &cl;
     fpsent *player1;
 
-    struct guninfo { short sound, attackdelay, damage, projspeed, part, kickamount; char *name; };
-
     static const int MONSTERDAMAGEFACTOR = 4;
-    static const int SGRAYS = 20;
-    static const int SGSPREAD = 4;
     static const int OFFSETMILLIS = 500;
     vec sg[SGRAYS];
 
-    guninfo *guns;
-    
-    IVAR(maxdebris, 10, 50, 1000);
+    IVARP(maxdebris, 10, 25, 1000);
+    IVARP(maxbarreldebris, 5, 10, 1000);
 
     weaponstate(fpsclient &_cl) : cl(_cl), player1(_cl.player1)
     {
-        static guninfo _guns[NUMGUNS] =
+        CCOMMAND(nextweapon, "i", (weaponstate *self, int *dir), self->nextweapon(*dir));
+        CCOMMAND(setweapon, "i", (weaponstate *self, int *gun), self->setweapon(*gun));
+        CCOMMAND(weapon, "sss", (weaponstate *self, char *w1, char *w2, char *w3),
         {
-            { S_PUNCH1,    250,  50, 0,   0,  1, "fist"            },
-            { S_SG,       1400,  10, 0,   0, 20, "shotgun"         },  // *SGRAYS
-            { S_CG,        100,  30, 0,   0,  7, "chaingun"        },
-            { S_RLFIRE,    800, 120, 80,  0, 10, "rocketlauncher"  },
-            { S_RIFLE,    1500, 100, 0,   0, 30, "rifle"           },
-            { S_FLAUNCH,   500,  75, 80,  0, 10, "grenadelauncher" },
-            { S_PISTOL,    500,  25, 0,   0,  7, "pistol"          },
-            { S_FLAUNCH,   200,  20, 50,  4,  1, "fireball"        },
-            { S_ICEBALL,   200,  40, 30,  6,  1, "iceball"         },
-            { S_SLIMEBALL, 200,  30, 160, 7,  1, "slimeball"       },
-            { S_PIGR1,     250,  50, 0,   0,  1, "bite"            },
-        };
-        guns = _guns;
-        
-        CCOMMAND(weaponstate, weapon, "sss",
-        {
-            self->weaponswitch(args[0][0] ? atoi(args[0]) : -1,
-                               args[1][0] ? atoi(args[1]) : -1,
-                               args[2][0] ? atoi(args[2]) : -1);
+            self->weaponswitch(w1[0] ? atoi(w1) : -1,
+                               w2[0] ? atoi(w2) : -1,
+                               w3[0] ? atoi(w3) : -1);
 
         });
     }
 
+    void gunselect(int gun)
+    {
+        if(gun!=player1->gunselect)
+        {
+            cl.cc.addmsg(SV_GUNSELECT, "ri", gun);
+            playsound(S_WEAPLOAD, &player1->o);
+        }
+        player1->gunselect = gun;
+    }
+
+    void nextweapon(int dir)
+    {
+        if(player1->state!=CS_ALIVE) return;
+        dir = (dir < 0 ? NUMGUNS-1 : 1);
+        int gun = player1->gunselect;
+        loopi(NUMGUNS)
+        {
+            gun = (gun + dir)%NUMGUNS;
+            if(player1->ammo[gun]) break;
+        }
+        gunselect(gun);
+    }
+
+    void setweapon(int gun)
+    {
+        if(player1->state!=CS_ALIVE || gun<GUN_FIST || gun>GUN_PISTOL) return;
+        gunselect(gun);
+    }
+
     void weaponswitch(int a = -1, int b = -1, int c = -1)
     {
+        if(player1->state!=CS_ALIVE || a<-1 || b<-1 || c<-1 || a>=NUMGUNS || b>=NUMGUNS || c>=NUMGUNS) return;
         int *ammo = player1->ammo;
-        if(a<-1 || b<-1 || c<-1 || a>=NUMGUNS || b>=NUMGUNS || c>=NUMGUNS) return;
         int s = player1->gunselect;
 
         if     (a>=0 && s!=a  && ammo[a])          s = a;
@@ -61,12 +71,7 @@ struct weaponstate
         else if(s!=GUN_PISTOL && ammo[GUN_PISTOL]) s = GUN_PISTOL;
         else                                       s = GUN_FIST;
 
-        if(s!=player1->gunselect) 
-        {
-            cl.cc.addmsg(SV_GUNSELECT, "ri", s);
-            playsound(S_WEAPLOAD, &player1->o);
-        }
-        player1->gunselect = s;
+        gunselect(s);
     }
     
     int reloadtime(int gun) { return guns[gun].attackdelay; }
@@ -92,13 +97,10 @@ struct weaponstate
 
     void createrays(vec &from, vec &to)             // create random spread of rays for the shotgun
     {
-        loopi(SGRAYS)
-        {
-            offsetray(from, to, SGSPREAD, sg[i]);
-        }
+        loopi(SGRAYS) offsetray(from, to, SGSPREAD, sg[i]);
     }
 
-    enum { BNC_GRENADE, BNC_GIBS, BNC_DEBRIS };
+    enum { BNC_GRENADE, BNC_GIBS, BNC_DEBRIS, BNC_BARRELDEBRIS };
 
     struct bouncent : physent
     {
@@ -109,6 +111,7 @@ struct weaponstate
         int bouncetype;
         vec offset;
         int offsetmillis;
+        int id;
     };
 
     vector<bouncent *> bouncers;
@@ -119,7 +122,7 @@ struct weaponstate
         bnc.reset();
         bnc.type = ENT_BOUNCE;
         bnc.o = from;
-        bnc.radius = bnc.type==BNC_DEBRIS ? 0.5f : 1.5f;
+        bnc.radius = type==BNC_DEBRIS ? 0.5f : 1.5f;
         bnc.eyeheight = bnc.radius;
         bnc.aboveeye = bnc.radius;
         bnc.lifetime = lifetime;
@@ -127,6 +130,7 @@ struct weaponstate
         bnc.local = local;
         bnc.owner = owner;
         bnc.bouncetype = type;
+        bnc.id = cl.lastmillis;
 
         vec dir(to);
         dir.sub(from).normalize();
@@ -142,23 +146,32 @@ struct weaponstate
 
     void bounceupdate(int time)
     {
-        static const int maxtime = 20; // run at least 50 times a second
         loopv(bouncers)
         {
             bouncent &bnc = *(bouncers[i]);
-            if(bnc.bouncetype==BNC_GRENADE && vec(bnc.vel).add(bnc.gravity).magnitude() > 50.0f) regular_particle_splash(5, 1, 150, bnc.o);
+            if(bnc.bouncetype==BNC_GRENADE && vec(bnc.vel).add(bnc.gravity).magnitude() > 50.0f) 
+            {
+                vec pos(bnc.o);
+                pos.add(vec(bnc.offset).mul(bnc.offsetmillis/float(OFFSETMILLIS)));
+                regular_particle_splash(5, 1, 150, pos);
+            }
             vec old(bnc.o);
             int rtime = time;
             while(rtime > 0)
             {
-                int qtime = min(maxtime, rtime);
+                int qtime = min(bnc.bouncetype==BNC_GRENADE ? 10 : 30, rtime);
                 rtime -= qtime;
                 if((bnc.lifetime -= qtime)<0 || bounce(&bnc, qtime/1000.0f, 0.6f))
                 {
                     if(bnc.bouncetype==BNC_GRENADE)
                     {
                         int qdam = guns[GUN_GL].damage*(bnc.owner->quadmillis ? 4 : 1);
+                        if(bnc.owner->type==ENT_AI) qdam /= MONSTERDAMAGEFACTOR;
+                        hits.setsizenodelete(0);
                         explode(bnc.local, bnc.owner, bnc.o, NULL, qdam, GUN_GL);                    
+                        if(bnc.local)
+                            cl.cc.addmsg(SV_EXPLODE, "ri3iv", cl.lastmillis-cl.maptime, GUN_GL, bnc.id-cl.maptime,
+                                    hits.length(), hits.length()*sizeof(hitmsg)/sizeof(int), hits.getbuf());
                     }
                     delete &bnc;
                     bouncers.remove(i--);
@@ -171,7 +184,12 @@ struct weaponstate
         }
     }
 
-    struct projectile { vec o, to, offset; float speed; fpsent *owner; int gun; bool local; int offsetmillis; };
+    void removebouncers(fpsent *owner)
+    {
+        loopv(bouncers) if(bouncers[i]->owner==owner) { delete bouncers[i]; bouncers.remove(i--); }
+    }
+
+    struct projectile { vec o, to, offset; float speed; fpsent *owner; int gun; bool local; int offsetmillis; int id; };
     vector<projectile> projs;
 
     void projreset() { projs.setsize(0); bouncers.deletecontentsp(); bouncers.setsize(0); }
@@ -188,15 +206,26 @@ struct weaponstate
         p.owner = owner;
         p.gun = gun;
         p.offsetmillis = OFFSETMILLIS;
+        p.id = cl.lastmillis;
     }
-    
+   
+    void removeprojectiles(fpsent *owner) 
+    { 
+        loopv(projs) if(projs[i].owner==owner) projs.remove(i--);
+    }
+
+    IVARP(blood, 0, 1, 1);
+
     void damageeffect(int damage, fpsent *d)
     {
         vec p = d->o;
         p.z += 0.6f*(d->eyeheight + d->aboveeye) - d->eyeheight;
-        particle_splash(3, damage, 1000, p);
-        s_sprintfd(ds)("@%d", damage);
-        particle_text(d->abovehead(), ds, 8);
+        if(blood()) particle_splash(3, damage/10, 1000, p);
+        if(d!=cl.player1)
+        {
+            s_sprintfd(ds)("@%d", damage);
+            particle_text(d->abovehead(), ds, 8);
+        }
     }
     
     void spawnbouncer(vec &p, vec &vel, fpsent *d, int type)
@@ -215,43 +244,77 @@ struct weaponstate
         loopi(min(d->superdamage/25, 40)+1) spawnbouncer(from, vel, d, BNC_GIBS);
     }
 
-    void hit(int target, int damage, fpsent *d, fpsent *at, vec &vel, bool isrl)
+    struct hitmsg
     {
-        d->lastpain = cl.lastmillis;
-        at->totaldamage += damage;
-        vel.mul(80*damage/d->weight);
-        d->superdamage = 0;
-        if(d==player1)           { if(isrl) vel.mul(5); d->vel.add(vel); cl.selfdamage(damage, at==player1 ? -1 : -2, at); } 
-        else if(d->type==ENT_AI) { if(isrl) vel.mul(3); d->vel.add(vel); ((monsterset::monster *)d)->monsterpain(damage, at); }
-        else                     { if(isrl) vel.mul(2); cl.cc.addmsg(SV_DAMAGE, "ri6", target, damage, d->lifesequence, (int)(vel.x*DVELF), (int)(vel.y*DVELF), (int)(vel.z*DVELF)); playsound(S_PAIN1+rnd(5), &d->o); }
-        damageeffect(damage, d);
-        if(d->type==ENT_AI && d->state==CS_DEAD) superdamageeffect(d->vel, d);
+        int target, lifesequence, info;
+        ivec dir;
+    };
+    vector<hitmsg> hits;
+
+    void hit(int damage, dynent *d, fpsent *at, const vec &vel, int gun, int info = 1)
+    {
+        if(d->type==ENT_INANIMATE) 
+        {
+            movableset::movable *m = (movableset::movable *)d;
+            m->hitpush(damage, vel, at, gun);
+            m->damaged(damage, at, gun);
+            return;
+        }
+
+        fpsent *f = (fpsent *)d;
+
+        f->lastpain = cl.lastmillis;
+        if(at->type==ENT_PLAYER) at->totaldamage += damage;
+        f->superdamage = 0;
+
+        if(f->type==ENT_AI || !m_mp(cl.gamemode) || f==player1) f->hitpush(damage, vel, at, gun);
+
+        if(f->type==ENT_AI) ((monsterset::monster *)f)->monsterpain(damage, at); 
+        else if(!m_mp(cl.gamemode)) cl.damaged(damage, f, at);
+        else 
+        { 
+            hitmsg &h = hits.add();
+            h.target = f->clientnum;
+            h.lifesequence = f->lifesequence;
+            h.info = info;
+            damageeffect(damage, f);
+            if(f==player1)
+            {
+                h.dir = ivec(0, 0, 0);
+                f->damageroll(damage);
+                damageblend(damage);
+                playsound(S_PAIN6);
+            }
+            else 
+            {
+                h.dir = ivec(int(vel.x*DNF), int(vel.y*DNF), int(vel.z*DNF));
+                playsound(S_PAIN1+rnd(5), &f->o); 
+            }
+        }
     }
 
-    void hitpush(int target, int damage, fpsent *d, fpsent *at, vec &from, vec &to)
+    void hitpush(int damage, dynent *d, fpsent *at, vec &from, vec &to, int gun, int rays)
     {
         vec v(to);
         v.sub(from);
         v.normalize();
-        hit(target, damage, d, at, v, false);
+        hit(damage, d, at, v, gun, rays);
     }
 
-    static const int RL_DAMRAD = 40;  
-
-    void radialeffect(fpsent *o, vec &v, int cn, int qdam, fpsent *at, int gun)
+    void radialeffect(dynent *o, vec &v, int qdam, fpsent *at, int gun)
     {
         if(o->state!=CS_ALIVE) return;
         vec dir;
         float dist = rocketdist(o, dir, v);
         if(dist<RL_DAMRAD) 
         {
-            int damage = (int)(qdam*(1-dist/1.5f/RL_DAMRAD));
-            if(gun==GUN_RL && o==at) damage /= 2; 
-            hit(cn, damage, o, at, dir, true);
+            int damage = (int)(qdam*(1-dist/RL_DISTSCALE/RL_DAMRAD));
+            if(gun==GUN_RL && o==at) damage /= RL_SELFDAMDIV; 
+            hit(damage, o, at, dir, gun, int(dist*DMF));
         }
     }
 
-    float rocketdist(fpsent *o, vec &dir, vec &v)
+    float rocketdist(dynent *o, vec &dir, const vec &v)
     {
         vec middle = o->o;
         middle.z += (o->aboveeye-o->eyeheight)/2;
@@ -265,23 +328,25 @@ struct weaponstate
     {
         particle_splash(0, 200, 300, v);
         playsound(S_RLHIT, &v);
-        particle_fireball(v, RL_DAMRAD, gun==GUN_RL ? 22 : 23);
-        int numdebris = rnd(maxdebris()-5)+5;
-        vec debrisvel = vec(owner->o).sub(v).normalize(), debrisorigin(v);
+        particle_fireball(v, RL_DAMRAD, gun==GUN_RL || gun==GUN_BARREL ? 22 : 23);
+        adddynlight(v, 1.15f*RL_DAMRAD, vec(1, 0.75f, 0.5f), 600, 400);
+        int numdebris = gun==GUN_BARREL ? rnd(max(maxbarreldebris()-5, 1))+5 : rnd(maxdebris()-5)+5;
+        vec debrisvel = owner->o==v ? vec(0, 0, 0) : vec(owner->o).sub(v).normalize(), debrisorigin(v);
         if(gun==GUN_RL) debrisorigin.add(vec(debrisvel).mul(8));
-        loopi(numdebris) spawnbouncer(debrisorigin, debrisvel, owner, BNC_DEBRIS);
+        loopi(numdebris)
+            spawnbouncer(debrisorigin, debrisvel, owner, gun==GUN_BARREL ? BNC_BARRELDEBRIS : BNC_DEBRIS);
         if(!local) return;
         loopi(cl.numdynents())
         {
-            fpsent *o = (fpsent *)cl.iterdynents(i);
+            dynent *o = cl.iterdynents(i);
             if(!o || o==notthis) continue;
-            radialeffect(o, v, i-1, qdam, owner, gun);
+            radialeffect(o, v, qdam, owner, gun);
         }
     }
 
     void splash(projectile &p, vec &v, dynent *notthis, int qdam)
     {
-        if(p.gun!=GUN_RL)
+        if(guns[p.gun].part)
         {
             particle_splash(0, 100, 200, v);
             playsound(S_FEXPLODE, &v);
@@ -293,14 +358,15 @@ struct weaponstate
         }
     }
 
-    bool projdamage(fpsent *o, projectile &p, vec &v, int i, int qdam)
+    bool projdamage(dynent *o, projectile &p, vec &v, int qdam)
     {
         if(o->state!=CS_ALIVE) return false;
         if(!intersect(o, p.o, v)) return false;
         splash(p, v, o, qdam);
         vec dir;
-        rocketdist(o, dir, v);
-        hit(i, qdam, o, p.owner, dir, p.gun==GUN_RL);
+        if(guns[p.gun].part) { dir = v; dir.normalize(); }
+        else rocketdist(o, dir, v);
+        hit(qdam, o, p.owner, dir, p.gun, 0);
         return true;
     }
 
@@ -319,50 +385,75 @@ struct weaponstate
             v.mul(time/dtime);
             v.add(p.o);
             bool exploded = false;
+            hits.setsizenodelete(0);
             if(p.local)
             {
                 loopj(cl.numdynents())
                 {
-                    fpsent *o = (fpsent *)cl.iterdynents(j);
+                    dynent *o = cl.iterdynents(j);
                     if(!o || p.owner==o || o->o.reject(v, 10.0f)) continue;
-                    if(projdamage(o, p, v, j-1, qdam)) exploded = true;
+                    if(projdamage(o, p, v, qdam)) { exploded = true; break; }
                 }
             }
             if(!exploded)
             {
                 if(dist<4)
                 {
-                    if(p.o!=p.to && raycubepos(p.o, vec(p.to).sub(p.o), p.to, 0, RAY_CLIPMAT|RAY_POLY)>=4) 
-                        continue; // if original target was moving, reevaluate endpoint
+                    if(p.o!=p.to) // if original target was moving, reevaluate endpoint
+                    {
+                        vec ray(p.to);
+                        ray.sub(p.o);
+                        if(raycubepos(p.o, ray, p.to, 0, RAY_CLIPMAT|RAY_POLY)>=4) continue;
+                    }
                     splash(p, v, NULL, qdam);
                     exploded = true;
                 }
                 else 
                 {   
-                    if(p.gun==GUN_RL) 
+                    vec pos(v);
+                    pos.add(vec(p.offset).mul(p.offsetmillis/float(OFFSETMILLIS)));
+                    if(guns[p.gun].part)
                     {
-                        regular_particle_splash(5, 2, 300, v); 
+                         regular_particle_splash(1, 2, 300, pos);
+                         particle_splash(guns[p.gun].part, 1, 1, pos);
                     }
-                    else 
+                    else
                     {
-                         regular_particle_splash(1, 2, 300, v); 
-                         particle_splash(guns[p.gun].part, 1, 1, v);
+                        regular_particle_splash(5, 2, 300, pos);
                     }
                 }   
             }
-            if(exploded) projs.remove(i--);
+            if(exploded) 
+            {
+                if(p.local)
+                    cl.cc.addmsg(SV_EXPLODE, "ri3iv", cl.lastmillis-cl.maptime, p.gun, p.id-cl.maptime,
+                            hits.length(), hits.length()*sizeof(hitmsg)/sizeof(int), hits.getbuf());
+                projs.remove(i--);
+            }
             else p.o = v;
         }
     }
 
     vec hudgunorigin(int gun, const vec &from, const vec &to, fpsent *d)
     {
-        if(d!=player1) return from;
         vec offset(from);
-        offset.add(vec(to).sub(from).normalize().mul(6));
+        if(d!=player1 || isthirdperson()) 
+        {
+            vec front, right;
+            vecfromyawpitch(d->yaw, d->pitch, 1, 0, front);
+            offset.add(front.mul(d->radius));
+            if(d->type!=ENT_AI || cl.ms.monstertypes[((monsterset::monster *)d)->mtype].vwepname)
+            {
+                offset.z += (d->aboveeye + d->eyeheight)*0.75f - d->eyeheight;
+                vecfromyawpitch(d->yaw, 0, 0, -1, right);
+                offset.add(right.mul(0.5f*d->radius));
+            }
+            return offset;
+        }
+        offset.add(vec(to).sub(from).normalize().mul(2));
         if(cl.hudgun())
         {
-            offset.sub(vec(camup).mul(0.2f));
+            offset.sub(vec(camup).mul(1.0f));
             offset.add(vec(camright).mul(0.8f));
         }
         return offset;
@@ -372,7 +463,6 @@ struct weaponstate
     {
         playsound(guns[gun].sound, d==player1 ? NULL : &d->o);
         int pspeed = 25;
-        vec behind = vec(from).sub(to).normalize().mul(4).add(from);
         switch(gun)
         {
             case GUN_FIST:
@@ -383,7 +473,7 @@ struct weaponstate
                 loopi(SGRAYS)
                 {
                     particle_splash(0, 20, 250, sg[i]);
-                    particle_flare(hudgunorigin(gun, behind, sg[i], d), sg[i], 300);
+                    particle_flare(hudgunorigin(gun, from, sg[i], d), sg[i], 300, 10);
                 }
                 break;
             }
@@ -393,7 +483,7 @@ struct weaponstate
             {
                 particle_splash(0, 200, 250, to);
                 //particle_trail(1, 10, from, to);
-                particle_flare(hudgunorigin(gun, behind, to, d), to, 600);
+                particle_flare(hudgunorigin(gun, from, to, d), to, 600, 10);
                 break;
             }
 
@@ -422,13 +512,13 @@ struct weaponstate
         }
     }
 
-    fpsent *intersectclosest(vec &from, vec &to, int &n, fpsent *at)
+    dynent *intersectclosest(vec &from, vec &to, fpsent *at)
     {
-        fpsent *best = NULL;
+        dynent *best = NULL;
         float bestdist = 1e16f;
         loopi(cl.numdynents())
         {
-            fpsent *o = (fpsent *)cl.iterdynents(i);
+            dynent *o = cl.iterdynents(i);
             if(!o || o==at || o->state!=CS_ALIVE) continue;
             if(!intersect(o, from, to)) continue;
             float dist = at->o.dist(o->o);
@@ -436,7 +526,6 @@ struct weaponstate
             {
                 best = o;
                 bestdist = dist;
-                n = i-1;
             }
         }
         return best;
@@ -452,8 +541,7 @@ struct weaponstate
         int qdam = guns[d->gunselect].damage;
         if(d->quadmillis) qdam *= 4;
         if(d->type==ENT_AI) qdam /= MONSTERDAMAGEFACTOR;
-        int i = -1, n = -1;
-        fpsent *o, *cl;
+        dynent *o, *cl;
         if(d->gunselect==GUN_SG)
         {
             bool done[SGRAYS];
@@ -461,27 +549,26 @@ struct weaponstate
             for(;;)
             {
                 bool raysleft = false;
-                int damage = 0;
+                int hitrays = 0;
                 o = NULL;
-                loop(r, SGRAYS) if(!done[r] && (cl = intersectclosest(from, sg[r], n, d)))
+                loop(r, SGRAYS) if(!done[r] && (cl = intersectclosest(from, sg[r], d)))
                 {
                     if((!o || o==cl) /*&& (damage<cl->health+cl->armour || cl->type!=ENT_AI)*/)
                     {
-                        damage += qdam;
+                        hitrays++;
                         o = cl;
                         done[r] = true;
-                        i = n;
                         shorten(from, o->o, sg[r]);
                     }
                     else raysleft = true;
                 }
-                if(damage) hitpush(i, damage, o, d, from, to);
+                if(hitrays) hitpush(hitrays*qdam, o, d, from, to, d->gunselect, hitrays);
                 if(!raysleft) break;
             }
         }
-        else if((o = intersectclosest(from, to, i, d)))
+        else if((o = intersectclosest(from, to, d)))
         {
-            hitpush(i, qdam, o, d, from, to);
+            hitpush(qdam, o, d, from, to, d->gunselect, 1);
             shorten(from, o->o, to);
         }
     }
@@ -491,14 +578,23 @@ struct weaponstate
         int attacktime = cl.lastmillis-d->lastaction;
         if(attacktime<d->gunwait) return;
         d->gunwait = 0;
-        if(!d->attacking) return;
+        if(d==player1 && !d->attacking) return;
         d->lastaction = cl.lastmillis;
         d->lastattackgun = d->gunselect;
-        if(!d->ammo[d->gunselect]) { cl.playsoundc(S_NOAMMO); d->gunwait = 600; d->lastattackgun = -1; weaponswitch(); return; }
+        if(!d->ammo[d->gunselect]) 
+        { 
+            if(d==player1)
+            {
+                cl.playsoundc(S_NOAMMO, d); 
+                d->gunwait = 600; 
+                d->lastattackgun = -1; 
+                weaponswitch(); 
+            }
+            return; 
+        }
         if(d->gunselect) d->ammo[d->gunselect]--;
         vec from = d->o;
         vec to = targ;
-        from.z -= 0.8f;    // below eye
 
         vec unitv;
         float dist = to.dist(from, unitv);
@@ -524,13 +620,21 @@ struct weaponstate
         if(d->gunselect==GUN_SG) createrays(from, to);
         else if(d->gunselect==GUN_CG) offsetray(from, to, 1, to);
             
-        if(d->quadmillis && attacktime>200) cl.playsoundc(S_ITEMPUP);
+        if(d->quadmillis && attacktime>200) cl.playsoundc(S_ITEMPUP, d);
+
+        hits.setsizenodelete(0);
 
         if(!guns[d->gunselect].projspeed) raydamage(from, to, d);
 
         shootv(d->gunselect, from, to, d, true);
 
-        if(d->type!=ENT_AI) cl.cc.addmsg(SV_SHOT, "ri7", d->gunselect, (int)(from.x*DMF), (int)(from.y*DMF), (int)(from.z*DMF), (int)(to.x*DMF), (int)(to.y*DMF), (int)(to.z*DMF));
+        if(d==player1)
+        {
+            cl.cc.addmsg(SV_SHOOT, "ri2i6iv", cl.lastmillis-cl.maptime, d->gunselect,
+                            (int)(from.x*DMF), (int)(from.y*DMF), (int)(from.z*DMF), 
+                            (int)(to.x*DMF), (int)(to.y*DMF), (int)(to.z*DMF),
+                            hits.length(), hits.length()*sizeof(hitmsg)/sizeof(int), hits.getbuf());
+        }
 
         d->gunwait = guns[d->gunselect].attackdelay;
 
@@ -559,9 +663,13 @@ struct weaponstate
             pitch = -bnc.roll;
             const char *mdl = "projectiles/grenade";
             string debrisname;
-            if(bnc.bouncetype==BNC_GIBS) mdl = ((int)(size_t)&bnc)&0x40 ? "gibc" : "gibh";
+            int cull = MDL_CULL_VFC|MDL_CULL_DIST;
+            if(bnc.bouncetype==BNC_GIBS) { mdl = ((int)(size_t)&bnc)&0x40 ? "gibc" : "gibh"; cull |= MDL_DYNSHADOW; }
             else if(bnc.bouncetype==BNC_DEBRIS) { s_sprintf(debrisname)("debris/debris0%d", ((((int)(size_t)&bnc)&0xC0)>>6)+1); mdl = debrisname; }
-            rendermodel(color, dir, mdl, ANIM_MAPMODEL|ANIM_LOOP, 0, 0, pos.x, pos.y, pos.z, yaw, pitch, 10.0f, 0, NULL, MDL_SHADOW);
+            else if(bnc.bouncetype==BNC_BARRELDEBRIS) { s_sprintf(debrisname)("barreldebris/debris0%d", ((((int)(size_t)&bnc)&0xC0)>>6)+1); mdl = debrisname; }
+            else cull = MDL_CULL_VFC|MDL_DYNSHADOW;
+                
+            rendermodel(color, dir, mdl, ANIM_MAPMODEL|ANIM_LOOP, 0, 0, pos, yaw, pitch, 0, 0, NULL, cull);
         }
         loopv(projs)
         {
@@ -579,7 +687,7 @@ struct weaponstate
             v.mul(3);
             v.add(pos);
             lightreaching(v, color, dir);
-            rendermodel(color, dir, "projectiles/rocket", ANIM_MAPMODEL|ANIM_LOOP, 0, 0, v.x, v.y, v.z, yaw, pitch, 10.0f, 0, NULL, MDL_SHADOW);
+            rendermodel(color, dir, "projectiles/rocket", ANIM_MAPMODEL|ANIM_LOOP, 0, 0, v, yaw, pitch, 0, 0, NULL, MDL_CULL_VFC|MDL_DYNSHADOW);
         }
     }  
 };
