@@ -3,7 +3,7 @@
 
 // ET_*: the only static entity types dictated by the engine... rest are gamecode dependent
 
-enum { ET_EMPTY=0, ET_LIGHT, ET_MAPMODEL, ET_PLAYERSTART, ET_ENVMAP, ET_PARTICLES, ET_SOUND, ET_GAMESPECIFIC };
+enum { ET_EMPTY=0, ET_LIGHT, ET_MAPMODEL, ET_PLAYERSTART, ET_ENVMAP, ET_PARTICLES, ET_SOUND, ET_SPOTLIGHT, ET_GAMESPECIFIC };
 
 struct entity                                   // persistent map entity
 {
@@ -27,36 +27,69 @@ struct extentity : entity                       // part of the entity that doesn
     uchar spawned, inoctanode, visible, triggerstate;        // the only dynamic state of a map entity
     vec color, dir;
     int lasttrigger;
+    extentity *attached;
 
-    extentity() : visible(false), triggerstate(TRIGGER_RESET), lasttrigger(0) {}
+    extentity() : visible(false), triggerstate(TRIGGER_RESET), lasttrigger(0), attached(NULL) {}
 };
 
 //extern vector<extentity *> ents;                // map entities
 
+enum 
+{ 
+    ANIM_DEAD = 0, ANIM_DYING, ANIM_IDLE, 
+    ANIM_FORWARD, ANIM_BACKWARD, ANIM_LEFT, ANIM_RIGHT, 
+    ANIM_PUNCH, ANIM_SHOOT, ANIM_PAIN, 
+    ANIM_JUMP, ANIM_SINK, ANIM_SWIM, 
+    ANIM_EDIT, ANIM_LAG, ANIM_TAUNT, ANIM_WIN, ANIM_LOSE, 
+    ANIM_GUNSHOOT, ANIM_GUNIDLE,
+    ANIM_VWEP, ANIM_SHIELD, ANIM_POWERUP, 
+    ANIM_MAPMODEL, ANIM_TRIGGER, 
+    NUMANIMS 
+};
+
+#define ANIM_INDEX       0xFF
+#define ANIM_LOOP        (1<<8)
+#define ANIM_START       (1<<9)
+#define ANIM_END         (1<<10)
+#define ANIM_REVERSE     (1<<11)
+#define ANIM_DIR         0xF00
+#define ANIM_SECONDARY   12
+#define ANIM_NOSKIN      (1<<24)
+#define ANIM_ENVMAP      (1<<25)
+#define ANIM_TRANSLUCENT (1<<26)
+#define ANIM_SHADOW      (1<<27)
+#define ANIM_SETTIME     (1<<28)
+#define ANIM_FLAGS       (0x7F<<24)
+
 struct animstate                                // used for animation blending of animated characters
 {
     int anim, frame, range, basetime;
     float speed;
     animstate() : anim(0), frame(0), range(0), basetime(0), speed(100.0f) { }
 
-    bool operator==(const animstate &o) const { return frame==o.frame && range==o.range && basetime==o.basetime && speed==o.speed; }
-    bool operator!=(const animstate &o) const { return frame!=o.frame || range!=o.range || basetime!=o.basetime || speed!=o.speed; }
+    bool operator==(const animstate &o) const { return frame==o.frame && range==o.range && (anim&(ANIM_SETTIME|ANIM_DIR))==(o.anim&(ANIM_SETTIME|ANIM_DIR)) && (anim&ANIM_SETTIME || basetime==o.basetime) && speed==o.speed; }
+    bool operator!=(const animstate &o) const { return frame!=o.frame || range!=o.range || (anim&(ANIM_SETTIME|ANIM_DIR))!=(o.anim&(ANIM_SETTIME|ANIM_DIR)) || (!(anim&ANIM_SETTIME) && basetime!=o.basetime) || speed!=o.speed; }
 };
 
-enum { ANIM_DYING = 0, ANIM_DEAD, ANIM_PAIN, ANIM_IDLE, ANIM_IDLE_ATTACK, ANIM_RUN, ANIM_RUN_ATTACK, ANIM_EDIT, ANIM_LAG, ANIM_JUMP, ANIM_JUMP_ATTACK, ANIM_GUNSHOOT, ANIM_GUNIDLE, ANIM_MAPMODEL, ANIM_TRIGGER, NUMANIMS };
+#define ANIM_INDEX       0xFF
+#define ANIM_LOOP        (1<<8)
+#define ANIM_START       (1<<9)
+#define ANIM_END         (1<<10)
+#define ANIM_REVERSE     (1<<11)
+#define ANIM_SECONDARY   12
+#define ANIM_NOSKIN      (1<<24)
+#define ANIM_ENVMAP      (1<<25)
+#define ANIM_TRANSLUCENT (1<<26)
+#define ANIM_SHADOW      (1<<27)
+#define ANIM_FLAGS       (0x7F<<24)
 
-#define ANIM_INDEX      0xFF
-#define ANIM_LOOP       (1<<8)
-#define ANIM_START      (1<<9)
-#define ANIM_END        (1<<10)
-#define ANIM_REVERSE    (1<<11)
-#define ANIM_NOSKIN     (1<<12)
-
-enum { CS_ALIVE = 0, CS_DEAD, CS_LAGGED, CS_EDITING, CS_SPECTATOR };
+enum { CS_ALIVE = 0, CS_DEAD, CS_SPAWNING, CS_LAGGED, CS_EDITING, CS_SPECTATOR };
 
 enum { PHYS_FLOAT = 0, PHYS_FALL, PHYS_SLIDE, PHYS_SLOPE, PHYS_FLOOR, PHYS_STEP_UP, PHYS_STEP_DOWN, PHYS_BOUNCE };
 
-enum { ENT_PLAYER = 0, ENT_AI, ENT_CAMERA, ENT_BOUNCE };
+enum { ENT_PLAYER = 0, ENT_AI, ENT_INANIMATE, ENT_CAMERA, ENT_BOUNCE };
+
+enum { COLLIDE_AABB = 0, COLLIDE_ELLIPSE };
 
 struct physent                                  // base entity type, can be affected by physics
 {
@@ -71,36 +104,53 @@ struct physent                                  // base entity type, can be affe
     bool inwater;
     bool jumpnext;
     bool blocked, moving;                       // used by physics to signal ai
+    physent *onplayer;
+    int lastmove, lastmoveattempt, collisions, stacks;
 
     char move, strafe;
 
     uchar physstate;                            // one of PHYS_* above
     uchar state;                                // one of CS_* above
     uchar type;                                 // one of ENT_* above
+    uchar collidetype;                          // one of COLLIDE_* above           
 
     physent() : o(0, 0, 0), yaw(270), pitch(0), roll(0), maxspeed(100), 
                radius(4.1f), eyeheight(14), aboveeye(1), xradius(4.1f), yradius(4.1f), 
-               inwater(false), blocked(false), moving(true), state(CS_ALIVE), type(ENT_PLAYER)
+               blocked(false), moving(true), 
+               onplayer(NULL), lastmove(0), lastmoveattempt(0), collisions(0), stacks(0),
+               state(CS_ALIVE), type(ENT_PLAYER),
+               collidetype(COLLIDE_ELLIPSE)
                { reset(); }
                
     void reset()
     {
+    	inwater = false;
         timeinair = strafe = move = 0;
         physstate = PHYS_FALL;
         vel = gravity = vec(0, 0, 0);
     }
 };
 
+struct occludequery;
+
 struct dynent : physent                         // animated characters, or characters that can receive input
 {
     bool k_left, k_right, k_up, k_down;         // see input code
     float targetyaw, rotspeed;                  // AI rotation
+    float lastyaw, lastpitch;                   // last yaw/pitch to interpolate from, MP only
+    int orientmillis;                           // time last yaw/pitch was recorded
 
     animstate prev[2], current[2];              // md2's need only [0], md3's need both for the lower&upper model
     int lastanimswitchtime[2];
     void *lastmodel[2];
-    
-    dynent() { reset(); loopi(2) { lastanimswitchtime[i] = -1; lastmodel[i] = NULL; } }
+    occludequery *query;
+    int occluded, lastrendered;
+
+    dynent() : lastyaw(0), lastpitch(0), orientmillis(0), query(NULL), occluded(0), lastrendered(0)
+    { 
+        reset(); 
+        loopi(2) { lastanimswitchtime[i] = -1; lastmodel[i] = NULL; } 
+    }
                
     void stopmoving()
     {
